
local Logger = TaggedLog("ActorManager", true)

ActorManager = {
    scene = nil,
    actorPool = {},
}

function ActorManager:init(scene)
    self.scene = scene
    self.scene.actors = self.scene.actors or {}
end

local actorPreCreateList = 
{
    "Slice_I",
    "Owl",
    "Slice_J",
    "Slice_H",
    "Slice_E",
    "Slice_F",
    "Slice_K",
    "Slice_D",
    "Ground",
    "Slice_A",
    "Slice_B",
    "Slice_G",
    "Skybox",
    "Slice_C",
    "Slice_L",
}

function ActorManager:startPreCreate()
    self.preCreateIndex = 1
    for _, name in ipairs(actorPreCreateList) do
        Global.resources:addActor( SceneData.actors[name] )
    end
    Global.resources:prefetch()
end

function ActorManager:preCreateUpdate()
    if self.preCreateIndex ~= nil and self.preCreateIndex <= #actorPreCreateList then
        local name = actorPreCreateList[ self.preCreateIndex ]
        Logger:message("Pre-creating actor object: "..name..", index: "..self.preCreateIndex)

        if self.scene.actors[name] == nil then
            local actorData = SceneData.actors[name]
            local actor     = self.scene:createActor(name, actorData)

            self.scene.actors[name] = nil
            self.actorPool[name]    = actor

            actor:setVisibility(false)
            actor:setActive(false)
        end

        self.preCreateIndex = self.preCreateIndex+1
    end
end

function ActorManager:getActor(name, createActor)
    -- Default creation to false unless overwritten
    createActor = createActor or false

    if(self.scene.actors[name] == nil and createActor) then
        self:instantiateActor(name)
    end

    return self.scene.actors[name]
end

function ActorManager:instantiateActor(name)
    if (name == nil) then
        return
    end

    local actorData = SceneData.actors[name]
    local actor = nil

    if (actorData == nil or actorData.model == nil) then
        -- If the name does not exist create empty actor
        Logger:message("Create transform actor object: "..name)
        actor = self.scene:createTransform(name)
        actor.materials = {}

        if actorData and actorData.viewFlags then
            actor:setViewFlags(actorData.viewFlags)
        end

        self.scene.actors[name] = actor
        actorData = actorData or {}
    else
        actor = self.actorPool[name]
        if actor == nil then
            Logger:message("Created actor object: "..name)
            actor = self.scene:createActor(name, actorData)
        else
            Logger:message("Created actor object from pre-create list: "..name)
			actor:setActive(true)
            self.scene.actors[name] = actor
            self.actorPool[name]    = nil
        end
    end

    actor.name = name
    actor:setVisibility(false)

    -- Set up position if it is specified
    local initialPos = actorData.position
    if (initialPos) then
        initialPos = Vector(initialPos[1], initialPos[2], initialPos[3])
        Logger:message("initial position = " .. tostring(initialPos))
        actor:setLocal({ position = initialPos})
    end

    -- Set up rotation if it is specified
    local initialRot = actorData.rotation
    if (initialRot) then
        local transform = actor:getLocal()
        transform.rotation = Quaternion.fromEulerAngles(Vector(initialRot[1], initialRot[2],initialRot[3]))
        actor:setLocal(transform)
    end

    -- Add any components needed
    if (actorData.components) then
        for _, compData in ipairs(actorData.components) do
            local class = _G[compData.name]
            if (class == nil) then
                Log.error("Unknown component type: " .. tostring(compData.name))
            else
                actor:addComponent(compData.name, class:new(compData.params))
                Logger:message("Added component: " .. compData.name .. " to " .. actor.name)
            end
        end
    end
    if (actorData.targetBone) then
        actor.targetBone = actor:getBone(actorData.targetBone)
    end

    self:addAnimationDriver(actor)

    Events.send("ActorManager", "onActorCreated", actor)
end


-- Where boneDef is a table with actorName and bonePath
function ActorManager:getBone(boneDef)
    local actor = self:getActor(boneDef.actorName, false)
    if (actor and boneDef.bonePath) then
        return actor:getBone(boneDef.bonePath)
    end
    return actor
end

function ActorManager:getTargetBone(actorName)
    local actor = self:getActor(actorName)
    if (actor) then
        return actor.targetBone
    end
    return nil
end

-- Return the world position of the actor's target bone if
-- one is set up on it, the actor's world position if there isn't
-- a target bone on the actor, or nil if there is no actor by that name
function ActorManager:getActorTrackingPos(actorName)
    local actor = self:getActor(actorName, false)
    if (actor) then
        -- Use the tracking bone if one is specified
        local targetBone = actor.targetBone
        if (targetBone) then
            return targetBone:getWorld().position
        end

        -- See if the actor has any components that might give a tracking position
        if actor.components then
            for _, comp in pairs(actor.components) do
                if comp.getTrackingPosition then
                    return comp:getTrackingPosition()
                end
            end
        end

        -- Just return the actor's position
        return actor:getWorld().position
    end
    return nil
end

function ActorManager:addAnimationDriver(actor)
    if (actor.animDriver == nil) then
        -- TODO: Re-create the AnimationDriver component
        actor.animDriver = actor:addComponent("animDriver", AnimationDriver.new())
    end
end


function ActorManager:destroyActor(name)
    if (self.scene.actors[name] and status(self.scene.actors[name]) ~= nil) then
        self.scene.actors[name]:destroy()
        self.scene.actors[name] = nil

        Log.message("ActorManager: destroyed "..name)
        Events.send("ActorManager", "onActorDestroyed", name)
    end
end
