
local Logger = TaggedLog("SoundControl", true)

SoundControl = {
    currentVolume = 0,
    positionalVolume = 1,
    audioType = "SFX",
    fading = false,
    startTimer = 0,
    fadeTimer = 1.1,
    fadeTime = 1,
    trackingActorName = nil,

    volumeStart = 0,
    volumeEnd = 0,
    time = 0,
    autorelease = true,
    timePlaying = 0,
}

function SoundControl:new(sound)
    local this = { sound = sound }
    self.__index = self
    setmetatable(this, self)

    sound:setVolume(this.currentVolume * SoundManager.masterVolumes[this.audioType])

    return this
end

function SoundControl:onUpdate()
    local startPlaying = false
    self:updateTiming()
    if (self.fading) then
        self.startTimer = self.startTimer + self.lastDeltaTime
        if (self.startTimer > 0) then
            if (not self.sound:isPlaying()) then
                self.timePlaying = 0
                startPlaying = true
                if (self.audioType == "Music") then
                    SoundManager.currentMusicBed = self.sound
                    Logger:message("Setting current music bed to: " .. self.sound.name)
                end
            end
            self.fadeTimer = self.fadeTimer + (self.lastDeltaTime / self.fadeTime)
            self.currentVolume = math.lerp(self.volumeStart, self.volumeEnd, self.fadeTimer)
            if (self.fadeTimer > 1) then
                self.fading = false
                if (self.currentVolume < 0.001 and self.autorelease) then
                    Logger:message("autoreleasing fading sound: " .. self.sound.name)
                    self.sound:stop()
                    self:releaseSound()
                    return
                end
            end
        end
    elseif (not self.sound:isPlaying() and self.autorelease) then
        Logger:message("autoreleasing sound: " .. self.sound.name)
        self:releaseSound()
        return
    end

    if (self.trackPosition) then
        -- Apply panning and attenuation for sounds that track an object
        local position = ActorManager:getActorTrackingPos(self.trackPosition)
        if position then
            local pan, volume = self.sound:getPanVolume3D(Global.scene.mainCamera, position, 0)
            self.sound:setPanBalance(pan)
            self.positionalVolume = volume
        end
    end

    local masterVolume = SoundManager.masterVolumes[self.audioType]
    self.sound:setVolume(self.currentVolume * self.positionalVolume * masterVolume)

    if startPlaying then
        self.sound:play()
    end
end

function SoundControl:forceFade()
    self.fading = false
    if (self.fadeTimer < 1) then
        self.fadeTimer = 1.1
        self.currentVolume = self.volumeEnd
    end
end

function SoundControl:setTimedFade(startDelay, newFadeTime, newFadeStart, newFadeEnd)
    self.fading = true
    self.startTimer = -startDelay
    self.fadeTimer = 0
    if (newFadeTime < 0.01) then
        self.fadeTimer = 0.99
        self.fadeTime = 1
    else
        self.fadeTime = newFadeTime;
    end

    self.volumeStart = newFadeStart or self.currentVolume
    self.volumeEnd = newFadeEnd
end

function SoundControl:setVolume(newVolume)
    if (self.fading) then Log.error("SoundControl: don't set the volume on a fading volume control: " .. self.sound.name .. tostring(debug.traceback())) end
    self.currentVolume = newVolume
end

function SoundControl:setAudioType(audioType)
    self.audioType = audioType
end

function SoundControl:releaseSound()
    Logger:message("releasing sound: " .. tostring(self.sound.name))
    SoundManager:releaseSound(self.sound)
end

-- To make up for the unreliable results from sound:getSeekPosition(),
-- we try to give a better estimate here.
function SoundControl:getSeekPosition()
    if (self.timePlaying) then
        return self.timePlaying % self.sound:getDurationSeconds()
    end
    return 0
end

function SoundControl:updateTiming()
    local updateTime = os.clock()
    self.lastUpdateTime = self.lastUpdateTime or updateTime
    self.lastDeltaTime = updateTime - self.lastUpdateTime
    self.lastUpdateTime = updateTime

    if (self.sound:isPlaying()) then
        self.timePlaying = self.timePlaying + self.lastDeltaTime
    end
end
