
local Logger = TaggedLog.new("SoundManager", true)

local function audioFileURI(baseName)
    if (not string.endsWith(baseName, ".mp3")) then
        return "buggynight:audio/" .. string.lower(baseName) .. ".ogg"
    else
        return "buggynight:audio/" .. string.lower(baseName)
    end
end

local function soundControlName(sound)
    return sound.name .. "-volume"
end

-- Compute the normalize pitch of the camera (-1 pointing down, 0 pointing at the horizon, 1 pointing up)
local function getCameraPitch()
    local forward = Vector.rotateBy(Vector.forward, Global.scene.mainCamera:getWorld().rotation)
    forward:normalize()
    return forward.y
end


SoundManager = Manager:new {
    name = "SoundManager",
    masterVolumes = {},

    emitter = nil,
    sounds = {},

    currentMusicBed = nil,
}

function SoundManager:prefetch(audioName, group)
    Logger:message("Prefetching: " .. tostring(audioName) .. " from segment: " .. tostring(SegmentManager:currentSegmentName()))
    group:addAudio(audioFileURI(audioName))
    group:prefetch()
end

function SoundManager:onStart()
    self.emitter = AudioEmitter.create()

    -- Set initial volume for each type
    for name, type in pairs(SceneData.AudioTypes) do
        if type.enabled then
            self.masterVolumes[name] = type.masterVolume
        else
            self.masterVolumes[name] = 0
        end
        Logger:message("Setting global volume for " .. name .. " to " .. tostring(type.masterVolume))
    end
end

function SoundManager:onDestroy()
    if (self.emitter) then
        self.emitter:destroy()
        self.emitter = nil
    end
end

function SoundManager:onUpdate()

    -- Attenuate any volumes that need it
    local cameraPitch = getCameraPitch()
    for name, type in pairs(SceneData.AudioTypes) do
        if type.enabled and type.verticalAttenuation and #type.verticalAttenuation > 0 then
            local attenuation = nil
            local prevEntry = nil
            for _, entry in ipairs(type.verticalAttenuation) do
                if cameraPitch > entry.pitch then
                    if prevEntry then
                        attenuation = math.lerp(entry.atten, prevEntry.atten, (cameraPitch - entry.pitch)  / (prevEntry.pitch - entry.pitch))
                    else
                        attenuation = entry.atten
                    end
                    break
                else
                    prevEntry = entry
                end
            end
            attenuation = attenuation or prevEntry.atten
            self.masterVolumes[name] = type.masterVolume * attenuation
        end
    end
end

function SoundManager:getSound(name)
    local sound = self.sounds[name]
    if (sound) then return sound end

    -- soundName, uriPath, preCache, startTimeMillis, stopTimeMillis
    sound = self.emitter:addSound(name, audioFileURI(name), false, 0, -1)
    if (sound) then
        sound.name = name
        sound.soundControl = self.managerGameObj:addComponent(soundControlName(sound), SoundControl:new(sound))
        self.sounds[name] = sound
        Logger:message("Created sound " .. tostring(name) .. " (" .. tostring(Util.numKeys(self.sounds)) .. " sounds allocated)")
    else
        Log.error("Unable to create sound for: " .. tostring(name))
    end
    return sound
end

function SoundManager:releaseSound(sound)
    if (sound and sound.name) then
        if (sound == self.currentMusicBed) then
            self.currentMusicBed = nil
        end
        if (sound.soundControl) then
            self.managerGameObj:removeComponent(soundControlName(sound))
            sound.soundControl = nil
        end
        self.emitter:removeSound(sound.name)
        self.sounds[sound.name] = nil
        Logger:message("Released sound " .. tostring(sound.name) .. " (" .. tostring(Util.numKeys(self.sounds)) .. " sounds allocated)")
    end
end

function SoundManager:getMusicBed(name, ambient, looping)
    local sound = self.sounds[name]
    if (sound) then return sound end

    sound = self:getSound(name)
    if (sound ~= nil) then
        sound:setLooping(looping)
        if (ambient) then
            sound.soundControl:setAudioType("Ambient")
        else
            sound.soundControl:setAudioType("Music")
        end
    end

    return sound
end

function SoundManager:playStab(name, volume)
    Logger:message("Playing stab: " .. name)
    return self:playSound(name, volume, false, "Stab")
end

function SoundManager:playSFX(name, volume, looping)
    Logger:message("Playing sound FX = " .. name)
    return self:playSound(name, volume, looping, "SFX")
end

function SoundManager:playAmbient(name, volume)
    Logger:message("Playing ambient: " .. name)
    return self:playSound(name, volume, false, "Ambient")
end

function SoundManager:playSound(name, volume, looping, type)
    local sound = self:getSound(name)
    if (sound == nil) then return end

    sound:setLooping(looping)
    sound.soundControl:setAudioType(type)
    sound.soundControl:setVolume(volume)
    sound:play()

    return sound
end
