
local Logger = TaggedLog("LookAtDetectorCore", true)
local debugMode = false

LookAtDetectorCore = {
    worldPosT = nil,
    isTriggered = false,

    -- The closer to 1 this value is the more directly you have to look at it to trigger it
    lookAtThreshold = nil,
    lookedAt = false,
    toTarget = nil,
    colliderRadius = 0,
    dotDifferential = 0,
    debugSphere = nil,
}

LookAtDetectorCore.new = function(self, params)
    local o = shallowcopy(params) or {}

    Logger:message("new : table = " .. table.tostring(o))
    setmetatable(o, self)
    self.__index = self

    return o
end

LookAtDetectorCore.onStart = function(self)
end

LookAtDetectorCore.getTransform = function(self)
    return self.worldPosT
end

LookAtDetectorCore.onUpdate = function(self)
    local posT = self.object:getWorld()

    -- Don't use full camera pos because it leads to parallax error (since the cam is on a boom)
    -- Do account for the camera height however, so get the rig
    if (Global.scene.cameraRig == nil) then
        return
    end

    local toTarget = posT.position - Global.scene.cameraRig:getWorld().position

    -- If the camera to target vector has changed then something moved, recalculate or goal value
    if (self.toTarget == nil or self.toTarget ~= toTarget) then
        self.toTarget = toTarget:copy()
        self.lookAtThreshold = self:calculateGoal()
    end

    toTarget:normalize()

    local look = Global.scene.mainCamera:getWorld():forward()

    self.dotDifferential = Vector.dot(look, toTarget)
    local dotDifferential = self.dotDifferential

    if (dotDifferential >= self.lookAtThreshold and not self.isTriggered) then
        self.isTriggered = true
        self.lookedAt = true
        self.object:setDrawDebug(true)
        Logger:message("Looking at: " .. self.label)
        Events.send(self.object, "lookingAt", self.label)
    elseif (dotDifferential < self.lookAtThreshold and self.isTriggered) then
        self.isTriggered = false
        self.lookedAt = false
        self.object:setDrawDebug(false)
        Logger:message("Not looking at: " .. self.label)
        Events.send(self.object, "notLookingAt", self.label)
    end

    -- Set up debug sphere if enabled
    if (debugMode and self.debugSphere == nil) then
        self.debugSphere = Global.scene:createActor(self.label.."_debugsphere", SceneData.actors.DebugSphere)
    elseif (not debugMode and debugSphere ~= nil) then
        self.debugSphere:destroy()
        self.debugSphere = nil
    end

    if (self.debugSphere) then
        self.debugSphere:setVisibility(true)

        local transform = self.debugSphere:getLocal()
        transform.position = self.object:getWorld().position
        transform.scale = Vector(self.colliderRadius, self.colliderRadius, self.colliderRadius)
        self.debugSphere:setLocal( transform )

--        Logger:message("LookAt Debug " .. self.label .." : ("..transform.position.x..", "..transform.position.y..", "..transform.position.z..")")
--        Logger:message("LookAt WorldCam at: " .. self.object.name)
    end
end

LookAtDetectorCore.onDestroy = function(self)
    if (self.debugSphere) then
        self.debugSphere:destroy()
        self.debugSphere = nil
    end
end


LookAtDetectorCore.calculateGoal = function(self)
    -- Calculate the minimum result of our dot product for the location and radius
    -- originally generated in the prototype.

    local targetLen	= self.toTarget:length()
    local maxAngle = math.atan2( self.colliderRadius, targetLen )

    -- Minimum dot-product we should need in our transition check
    local dotGoal = math.cos(maxAngle)

    return dotGoal
end

LookAtDetectorCore.getThreshold = function(self)
    return (self and self.lookAtThreshold) or ""
end

LookAtDetectorCore.debugDot = function(self)
    return self.dotDifferential
end
