
clamp = function(x, min, max)
	if x > max then
		return max
	elseif x < min then
		return min
	end
	return x
end

-- returns the angle between two vectors
Vector.angle = function(a, b)
	local aDotB = Vector.dot(a:normalized(), b:normalized());
	local denom = a:length()*b:length(); 
	if (denom > 0.0) then
		local angle = math.acos(clamp(aDotB/denom, -1, 1));
		return angle
	end
	
	return 0
end

Vector.length = function(self) 
	return math.sqrt(self.x*self.x+self.y*self.y+self.z*self.z);	
end

Vector.lengthSq = function(self)
	return self.x*self.x+self.y*self.y+self.z*self.z;	
end

Vector.dot = function(a, b) 
	return a.x * b.x + a.y * b.y + a.z * b.z;
end

-- add new method to quaternion
Quaternion.getEulerAngles = function(self)

	self:normalize()
	local sqw = self.w*self.w
    local sqx = self.x*self.x
    local sqy = self.y*self.y
    local sqz = self.z*self.z
	local unit = sqx + sqy + sqz + sqw -- if normalized is one, otherwise is correction factor
	local test = self.x*self.y + self.z*self.w
	
	local angles = Vector.new(0,0,0)
	
	if (test >= 0.4999*unit) then -- singularity at north pole
		--Log.error("North pole")
		angles.y = math.deg(2.0 * math.atan2(self.x,self.w))
		angles.z = math.deg(math.pi/2.0)
		angles.x = 0.0
	elseif (test <= -0.4999*unit) then -- singularity at south pole
		--Log.error("South pole")
		angles.y = math.deg(-2.0 * math.atan2(self.x,self.w))
		angles.z = math.deg(-math.pi/2.0)
		angles.x = 0.0
	else
	    angles.y = math.deg(math.atan2(2.0*self.y*self.w-2.0*self.x*self.z , sqx - sqy - sqz + sqw))
	    if(unit == 0.0) then
	    	Log.error("Divide by 0!!!!")
	    	angles.z = 0.0
	    else
			angles.z = math.deg(math.asin(2.0*test/unit))
		end
		angles.x = math.deg(math.atan2(2.0*self.x*self.w-2.0*self.y*self.z , -sqx + sqy - sqz + sqw))
	end
	
	
	return angles
	
end

Quaternion.axisAngle = function(axis, angle)
	-- convert to rad
	local rad = math.rad(angle)
	
	local q = Quaternion.new()
	q:setAxisAngle(axis, rad)
	return q
end

Quaternion.fromEulerAngles = function(angles)
	-- convert to rad
	local rad = Vector(0,0,0)
	rad.x = math.rad(angles.x)
	rad.y = math.rad(angles.y)
	rad.z = math.rad(angles.z)
	
	local c1 = math.cos(rad.y/2.0)
 	local s1 = math.sin(rad.y/2.0)
 	local c2 = math.cos(rad.z/2.0)
 	local s2 = math.sin(rad.z/2.0)
 	local c3 = math.cos(rad.x/2.0)
 	local s3 = math.sin(rad.x/2.0)
    
    local c1c2 	= c1*c2
    local s1s2 	= s1*s2
	local w		= c1c2*c3 - s1s2*s3
	local x 	= c1c2*s3 + s1s2*c3
	local y 	= s1*c2*c3 + c1*s2*s3
	local z 	= c1*s2*c3 - s1*c2*s3
	
	return Quaternion.new(x, y, z, w)
end

Quaternion.lerp = function(t, a, b) 
	t = clamp( t, 0.0, 1.0 )
	local a_copy = Quaternion.copy(a)
	if( Quaternion.dot(a_copy, b) < 0 ) then
		a_copy:invert()
		a_copy.w = -a_copy.w
	end
	return Quaternion.lerp_internal(t, a_copy, b)
end

Quaternion.slerp = function(t, a, b) 
	t = clamp( t, 0.0, 1.0 )
	local a_copy = Quaternion.copy(a)
	if( Quaternion.dot(a_copy, b) < 0 ) then
		a_copy:invert()
		a_copy.w = -a_copy.w
	end
	return Quaternion.slerp_internal(t, a_copy, b)
end

Quaternion.lerp_internal = function(t, a, b) 
	local out = a + (b - a):scale(t);
	out:normalize();
	return out;
end

Quaternion.slerp_internal = function(t, a, b) 
	local c = Quaternion.dot(a, b)
	if (c >= 0.9) then
		return Quaternion.lerp_internal(t, a, b)
	end
	
	local s = math.sqrt(math.abs(1.0 - c*c)) -- sine of the angle
	if (s < 0.1) then
		set(a)
		return
	end

	local sign = 1.0
	if ( c < 0.0 ) then
		sign = -1.0
	end	
	
	local angle = math.asin(s)
	local invs = 1.0 / s
	local coef0 = math.sin((1.0 - t) * angle) * invs
	local coef1 = math.sin(t * angle) * invs * sign

	local x = a.x * coef0 + b.x * coef1
	local y = a.y * coef0 + b.y * coef1
	local z = a.z * coef0 + b.z * coef1
	local w = a.w * coef0 + b.w * coef1
	
	return Quaternion.new(x,y,z,w)
end

Quaternion.inverse = function(self)
	local out = Quaternion.new(self.x,self.y,self.z,self.w)
	out:invert()
	return out
end

Quaternion.lookAt = function(lookDir)
	local look = Vector(0, 0, 1, 0);
    local right = Vector(1, 0, 0, 0);
    local up = Vector(0, 1, 0, 0);

    -- flatten vectors to x-z plane
    local toObjectXZ = Vector(lookDir.x, 0, lookDir.z);
    local camLookXZ =  Vector(look.x, 0, look.z);

    local angle = Vector.angle(toObjectXZ, camLookXZ);
    local signY = math.sign(Vector.dot(right, toObjectXZ));
    local lookYRotateAngle = signY * angle;
    local rotationY = Quaternion.axisAngle(up, math.deg(lookYRotateAngle));

    ----------------------------------------------------
	
    -- flatten vectors to z-y plane
    local toObjectZY = Vector(lookDir.x, 0, lookDir.z);

    local sign = math.sign(Vector.dot(up, lookDir));
    local lookXRotateAngle = Vector.angle(toObjectZY, lookDir);
	local rotationX = Quaternion.axisAngle(right, math.deg(-sign*lookXRotateAngle));

    return (rotationY*rotationX);
end

Quaternion.shortestDistance = function(a, b)
	local a_copy = Quaternion.copy(a)
	if( Quaternion.dot(a_copy, b) < 0.0 ) then
		a_copy:invert()
		a_copy.w = -a_copy.w
		return a_copy
	end
	
	return a
end

Quaternion.getRoll = function(self, axisAngles)

	local rotation = self:copy()
	
	local right = Vector.rotateBy(Vector.right, rotation)
	right:normalize()
	local up = Vector.rotateBy(Vector.up, rotation)
	up:normalize()
	
	-- project onto the XY plane
	local xyUp = Vector(Vector.dot(right, Vector.up), Vector.dot(up, Vector.up), 0, 0)
	xyUp:normalize()
	
	-- calculate roll from this isolated case
	local roll = math.deg(Vector.angle(Vector.up, xyUp) or 0)
	
	-- half space test to capture which half of the unit-circle/plane we are in ( quadrant 1 & 2 or quadrant 3 & 4)
	local rollHS = Vector.dot(Vector.right, xyUp*-1)
	
	local finalRoll = math.sign(rollHS)*roll
	
	if(finalRoll < 0) then
		finalRoll = finalRoll + 360
	end
	
	return finalRoll, Quaternion.axisAngle(Vector.forward, finalRoll)
end

Quaternion.getPitch = function(self, axisAngles)

	local rotation = self:copy()
	
	local roll, rollQuat = rotation:getRoll()
	local yaw, yawQuat = rotation:getYaw()
	rotation = rotation*rollQuat:inverse()*yawQuat:inverse()
	
	local up = Vector.rotateBy(Vector.up, rotation)
	up:normalize()
	
	-- project onto the XZ plane
	local yzUp = Vector(0, Vector.dot(up, Vector.up), Vector.dot(up, Vector.forward), 0)
	yzUp:normalize()
	
	-- calculate roll from this isolated case
	local pitch = math.deg(Vector.angle(Vector.up, up) or 0)
	
	-- half space test to capture which half of the unit-circle/plane we are in ( quadrant 1 & 2 or quadrant 3 & 4)
	local pitchHS = Vector.dot(Vector.forward, up*-1)
	
	local final = math.sign(pitchHS)*pitch
	
	if(final < 0) then
		final = final + 360
	end
	
--	Renderer.drawText(10, 100, textStyle, string.format("pitch %.2f", final))
--	Renderer.drawText(10, 130, textStyle, string.format("pitchHS %.2f", pitchHS))
--	Renderer.drawText(10, 160, textStyle, string.format("yzUp %.2f, %.2f, %.2f", up.x, up.y, up.z))
	
	return final, rotation
end

Quaternion.getYaw = function(self, axisAngles)

	local rotation = self:copy()
	
	local roll, rollQuat = self:getRoll()
	
	rotation = rotation*rollQuat:inverse()
	
	local right = Vector.rotateBy(Vector.right, rotation)
	right:normalize()
	
	-- project onto the XZ plane
	local xzRight = Vector(Vector.dot(right, Vector.right), 0, Vector.dot(right, Vector.forward), 0)
	xzRight:normalize()
	
	-- calculate roll from this isolated case
	local yaw = math.deg(Vector.angle(Vector.right, xzRight) or 0)
	
	-- half space test to capture which half of the unit-circle/plane we are in ( quadrant 1 & 2 or quadrant 3 & 4)
	local yawHS = Vector.dot(Vector.forward, xzRight*-1)
	
	local final = math.sign(yawHS)*yaw
	
	if(final < 0) then
		final = final + 360
	end
	
--	Renderer.drawText(10, 100, textStyle, string.format("yaw %.2f", final))
--	Renderer.drawText(10, 130, textStyle, string.format("yawHS %.2f", yawHS))
--	Renderer.drawText(10, 160, textStyle, string.format("xzRight %.2f, %.2f, %.2f", xzRight.x, xzRight.y, xzRight.z))
	
	return final, Quaternion.axisAngle(Vector.up, final)
end

Quaternion.toMatrix = function(self)
	local q = self
    local sqw = q.w*q.w
    local sqx = q.x*q.x
    local sqy = q.y*q.y
    local sqz = q.z*q.z
    
    local mat = Matrix.new()

    -- invs (inverse square length) is only required if quaternion is not already normalised
    local invs = 1 / (sqx + sqy + sqz + sqw)
    mat.m00 = ( sqx - sqy - sqz + sqw)*invs  -- since sqw + sqx + sqy + sqz =1/invs*invs
    mat.m11 = (-sqx + sqy - sqz + sqw)*invs
    mat.m22 = (-sqx - sqy + sqz + sqw)*invs
    
    local tmp1 = q.x*q.y
    local tmp2 = q.z*q.w
    mat.m10 = 2.0 * (tmp1 + tmp2)*invs
    mat.m01 = 2.0 * (tmp1 - tmp2)*invs
    
    tmp1 = q.x*q.z
    tmp2 = q.y*q.w
    mat.m20 = 2.0 * (tmp1 - tmp2)*invs
    mat.m02 = 2.0 * (tmp1 + tmp2)*invs
    tmp1 = q.y*q.z
    tmp2 = q.x*q.w
    mat.m21 = 2.0 * (tmp1 + tmp2)*invs
    mat.m12 = 2.0 * (tmp1 - tmp2)*invs
    
    return mat    
end

Matrix.setColumn = function(self, col, vector)
	self["m".."0"..(col - 1)] = vector.x
	self["m".."1"..(col - 1)] = vector.y
	self["m".."2"..(col - 1)] = vector.z
	self["m".."3"..(col - 1)] = vector.w
end

Matrix.axisX = function(self)
	return Vector(self.m00, self.m10, self.m20)
end

Matrix.axisY = function(self)
	return Vector(self.m01, self.m11, self.m21)
end

Matrix.axisZ = function(self)
	return Vector(self.m02, self.m12, self.m22)
end

Matrix.mulVec = function(self, vector)
	local ret = Vector.new()
	ret.x = vector.x*self.m00 + vector.y*self.m01 + vector.z*self.m02 + vector.w*self.m03
	ret.y = vector.x*self.m10 + vector.y*self.m11 + vector.z*self.m12 + vector.w*self.m13
	ret.z = vector.x*self.m20 + vector.y*self.m21 + vector.z*self.m22 + vector.w*self.m23
	ret.w = vector.x*self.m30 + vector.y*self.m31 + vector.z*self.m32 + vector.w*self.m33
	return ret
end

Matrix.new = function(m00, m01, m02, m03, m10, m11, m12, m13, m20, m21, m22, m23, m30, m31, m32, m33)
	local self = setmetatable({	m00 = m00 or 1.0, m01 = m01 or 0.0, m02 = m02 or 0.0, m03 = m03 or 0.0,
							m10 = m10 or 0.0, m11 = m11 or 1.0, m12 = m12 or 0.0, m13 = m13 or 0.0,
							m20 = m20 or 0.0, m21 = m21 or 0.0, m22 = m22 or 1.0, m23 = m23 or 0.0,
							m30 = m30 or 0.0, m31 = m31 or 0.0, m32 = m32 or 0.0, m33 = m33 or 1.0 }, Matrix)
							
	self[1] = self.m00
	self[2] = self.m10
	self[3] = self.m20
	self[4] = self.m30
	
	self[5] = self.m01
	self[6] = self.m11
	self[7] = self.m21
	self[8] = self.m31
	
	self[9] = self.m02
	self[10] = self.m12
	self[11] = self.m22
	self[12] = self.m32
	
	self[13] = self.m03
	self[14] = self.m13
	self[15] = self.m23
	self[16] = self.m33
	
	return self
end

Matrix.angleAxis = function (_angle, axis)
    -- angles are in degrees. Switch to radians
    local angle = math.rad(_angle)

    angle = angle/2
    local sinA = math.sin(angle)
    local cosA = math.cos(angle)
    local sinA2 = sinA * sinA

    -- normalize
    axis:normalize()
    
    if (axis:lengthSq() <= 0.0) then
        axis = Vector(0,0,0,1)
    end

    local matR = Matrix.new()

    -- optimize case where axis is along major axis
    if (axis.x == 1 and axis.y == 0 and axis.z == 0) then
        matR[5+1] = 1 - 2 * sinA2
        matR[6+1] = 2 * sinA * cosA
        matR[9+1] = -2 * sinA * cosA
        matR[10+1] = 1 - 2 * sinA2
    elseif (axis.x == 0 and axis.y == 1 and axis.z == 0) then
        matR[0+1] = 1 - 2 * sinA2
        matR[2+1] = -2 * sinA * cosA
        matR[8+1] = 2 * sinA * cosA
        matR[10+1] = 1 - 2 * sinA2
    elseif (axis.x == 0 and axis.y == 0 and axis.z == 1) then
        matR[0+1] = 1 - 2 * sinA2
        matR[1+1] = 2 * sinA * cosA
        matR[4+1] = -2 * sinA * cosA
        matR[5+1] = 1 - 2 * sinA2
    else 
        local x = axis.x
        local y = axis.y
        local z = axis.z
        local x2 = x * x
        local y2 = y * y
        local z2 = z * z

        matR[0+1] = 1 - 2 * (y2 + z2) * sinA2
        matR[1+1] = 2 * (x * y * sinA2 + z * sinA * cosA)
        matR[2+1] = 2 * (x * z * sinA2 - y * sinA * cosA)
        matR[4+1] = 2 * (y * x * sinA2 - z * sinA * cosA)
        matR[5+1] = 1 - 2 * (z2 + x2) * sinA2
        matR[6+1] = 2 * (y * z * sinA2 + x * sinA * cosA)
        matR[8+1] = 2 * (z * x * sinA2 + y * sinA * cosA)
        matR[9+1] = 2 * (z * y * sinA2 - x * sinA * cosA)
        matR[10+1] = 1 - 2 * (x2 + y2) * sinA2
    end

    return matR
end

Matrix.__mul = function (a, b)
    local a00 = a[1]
    local a01 = a[2]
    local a02 = a[3]
    local a03 = a[4]
    local a04 = a[5]
    local a05 = a[6]
    local a06 = a[7]
    local a07 = a[8]
    local a08 = a[9]
    local a09 = a[10]
    local a10 = a[11]
    local a11 = a[12]
    local a12 = a[13]
    local a13 = a[14]
    local a14 = a[15]
    local a15 = a[16]

    local b00 = b[1]
    local b01 = b[2]
    local b02 = b[3]
    local b03 = b[4]
    local b04 = b[5]
    local b05 = b[6]
    local b06 = b[7]
    local b07 = b[8]
    local b08 = b[9]
    local b09 = b[10]
    local b10 = b[11]
    local b11 = b[12]
    local b12 = b[13]
    local b13 = b[14]
    local b14 = b[15]
    local b15 = b[16]

    local ret = Matrix.new()
    ret[1]	= a00 * b00 + a01 * b04 + a02 * b08 + a03 * b12
    ret[2]	= a00 * b01 + a01 * b05 + a02 * b09 + a03 * b13
    ret[3] = a00 * b02 + a01 * b06 + a02 * b10 + a03 * b14
    ret[4] = a00 * b03 + a01 * b07 + a02 * b11 + a03 * b15

    ret[5] = a04 * b00 + a05 * b04 + a06 * b08 + a07 * b12
    ret[6] = a04 * b01 + a05 * b05 + a06 * b09 + a07 * b13
    ret[7] = a04 * b02 + a05 * b06 + a06 * b10 + a07 * b14
    ret[8] = a04 * b03 + a05 * b07 + a06 * b11 + a07 * b15

    ret[9] = a08 * b00 + a09 * b04 + a10 * b08 + a11 * b12
    ret[10] = a08 * b01 + a09 * b05 + a10 * b09 + a11 * b13
    ret[11] = a08 * b02 + a09 * b06 + a10 * b10 + a11 * b14
    ret[12] = a08 * b03 + a09 * b07 + a10 * b11 + a11 * b15

    ret[13] = a12 * b00 + a13 * b04 + a14 * b08 + a15 * b12
    ret[14] = a12 * b01 + a13 * b05 + a14 * b09 + a15 * b13
    ret[15] = a12 * b02 + a13 * b06 + a14 * b10 + a15 * b14
    ret[16] = a12 * b03 + a13 * b07 + a14 * b11 + a15 * b15
    
    return ret
end

Matrix.toEuler = function(self)
	local m = self
	
	local heading = nil
	local bank = nil
	local attitude = nil
	
    -- Assuming the angles are in radians.
	if (m.m10 > 0.998) then -- singularity at north pole
		heading = math.atan2(m.m02,m.m22)
		attitude = math.pi/2
		bank = 0
	elseif (m.m10 < -0.998) then -- singularity at south pole
		heading = math.atan2(m.m02,m.m22)
		attitude = -math.pi/2
		bank = 0
	else
		heading = math.atan2(-m.m20,m.m00)
		bank = math.atan2(-m.m12,m.m11)
		attitude = math.asin(m.m10)
	end
	
	return Vector(math.deg(bank), math.deg(heading), math.deg(attitude))
end

Matrix.transpose = function (self)
	local m = self
    return Matrix.new(	m[1], m[5], m[9],  m[13],m[2], m[6], m[10], m[14],m[3], m[7], m[11], m[15],m[4], m[8], m[12], m[16])
end


Matrix.rotate = function (self, angle, axis)
    return Matrix.__mul(self, Matrix.angleAxis(angle, axis))
end


Matrix.rotateX = function (self, angle) 
    return Matrix.__mul(self, Matrix.angleAxis(angle, self:axisX()))
end

Matrix.rotateY = function (self, angle)
    return Matrix.__mul(self, Matrix.angleAxis(angle, self:axisY()))
end

Matrix.rotateZ = function (self, angle)
    return Matrix.__mul(self, Matrix.angleAxis(angle, self:axisZ()))
end


Vector.normalized = function(self) 
	local out = self
	out:normalize()
	return out
end

Vector.setLength = function(self, len) 
	self:normalize()
	self = self * len
end

-- Add getParameters functionality to the materials, as well as some
-- utility functions to make it easier to change commonly used parameters
Material.super = {}
Material.super.setParameters = Material.setParameters

function Material:getParameters()
    self.parameters = self.parameters or {}
    return self.parameters
end

function Material:setParameters(params)
    -- Merge new params with cache of existing ones
    self.parameters = self.parameters or {}
    for key, value in pairs(params) do self.parameters[key] = value end

    Material.super.setParameters(self, params)
end

function Material:setParameter(key, value)
    local params = {}
    params[key] = value
    self:setParameters(params)
end

function Material:setTextureOffset(textureName, u, v)
    local paramName = "u_" .. textureName .. "_ST"
    if (self.parameters[paramName]) then
        local param = self.parameters[paramName]:copy()
        param.z = u or param.z
        param.w = v or param.w
        local params = {}
        params[paramName] = param
        self:setParameters(params)
    end
end

Transform.superVisible = {}
Transform.superVisible.setVisibility = Transform.setVisibility

Transform.setVisibility = function(self, visibility)
	local wasVisible = self:isVisible()
	Transform.superVisible.setVisibility(self, visibility)

	if (self.components == nil) then
		return
	end
	for key, comp in pairs(self.components) do
		if (comp.onVisible ~= nil and not wasVisible) then
			comp:onVisible(visibility)
		end
	end

end

Transform.removeComponent = function(self, name)
    if (self.components) then
        self.components[name] = nil
    end
end

SRT.invert = function(self)
	if(math.abs(self.scale.x) > 0) then
		self.scale.x = 1.0 / self.scale.x
	end
	if(math.abs(self.scale.y) > 0) then
		self.scale.y = 1.0 / self.scale.y
	end
	if(math.abs(self.scale.z) > 0) then
		self.scale.z = 1.0 / self.scale.z
	end

	self.rotation.w = self.rotation.w
	self.rotation.x = -self.rotation.x
	self.rotation.y = -self.rotation.y
	self.rotation.z = -self.rotation.z

	self.position.x = -self.position.x
	self.position.y = -self.position.y
	self.position.z = -self.position.z

	self.position:rotateBy(self.rotation)
end

SRT.inverted = function(self)
	local invertedTransfrom = self:copy()
	if(math.abs(invertedTransfrom.scale.x) > 0) then
		invertedTransfrom.scale.x = 1.0 / invertedTransfrom.scale.x
	end
	if(math.abs(invertedTransfrom.scale.y) > 0) then
		invertedTransfrom.scale.y = 1.0 / invertedTransfrom.scale.y
	end
	if(math.abs(invertedTransfrom.scale.z) > 0) then
		invertedTransfrom.scale.z = 1.0 / invertedTransfrom.scale.z
	end

	invertedTransfrom.rotation.w = invertedTransfrom.rotation.w
	invertedTransfrom.rotation.x = -invertedTransfrom.rotation.x
	invertedTransfrom.rotation.y = -invertedTransfrom.rotation.y
	invertedTransfrom.rotation.z = -invertedTransfrom.rotation.z

	invertedTransfrom.position.x = -invertedTransfrom.position.x
	invertedTransfrom.position.y = -invertedTransfrom.position.y
	invertedTransfrom.position.z = -invertedTransfrom.position.z

	invertedTransfrom.position:rotateBy(invertedTransfrom.rotation)
	
	return invertedTransfrom
end

SRT.__mul = function(a, b)
	local sc = a.scale * b.scale
	local rot = a.rotation * b.rotation
	local pos = (a.position * sc):rotateBy(rot) + b.position
	return SRT.new(pos, rot, scale)
end


-- Make the sound API a little easier to use
--
-- Sound play states -> stopped = 0, paused = 1, playing = 2
Sound.stop = Sound.stop or function(self)
    self:setPlayState(0)
end

Sound.pause = Sound.pause or function(self)
    self:setPlayState(1)
end

Sound.play = Sound.play or function(self)
    self:setPlayState(2)
end

Sound.isPlaying = Sound.isPlaying or function(self)
    return (self:getPlayState() == 2)
end

Sound.getDurationSeconds = Sound.getDurationSeconds or function(self)
    return self:getDuration() / 1000
end

Sound.getSeekPositionSeconds = Sound.getSeekPositionSeconds or function(self)
    return self:getSeekPosition() / 1000
end

-- camera: 			camera object
-- position: 		sound position: Vector3
-- headOcclusion:	how much to dampen volume when audio is behind listener
Sound.getPanVolume3D = function( self, camera, position, headOcclusion )
    local middleVolume = 1
    local extentVolume = 0.3
    local cameraWorld 	= camera:getWorld()
    local cameraForward = cameraWorld:forward()

    local cameraParent = camera.parent or camera

    local targetForward = position - cameraParent:getWorld().position
    targetForward.w 	= 0.0
    targetForward:normalize()

    -- Camera right without roll
    local cameraRight 	= Vector.cross( cameraForward, Vector.up )
    cameraRight.w		= 0.0
    cameraRight:normalize()

    local forwardDot 	= Vector.dot( targetForward, cameraForward )
    local rightDot 		= Vector.dot( targetForward, cameraRight )

    local pan = (1 - math.abs(forwardDot)) * math.sign(rightDot)
    local volume = 1

    local middle = middleVolume
    local extent = extentVolume
    if forwardDot < 0 then
        -- dampen volume if position is behind head
        middle = headOcclusion
    end
    volume = volume * math.lerp(middle, extent, math.abs(pan))

    local compressedPan = math.clamp(pan * 1.42, -1, 1)

    return compressedPan, volume
end

Color.lerp = function(t, a, b)
	local out = a + (b - a):scale(t)
	out:normalize()
	return out
end

Color.normalize = function(self)
	local len = self:length()
	if (len > 0.001) then
		local inv_len = 1.0/len
		self.r = self.r * inv_len
		self.g = self.g * inv_len
		self.b = self.b * inv_len
		self.a = self.a * inv_len
	end
	return len
end

Color.length = function(self) 
	return math.sqrt(self.r*self.r+self.g*self.g+self.b*self.b+self.a*self.a)
end


Color.scale = function(self, s)
	if (type(s) == "table") then
		self.r = self.r * s.r
		self.g = self.g * s.g
		self.b = self.b * s.b
		self.a = self.a * s.a
	elseif (type(s) == "number") then
		self.r = self.r * s
		self.g = self.g * s
		self.b = self.b * s
		self.a = self.a * s
	end
	
	return self
end
