local Logger = TaggedLog("AsyncLoader", true)

AsyncLoader = {}

local segmentGroups = {
	{
		"Intro",
        "IntroB",
    },
    {
		"RockA",
    },
    {        
		"RockB",
		"RockC",
		"RockD",
    },
    {
        "LogA",
    },
    {        
        "LogB",
        "LogC",
        "LogD",
    },
    {
        "TwigA",
    },
    {        
        "TwigB",
        "TwigC",
        "TwigD",
    },
    {
        "SquirrelA",
    },
    {        
        "SquirrelB",
        "SquirrelC",
        "SquirrelD",
        "SquirrelE",
    },
    {
        "FrogA",
    },
    {        
        "FrogB",
        "FrogC",
        "FrogD",
    },
    {
        "EndA",
    },
    {        
        "EndB",
        "EndC",
        "ThatsAllFolks",
    },
}

-- create a name to index mapping for segments
local segmentmapping = {}
for i, seg in ipairs(SceneData.segments) do
    segmentmapping[seg.name] = i
end

AsyncLoader.prefetchActors = function(actorList, group)
    if actorList then
        for _, actor in ipairs(actorList) do
            local actordef = SceneData.actors[actor]
            if actordef then
                Logger:message("fetching actor: " .. tostring(actor))
                group:addActor(actordef)
            end
        end
    end
end

AsyncLoader.prefetchAudio = function(segmentName, group)
    local segmentData = SceneData.audio[segmentName]
    if segmentData then
        for _, sound in ipairs(segmentData.sounds) do
            SoundManager:prefetch(sound.name, group)
        end
    end
end

-- prefetch a segment - prefetches all the resources
-- for all the actors, sounds, materials, etc in the segment
AsyncLoader.prefetchSegment = function(segment, group)
    if segment then
        Logger:message("prefetching segment " .. tostring(segment.name) .. " during " .. SegmentManager:currentSegmentName())
        AsyncLoader.prefetchActors(table.flatten(segment.active), group)
        AsyncLoader.prefetchActors(table.flatten(segment.active_postRot), group)
        AsyncLoader.prefetchActors(table.flatten(segment.prefetch), group)
        AsyncLoader.prefetchAudio(segment.name, group)
        if segment.particleSystems then
            for _, uri in ipairs(segment.particleSystems) do
                group:addParticleSystem(uri)
            end
        end

		group:prefetch()

        if Global.storyManager then
            Global.storyManager:prefetchState(segment.name)

            if segment.transitions then
                for _, transition in pairs(segment.transitions) do
                    Global.storyManager:prefetchState(transition)
                end
            end
        end
    end
end

AsyncLoader.cleanup = function(self)
    System.gc()
end

local prefetchIndex = 1
local prefetchCounter = 1

local resourceGroups = {}

AsyncLoader.prefetch = function()
    prefetchCounter = prefetchCounter - 1
    if(prefetchCounter == 0) then
        local batch = segmentGroups[prefetchIndex]
        if (batch) then
            prefetchCounter = table.getn(batch)
            for i, segname in ipairs(batch) do
                local segindex = segmentmapping[segname]

				-- track resources in resource group
				local group = System.createResourceGroup( segname )
				resourceGroups[segname] = group
				Log.warning( "Prefetching into resource group: "..segname )

                AsyncLoader.prefetchSegment(SceneData.segments[segindex], group)
            end
        end
        prefetchIndex = prefetchIndex + 1
    end
end

AsyncLoader.prefetchForSegment = function(segmentName)
    -- Find the segment in the batches and update the indices
    for batchIndex, batch in ipairs(segmentGroups) do
        local counter = #batch
        for _, seg in ipairs(batch) do
            if (seg == segmentName) then
                -- Prefetch the block for the given segment
                prefetchIndex = batchIndex
                prefetchCounter = 1
                AsyncLoader.prefetch()
                prefetchCounter = counter
                return
            end
            counter = counter - 1
        end
    end
end

-- Call this to move the prefetch index backward (<0) or forward (>0)
-- Should only be used if manually changing the current segment.
AsyncLoader.jump = function(offset)
    prefetchIndex = prefetchIndex + offset
end

function AsyncLoader.freeResourceGroup( name )
	local group = resourceGroups[name]
	if group then
		Log.warning( "Freeing resource group: "..name )
		group:destroy()
		resourceGroups[name] = nil
	end
end

function AsyncLoader.shutdown()
	-- free any remaining resource groups
	for name, group in pairs(resourceGroups) do
		Log.warning( "Freeing resource group: "..name )
		group:destroy()
	end
end

function AsyncLoader.generateStreamingData()
    local states = {}
    local resources = {}

    local scene = Scene.create()
    local soundEmitter = AudioEmitter.create()

    local function resourceDuration(uri)
        local duration
        if string.endsWith(uri, ".mfb") then
            local fbSampler = FlipbookSampler.create()
            fbSampler:setFlipbook(uri)
            duration = fbSampler:getNumFrames() / 60
            fbSampler:destroy()
        elseif string.endsWith(uri, ".ogg") then
            local sound = soundEmitter:addSound(uri, uri, false, 0, -1)
            duration = sound:getDurationSeconds()
            soundEmitter:removeSound(uri)
        end
        return duration
    end

    local function animationDuration(actorName, animName)
        local actorData = def.actors[actorName]
        if actorData and actorData.animation and actorData.animation[animName] then
            local anim = actorData.animation[animName]
            local startFrame = anim[2]
            local endFrame = anim[3]
            if endFrame == -1 then
                -- load the actor so we can get the animation's real end frame
                local actor = scene:createActor(actorName, actorData)
                if actor and actor.animation and actor.animation[animName] then
                    endFrame = actor.animation[animName]:getStopFrame()
                    actor:destroy()
                    Util.fullGC()
                end
            end
            return (endFrame - startFrame) / 60
        end
    end

    local function minAnimationDuration(actorNames, segmentName)
        local minDuration
        for _, actorName in ipairs(actorNames) do
            local animDuration = animationDuration(actorName, segmentName)
            if minDuration then
                minDuration = math.min(minDuration, animDuration)
            else
                minDuration = animDuration
            end
        end
        return minDuration
    end

    local function convertSegmentResources(resourceGroup)
        local segmentResources = {}
        if resourceGroup.uris then
            for uri, _ in pairs(resourceGroup.uris) do
                local path = uri:gsub(":", "/")
                if not resources[path] then
                    resources[path] = { path = path, duration = resourceDuration(uri) }
                end
                table.insert(segmentResources, path)
            end
        end
        return segmentResources
    end

    local function resourcesForSegment(segmentName)
        AsyncLoader.prefetchForSegment(segmentName)
        local group = resourceGroups[segmentName]
        if group then
            return convertSegmentResources(group)
        end
        return {}
    end

--    AsyncLoader.prefetchIntroMusic()

    for order, segmentData in ipairs(SceneData.segments) do
        local segmentName = segmentData.name
        if not states[segmentName] then
            local state = {}
            states[segmentName] = state

            state.order = order
            state.resources = resourcesForSegment(segmentName)

            state.transitions = {}
            local minSegmentDuration
            if segmentData.nextSegments then
                for _, transition in ipairs(segmentData.nextSegments) do
                    local minTransitionDuration = minAnimationDuration(segmentData.actors, segmentName)
                    state.transitions.default = {
                        toSegment = transition,
                        minDuration = minTransitionDuration,
                    }
                    if minSegmentDuration then
                        minSegmentDuration = math.min(minSegmentDuration, minTransitionDuration)
                    else
                        minSegmentDuration = minTransitionDuration
                    end
                end
            end
            state.minDuration = minSegmentDuration
        end
    end

    -- Compute the base resources needed to start the show
    local baseResourceGroup = Global.resources

    local rootState = {}
    rootState.order = 0
    rootState.resources = convertSegmentResources(baseResourceGroup)
    states["BaseResources"] = rootState

    local filename = "buggynight.xml"
    local file, error = io.open(filename, "w")

    file:write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
    file:write("<StoryManifest xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:noNamespaceSchemaLocation=\"ResourceManifestSchema.xsd\">\n")

    file:write("\t<Chunks>\n")
    for _, state in pairs(states) do
        local url = string.format("http://dl.google.com/spotlight/buggynight/1/buggynight_%03d.zip", state.order)
        file:write("\t\t<Chunk ID=\"" .. tostring(state.order) .. "\" URL=\"" .. url .. "\">\n")
        if state.resources then
            for _, name in ipairs(state.resources) do
                local resource = resources[name]
                if resource then
                    local duration = ""
                    if resource.duration then
                        duration = " duration=\"" .. string.format("%.2f", resource.duration) .. "\""
                    end
                    file:write("\t\t\t<Resource Name=\"" .. resource.path ..  "\"" .. duration .. "/>\n")
                end
            end
        end
        file:write("\t\t</Chunk>\n")
    end
    file:write("\t</Chunks>\n")

    file:write("\t<StateMachines>\n")
    file:write("\t\t<StateMachine Name=\"buggynight\" Root=\"BaseResources\">\n")

    file:write("\t\t\t<States>\n")
    for name, state in pairs(states) do
        local minDuration = ""
        if state.minDuration then
            minDuration = string.format(" MinDuration=\"%.2f\"", state.minDuration)
        end
        file:write("\t\t\t\t<State Name=\"" .. name .. "\" Order=\"" .. tostring(state.order) .. "\"" .. minDuration .. ">\n")
        file:write("\t\t\t\t\t<Chunks>\n")
        file:write("\t\t\t\t\t\t<ChunkID>0</ChunkID>\n")
        file:write("\t\t\t\t\t\t<ChunkID>" .. tostring(state.order) .. "</ChunkID>\n")
        file:write("\t\t\t\t\t</Chunks>\n")
        if state.transitions then
            file:write("\t\t\t\t\t<Transitions>\n")
            for targetName, transition in pairs(state.transitions) do
                file:write("\t\t\t\t\t\t<Transition To=\"" .. transition.toSegment .. "\" MinDuration=\"" .. string.format("%.2f", transition.minDuration or 0) .. "\" Target=\"" .. targetName .. "\"/>\n")
            end
            file:write("\t\t\t\t\t</Transitions>\n")
        end
        file:write("\t\t\t\t</State>\n")
    end
    file:write("\t\t\t</States>\n")

    file:write("\t\t</StateMachine>\n")
    file:write("\t</StateMachines>\n")

    file:write("</StoryManifest>\n")
    file:close()

    scene:destroy()
    soundEmitter:destroy()

end
