local Logger = TaggedLog.new("Segment", true)

local activatedObjs = {}

local ExploreRotationThreshold  = 250

Segment = {
    hasBegun = false,
    playing = false,

    delayTimer      = 0,
    breakoutTimer   = 0,
    motionMin       = 0,
    exploreRotation = 0,
}

function Segment:new(data)
    local o = data or {}

    o.animated = table.flatten(o.animated)
    o.active = table.flatten(o.active)


    Logger:message(o.name .. " active = " .. table.tostring(o.active))
    Logger:message(o.name .. " animated = " .. table.tostring(o.animated))

    if (o.active_postRot) then
        o.active_postRot = table.flatten(o.active_postRot)
        Logger:message(o.name .. " active_postRot = " .. table.tostring(o.active_postRot))
    end
    if (o.counterRot) then
        o.active_postRot = table.flatten(o.active_postRot)
        Logger:message(o.name .. " active_postRot = " .. table.tostring(o.active_postRot))
    end


    setmetatable(o, self)
    self.__index = self

    return o
end


function Segment:update()
    if (self.animated[1] == nil) then
        return
    end

    local lookedAt = self:lookingAtTrigger()

    if (self.playing) then
        local dt = Time.dt()

        local currentFinished = false

        if (self.breakoutSec) then
            if (lookedAt) then
                self.breakoutTimer = self.breakoutTimer - dt
            end
            if (self.breakoutTimer <= 0) then
                currentFinished = true
            end

        else
            if ( self.animated and #self.animated > 0) then
                local actor = ActorManager:getActor(self.animated[1])
                currentFinished = actor.animDriver:isFinished(true)
            end
        end

        -- If the following conditions are met then we're ready for the next
        self.exploreRotation = self.exploreRotation + SegmentManager.deltaYaw
        if (currentFinished) then
            self.delayTimer = self.delayTimer - dt
            if (self.delay ~= nil) then -- this is a segment where a hunt for the bugs take place
                AnalyticsManager.bugHuntBegan()
            end

            local exploredLongEnough = not self.delay or (self.delayTimer <= 0 and (math.abs(SegmentManager.deltaYaw) / dt >= self.motionMin))
            local exploredFarEnough = math.abs(self.exploreRotation) >= ExploreRotationThreshold
            if ((not self.needsFocus or lookedAt) and (exploredLongEnough or exploredFarEnough) )then
                Logger:message(self.name .. ": currentFinished")
                self:prepareNextSegment(lookedAt)
            end
        end
    end
end


function Segment:startPlaying(startingCheckpoint)
    Logger:message("Starting to play: " .. self.name)

    self:activateObjects(self.active, true)

    if (self.active_postRot and #self.active_postRot > 0) then
       self:changeWorldRot()
    end

    self.playing = true

    self:initializeAnimDrivers()
    self:playAnimations()

    self.hasBegun = true

    self.delayTimer      = self.delay or 0
    self.breakoutTimer   = self.breakoutSec or 0
    self.exploreRotation = 0
end

function Segment:stopPlaying()
    self.playing = false
end


function Segment:initializeAnimDrivers()
    for i, actorName in ipairs(self.animated) do
        if (actorName) then
            local actor = ActorManager:getActor(actorName, false)
            if (actor and actor.animDriver) then
                local anim = actor.animDriver:init(self.name, actor.animation, i == 1)
            end
        end
    end
end

function Segment:playAnimations()
    for _, actorName in ipairs(self.animated) do
        if (actorName) then
            local actor = ActorManager:getActor(actorName, false)
            if (actor and actor.animDriver) then
                Logger:message("Playing " .. self.name .. " on " .. actorName)
                actor.animDriver:play()
            end
        end
    end
end


function Segment:lookAtMainActor()
    if (self.animated[1] and Global.scene.mainCamera and Global.scene.mainCamera.controller) then
        local actor = ActorManager:getActor(self.animated[1])
        if (actor) then
            -- TODO: this might need to do a lookup on bone, rather than the actor root
            Global.scene.mainCamera.controller:lookAt(actor:getWorld())
        end
    end
end


function Segment:lookingAtTrigger()
    if (self.transitions == nil) then
        return nil
    end

    for target,_ in pairs(self.transitions) do
        if (target) then
            local detector = ActorManager:getActor(target, false)
            if (detector and detector.components) then
                detector = detector.components["LookAtDetector"]
                if (detector and detector:lookedAt()) then
                    return target
                end
            end
        end
    end

    return nil
end


function Segment:prepareNextSegment(target)
    if (target == nil) then
        target = "default"
    end
    
    if (self.checkpoint) then  -- I am using "checkPoint" as an indicator of this being a 'A' segment so we can track the using finding the bugs
        AnalyticsManager.bugHuntEnd()
    end

    if (self.transitions[target]) then
        SegmentManager:startSegment(self.transitions[target])
    else
        self:stopPlaying()
        Events.send(self, "FinalSegment")
        Buggynight:showEnded()
    end
end

function Segment:currentTime()
    if (self.animated and #self.animated > 0) then
        local actor = self.animated[1]
        if (actor and actor.animDriver) then
            return actor.animDriver:currentTime()
        end
    end
    return 0
end

function Segment:activateObjects(names, removeExceptions)
    if (names == nil) then
        return
    end

    local remove = shallowcopy(activatedObjs)
    local activate = {}

    -- Find the objects that need to be activated (and not removed)
    for _, name in ipairs(names) do
        if (name) then
            activate[name] = true
            remove[name] = nil
        end
    end

    -- Activate any new objects
    if (activate) then
        for name, value in pairs(activate) do
            local actor = ActorManager:getActor(name, true)
            if (actor) then
                actor:setVisibility(true)
            end
            if (not activatedObjs[name]) then
                activatedObjs[name] = actor
            end
            Logger:message("Activating: " .. name)
        end
    end

    -- Apply global rotation to active objects
    for _, name in ipairs(names) do
        if (name) then
            local actor = ActorManager:getActor(name, false)
            if (actor) then
                local yRot = Global.worldRot

                -- This is a total hack to compensate for the fact the we authored the animations with rotation built into them
                if (self.counterRot and self.counterRot) then
                    for _, exception in ipairs(self.counterRot) do
                        if (exception == name) then
                            yRot = yRot - Global.worldRotCompensation
                            break
                        end
                    end
                end

                --if (Global.worldRot) then
                --    Logger:message("Global rotation " .. name .. " : yRot " .. yRot.. " : globalYRot "..Global.worldRot.." : "..Global.scene.mainCamera:getWorld().rotation:getYaw() )
                --end

                actor:setRotation( Quaternion.axisAngle( Vector.up, yRot ) )
            end
        end
    end

    -- Remove any unneeded objects
    if (removeExceptions and remove) then
        for name, _ in pairs(remove) do
            ActorManager:destroyActor(name)
            activatedObjs[name] = nil
            Logger:message("Removing: " .. name)
        end
    end
end

function Segment:changeWorldRot()
    local wedgeSize     = 30
    local initialOffset = 90

    local currentYaw    = SegmentManager.currentYaw
    local prevYaw       = SegmentManager.prevYaw

    local preApplyYRot  = Global.worldRot
    Global.worldRotCompensation = self.yCompensation

    local numSlices     = math.floor(currentYaw / wedgeSize)
    -- Attempt to bias the spawn positions (left or right of cam) by user motion
    -- (this could definitely use some tuning and refinement)
    local isRecording = (Platform.record~="" or Platform.replay~="")
    if (SegmentManager.deltaYaw < 0 and not isRecording) then
        initialOffset = initialOffset * -1
        numSlices = math.ceil(currentYaw / wedgeSize)
    end

    local validRotation = false

    while (not validRotation) do
        local newWorldYRot = (numSlices * wedgeSize + initialOffset) % 360

        Logger:message(" == Change World Rot == : "..preApplyYRot.." => "..newWorldYRot)

        -- Remove the two slices that will overlap the homeslice
        local hiddenSliceNames = {}
        -- we need to calculate the lowest number slice first so this is actually: ((slice_index - 1) + 12) %12
        local sliceIndex = ((newWorldYRot / wedgeSize) - (preApplyYRot / wedgeSize) + 11) % 12

        hiddenSliceNames[1] = SceneData[self.sliceTable][sliceIndex+1]
        sliceIndex = (sliceIndex+1) % 12
        hiddenSliceNames[2] = SceneData[self.sliceTable][sliceIndex+1]

--        validRotation = true
--        for _, namesList in ipairs(hiddenSliceNames) do
--           if (namesList == nil) then
--               validRotation = false
--               break
--           end
--        end

--        if (validRotation == true) then
        if (hiddenSliceNames[1] ~= nil and hiddenSliceNames[2] ~= nil) then
            validRotation = true

            Global.worldRot = newWorldYRot

            -- Go through our two lists of actors-per-slice and remove those actors
            for _, namesList in ipairs(hiddenSliceNames) do
                for _, name in ipairs(namesList) do
                    --Logger:message(" == Removing Actor == "..name)

                    local actor = ActorManager:getActor(name, false)
                    if (actor) then
                        actor:setVisibility(false)
                        ActorManager:destroyActor(name)
                    end
                end
            end

        else
            -- This positioning overlaps with the action slice, look for a better fit
            numSlices = (numSlices + (1*math.sign(initialOffset)) + 12) % 12
        end
    end

    self:activateObjects(self.active_postRot, false)
end
