local Logger = TaggedLog.new("SegmentManager", false)

-- Used to decide whether to reset to beginning of current segment
-- or to jump to the one that came before when using "Previous"
local goToPrevThreshold = 1.5

SegmentManager = Manager:new {
    name = "SegmentManager",

    currentYaw  = 0,
    prevYaw     = 0,
    deltaYaw    = 0,
    paused      = false,
    waitForStreamingState  = nil,
}

function SegmentManager:onStart()
    self.current = nil
    self.segments = {}
    self.history = {}
    self.starterSegment = nil

    local firstSegmentName
    -- Initialize segments from scene data
    if (SceneData.segments) then
        for _, seg in ipairs(SceneData.segments) do
            local segment = Segment:new(seg)
            self.segments[segment.name] = segment
            firstSegmentName = firstSegmentName or segment.name
        end
    end

    -- Listen for control events
    Events.listen(self, "nextAnimation")
    Events.listen(self, "previousAnimation")

    self.starterSegment = firstSegmentName
    if (Context.getCheckpoint()) then
        self.starterSegment = Context.getCheckpoint()
    end

    if (self.starterSegment) then
        self:startSegment(self.starterSegment)
    end

    Toolbox:listen("TransportControl", {
        onTransportNext = function() self:next() end,
        onTransportPrev = function() self:previous() end,
        onTransportPause = function() self:pause() end,
        onTransportResume = function() self:resume() end,
    })
end

function SegmentManager:onUpdate()
    -- Used for determining user "intention" by camera direction
    self.prevYaw = self.currentYaw
    self.currentYaw = Global.scene.mainCamera:getWorld().rotation:getYaw()

    local prevRot       = Quaternion.axisAngle(Vector(0.0, 1.0, 0.0, 0.0), self.prevYaw)
    local currentRot    = Quaternion.axisAngle(Vector(0.0, 1.0, 0.0, 0.0), self.currentYaw)
    local deltaRot      = prevRot:inverse() * currentRot

    self.deltaYaw = deltaRot:getYaw()
    if (self.deltaYaw > 180) then
        self.deltaYaw = -360 + self.deltaYaw
    end

    if (Input.isKeyDown(KB_N, true)) then
        Logger:message("going next")
        self:next()
    elseif (Input.isKeyDown(KB_P, true)) then
        Logger:message("going previous")
        self:previous()
    elseif (Input.isKeyDown(KB_SPACE, true)) then
        Logger:message("toggling pause")
        self:togglePause()
    end

    if self.waitForStreamingState then
        self:goToSegment(self.waitForStreamingState)
    end

    if (self.current) then
        self.current:update()
    end
	
	if(System.isStatManagerEnabled()) then
		local segTime = Stat.segmentTime:get() + Time.dt();
		Stat.segmentTime:set(segTime);
	end
end

function SegmentManager:next()
    if (self.current == nil or self.current.transitions == nil) then return end

    local firstTransitionKey, firstTransitionName = next(self.current.transitions)
    if (firstTransitionName) then
        Logger:message("next transition to: " .. firstTransitionName)
        self:startSegment(firstTransitionName)

        AsyncLoader.prefetch()
    end
end


function SegmentManager:previous()
    if (self.current == nil) then return end

    local goToPrevious = (self.current:currentTime() < goToPrevThreshold)
    local numHistory = #self.history

    -- If we are below the threshold and we have previous segment history then go to it
    if (numHistory > 1 and goToPrevious) then
        local prevSegment = self.history[numHistory - 1]

        table.remove(self.history)
        numHistory = numHistory - 1
        self.current = prevSegment
        if (System.isStatManagerEnabled()) then
            Stat.segment:set(self:currentSegmentName())
        end
    end

    -- Play normally
    self.current:startPlaying()
    --self:lookAtCurrent()

    self:notifySegmentPlaying()

    -- jump the async loader back
    AsyncLoader.jump(-1)
end

function SegmentManager:togglePause()
    if self.paused then
        self:resume()
    else
        self:pause()
    end
end

function SegmentManager:pause()
    Time.pause()
    self.paused = true
end

function SegmentManager:resume()
    Time.resume()
    self.paused = false
end

function SegmentManager:startSegment(name)
    local lastName = self.current and self.current.name or nil

    if Global.storyManager then
        local resourcesLoaded = Global.storyManager:setState(name)
        if not resourcesLoaded then
            Logger:message("waiting for segment resources: " .. name)
            self.waitForStreamingState = name
            return
        else
            self.waitForStreamingState = nil
        end
    end

    Logger:message("startSegment: " .. name)
    local segment = self.segments[name]
    if (segment == nil) then
        Log.error("startSegment fail, segment not found: " .. name)
        return
    end

    self.current = segment
    table.insert(self.history, self.current)

    local isStartingCheckpoint = (Context and Context.getCheckpoint() == self.current.name)
    self.current:startPlaying(isStartingCheckpoint)
    self:notifySegmentPlaying()

    if (System.isStatManagerEnabled()) then
        Stat.segment:set(name)
        Stat.segmentTime:set(0.0)
    end

    AsyncLoader.prefetch()

    if lastName then
        AsyncLoader.freeResourceGroup( lastName )
    end
end

function SegmentManager:lookAtCurrent()
    if (self.current) then
        self.current:lookAtMainActor()
    end
end

function SegmentManager:currentSegmentName()
    return (self.current and self.current.name) or "EMPTY_SEGMENT"
end

function SegmentManager:notifySegmentPlaying()
    if (self.current and self.current.checkpoint) then
        Context.onCheckpoint(self.current.name)
    end

    Logger:message("Segment playing: " .. self:currentSegmentName())
    Events.send(self, "segmentPlaying", self:currentSegmentName())
end
