
--cache lua functions.
local sin   = math.sin;
local cos   = math.cos;
local asin  = math.asin;
local acos  = math.acos;
local atan2 = math.atan2;
local sqrt  = math.sqrt;
local abs   = math.abs;
local rad   = math.rad;
local deg   = math.deg;
local sign  = math.sign;

FastQuat = {}
FastQuat.__index = FastQuat

FastQuat.new = function(x, y, z, w)
	return setmetatable({ x or 0.0, y or 0.0, z or 0.0, w or 1.0 }, FastQuat);
end

--temporary memory which can be used by functions to avoid allocating memory.
local _tmpVec   = Vector.new();
local _tmpFwd   = Vector.new();
local _tmpRight = Vector.new();
local _tmpUp    = Vector.new();
local _tmpArr   = { 0.0, 0.0, 0.0, 0.0 };
local _tmpQuat  = FastQuat.new();


FastQuat.copy = function(self)
	return FastQuat.new(self[1], self[2], self[3], self[4]);
end

FastQuat.copyInline = function(self, ob) 
	self[1] = ob[1];
	self[2] = ob[2];
	self[3] = ob[3];
	self[4] = ob[4];
end

FastQuat.length = function(self)
	return sqrt( self[0]*self[0] + self[1]*self[1] + self[2]*self[2] + self[3]*self[3] )
end

FastQuat.lengthSq = function(self)
	return self[0]*self[0] + self[1]*self[1] + self[2]*self[2] + self[3]*self[3]
end

FastQuat.normalize = function(self)
	local len = sqrt( self[0]*self[0] + self[1]*self[1] + self[2]*self[2] + self[3]*self[3] );
	if len > 0.0 then
		local inv_len = 1.0/len;
		self[0] = self[0] * inv_len;
		self[1] = self[1] * inv_len;
		self[2] = self[2] * inv_len;
		self[3] = self[3] * inv_len;
	end
	return len;
end

FastQuat.set = function(x, y, z, w)
	self[1] = x
	self[2] = y
	self[3] = z
	self[4] = w
end

FastQuat.unpack = function()
	return self[1], self[2], self[3], self[4];
end

FastQuat.getX = function()
	return self[1];
end

FastQuat.getY = function()
	return self[2];
end

FastQuat.getZ = function()
	return self[3];
end

FastQuat.getW = function()
	return self[4];
end

--q2 = q0*q1
FastQuat.mul = function(self, a, b)
	self[1] = a[4]*b[1] + a[1]*b[4] + a[2]*b[3] - a[3]*b[2];
	self[2] = a[4]*b[2] - a[1]*b[3] + a[2]*b[4] + a[3]*b[1];
	self[3] = a[4]*b[3] + a[1]*b[2] - a[2]*b[1] + a[3]*b[4];	
	self[4] = a[4]*b[4] - a[1]*b[1] - a[2]*b[2] - a[3]*b[3];
end

--q2 = q0+q1
FastQuat.add = function(self, a, b)
	self[1] = a[1] + b[1];
	self[2] = a[2] + b[2];
	self[3] = a[3] + b[3];
	self[4] = a[4] + b[4];
end

--q2 = q0-q1
FastQuat.sub = function(self, a, b)
	self[1] = a[1] - b[1];
	self[2] = a[2] - b[2];
	self[3] = a[3] - b[3];
	self[4] = a[4] - b[4];
end

--q1 *= q0
FastQuat.mod = function(self, mod)
	--unfortunate local variable, but no meta data is required - only the actual array values.
	_tmpArr[1] = self[1]; 
	_tmpArr[2] = self[2];
	_tmpArr[3] = self[3]; 
	_tmpArr[4] = self[4];

	self[1] = _tmpArr[4]*mod[1] + _tmpArr[1]*mod[4] + _tmpArr[2]*mod[3] - _tmpArr[3]*mod[2];
	self[2] = _tmpArr[4]*mod[2] - _tmpArr[1]*mod[3] + _tmpArr[2]*mod[4] + _tmpArr[3]*mod[1];
	self[3] = _tmpArr[4]*mod[3] + _tmpArr[1]*mod[2] - _tmpArr[2]*mod[1] + _tmpArr[3]*mod[4];	
	self[4] = _tmpArr[4]*mod[4] - _tmpArr[1]*mod[1] - _tmpArr[2]*mod[2] - _tmpArr[3]*mod[3];
end

--q1 += q0
FastQuat.inc = function(self, inc)
	self[1] = self[1] + inc[1];
	self[2] = self[2] + inc[2];
	self[3] = self[3] + inc[3];
	self[4] = self[4] + inc[4];
end

--q1 -= q0
FastQuat.dec = function(self, dec)
	self[1] = self[1] - dec[1];
	self[2] = self[2] - dec[2];
	self[3] = self[3] - dec[3];
	self[4] = self[4] - dec[4];
end

--q2 = q2 + q0*q1
FastQuat.madd = function(self, a, b)
	self[1] = self[1] + a[4] * b[1] + a[1] * b[4] + a[2] * b[3] - a[3] * b[2];
	self[2] = self[2] + a[4] * b[2] - a[1] * b[3] + a[2] * b[4] + a[3] * b[1];
	self[3] = self[3] + a[4] * b[3] + a[1] * b[2] - a[2] * b[1] + a[3] * b[4];	
	self[4] = self[4] + a[4] * b[4] - a[1] * b[1] - a[2] * b[2] - a[3] * b[3];
end

FastQuat.scale = function(x, y, z, w)
	self[1] = self[1] * x;
	self[2] = self[2] * y;
	self[3] = self[3] * z;
	self[4] = self[4] * w;
end

FastQuat.scalarMul = function(a, s)
	return FastQuat.new(a[1]*s, a[2]*s, a[3]*s, a[4]*s);
end

FastQuat.scalarMulInline = function(self,s)
	self[1] = self[1]*s;
	self[2] = self[2]*s;
	self[3] = self[3]*s;
	self[4] = self[4]*s;
end

FastQuat.dot = function(a, b) 
	return a[1]*b[1] + a[2]*b[2] + a[3]*b[3] + a[4]*b[4];
end

FastQuat.fromEulerAngles = function(angles)
	-- convert to rad
	_tmpVec.x = rad(angles.x);
	_tmpVec.y = rad(angles.y);
	_tmpVec.z = rad(angles.z);
	
	local c1 = cos(_tmpVec.y/2.0);
 	local s1 = sin(_tmpVec.y/2.0);
 	local c2 = cos(_tmpVec.z/2.0);
 	local s2 = sin(_tmpVec.z/2.0);
 	local c3 = cos(_tmpVec.x/2.0);
 	local s3 = sin(_tmpVec.x/2.0);
    
    local c1c2 	= c1*c2;
    local s1s2 	= s1*s2;
	local w		= c1c2*c3 - s1s2*s3;
	local x 	= c1c2*s3 + s1s2*c3;
	local y 	= s1*c2*c3 + c1*s2*s3;
	local z 	= c1*s2*c3 - s1*c2*s3;
	
	return FastQuat.new(x, y, z, w);
end

FastQuat.fromEulerAnglesInline = function(self, angles)
	-- convert to rad
	_tmpVec.x = rad(angles.x);
	_tmpVec.y = rad(angles.y);
	_tmpVec.z = rad(angles.z);
	
	local c1 = cos(_tmpVec.y/2.0);
 	local s1 = sin(_tmpVec.y/2.0);
 	local c2 = cos(_tmpVec.z/2.0);
 	local s2 = sin(_tmpVec.z/2.0);
 	local c3 = cos(_tmpVec.x/2.0);
 	local s3 = sin(_tmpVec.x/2.0);
    
    local c1c2 = c1*c2;
    local s1s2 = s1*s2;
	self[4]	   = c1c2*c3 - s1s2*s3;
	self[1]    = c1c2*s3 + s1s2*c3;
	self[2]    = s1*c2*c3 + c1*s2*s3;
	self[3]    = c1*s2*c3 - s1*c2*s3;
end

-- create a quaternion from an axis angle, the angle is in radians.
FastQuat.setAxisAngle = function(self, axis, angle)
	local theta 		= angle*0.5;
	local sinTheta     	= sin(theta);				
	self[1]				= axis.x*sinTheta;
	self[2]				= axis.y*sinTheta;
	self[3]				= axis.z*sinTheta;
	self[4]				= cos(theta);

	--inline normalize.
	local len = sqrt( self[1]*self[1] + self[2]*self[2] + self[3]*self[3] + self[4]*self[4] );
	if len > 0.0 then
		local inv_len = 1.0/len;
		self[1] = self[1] * inv_len;
		self[2] = self[2] * inv_len;
		self[3] = self[3] * inv_len;
		self[4] = self[4] * inv_len;
	end
end

FastQuat.setEuler = function(self, x, y, z) 
	local c1 = cos(x*0.5);
	local c2 = cos(y*0.5);
	local c3 = cos(z*0.5);

	local s1 = sin(x*0.5);
	local s2 = sin(y*0.5);
	local s3 = sin(z*0.5);

	self[1] = s1 * c2 * c3 - c1 * s2 * s3;
	self[2] = c1 * s2 * c3 + s1 * c2 * s3;
	self[3] = c1 * c2 * s3 - s1 * s2 * c3;
	self[4] = c1 * c2 * c3 + s1 * s2 * s3;
end

FastQuat.getAxisAngle = function(self, axis)
	local angle = 2.0*acos(self[4]);
	local s = sqrt(1.0 - self[4]*self[4]);

	if (abs(s) < 0.0005) then
		s = 1.0;
	end

	local scale = 1.0 / s;
	axis.x = self[1]*scale;
	axis.y = self[2]*scale;
	axis.z = self[3]*scale;

	return angle;
end

FastQuat.getEuler = function(self, euler)
	_tmpVec.x = self[1] * self[1];
	_tmpVec.y = self[2] * self[2]; 
	_tmpVec.z = self[3] * self[3]; 
	_tmpVec.w = self[4] * self[4]; 

	local unit = _tmpVec.x + _tmpVec.y + _tmpVec.z + _tmpVec.w; 
	local test = self[1] * self[2] + self[3] * self[4]; 

	euler.x = atan2(2.0*self[1]*self[4]-2.0*self[2]*self[3],-_tmpVec.x + _tmpVec.y - _tmpVec.z + _tmpVec.w);
	euler.y = atan2(2.0*self[2]*self[4]-2.0*self[1]*self[3], _tmpVec.x - _tmpVec.y - _tmpVec.z + _tmpVec.w);
	euler.z = asin( ( 2.0 * test ) / unit ); 
end

FastQuat.inverse = function(self)
	return FastQuat.new(-self[1],-self[2],-self[3],self[4]);
end

FastQuat.invert = function(self)
	self[1] = -self[1];
	self[2] = -self[2];
	self[3] = -self[3];
end

FastQuat.angle = function(a, b)
	local w = a[4]*b[4] + a[1]*b[1] + a[2]*b[2] + a[3]*b[3];
	return 2.0*acos(w);
end

--cosAngle is MUCH faster then FastQuat.angle() since
--acos() is not needed.
FastQuat.cosAngle = function(a, b)
	local w = a[4]*b[4] + a[1]*b[1] + a[2]*b[2] + a[3]*b[3];
	return 2.0*w;
end

FastQuat.fromTo = function(from, to)
	local q = FastQuat.new();
	q[1] = from[4]*to[1] - from[1]*to[4] - from[2]*to[3] + from[3]*to[2];
	q[2] = from[4]*to[2] + from[1]*to[3] - from[2]*to[4] - from[3]*to[1];
	q[3] = from[4]*to[3] - from[1]*to[2] + from[2]*to[1] - from[3]*to[4];
	q[4] = from[4]*to[4] + from[1]*to[1] + from[2]*to[2] + from[3]*to[3];

	return q;
end

FastQuat.fromToInline = function(self, from, to)
	self.x = from[4]*to[1] - from[1]*to[4] - from[2]*to[3] + from[3]*to[2];
	self.y = from[4]*to[2] + from[1]*to[3] - from[2]*to[4] - from[3]*to[1];
	self.z = from[4]*to[3] - from[1]*to[2] + from[2]*to[1] - from[3]*to[4];
	self.w = from[4]*to[4] + from[1]*to[1] + from[2]*to[2] + from[3]*to[3];
end

-- Based on: http://lolengine.net/blog/2013/09/18/beautiful-maths-quaternion-from-vectors
FastQuat.lookAt = function( lookAt )
	--updated from "fixemups"
	_tmpFwd   = lookAt:normalized();
	_tmpRight = Vector.cross(Vector.up, forward);
	_tmpUp    = Vector.cross(forward, right);
	
	local ret = FastQuat.new();
	ret.w = sqrt(1.0 + _tmpRight.x + _tmpUp.y + _tmpFwd.z) * 0.5;
	local w_inv = 1.0 / (4.0 * ret.w);
	ret.x = (forward.y   - _tmpUp.z)    * w_inv;
	ret.y = (_tmpRight.z - _tmpFwd.x)   * w_inv;
	ret.z = (_tmpUp.x    - _tmpRight.y) * w_inv;

	return ret;
end

FastQuat.lookAtInline = function( self, lookAt )
	--updated from "fixemups"
	_tmpFwd   = lookAt:normalized();
	_tmpRight = Vector.cross(Vector.up, forward);
	_tmpUp    = Vector.cross(forward, right);
	
	self.w = sqrt(1.0 + _tmpRight.x + _tmpUp.y + _tmpFwd.z) * 0.5;
	local w_inv = 1.0 / (4.0 * self.w);
	self.x = (_tmpFwd.y   - _tmpUp.z)    * w_inv;
	self.y = (_tmpRight.z - _tmpFwd.x)   * w_inv;
	self.z = (_tmpUp.x    - _tmpRight.y) * w_inv;
end

FastQuat.lookAtWithUp = function( dir, up )
    _tmpRight = Vector.cross( dir, up )
    _tmpRight:normalize()
    
    _tmpUp = Vector.cross( _tmpRight, dir )
    _tmpUp:normalize()

	--avoid the extra copies...
	local basis = Matrix.new(_tmpRight.x, _tmpUp.x, -dir.x, 0.0,
							 _tmpRight.y, _tmpUp.y, -dir.y, 0.0,
							 _tmpRight.z, _tmpUp.z, -dir.z, 0.0,
							     0.0,  0.0,    0.0, 1.0)

    local q = FastQuat.new()
    q:setMatrix( basis )

    return q
end

FastQuat.lookAtWithUpInline = function( self, dir, up )
    _tmpRight = Vector.cross( dir, up )
    _tmpRight:normalize()
    
    _tmpUp = Vector.cross( _tmpRight, dir )
    _tmpUp:normalize()

	--avoid the extra copies...
	local basis = Matrix.new(_tmpRight.x, _tmpUp.x, -dir.x, 0.0,
							 _tmpRight.y, _tmpUp.y, -dir.y, 0.0,
							 _tmpRight.z, _tmpUp.z, -dir.z, 0.0,
							     0.0,  0.0,    0.0, 1.0)

    self:setMatrix( basis )
end

FastQuat.setMatrix = function(self, mat)
	local s, t
	if mat.m00 + mat.m11 + mat.m22 > 0.0 then
		t = mat.m00 + mat.m11 + mat.m22 + 1.0;
		s = 1.0 / sqrt(t) * 0.5;
		self[4] = s * t;
		self[3] = ( mat.m10 - mat.m01 ) * s;
		self[2] = ( mat.m02 - mat.m20 ) * s;
		self[1] = ( mat.m21 - mat.m12 ) * s;
	elseif mat.m00 > mat.m11 and mat.m00 > mat.m22 then
		t = mat.m00 - mat.m11 - mat.m22 + 1.0;
		s = 1.0 / sqrt(t) * 0.5;
		self[1] = s * t;
		self[2] = ( mat.m10 + mat.m01 ) * s;
		self[3] = ( mat.m02 + mat.m20 ) * s;
		self[4] = ( mat.m21 - mat.m12 ) * s;
	elseif mat.m11 > mat.m22 then
		t = -mat.m00 + mat.m11 - mat.m22 + 1.0;
		s = 1.0 / sqrt(t) * 0.5;
		self[2] = s * t;
		self[1] = ( mat.m10 + mat.m01 ) * s;
		self[4] = ( mat.m02 - mat.m20 ) * s;
		self[3] = ( mat.m21 + mat.m12 ) * s;
	else
		t = -mat.m00 - mat.m11 + mat.m22 + 1.0;
		s = 1.0 / sqrt(t) * 0.5;
		self[3] = s * t;
		self[4] = ( mat.m10 - mat.m01 ) * s;
		self[1] = ( mat.m02 + mat.m20 ) * s;
		self[2] = ( mat.m21 + mat.m12 ) * s;
	end
	
	--inline normalize.
	local len = sqrt( self[1]*self[1] + self[2]*self[2] + self[3]*self[3] + self[4]*self[4] );
	if len > 0.0 then
		local inv_len = 1.0/len;
		self[1] = self[1] * inv_len;
		self[2] = self[2] * inv_len;
		self[3] = self[3] * inv_len;
		self[4] = self[4] * inv_len;
	end
end

FastQuat.lerp = function(t, a, b) 
	t = clamp( t, 0.0, 1.0 )
	_tmpQuat:copyInline(a)
	if( FastQuat.dot(a, b) < 0 ) then
		_tmpQuat:invert()
		_tmpQuat.w = -_tmpQuat.w;
	end
	local q = FastQuat.new();
	q:lerp_internal(t, _tmpQuat, b);

	return q;
end

FastQuat.lerpInline = function(self, t, a, b) 
	t = clamp( t, 0.0, 1.0 )
	_tmpQuat:copyInline(a)
	if( FastQuat.dot(a, b) < 0 ) then
		_tmpQuat:invert()
		_tmpQuat.w = -_tmpQuat.w;
	end
	self:lerp_internal(t, _tmpQuat, b);
end

FastQuat.lerp_internal = function(self, t, a, b) 
	self[1] = a[1]+(b[1]-a[1])*t;
	self[2] = a[2]+(b[2]-a[2])*t;
	self[3] = a[3]+(b[3]-a[3])*t;
	self[4] = a[4]+(b[4]-a[4])*t;

	--inline normalize.
	local len = sqrt( self[1]*self[1] + self[2]*self[2] + self[3]*self[3] + self[4]*self[4] );
	if len > 0.0 then
		local inv_len = 1.0/len;
		self[1] = self[1] * inv_len;
		self[2] = self[2] * inv_len;
		self[3] = self[3] * inv_len;
		self[4] = self[4] * inv_len;
	end
end

FastQuat.slerp = function(t, a, b) 
	t = clamp( t, 0.0, 1.0 )
	_tmpQuat:copyInline(a)
	if( FastQuat.dot(a, b) < 0 ) then
		_tmpQuat:invert()
		_tmpQuat.w = -a_copy.w;
	end
	local q = FastQuat.new();
	q:slerp_internal(t, _tmpQuat, b);

	return q;
end

FastQuat.slerpInline = function(self, t, a, b) 
	t = clamp( t, 0.0, 1.0 )
	_tmpQuat:copyInline(a)
	if( FastQuat.dot(a, b) < 0 ) then
		_tmpQuat:invert()
		_tmpQuat.w = -_tmpQuat.w;
	end
	self:slerp_internal(t, _tmpQuat, b);
end

FastQuat.slerp_internal = function(self, t, a, b) 
	local c = FastQuat.dot(a, b);
	if (c >= 0.9) then
		self:lerp_internal(t, a, b);
		return;
	end
	
	local s = sqrt(abs(1.0 - c*c)); -- sine of the angle
	if (s < 0.1) then
		self:copy_inplace(a);
		return;
	end

	local sign = 1.0;
	if ( c < 0.0 ) then
		sign = -1.0;
	end	
	
	local angle = asin(s);
	local invs = 1.0 / s;
	local coef0 = sin((1.0 - t) * angle) * invs;
	local coef1 = sin(t * angle) * invs * sign;

	self[1] = a[1]*coef0 + b[1]*coef1;
	self[2] = a[2]*coef0 + b[2]*coef1;
	self[3] = a[3]*coef0 + b[3]*coef1;
	self[4] = a[4]*coef0 + b[4]*coef1;
end

FastQuat.shortestDistance = function(a, b)
	if( FastQuat.dot(a, b) < 0.0 ) then
		_tmpQuat:copyInline(a)
		_tmpQuat:invert()
		_tmpQuat.w = -_tmpQuat.w;
		return _tmpQuat
	end
	
	return a;
end

local twoPI = 6.283185307179586476925286766559;

FastQuat.getRoll = function(self)
	_tmpRight = Vector.rotateBy(Vector.right, self)
	_tmpUp    = Vector.rotateBy(Vector.up, self)

	_tmpRight:normalize()
	_tmpUp:normalize()
	
	-- project onto the XY plane
	local xyUp = Vector(_tmpRight.y, _tmpUp.y, 0, 0);
	xyUp:normalize()
	
	-- calculate roll from this isolated case
	local roll = Vector.angle(Vector.up, xyUp) or 0;
	
	-- half space test to capture which half of the unit-circle/plane we are in ( quadrant 1 & 2 or quadrant 3 & 4)
	local rollHS = Vector.dot(Vector.right, xyUp*-1)
	
	local finalRoll = sign(rollHS)*roll;
	
	if(finalRoll < 0) then
		finalRoll = finalRoll + twoPI
	end
	
	local axis = FastQuat.new();
	axis:setAxisAngle(Vector.forward, finalRoll);

	return finalRoll, axis
end

FastQuat.getPitch = function(self)
	local roll, rollQuat = self:getRoll();
	local yaw, yawQuat = self:getYaw();
	local rotation = self * rollQuat:inverse() * yawQuat:inverse()
	
	_tmpUp = Vector.rotateBy(Vector.up, rotation)
	_tmpUp:normalize()
	
	-- project onto the XZ plane
	local yzUp = Vector(0, up.y, -up.z, 0)
	yzUp:normalize()
	
	-- calculate roll from this isolated case
	local pitch = Vector.angle(Vector.up, up) or 0;
	
	-- half space test to capture which half of the unit-circle/plane we are in ( quadrant 1 & 2 or quadrant 3 & 4)
	local pitchHS = Vector.dot(Vector.forward, up*-1)
	
	local final = math.sign(pitchHS)*pitch;
	
	if(final < 0) then
		final = final + twoPI
	end

	--callers are expecting degrees...	
	return deg(final), rotation
end

FastQuat.getYaw = function(self)
	local roll, rollQuat = self:getRoll()
	local rotation = self*rollQuat:inverse()
	
	_tmpRight = Vector.rotateBy(Vector.right, rotation)
	_tmpRight:normalize()
	
	-- project onto the XZ plane
	local xzRight = Vector(_tmpRight.x, 0, -_tmpRight.z, 0);
	xzRight:normalize()
	
	-- calculate roll from this isolated case
	local yaw = Vector.angle(Vector.right, xzRight) or 0
	
	-- half space test to capture which half of the unit-circle/plane we are in ( quadrant 1 & 2 or quadrant 3 & 4)
	local yawHS = Vector.dot(Vector.forward, xzRight*-1)
	local final = sign(yawHS)*yaw;
	
	if(final < 0) then
		final = final + twoPI
	end

	local axis = FastQuat.new();
	axis:setAxisAngle(Vector.up, final);
	
	--this is expecting degrees...	
	return deg(final), axis
end

FastQuat.toMatrix = function(self)
    local sqw = self[4]*self[4];
    local sqx = self[1]*self[1];
    local sqy = self[2]*self[2];
    local sqz = self[3]*self[3];
    
    local mat = Matrix.new()

    -- invs (inverse square length) is only required if quaternion is not already normalised
    local invs = 1 / (sqx + sqy + sqz + sqw)
    mat.m00 = ( sqx - sqy - sqz + sqw)*invs; -- since sqw + sqx + sqy + sqz =1/invs*invs
    mat.m11 = (-sqx + sqy - sqz + sqw)*invs;
    mat.m22 = (-sqx - sqy + sqz + sqw)*invs;
    
    local tmp1 = self[1]*self[2];
    local tmp2 = self[3]*self[4];
    mat.m10 = 2.0 * (tmp1 + tmp2)*invs ;
    mat.m01 = 2.0 * (tmp1 - tmp2)*invs ;
    
    tmp1 = self[1]*self[3];
    tmp2 = self[2]*self[4];
    mat.m20 = 2.0 * (tmp1 - tmp2)*invs ;
    mat.m02 = 2.0 * (tmp1 + tmp2)*invs ;

    tmp1 = self[2]*self[3];
    tmp2 = self[1]*self[4];
    mat.m21 = 2.0 * (tmp1 + tmp2)*invs ;
    mat.m12 = 2.0 * (tmp1 - tmp2)*invs ;  
    
    return mat    
end

FastQuat.toMatrixInline = function(self, mat)
    local sqw = self[4]*self[4];
    local sqx = self[1]*self[1];
    local sqy = self[2]*self[2];
    local sqz = self[3]*self[3];
    
    -- invs (inverse square length) is only required if quaternion is not already normalised
    local invs = 1 / (sqx + sqy + sqz + sqw)
    mat.m00 = ( sqx - sqy - sqz + sqw)*invs; -- since sqw + sqx + sqy + sqz =1/invs*invs
    mat.m11 = (-sqx + sqy - sqz + sqw)*invs;
    mat.m22 = (-sqx - sqy + sqz + sqw)*invs;
    
    local tmp1 = self[1]*self[2];
    local tmp2 = self[3]*self[4];
    mat.m10 = 2.0 * (tmp1 + tmp2)*invs;
    mat.m01 = 2.0 * (tmp1 - tmp2)*invs;
    
    tmp1 = self[1]*self[3];
    tmp2 = self[2]*self[4];
    mat.m20 = 2.0 * (tmp1 - tmp2)*invs;
    mat.m02 = 2.0 * (tmp1 + tmp2)*invs;

    tmp1 = self[2]*self[3];
    tmp2 = self[1]*self[4];
    mat.m21 = 2.0 * (tmp1 + tmp2)*invs;
    mat.m12 = 2.0 * (tmp1 - tmp2)*invs;  

	--make sure all non-set components are set to defaults.
	mat.m03 = 0;
	mat.m13 = 0;
	mat.m23 = 0;

	mat.m30 = 0;
	mat.m31 = 0;
	mat.m32 = 0;

	mat.m33 = 1;
    
    return mat    
end

FastQuat.__tostring = function(self)
	return "{"..self[1]..","..self[2]..","..self[3]..","..self[4].."}";
end

FastQuat.__eq = function(a, b)
	return a[1] == b[1] and a[2] == b[2] and a[3] == b[3] and a[4] == b[4];
end

FastQuat.__lt = function(a, b)
	return a[1] < b[1] and a[2] < b[2] and a[3] < b[3] and a[4] < b[4];
end

FastQuat.__le = function(a, b)
	return a[1] <= b[1] and a[2] <= b[2] and a[3] <= b[3] and a[4] <= b[4];
end

FastQuat.identity = FastQuat.new(0,0,0,1);

setmetatable(FastQuat, { __call = function(_, ...) return FastQuat.new(...) end })