
Matrix = {}
Matrix.__index = Matrix

Matrix.copy = function(self) 
	return Matrix.new( 
		self.m00, self.m01, self.m02, self.m03,
		self.m10, self.m11, self.m12, self.m13,
		self.m20, self.m21, self.m22, self.m23,
		self.m30, self.m31, self.m32, self.m33
	)
end

Matrix.new = function(m00, m01, m02, m03, m10, m11, m12, m13, m20, m21, m22, m23, m30, m31, m32, m33)
	return setmetatable({	m00 = m00 or 1.0, m01 = m01 or 0.0, m02 = m02 or 0.0, m03 = m03 or 0.0,
							m10 = m10 or 0.0, m11 = m11 or 1.0, m12 = m12 or 0.0, m13 = m13 or 0.0,
							m20 = m20 or 0.0, m21 = m21 or 0.0, m22 = m22 or 1.0, m23 = m23 or 0.0,
							m30 = m30 or 0.0, m31 = m31 or 0.0, m32 = m32 or 0.0, m33 = m33 or 1.0 }, Matrix);
end

Matrix.setTranslation = function(self, x, y, z) 
	self.m03 = x;
	self.m13 = y;
	self.m23 = z;
end

Matrix.translate = function(self, x, y, z) 
	self.m03 = self.m03 + x;
	self.m13 = self.m13 + y;
	self.m23 = self.m23 + z;
end

Matrix.lookAt = function(eye, at, up)
	local forward = at - eye
	forward:normalize()
	
	local side = Vector.cross( forward, up )
	side:normalize()
	
	local yup = Vector.cross( side, forward )
	
	local result = Matrix.identity:copy()
	result.m00 = side.x
	result.m01 = side.y
	result.m02 = side.z
		
	result.m10 = yup.x
	result.m11 = yup.y
	result.m12 = yup.z
	
	result.m20 = -forward.x
	result.m21 = -forward.y
	result.m22 = -forward.z
	
	local px = -eye.x
	local py = -eye.y
	local pz = -eye.z
	
	result.m30 = result.m00*px + result.m01*py + result.m02*pz + result.m03
	result.m31 = result.m10*px + result.m11*py + result.m12*pz + result.m13
	result.m32 = result.m20*px + result.m21*py + result.m22*pz + result.m23
	
	return result
end

function Matrix.perspectiveFovX( fovx, aspect, znear, zfar )
	local xmax = znear * math.tan( fovx * 0.5 )
	local ymax = xmax / aspect

	local left = -xmax
	local right = xmax
	local bottom = -ymax 
	local top = ymax

	local t1 = 2.0 * znear
	local t2 = right - left
	local t3 = top - bottom
	local t4 = zfar - znear

	return Matrix.new(
		t1 / t2,
		0.0,
		( right + left ) / t2,
		0.0,
		0.0,
		t1 / t3,
		( top + bottom ) / t3,
		0.0,
		0.0,
		0.0,
		( -zfar - znear ) / t4,
		( -t1 * zfar ) / t4,
		0.0,
		0.0,
		-1.0,	
		0.0 )
end

function Matrix.perspectiveFovY( fovy, aspect, znear, zfar )
	local ymax = znear * math.tan( fovy * 0.5 )
	local xmax = ymax * aspect

	local left = -xmax
	local right = xmax
	local bottom = -ymax 
	local top = ymax

	local t1 = 2.0 * znear
	local t2 = right - left
	local t3 = top - bottom
	local t4 = zfar - znear

	return Matrix.new(
		t1 / t2,
		0.0,
		( right + left ) / t2,
		0.0,
		0.0,
		t1 / t3,
		( top + bottom ) / t3,
		0.0,
		0.0,
		0.0,
		( -zfar - znear ) / t4,
		( -t1 * zfar ) / t4,
		0.0,
		0.0,
		-1.0,	
		0.0 )
end

Matrix.invert = function(self)
	local inv = Matrix.identity:copy()

	inv.m00 =	self.m11*self.m22*self.m33 -
				self.m11*self.m32*self.m23 -
				self.m12*self.m21*self.m33 +
				self.m12*self.m31*self.m23 +
				self.m13*self.m21*self.m32 -
				self.m13*self.m31*self.m22

	inv.m01 =  -self.m01*self.m22*self.m33 +
				self.m01*self.m32*self.m23 +
				self.m02*self.m21*self.m33 -
				self.m02*self.m31*self.m23 -
				self.m03*self.m21*self.m32 +
				self.m03*self.m31*self.m22

	inv.m02 =	self.m01*self.m12*self.m33 -
				self.m01*self.m32*self.m13 -
				self.m02*self.m11*self.m33 +
				self.m02*self.m31*self.m13 +
				self.m03*self.m11*self.m32 -
				self.m03*self.m31*self.m12

	inv.m03 =  -self.m01*self.m12*self.m23 +
				self.m01*self.m22*self.m13 +
				self.m02*self.m11*self.m23 -
				self.m02*self.m21*self.m13 -
				self.m03*self.m11*self.m22 +
				self.m03*self.m21*self.m12

	inv.m10 =  -self.m10*self.m22*self.m33 +
				self.m10*self.m32*self.m23 +
				self.m12*self.m20*self.m33 -
				self.m12*self.m30*self.m23 -
				self.m13*self.m20*self.m32 +
				self.m13*self.m30*self.m22

	inv.m11 =	self.m00*self.m22*self.m33 -
				self.m00*self.m32*self.m23 -
				self.m02*self.m20*self.m33 +
				self.m02*self.m30*self.m23 +
				self.m03*self.m20*self.m32 -
				self.m03*self.m30*self.m22

	inv.m12 =  -self.m00*self.m12*self.m33 +
				self.m00*self.m32*self.m13 +
				self.m02*self.m10*self.m33 -
				self.m02*self.m30*self.m13 -
				self.m03*self.m10*self.m32 +
				self.m03*self.m30*self.m12

	inv.m13 =	self.m00*self.m12*self.m23 -
				self.m00*self.m22*self.m13 -
				self.m02*self.m10*self.m23 +
				self.m02*self.m20*self.m13 +
				self.m03*self.m10*self.m22 -
				self.m03*self.m20*self.m12

	inv.m20 =	self.m10*self.m21*self.m33 -
				self.m10*self.m31*self.m23 -
				self.m11*self.m20*self.m33 +
				self.m11*self.m30*self.m23 +
				self.m13*self.m20*self.m31 -
				self.m13*self.m30*self.m21

	inv.m21 =  -self.m00*self.m21*self.m33 +
				self.m00*self.m31*self.m23 +
				self.m01*self.m20*self.m33 -
				self.m01*self.m30*self.m23 -
				self.m03*self.m20*self.m31 +
				self.m03*self.m30*self.m21

	inv.m22 =	self.m00*self.m11*self.m33 -
				self.m00*self.m31*self.m13 -
				self.m01*self.m10*self.m33 +
				self.m01*self.m30*self.m13 +
				self.m03*self.m10*self.m31 -
				self.m03*self.m30*self.m11

	inv.m23 =  -self.m00*self.m11*self.m23 +
				self.m00*self.m21*self.m13 +
				self.m01*self.m10*self.m23 -
				self.m01*self.m20*self.m13 -
				self.m03*self.m10*self.m21 +
				self.m03*self.m20*self.m11

	inv.m30 =  -self.m10*self.m21*self.m32 +
				self.m10*self.m31*self.m22 +
				self.m11*self.m20*self.m32 -
				self.m11*self.m30*self.m22 -
				self.m12*self.m20*self.m31 +
				self.m12*self.m30*self.m21

	inv.m31 =	self.m00*self.m21*self.m32 -
				self.m00*self.m31*self.m22 -
				self.m01*self.m20*self.m32 +
				self.m01*self.m30*self.m22 +
				self.m02*self.m20*self.m31 -
				self.m02*self.m30*self.m21

	inv.m32 =  -self.m00*self.m11*self.m32 +
				self.m00*self.m31*self.m12 +
				self.m01*self.m10*self.m32 -
				self.m01*self.m30*self.m12 -
				self.m02*self.m10*self.m31 +
				self.m02*self.m30*self.m11

	inv.m33 =	self.m00*self.m11*self.m22 -
				self.m00*self.m21*self.m12 -
				self.m01*self.m10*self.m22 +
				self.m01*self.m20*self.m12 +
				self.m02*self.m10*self.m21 -
				self.m02*self.m20*self.m11
	
	local det = self.m00*inv.m00 +
				self.m10*inv.m01 +
				self.m20*inv.m02 +
				self.m30*inv.m03
				
	if det ~= 0.0 then
		det = 1.0 / det
		self.m00 = inv.m00 * det
		self.m01 = inv.m01 * det
		self.m02 = inv.m02 * det
		self.m03 = inv.m03 * det
				    		
		self.m10 = inv.m10 * det
		self.m11 = inv.m11 * det
		self.m12 = inv.m12 * det
		self.m13 = inv.m13 * det
				    		
		self.m20 = inv.m20 * det
		self.m21 = inv.m21 * det
		self.m22 = inv.m22 * det
		self.m23 = inv.m23 * det
				    		
		self.m30 = inv.m30 * det
		self.m31 = inv.m31 * det
		self.m32 = inv.m32 * det
		self.m33 = inv.m33 * det
	end
end

Matrix.applyTransform = function(self, v)
	local r = v.new()
	r.x = v.x*self.m00 + v.y*self.m01 + v.z*self.m02 + v.w*self.m03
	r.y = v.x*self.m10 + v.y*self.m11 + v.z*self.m12 + v.w*self.m13
	r.z = v.x*self.m20 + v.y*self.m21 + v.z*self.m22 + v.w*self.m23
	r.w = v.x*self.m30 + v.y*self.m31 + v.z*self.m32 + v.w*self.m33
	return r
end

Matrix.__add = function(a, b)
	return Matrix.new(a.m00 + b.m00, a.m01 + b.m01, a.m02 + b.m02, a.m03 + b.m03, 
					  a.m10 + b.m00, a.m11 + b.m11, a.m12 + b.m12, a.m13 + b.m13, 
					  a.m20 + b.m00, a.m21 + b.m21, a.m22 + b.m22, a.m23 + b.m23, 
					  a.m30 + b.m00, a.m31 + b.m31, a.m32 + b.m32, a.m33 + b.m33);
end

Matrix.__tostring = function(self)
	return "{"..self.m00..","..self.m01..","..self.m02..","..self.m03..","..self.m10..","..self.m11..","..self.m12..","..self.m13..","..self.m20..","..self.m21..","..self.m22..","..self.m23..","..self.m30..","..self.m31..","..self.m32..","..self.m33.."}"
end

Matrix.identity = Matrix.new( 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 )

setmetatable(Matrix, { __call = function(_, ...) return Matrix.new(...) end })
