
Quaternion = {}
Quaternion.__index = Quaternion

Quaternion.length = function(self)
	return math.sqrt( self.x*self.x + self.y*self.y + self.z*self.z + self.w*self.w )
end

Quaternion.lengthSq = function(self)
	return self.x*self.x + self.y*self.y + self.z*self.z + self.w*self.w
end

Quaternion.normalize = function(self)
	local len = self:length();

	if len > 0.0 then
		local inv_len = 1.0/len;
		self.x = self.x * inv_len;
		self.y = self.y * inv_len;
		self.z = self.z * inv_len;
		self.w = self.w * inv_len;
	end

	return len;
end

Quaternion.unpack = function(self)
	return self.x, self.y, self.z, self.w;
end

Quaternion.new = function(x, y, z, w)
	return setmetatable({ x = x or 0.0, y = y or 0.0, z = z or 0.0, w = w or 1.0 }, Quaternion);
end

Quaternion.__add = function(a, b)
	return Quaternion.new(a.x + b.x, a.y + b.y, a.z + b.z, a.w + b.w);
end

Quaternion.__sub = function(a, b)
	return Quaternion.new(a.x - b.x, a.y - b.y, a.z - b.z, a.w - b.w);
end

Quaternion.__mul = function(a, b)
	local w = a.w * b.w - a.x * b.x - a.y * b.y - a.z * b.z;
	local x = a.w * b.x + a.x * b.w + a.y * b.z - a.z * b.y;
	local y = a.w * b.y - a.x * b.z + a.y * b.w + a.z * b.x;
	local z = a.w * b.z + a.x * b.y - a.y * b.x + a.z * b.w;	
	local out = Quaternion.new(x, y, z, w)
	return out;
end

Quaternion.fastMul = function(self, a, b)
	self.w = a.w * b.w - a.x * b.x - a.y * b.y - a.z * b.z;
	self.x = a.w * b.x + a.x * b.w + a.y * b.z - a.z * b.y;
	self.y = a.w * b.y - a.x * b.z + a.y * b.w + a.z * b.x;
	self.z = a.w * b.z + a.x * b.y - a.y * b.x + a.z * b.w;	
end

Quaternion.fastAdd = function(self, a, b)
	self.w = a.w + b.w;
	self.x = a.x + b.x;
	self.y = a.y + b.y;
	self.z = a.z + b.z;
end

Quaternion.fastMadd = function(self, a, b)
	self.w = self.w + a.w * b.w - a.x * b.x - a.y * b.y - a.z * b.z;
	self.x = self.x + a.w * b.x + a.x * b.w + a.y * b.z - a.z * b.y;
	self.y = self.y + a.w * b.y - a.x * b.z + a.y * b.w + a.z * b.x;
	self.z = self.z + a.w * b.z + a.x * b.y - a.y * b.x + a.z * b.w;	
end

Quaternion.__eq = function(a, b)
	return a.x == b.x and a.y == b.y and a.z == b.z and a.w == b.w;
end

Quaternion.__lt = function(a, b)
	return a.x < b.x and a.y < b.y and a.z < b.z and a.w < b.w;
end

Quaternion.__le = function(a, b)
	return a.x <= b.x and a.y <= b.y and a.z <= b.z and a.w <= b.w;
end

Quaternion.scale = function(self, x, y, z, w)
	self.x = self.x * x;
	self.y = self.y * y;
	self.z = self.z * z;
	self.w = self.w * w;
end

Quaternion.copy = function(self)
	return Quaternion.new(self.x, self.y, self.z, self.w);
end

Quaternion.copy_inplace = function(self, ob) 
	self.x = ob.x;
	self.y = ob.y;
	self.z = ob.z;
	self.w = ob.w;
end

Quaternion.__tostring = function(self)
	return "{"..self.x..","..self.y..","..self.z..","..self.w.."}";
end

Quaternion.set = function(self, x, y, z, w)
	self.x = x
	self.y = y
	self.z = z
	self.w = w
end

Quaternion.scale = function(a, s)
	return Quaternion.new(a.x*s, a.y*s, a.z*s, a.w*s);
end

Quaternion.dot = function(a, b) 
	return a.x * b.x + a.y * b.y + a.z * b.z + a.w * b.w;
end

-- create a quaternion from an axis angle, the angle is in radians.
Quaternion.setAxisAngle = function(self, axis, angle)
	local theta 		= angle*0.5;
	local sinTheta     	= math.sin(theta);				
	self.x				= axis.x*sinTheta;
	self.y				= axis.y*sinTheta;
	self.z				= axis.z*sinTheta;
	self.w				= math.cos(theta);
	self:normalize()
end

Quaternion.axisAngle = function(axis, angle)
	local q = Quaternion.new()
	q:setAxisAngle(axis, angle)
	return q
end

Quaternion.setEuler = function(self, x, y, z) 
	local c1 = math.cos(x*0.5);
	local c2 = math.cos(y*0.5);
	local c3 = math.cos(z*0.5);

	local s1 = math.sin(x*0.5);
	local s2 = math.sin(y*0.5);
	local s3 = math.sin(z*0.5);

	self.w = c1 * c2 * c3 + s1 * s2 * s3;
	self.x = s1 * c2 * c3 - c1 * s2 * s3;
	self.y = c1 * s2 * c3 + s1 * c2 * s3;
	self.z = c1 * c2 * s3 - s1 * s2 * c3;
end

Quaternion.getAxisAngle = function(self)
	local angle = 2.0*math.acos(self.w);

	local s = math.sqrt(1.0 - self.w*self.w);
	if (math.abs(s) < 0.0005) then
		s = 1.0;
	end

	return Vector.new(self.x/s, self.y/s, self.z/s), angle;
end

Quaternion.getEuler = function(self)
	local xx = self.x * self.x; 
	local yy = self.y * self.y; 
	local zz = self.z * self.z; 
	local ww = self.w * self.w; 

	local unit = xx + yy + zz + ww; 
	local test = self.x * self.y + self.z * self.w; 

	local outx = math.atan2(2.0*self.x*self.w-2.0*self.y*self.z , -xx + yy - zz + ww);
	local outy = math.atan2(2.0*self.y*self.w-2.0*self.x*self.z , xx - yy - zz + ww);
	local outz = math.asin( ( 2.0 * test ) / unit ); 

	return Vector.new(outx, outy, outz);
end

-- supported by unity
--Quaternion.rotateTowards = function(dir)
--end

Quaternion.inverse = function(self)
	return Quaternion.new(-self.x,-self.y,-self.z,self.w);
end

Quaternion.invert = function(self)
	self.x = -self.x;
	self.y = -self.y;
	self.z = -self.z;
	self.w = self.w;
--	self:normalize();
end

Quaternion.conjugate = function(self)
	return Quaternion.new(-self.x,-self.y,-self.z,self.w);
end

Quaternion.angle = function(a, b)
	local diff = a:inverse() * b;
	return 2.0*(math.acos(diff.w));
end

Quaternion.fromTo = function(from, to)
	local diff = from:inverse() * to; 
	return diff;
end

-- Based on: http://lolengine.net/blog/2013/09/18/beautiful-maths-quaternion-from-vectors
Quaternion.lookAt = function( from, to )
	local x = Vector.cross( from, to )
	local w = 1.0 + Vector.dot( from, to )
	local q = Quaternion.new( x.x, x.y, x.z, w )
	q:normalize()	
	return q
end

Quaternion.lookAtWithUp = function( dir, up )
    local right             = Vector.cross( dir, up )
    right:normalize()
    
    up                      = Vector.cross( right, dir )
    up:normalize()

	--avoid the extra copies...
	local basis = Matrix.new(right.x, up.x, -dir.x, 0.0,
							 right.y, up.y, -dir.y, 0.0,
							 right.z, up.z, -dir.z, 0.0,
							     0.0,  0.0,    0.0, 1.0)

    local q                 = Quaternion.new()
    q:setMatrix( basis )

    return q
end


Quaternion.setMatrix = function(self, mat)
	local s, t
	if mat.m00 + mat.m11 + mat.m22 > 0.0 then
		t = mat.m00 + mat.m11 + mat.m22 + 1.0;
		s = 1.0 / math.sqrt(t) * 0.5;
		self.w = s * t;
		self.z = ( mat.m10 - mat.m01 ) * s;
		self.y = ( mat.m02 - mat.m20 ) * s;
		self.x = ( mat.m21 - mat.m12 ) * s;
	elseif mat.m00 > mat.m11 and mat.m00 > mat.m22 then
		t = mat.m00 - mat.m11 - mat.m22 + 1.0;
		s = 1.0 / math.sqrt(t) * 0.5;
		self.x = s * t;
		self.y = ( mat.m10 + mat.m01 ) * s;
		self.z = ( mat.m02 + mat.m20 ) * s;
		self.w = ( mat.m21 - mat.m12 ) * s;
	elseif mat.m11 > mat.m22 then
		t = -mat.m00 + mat.m11 - mat.m22 + 1.0;
		s = 1.0 / math.sqrt(t) * 0.5;
		self.y = s * t;
		self.x = ( mat.m10 + mat.m01 ) * s;
		self.w = ( mat.m02 - mat.m20 ) * s;
		self.z = ( mat.m21 + mat.m12 ) * s;
	else
		t = -mat.m00 - mat.m11 + mat.m22 + 1.0;
		s = 1.0 / math.sqrt(t) * 0.5;
		self.z = s * t;
		self.w = ( mat.m10 - mat.m01 ) * s;
		self.x = ( mat.m02 + mat.m20 ) * s;
		self.y = ( mat.m21 + mat.m12 ) * s;
	end
	
	self:normalize();
end

Quaternion.lerp = function(t, a, b) 
	local out = a + (b - a) * t;
	out:normalize();
	return out;
end

Quaternion.slerp = function(t, a, b) 
	local c = Quaternion.dot(a, b);
	if (c >= 0.9) then
		return Quaternion.lerp(t, a, b);
	end
	
	local s = math.sqrt(math.abs(1.0 - c*c)); -- sine of the angle
	if (s < 0.1) then
		return Quaternion.new(a.x, a.y, a.z, a.w)
	end

	local sign = 1.0;
	if ( c < 0.0 ) then
		sign = -1.0;
	end	
	
	local angle = math.asin(s);
	local invs = 1.0 / s;
	local coef0 = math.sin((1.0 - t) * angle) * invs;
	local coef1 = math.sin(t * angle) * invs * sign;

	local x = a.x * coef0 + b.x * coef1;
	local y = a.y * coef0 + b.y * coef1;
	local z = a.z * coef0 + b.z * coef1;
	local w = a.w * coef0 + b.w * coef1;
	
	return Quaternion.new(x,y,z,w);
end

Quaternion.identity = Quaternion.new(0,0,0,1);

setmetatable(Quaternion, { __call = function(_, ...) return Quaternion.new(...) end })
