
-- some native functions are overridden here...
Scene.super = {}
Scene.super.createActor = Scene.createActor
Scene.super.createLight = Scene.createLight
Scene.super.createCamera = Scene.createCamera
Scene.super.createTransform = Scene.createTransform
Scene.super.createParticleEmitter = Scene.createParticleEmitter

persp = function(fov, n, f) 
	return function(cam)
		cam:setPerspective(fov, n, f);
	end
end

ortho = function(l, r, b, t, n, f)
	return function(cam)
		cam:setOrthographic(l, r, b, t, n, f);
	end
end

Scene.createTransform = function(self, name)
	local transform = Scene.super.createTransform(self, name);
	self.transforms = self.transforms or {};
	self.transforms[name] = transform;
	return transform;
end

Scene.createCamera = function(self, name, config)
	if(config == nil) then
		config = persp(math.rad(60.0), 0.01, 10000.0);
	end
	
	local cam = Scene.super.createCamera(self, name);
	config(cam);
		
	self.cameras = self.cameras or {};
	self.cameras[name] = cam;
	
	return cam;
end

Scene.setupActor = function(self, out, def, dontAddAnims)
	if(def.transform ~= nil) then
		out:setLocal(def.transform);
	end
	
	if(def.boneRadius and Actor.setBoneRadius) then
        out:setBoneRadius(def.boneRadius);
	end

	if ( def.animation ) then
        out.animation = {};
        if ( not dontAddAnims ) then
            for key,value in pairs(def.animation) do
                out.animation[key] = out:addAnimation(key, value[1], value[2], value[3], value[4], value[5] or true);
                out.animation[key].uri = value[1];
                if(key == "default") then
                    out.animation[key]:play();
                end
            end
        end
	end
	
	if(def.flipbooks ~= nil) then
		out.flipbooks = {};
		for key,value in pairs(def.flipbooks) do
			-- name, uri, startFrame, stopFrame, looping, sync, material, inputs
			out.flipbooks[key] = out:addFlipbook(key, value[1], value[2], value[3], value[4], value[5], value[6], value[7]);
		end		
	end

	if (def.viewFlags ~= nil) then
		out:setViewFlags(def.viewFlags)
	end

	if (def.materials) then
		out.materials = {}
		for _, matData in ipairs(def.materials) do
			local material = out:setMaterial( matData.name, matData )
			if material then
				out.materials[matData.name] = material
			else
				Log.error("Unable to find material " .. matData.name)
			end
		end
	end
end	

-- parses and initializes the actor based on the given table
Scene.createActor = function(self, name, def, deferSetup, dontAddAnims)
	local out = nil;
	if(type(def) == "table") then
		out = Scene.super.createActor(self, name, def.model); 
		out.uri = def.model;
		out.name = name;
		
		if (self.actors == nil) then
			self.actors = {}
		end
		
		out.destroy = function(actor)
            self.actors[name] = nil;
			Actor.destroy(actor);
		end
			
		self.actors[name] = out;
		
		if(deferSetup == nil) then
			Scene.setupActor(self, out, def, dontAddAnims);
		end
	end
	return out;
end

Scene.createParticleEmitter = function(self, name, uri)
	local actors = Scene.super.createParticleEmitter(self, name, uri)
	actors.name = name
	actors.uri = uri
	actors.emitter = true
	self.actors = self.actors or {}
	self.actors[name] = actors
	return actors
end

Scene.createLight = function(self, name, def)
	local out = nil;
	if(type(def) == "table") then
		out = Scene.super.createLight(self, name); 
		
		if (self.lights == nil) then
			self.lights = {}
		end

		self.lights[name] = out;

		if(def.transform) then
			out:setLocal(def.transform);
		end
		
		if(def.type) then
			out:setType(def.type);
		end

		if(def.range) then
			out:setRange(def.range[1], def.range[2]);
		end

		if(def.spotAngle) then
			out:setSpotAngle(def.spotAngle[1], def.spotAngle[2]);
		end

		if(def.wrap) then
			out:setWrap(def.wrap);
		end

		if(def.ambient) then
			out:setAmbient(def.ambient);
		end

		if(def.diffuse) then
			out:setDiffuse(def.diffuse);
		end
	end
	return out;
end

Transform.super = {}
Transform.super.addComponent = Transform.addComponent

Transform.__onLateUpdate = function(self)
--	Log.message("__onLateUpdate");
	if(self.components == nil) then
		return;
	end
	for key, comp in pairs(self.components) do
		if (comp.onLateUpdate ~= nil and comp.active) then
			comp:onLateUpdate();
		end
	end
end

Transform.__onUpdate = function(self)
--	Log.message("__onUpdate");
	if(self.components == nil) then
		return;
	end
	for key, comp in pairs(self.components) do
		if(comp.__started == nil) then
			comp.__started = true;
			if (comp.onStart ~= nil and comp.active) then
				comp:onStart();
			end
		end
		if (comp.onUpdate ~= nil and comp.active) then
			comp:onUpdate();
		end
	end
end

--[[
Transform.__onStart = function(self)
--	Log.message("__onStart");
	if(self.components == nil) then
		return;
	end
	for key, comp in pairs(self.components) do
		if(comp.__started == nil) then
			if (comp.onStart ~= nil and comp.active) then
				comp:onStart();
			end
			comp.__started = true;
		end
	end
end
]]

Transform.__onDestroy = function(self)
	if(self.components == nil) then
		return;
	end
	for key, comp in pairs(self.components) do
		if (comp.onDestroy ~= nil) then
			comp:onDestroy()
		end
	end
end

Transform.addComponent = function(self, name, t)
	if (self.components == nil) then
		self.components = {};
		Transform.super.addComponent(self, "LComponent");
	end
	
	self.components[name] = t;--shallowcopy(t);--setmetatable({ object = self, active = true }, t);
	self.components[name].object = self;
	self.components[name].active = true;
	self.components[name].base_name = name;
--	self.components[name].__index = t;
	
	return self.components[name];
end

Transform.sendEvent = function(self, target, evt)	
	if (target.components == nil) then
		return false;
	end
		
	evt.sender = self;
	for key, comp in pairs(target.components) do
		if (comp.onEvent ~= nil) then
			if(comp:onEvent(evt)) then
				return true;
			end
		end
	end
	return false;
end

-- material flags
enum("MT_TEXTURE", 0)
enum("MT_FLIPBOOK")
enum("MT_RAW")

-- deprecated
enum("MTT_TEXTURE", 0)
enum("MTT_FLIPBOOK")
enum("MTT_RAW")

