-- The Toolbox is a Global API and serves as a central hub for 
-- MOXIE engine tools. The API is designed as a pass-through API. 
-- All function calls should gracefully handle the case where 
-- no tools are available (ex: candidate mode). Directly accessing
-- tools inside the Toolbox should be avoided unless properly 
-- guarded.

-- PROJECT INTEGRATION: 
--		Your project must call Toolbox:init() somewhere. 
--		Call Toolbox:update() once per frame.

Persist = Persist or {}
Persist.toolbox = Persist.toolbox or {}
Persist.toolbox.active = Persist.toolbox.active or {}
Persist.toolbox.options = Persist.toolbox.options or {}

Toolbox = {
	enabled = not Platform.candidate,
	tools = {},
    options = {},
}
Toolbox.__index = Toolbox

Toolbox.init = function(self, tools, options)
    if not Toolbox.enabled then return end

    Toolbox:_initGUI()

    self:add("Menu", ToolboxMenus:new(), true)
    self:_addTools(tools)
    self:_addOptions(options)

    self:hide()
end

Toolbox._initGUI = function(self)
    if Platform.android  or Platform.iOS then
        GUI.setScale(3)
    end
    GUI.loadTheme()
    GUI.show()

    -- Create UI toggle
    -- Clicking or touching the 'gear' icon toggles tool visibility.
    self.uitoggle = GUI.gear(nil, GUIAligned(ALIGN_TOP_RIGHT), GUISize(SZ_WRAP, SZ_WRAP), true, "")
    self.uitoggle:setLayer(LAYER1)
    local pos = self.uitoggle:getLocalPosition()
    self.uitoggle:moveTo(pos.x - self.uitoggle:getWidth(), 0)
    self.uitoggle:anchor(false, true, true, false)

    function self.uitoggle:onclick()
        -- NOTE: self is refering to Toolbox.toggle
        if self.checked then
            for toolName, tool in pairs(Toolbox.tools) do
                if tool.active and tool.show then
                    tool:show()
                end
            end
        else
            for _, tool in pairs(Toolbox.tools) do
                if tool.hide then
                    tool:hide()
                end
            end
        end
    end
end

-- show all tools
Toolbox.show = function(self)
	if (self.uitoggle ~= nil) then
		self.uitoggle:check()
	end
end

-- hide all tools
Toolbox.hide = function(self)
	if(self.uitoggle ~= nil) then
		self.uitoggle:uncheck()
	end
end

-- add a tool to the toolbox.
Toolbox.add = function(self, name, tool, active)
	self.tools = self.tools or {}
	self.active = self.active or {}

	self.tools[name] = tool
    if Persist.toolbox.active[name] ~= nil then
        active = Persist.toolbox.active[name]
    end
    tool.active = active
    local toolbox = self
    local onClose = tool.onClose
    tool.onClose = function(self)
        toolbox:activateTool(name, false)
        if onClose then
            onClose(tool)
        end
    end
    self:activateTool(name, active)
end

Toolbox.activateTool = function(self, name, active)
    local tool = self.tools[name]
    if tool then
        tool.active = active
        if tool.active then
            tool:show()
        else
            tool:hide()
        end
        Persist.toolbox.active[name] = active
        System.flushPersist()
    end
end

-- listen for events from the given tool.
Toolbox.listen = function(self, toolName, eventhandler)
	if self.tools[toolName] then
		self.tools[toolName]:addListener(eventhandler)
	end
end

-- commands are a way to tell tools to do something programmatically.
Toolbox.command = function(self, command, ...) 
	for _, tool in pairs(self.tools) do
		if type(tool.onCommand) == 'function' then
			if tool:onCommand(command, ...) then
				return
			end
		end
	end
end

Toolbox.addOption = function(self, params)
    local option = ToolboxOption:new(params)
    table.insert(self.options, option)
    if Persist.toolbox.options[option.name] ~= nil then
        option.set(Persist.toolbox.options[option.name])
    end
end

Toolbox._addOptions = function(self, options)
    if options then
        for _, option in ipairs(options) do
            self:addOption(option)
        end
    end
end

Toolbox.update = function(self)
	for _, tool in pairs(self.tools) do
		if type(tool.onUpdate) == 'function' then
			tool:onUpdate()
		end
	end	
end

Toolbox._addTools = function(self, tools)
    if tools then
        for _, tool in ipairs(tools) do
            Log.message("Toolbox adding: " .. tostring(tool.name))
            self:add(tool.name, _G[tool.toolbox]:new(tool.params), tool.active)
        end
    end
end

-- Look up the global var name (in "foo.bar.baz" form) and
-- return the global object (foo.bar) and the name of its
-- property ("baz")
local function lookupGlobalVar(name)
    local obj = _G
    local objPartNames = name:split("%.")

    if #objPartNames > 1 then
        for i = 1, #objPartNames - 1 do
            obj = obj[objPartNames[i]]
        end
    end

    return obj, objPartNames[#objPartNames]
end

ToolboxOption = {}

function ToolboxOption:new(params)
    local instance = shallowcopy(params) or {}
    setmetatable(instance, self)
    self.__index = self

    assert(instance.accessorFns or instance.globalVar)

    if instance.accessorFns then
        instance.get = instance.accessorFns.get
        instance.set = instance.accessorFns.set
    else
        instance.get = function()
            local obj, property = lookupGlobalVar(instance.globalVar)
            return obj[property]
        end
        instance.set = function(value)
            local obj, property = lookupGlobalVar(instance.globalVar)
            obj[property] = value
        end
    end

    return instance
end
