
local Logger = TaggedLog("ActorManager", true)

ActorManager = {
    scene = nil,
    showBones = false,
    debugBoneScale = nil,
    actorGroup = {},
    proxyActors = {},
    disableGlow = false,
    enableParticles = true,
    enableDebugging = false,
    enableNativeFlipbookDriver = false,
}

function ActorManager:init(scene)
    self.scene = scene
    self.scene.actors = self.scene.actors or {}
end


function ActorManager:getActor(name, createActor)
    -- Default creation to false unless overwritten
    createActor = createActor or false
    
    if(self.scene.actors[name] == nil and createActor) then
        self:instantiateActor(name)
        
        -- create proxy actor for the group
        if(self.actorGroup[name]) then
            Logger:message("Setting proxy actor: " .. tostring(self.actorGroup[name]) .. " to " .. tostring(name))
            self.proxyActors[self.actorGroup[name]] = name
        end
        
    end

    -- If they were asking for a proxy actor, return it instead
    if not self.scene.actors[name] and self.proxyActors[name] then
        return self.scene.actors[self.proxyActors[name]]
    end

    return self.scene.actors[name]
end


function ActorManager:instantiateActor(name)
    if (name == nil) then
        return
    end

    local actorData = SceneData.actors[name] or LaunchData.actors[name]
    local actor = nil

    -- Simple flag to disable an actor for debug purposes
    if (actorData ~= nil and actorData.debugDisable) then
        return
    end
    
    if (actorData == nil or actorData.model == nil) then
    
    	if(actorData and actorData.mpxUri) then
    		actor = self.scene:createParticleEmitter(name, actorData.mpxUri)
    	else
	        -- If the name does not exist create empty actor
	        Logger:message("Create transform actor object: "..name)
	        actor = self.scene:createTransform(name)
        end
        
        actor.materials = {}

        self.scene.actors[name] = actor
        actorData = actorData or {}
    
    else
        -- force to lower case
        actorData.model = string.lower(actorData.model)
        
        Logger:message("Created actor object: "..name)
        actor = self.scene:createActor(name, actorData, true)

        PostProcessActorData(actor, actorData)
    end

    actor.name = name
    
--  self:printDebug(actor, actorData)
    
    self.scene:setupActor(actor, actorData)

    actor:setVisibility(false)

    -- Set up position if it is specified
    local initialPos = actorData.position
    if (initialPos) then
        initialPos = Vector(initialPos[1], initialPos[2], initialPos[3])
        Logger:message("initial position = " .. tostring(initialPos))
		local localTr = actor:getLocal()
		localTr.position = initialPos
        actor:setLocal(localTr)
    end

    -- Set up rotation if it is specified
    local initialRot = actorData.rotation
    if (initialRot) then
    	local transform = actor:getLocal()
    	if(initialRot.axisAngle) then
    		-- its a quaternion
    		transform.rotation = initialRot:copy()
    	else
    		-- its yaw, pitch, roll
        	transform.rotation = Quaternion.fromEulerAngles(Vector(initialRot[1], initialRot[2],initialRot[3]))
        end
        actor:setLocal(transform)
    end

    -- Add any components needed
    if (actorData.components) then
        for _, compData in ipairs(actorData.components) do
            local class = _G[compData.name]
            if (class == nil) then
                Log.error("Unknown component type: " .. tostring(compData.name))
            else
                -- add conditional segment data (makes this component active for given segments)
                if(compData.segments) then
                    for _, segName in ipairs(compData.segments) do
                        --SegmentManager:addSegmentComponent(segName, actor.name, compData)
                        actor.segmentComponents = actor.segmentComponents or {}
                        actor.segmentComponents[segName] = actor.segmentComponents[segName] or {}
                        table.insert(actor.segmentComponents[segName], compData)
                    end
                else
                    actor:addComponent(compData.alias or compData.name, class:new(compData.params))
                    Logger:message("Added component: " .. compData.name .. " to " .. actor.name)
                end
            end
        end
    end

    if(type(actorData.animationRate) =="number" and actorData.animation) then
        for name, anim in pairs(actorData.animation) do
            actor.animation[name]:setRate(actorData.animationRate);
        end
    end
    
    if(actorData.disableInterp and actorData.animation) then
        for name, anim in pairs(actorData.animation) do
            actor.animation[name]:setInterpolate(not actorData.disableInterp);
        end
    end
    
    if(actorData.actorGroup) then
        -- assign the actor to the group
        self.actorGroup[actor.name] = actorData.actorGroup       

    end
    
--  if(actorData.offsetBones) then
--      actor.offsetBones = shallowcopy(actorData.offsetBones);
--  end

    self:showDebug(actor, actorData)
    
    self:addAnimationDriver(actor)
    
    self:addFlipbookSampler(actor, actorData)
    
    self:createAttachment(actor, actorData)
        
    for name, value in pairs(actorData) do
        if(actor[name] == nil) then
            actor[name] = value;
        end
    end

    -- Call any post process functions
    if actorData.postProcess then
        for _, postFn in ipairs(actorData.postProcess) do
            postFn(actor)
        end
    end
    
    -- Dump actor tables to file in /dist folder
    --[[
    local matStr = "actors = "..Serpent.dumppretty(actor);
    local file,err = io.open( actor.name..".lua", "w" )
    file:write(matStr);
    file:close();
    --]]

    Events.send("ActorManager", "onActorCreated", actor)
end


-- Where boneDef is a table with actorName and bonePath
function ActorManager:getBone(boneDef)
    local actor = self:getActor(boneDef.actorName, false)
    if (actor and boneDef.bonePath) then
        return actor:getBone(boneDef.bonePath)
    end
    return actor
end

function ActorManager:getTargetBone(actorName)
    local actor = self:getActor(actorName)
    if (actor) then
        return actor.targetBone
    end
    return nil
end

-- Return the world position of the actor's target bone if
-- one is specified, nil otherwise
function ActorManager:getActorTrackingPos(actorName)
    local pos = nil
    local targetBone = self:getTargetBone(actorName)
    if (targetBone) then
        pos = targetBone:getWorld().position
    end
    return pos
end

function ActorManager:getActorsInGroup(groupName)
    local actors = {}
    for actorName, group in pairs(self.actorGroup) do
        if groupName == group then
            local actor = self:getActor(actorName, false)
            if actor then
                table.insert(actors, actor)
            end
        end
    end
    return actors
end

function ActorManager:addAnimationDriver(actor)
    if (actor.animDriver == nil) then
        actor.animDriver = actor:addComponent("animDriver", AnimationDriver.new())
    end
end

function ActorManager:addFlipbookSampler(actor, actorData)
    actor.fbdrivers = actor.fbdrivers or {}
    
    if(actorData.fbdrivers) then
        for name, fbarray in pairs(actorData.fbdrivers) do
            
            actor.fbdrivers[name] = actor.fbdrivers[name] or {}
            
            local animationData = actorData.animation[name] or {}
            local startFrame = animationData[2] and animationData[2] or 1;
            
            for material, data in pairs(fbarray) do
                
                if(data ~= nil and (actor.fbdrivers[name] == nil or actor.fbdrivers[name][material] == nil)) then
                    data.mat = material;
                    data.frameStartOffset = math.max(0, startFrame-1);
                    if(not self.enableNativeFlipbookDriver) then
                    	actor.fbdrivers[name][material] = actor:addComponent(name..material.."FlipbookDriver", FlipbookDriver:new(name, data))
                    else
	                    data.animName = name
	                    actor.fbdrivers[name][material] = actor:addFlipbookDriver(data)
	                    
	                     -- add callback method
	                    actor.fbdrivers[name][material].onFinished = data.onFinished
                    end
                end
            end
        end
    end
end

function ActorManager:createAttachment(actor, actorData)
    if(actorData.attachTo) then
        local parent = self:getActor(actorData.attachTo.parentName, true)
        
        -- life cycle must be bound to parent now
        actor.attachedTo = parent
        actor.attachedToName = parent.name
        self.scene.actors[actor.name].attachments = self.scene.actors[actor.name].attachments  or {}
        table.insert(self.scene.actors[actor.name].attachments, parent)

        if (actorData.attachTo.targetBone and parent.findBonePath) then
            actorData.attachTo.bonePath = actorData.attachTo.bonePath or parent:findBonePath(actorData.attachTo.targetBone)
            actor.attachmentBone = actorData.attachTo.bonePath
        end

        if (actorData.attachTo.bonePath) then
            Log.message("Attaching "..actor.name.." to "..parent.name.." at "..actorData.attachTo.bonePath)

            actor.attachmentBone = actorData.attachTo.bonePath
            parent:addAttachment(actorData.attachTo.bonePath, actor)
        end
    end
end
--[[
function ActorManager:printDebug(actor, actorData)

    if(actorData.dumpBones) then
        local trs = actor:listTransforms();
        Log.message("Logging "..actor.name.." hierarchy")
        if(trs ~= nil) then
            for _, tr in ipairs(trs) do
                Log.message("\t"..tr);
            end
        end
    end
end
]]
function ActorManager:showDebug(actor, actorData)
    if(self.showBones and not actorData.hideDebug) then -- 
        local isAxis = string.find(actor.name, "Axis_") ~= nil;
        if(actor.listTransforms ~= nil) then
            local trs = actor:listTransforms();
            if(trs ~= nil and not isAxis) then
                for _, tr in ipairs(trs) do
                    Log.message("Creating axis_"..tr)
                    local trackHistory = string.find(tr, "target") ~= nil
                    actor:addComponent("debugAxis_"..tr, Axis.new(Axis, {attachTo = {parentName=actor.name, bonePath=tr}, track=trackHistory}))
                end
            end
        end
    end
end

function ActorManager:destroyActor(name)
    if (name and self.scene.actors[name] and status(self.scene.actors[name]) ~= nil) then
        if(not self.scene.actors[name].attachedTo) then
            self.scene.actors[name]:destroy()
        else
        	if(self.scene.actors[name].attachedTo.removeAttachment)then
            	self.scene.actors[name].attachedTo:removeAttachment(self.scene.actors[name])
            end
            self.scene.actors[name]:destroy();
        end
        self.scene.actors[name] = nil

        Log.message("ActorManager: destroyed "..name)
        Events.send("ActorManager", "onActorDestroyed", name)
    end
end
