
AudioPlayer = {
    enabled = true,  -- turn this off to turn off all audio for the show

    emitter = nil,
    sounds = {},
    segmentEventData = nil,
    segmentTimer = 0,
}
AudioPlayer.__index = AudioPlayer

local defaultType = { enabled = true, looping = false, volume = 0.8 }


-- Compute the normalized pitch of the camera (-1 pointing down, 0 pointing at the horizon, 1 pointing up)
local function getCameraPitch()
    local forward = Vector.rotateBy(Vector.forward, Global.scene.mainCamera:getWorld().rotation)
    forward:normalize()
    return forward.y
end

function AudioPlayer.new(screenplay)
    if AudioPlayer.enabled then
        screenplay = screenplay or {}
        screenplay.types = screenplay.types or {}
        screenplay.sounds = screenplay.sounds or {}
        screenplay.segments = screenplay.segments or {}

        local instance = { screenplay = screenplay }
        setmetatable(instance, AudioPlayer)
        instance:_init()

        return instance
    else
        -- Return a dummy object that does nothing, as audio is disabled
        return {
            destroy = function() end,
            update = function() end,
            renderInfo = function() end,
            prefetchSegment = function() end,
        }
    end
end

function AudioPlayer:_init()
    self.emitter = AudioEmitter.create()
    self.frameRate = self.screenplay.frameRate or 60

    Events.listen(self, "segmentPlaying")
end

function AudioPlayer:destroy()
    if self.emitter then
        self.emitter:destroy()
        self.emitter = nil
    end
    Events.stopListening(self, "segmentPlaying")
end

function AudioPlayer:update()
    local dt = Time.dt()
    PositionalSound:clearCache()
    self:_processEvents(dt)
    self:_updateMasterVolumes(dt)
    self:_updateSounds(dt)
end

function AudioPlayer:renderInfo(nextLine, fontStyle)
    if self.sounds then
        nextLine()
        Renderer.drawText(10, nextLine(), fontStyle, "Sounds: ")
        local masterVolumes = ""
        for name, type in pairs(self.screenplay.types) do
            masterVolumes = string.format("%s [ %s: vol = %.2f", masterVolumes, name, type.masterVolume)
            if type.masterVolumeGoal then
                masterVolumes = string.format("%s, goal = %.2f", masterVolumes, type.masterVolumeGoal)
            end
            masterVolumes = masterVolumes .. "] "
        end
        Renderer.drawText(10, nextLine(), fontStyle, masterVolumes)

        for name, sound in pairs(self.sounds) do
            local state = sound:getPlayState()
            local status = state == PS_STOPPED and "S" or state == PS_PAUSED and "P" or ""
            local tracking = ""
            if sound.control.trackingTargets then tracking = string.format(" [trackVol = %.2f, targets = %s]", sound.control.trackingVolumeGoal, table.tostring(sound.control.trackingTargets)) end
            if sound:getLooping() then status = status .. " L" end
            Renderer.drawText(10, nextLine(), fontStyle, string.format("  %s[%0.2f%s] : vol = %.2f %s", name, sound.control:getSeekPosition(), status, sound:getVolume(), tracking))
        end
        nextLine()
    end
end

function AudioPlayer:prefetchSegment(segmentName, resourceGroup)
    local segmentData = self.screenplay.segments[segmentName]
    if not segmentData then return end

    if segmentData.events then
        for _, event in ipairs(segmentData.events) do
            self:_prefetchSound(event.play or event.fade, resourceGroup)
        end
    end
    if segmentData.checkpoint then
        for _, event in ipairs(segmentData.checkpoint) do
            self:_prefetchSound(event.play or event.fade, resourceGroup)
        end
    end
end

function AudioPlayer:_prefetchSound(name, resourceGroup)
    local soundDef = self.screenplay.sounds[name]
    if soundDef then
        local type = (soundDef.type and self.screenplay.types[soundDef.type]) or defaultType
        if type.enabled and soundDef.uri then
            resourceGroup:addAudio(soundDef.uri)
        end
    end
end

function AudioPlayer:segmentPlaying(segmentName)
    local segmentData = self.screenplay.segments[segmentName]

    local pendingEvents = {}
    if segmentData and segmentData.events then
        for _, event in ipairs(segmentData.events) do
            table.insert(pendingEvents, deepcopy(event))
        end
    end

    local segment = SegmentManager.segments[segmentName]
    local checkpoint = segment and segment.checkpointTriggered
    if checkpoint and segmentData and segmentData.checkpoint then
        for _, event in ipairs(segmentData.checkpoint) do
            table.insert(pendingEvents, deepcopy(event))
        end
    end

    self.pendingEvents = pendingEvents
    self.segmentStartFrame = ((segmentData and segmentData.startFrame) or 0) + (self.screenplay.frameLatency or 0)
    self.segmentTimer = 0
end

-- Manually fade a sound outside of the script (used for GWI)
function AudioPlayer:fadeSound(soundName, duration, fromVolume, toVolume, seekIndex)
    local sound = self:_getSound(soundName)
    if sound then
        if seekIndex then
            sound.control:setSeekPosition(seekIndex)
        end
        sound.control:fade(duration, fromVolume, toVolume)

    end
end

-- Manually play a sound outside of the script (used for GWI)
function AudioPlayer:playSound(soundName, volume)
    local sound = self:_getSound(soundName)
    if sound then
        sound.control:setVolume(volume)
        sound:play()
    end
end

function AudioPlayer:_processEvents(dt)
    if not self.pendingEvents or #self.pendingEvents <= 0 then return end

    self.segmentTimer = self.segmentTimer + dt
    local frame =  self.segmentStartFrame + self.segmentTimer * self.screenplay.frameRate

    local stillPending = {}
    for _, event in ipairs(self.pendingEvents) do
        local gwiHackedOffset = ((event.includeGlobalIntroAnimDelay and Global.introAnimDelay) or 0) *  self.screenplay.frameRate
        local eventFrame = (event.frame or 0) + (event.frameOffset or 0) + gwiHackedOffset
        if eventFrame <= frame then
            if event.play then
                self:_playSound(event)
            elseif event.fade then
                self:_fadeSound(event)
            elseif event.stop then
                self:_stopSound(event)
            elseif event.masterVolume then
                self:_masterVolume(event)
            else
                Log.message("Unknown sound event: " .. table.tostring(event))
            end
        else
            table.insert(stillPending, event)
        end
    end
    self.pendingEvents = stillPending
end

function AudioPlayer:_updateMasterVolumes(dt)
    local cameraPitch
    for name, type in pairs(self.screenplay.types) do
        if type.enabled then
            if type.verticalAttenuationCurve then
                -- Track any vertical attenuations based on the camera pitch
                cameraPitch = cameraPitch or getCameraPitch()
                local volume = self:_evaluateVerticalAttenuation(cameraPitch, type.verticalAttenuationCurve)
                type.masterVolumeGoal = volume
            end

            -- Track any actors for this master volume and attenuate according to their heading relative to the camera
            if type.masterTrackingTargets and type.masterTrackingCurve then
                local volume, pan = PositionalSound:trackingVolumeFor(type.masterTrackingTargets, type.masterTrackingCurve)
                -- TODO: hook up a pan goal?
                if volume then
                    type.masterVolumeGoal = volume
                end
            end

            -- Adjust the master volume towards its goal
            if type.masterVolumeGoal then
                local trackingSpeed = (type.masterTrackingCurve and type.masterTrackingCurve.speed) or
                                      (type.verticalAttenuationCurve and type.verticalAttenuationCurve.speed)
                if trackingSpeed then
                    type.masterVolume = type.masterVolume or type.masterVolumeGoal
                    type.masterVolume = math.lerp(type.masterVolume, type.masterVolumeGoal, dt * trackingSpeed)
                else
                    type.masterVolume = type.masterVolumeGoal
                end
            end
        end
    end
end

function AudioPlayer:_getSound(name)
    local sound = self.sounds[name]
    if not sound then
        local soundDef = self.screenplay.sounds[name]
        if not soundDef then
            Log.error("Undefined sound: " .. tostring(name))
            return nil
        end

        local type = (soundDef.type and self.screenplay.types[soundDef.type]) or defaultType
        if not type.enabled then return nil end

        -- soundName, uriPath, preCache, startTimeMillis, stopTimeMillis
        sound = self.emitter:addSound(name, soundDef.uri, false, 0, -1)
        if sound then
            sound.name = name
            sound.control = SoundControl.new(sound)
            sound.control:setType(type)
            sound.control.autorelease = soundDef.autorelease or type.autorelease or false
            sound.control.minVolume = soundDef.minVolume or type.minVolume or nil
            sound.control.maxVolume = soundDef.maxVolume or type.maxVolume or nil
            sound:setLooping(soundDef.looping or type.looping or false)
            sound.control:track(soundDef.trackingTargets or type.trackingTargets or nil,
                                soundDef.trackingCurve or type.trackingCurve or nil)

            local volume = soundDef.volume or (type and type.volume)
            if volume then sound.control:setVolume(volume) end

            self.sounds[name] = sound
        else
            Log.error("Unable to create sound for " .. table.tostring(soundDef))
        end
    end
    return sound
end

function AudioPlayer:_updateSoundFromEvent(sound, event)
    local control = sound.control
    if event.autorelease then
        control.autorelease = true
    end
    if event.volume then
        control:setVolume(event.volume)
    end
    if event.trackingTargets or event.trackingCurve then
        control:track(event.trackingTargets, event.trackingCurve)
    end
end

function AudioPlayer:_releaseSound(sound)
    if sound and sound.name then
        self.emitter:removeSound(sound.name)
        self.sounds[sound.name] = nil
    end
end

function AudioPlayer:_playSound(event)
    local sound = self:_getSound(event.play)
    if sound then
        self:_updateSoundFromEvent(sound, event)
        sound:play()
    end
end

function AudioPlayer:_fadeSound(event)
    local sound = self:_getSound(event.fade)
    if sound then
        self:_updateSoundFromEvent(sound, event)
        sound.control:fade(event.duration, event.from, event.to)
    end
end

function AudioPlayer:_stopSound(event)
    local sound = self:_getSound(event.stop)
    if sound then
        sound:stop()
        if sound.control.autorelease then
            self:_releaseSound(sound)
        end
    end
end

function AudioPlayer:_masterVolume(event)
    local type = self.screenplay.types[event.masterVolume]
    if type then
        type.masterTrackingTargets = event.masterTrackingTargets
        type.masterTrackingCurve = event.masterTrackingCurve
        if not type.masterTrackingTargets then
            type.masterVolumeGoal = nil
        end
    end
end

function AudioPlayer:_updateSounds(dt)
    for _, sound in pairs(self.sounds) do
        sound.control:update(dt)
        if sound.release then
            self:_releaseSound(sound)
        end
    end
end

function AudioPlayer:_evaluateVerticalAttenuation(cameraPitch, curve)
    -- walk the attenuation table and find where our pitch fits
    local upperEntry, lowerEntry
    for _, entry in ipairs(curve) do
        if cameraPitch < entry.pitch then
            upperEntry = upperEntry or entry
        else
            lowerEntry = entry
        end
    end

    -- If we are lower than the lowest or higher than the highest use the same for both
    lowerEntry = lowerEntry or upperEntry
    upperEntry = upperEntry or lowerEntry

    -- lerp between the upper and lower to get the volume and panning balance
    local t
    local range = upperEntry.pitch - lowerEntry.pitch
    if range ~= 0 then
        t = (cameraPitch - lowerEntry.pitch) / range
    end

    local volume
    if t then
        volume = math.lerp(lowerEntry.volume, upperEntry.volume, t)
    else
        volume = upperEntry.volume
    end

    return volume
end
