
PositionalSound = {
    headingCache = {},
}

function PositionalSound:clearCache()
    self.headingCache = {}
end

function PositionalSound:trackingVolumeFor(targets, positionalSoundCurve, targetVolumes)
    local volume = 0
    local pan = 0
    local sumHeadings = 0
    local anyTracked = false
    local closestHeading

    for _, target in ipairs(targets) do
        local targetName = self:_targetName(target)
        local currentTargetVolume = (targetVolumes and targetVolumes[targetName] and targetVolumes[targetName].currentVolume) or 1
        local targetVolume, targetPan, heading = self:_getVolumeFor(target, positionalSoundCurve)
        if targetVolume then
            anyTracked = true
            if targets.closest then
                -- just take the closest actor for its volume
                local absHeading = math.abs(heading)
                if not closestHeading or absHeading < closestHeading then
                    volume = targetVolume
                    pan = targetPan
                    closestHeading = absHeading
                end
            else
                -- do a weighted average of the volumes
                local weightedHeading = 1 - math.abs(heading)
                sumHeadings = sumHeadings + weightedHeading
                volume = volume + (targetVolume * currentTargetVolume * weightedHeading)
                pan = pan + (targetPan * weightedHeading)
            end
        end
    end

    if anyTracked then
        if not targets.closest then
            volume = volume / sumHeadings
            pan = pan / sumHeadings
        end
    else
        volume = 0
    end

    return volume, pan
end

function PositionalSound:_targetName(targetDef)
    if type(targetDef) == "table" then
        -- use the first actor name from the target table
        return targetDef[1].actor
    end
    return targetDef
end

function PositionalSound:_getVolumeFor(target, positionalSoundCurve)
    local heading = self:_getCameraHeadingFor(target)
    if not heading then return nil, nil, nil end

    return self:_evaluatePositionalSoundCurve(positionalSoundCurve, heading)
end

function PositionalSound:_getCameraHeadingFor(target)
    local actorName = self:_targetName(target)

    -- Returned cached heading if we have already computed it
    local heading = self.headingCache[actorName]
    if heading then
        return heading
    end

    local position = self:_getTargetPosition(target)
    if not position then
--        Log.error("Unable to get tracking position for actor: " .. actorName)
        return nil
    end

    local camera        = Global.scene.mainCamera
    local cameraParent  = camera.parent or camera
    local cameraWorld 	= camera:getWorld()
    local cameraForward = cameraWorld:forward()

    -- Direction to the actor
    local targetForward = position - cameraParent:getWorld().position
    targetForward.w 	= 0
    targetForward:normalize()

    -- Camera right without roll
    local cameraRight 	= Vector.cross(cameraForward, Vector.up)
    cameraRight.w		= 0
    cameraRight:normalize()

    local forwardDot 	= Vector.dot(targetForward, cameraForward)
    local rightDot 		= Vector.dot(targetForward, cameraRight)

    -- Compute the heading with:
    -- heading =   0 -> camera is looking right at the target
    -- heading = 0.5 -> target is directly to the left of the camera
    -- heading =   1 -> target is directly behind the camera
    -- negative of these are on the left side
    heading = (1 - forwardDot) * 0.5
    if rightDot < 0 then heading = -heading end

    self.headingCache[actorName] = heading
    return heading
end

function PositionalSound:_getTargetPosition(target)
    if type(target) == "table" then
        -- handle the complicated case where we need to check multiple actors for a given target
        for _, actorInfo in ipairs(target) do
            local pos = self:_getActorPosition(actorInfo.actor, actorInfo.bone)
            if pos then return pos end
        end
    else
        return self:_getActorPosition(target)
    end
end

function PositionalSound:_getActorPosition(actorName, boneName)
    local actor = ActorManager:getActor(actorName, false)
    if actor then
        local bone = actor[boneName or "targetBone"]
        if bone then
            return bone:getWorld().position
        end
    end
end

function PositionalSound:_evaluatePositionalSoundCurve(curve, heading)
    if not curve or #curve == 0 then return end

    local volume, pan
    local lookupHeading = heading
    if curve.symmetric then
        lookupHeading = math.abs(heading)
    end

    -- walk the curve data table and find where our heading fits
    local upperEntry, lowerEntry
    for _, entry in ipairs(curve) do
        if lookupHeading < entry.heading then
            upperEntry = upperEntry or entry
        else
            lowerEntry = entry
        end
    end

    -- If we are lower than the lowest or higher than the highest use the same for both
    lowerEntry = lowerEntry or upperEntry
    upperEntry = upperEntry or lowerEntry

    -- lerp between the upper and lower to get the volume and panning balance
    local t
    local range = upperEntry.heading - lowerEntry.heading
    if range ~= 0 then
        t = (lookupHeading - lowerEntry.heading) / range
    end
    if t then
        pan = math.lerp(lowerEntry.pan, upperEntry.pan, t)
        volume = math.lerp(lowerEntry.volume, upperEntry.volume, t)
    else
        pan = upperEntry.pan
        volume = upperEntry.volume
    end
    if heading < 0 then pan = -pan end

    return volume, pan, heading
end
