
SoundControl = {
    currentVolume = 0,

    tracking = nil,
    trackingVolume = nil,
    trackingVolumeGoal = nil,
    trackingPan = nil,
    trackingPanGoal = nil,

    fading = false,
    startTimer = 0,
    fadeTimer = 1.1,
    fadeTime = 1,

    volumeStart = 0,
    volumeEnd = 0,
    time = 0,
    autorelease = true,
    timePlaying = 0,

    -- Actual sound values used just to cache the values
    -- so we don't have to call native functions unless the
    -- final volume or pan actually changes
    _soundVolume = 0,
    _soundPan = 0,

}
SoundControl.__index = SoundControl

function SoundControl.new(sound)
    local instance = { sound = sound }
    setmetatable(instance, SoundControl)

    instance._soundVolume = instance.currentVolume
    sound:setVolume(instance._soundVolume)

    return instance
end

function SoundControl:update(dt)
    local sound = self.sound
    local isPlaying = sound:isPlaying()
    local startPlaying = false

    if isPlaying and not Time.isPaused() then
        self.timePlaying = self.timePlaying + dt
    end

    if self.fading or self.trackingVolumes then
        self:_updateFade(dt)
        if not isPlaying then startPlaying = true end
    end

    if self.trackingTargets then
        self:_updateTrackingVolume(dt)
    end
    local masterVolume = (self.type and self.type.masterVolume) or 1
    local volume = self.currentVolume * (self.trackingVolume or 1) * masterVolume
    if self.minVolume then volume = math.max(self.minVolume, volume) end
    if self.maxVolume then volume = math.min(self.maxVolume, volume) end

    if volume ~= self._soundVolume then
        self._soundVolume = volume
        self.sound:setVolume(volume)
    end
    local pan = self.trackingPan or 0
    if pan ~= self._soundPan then
        self._soundPan = pan
        self.sound:setPanBalance(pan)
    end

    if startPlaying then
        sound:play()
    elseif not isPlaying and self.autorelease then
        sound.release = true
    end
end

function SoundControl:setType(type)
    self.type = type
end

function SoundControl:setVolume(newVolume)
    if (self.fading) then Log.error("SoundControl: don't set the volume on a fading volume control: " .. self.sound.name .. tostring(debug.traceback())) end
    if type(newVolume) == "table" then
        if self.trackingVolumes then
            for targetName, volume in pairs(newVolume) do
                local targetVolume = self.trackingVolumes[targetName]
                if targetVolume then
                    targetVolume.currentVolume = volume
                end
            end
        else
            Log.error("Setting a single volume on a track that has multiple volumes: " .. tostring(self.sound.name))
            return
        end
    else
        self.currentVolume = newVolume
    end
end

function SoundControl:fade(duration, from, to)
    if self.trackingVolumes then
        if type(from) ~= "table" and type(to) ~= "table" then
            Log.message("Trying to fade a multi-volume sound with single volumes: " .. tostring(self.sound.name))
            assert(false)
        end

        for target, volume in pairs(self.trackingVolumes) do
            local fromVol = from and from[target]
            local toVol = to and to[target]
            if fromVol or toVol then
                volume.fading = true
                volume.fadeTimer = 0
                volume.fadeDuration = duration
                volume.fadeFromVolume = fromVol or volume.currentVolume
                volume.fadeToVolume = toVol or 0
            end
        end
    else
        self.fading = true
        self.fadeTimer = 0
        self.fadeDuration = duration
        self.fadeFromVolume = from or self.currentVolume
        self.fadeToVolume = to or 0
    end
end

function SoundControl:track(trackingTargets, trackingCurve)
    self.trackingTargets = trackingTargets or self.trackingTargets
    self.trackingCurve = trackingCurve or self.trackingCurve

    if trackingTargets then
        if #trackingTargets > 1 then
            local trackingVolumes = {}
            local volume = self.currentVolume
            for _, target in ipairs(trackingTargets) do
                local targetName = target
                if type(target) == "table" then
                    -- handle the complicated case where we need to check multiple actors for a given target
                    targetName = target[1].actor
                end
--                Log.message("tracking multiple targets: " .. tostring(self.sound.name) .. " -> " .. targetName)
                trackingVolumes[targetName] = { fading = false, currentVolume = volume }
            end
            self.trackingVolumes = trackingVolumes
            self.currentVolume = 1
        else
            self.trackingVolume = nil
        end
    else
        self.trackingVolumes = nil
    end
end

-- To make up for the unreliable results from sound:getSeekPosition(),
-- we try to give a better estimate here.
function SoundControl:getSeekPosition()
    if self.timePlaying then
        return self.timePlaying % self.sound:getDurationSeconds()
    end
    return 0
end

function SoundControl:setSeekPosition(seconds)
    self.sound:setSeekPosition(seconds * 1000)
    self.timePlaying = seconds
end

function SoundControl:_updateFade(dt)
    if self.trackingVolumes then
        local allReleased = true
        for _, volume in pairs(self.trackingVolumes) do
            if volume.fading then
                volume.fadeTimer = volume.fadeTimer + (dt / volume.fadeDuration)
                volume.currentVolume = math.lerp(volume.fadeFromVolume, volume.fadeToVolume, volume.fadeTimer)
                if (volume.fadeTimer > 1) then
                    volume.fading = false
                    if volume.currentVolume < 0.001 and self.autorelease then
                        volume.released = true
                    end
                end
            end
            allReleased = allReleased and volume.released
        end
        if allReleased then
            self.sound:stop()
            self.sound.release = true
        end
    else
        self.fadeTimer = self.fadeTimer + (dt / self.fadeDuration)
        self.currentVolume = math.lerp(self.fadeFromVolume, self.fadeToVolume, self.fadeTimer)
        if (self.fadeTimer > 1) then
            self.fading = false
            if self.currentVolume < 0.001 and self.autorelease then
                self.sound:stop()
                self.sound.release = true
                return
            end
        end
    end
end

function SoundControl:_updateTrackingVolume(dt)
    self.trackingVolumeGoal, self.trackingPanGoal = PositionalSound:trackingVolumeFor(self.trackingTargets, self.trackingCurve, self.trackingVolumes)

    local trackingSpeed = self.trackingCurve.speed
    if self.trackingVolumeGoal then
        if trackingSpeed then
            local easeInOut = 1.7
            self.trackingVolume = self.trackingVolume or self.trackingVolumeGoal
            if self.trackingVolume > self.trackingVolumeGoal then
                easeInOut = 1 / easeInOut
            end
            self.trackingVolume = math.lerp(self.trackingVolume, self.trackingVolumeGoal, dt * trackingSpeed * easeInOut)
        else
            self.trackingVolume = self.trackingVolumeGoal
        end
    end

    if self.trackingPanGoal then
        if trackingSpeed then
            self.trackingPan = self.trackingPan or self.trackingPanGoal
            self.trackingPan = math.lerp(self.trackingPan, self.trackingPanGoal, dt * trackingSpeed)
        else
            self.trackingPan = self.trackingPanGoal
        end
    end
end
