
local ActorBonePaths = {}

local techniqueToFlipbookSampler = {
    flipbookVideoColor = {"u_FlipbookVideoTex"},
    flipbookVideoColorAlpha = {"u_FlipbookVideoTex", "u_FlipbookVideoAlpha"},
    flipbookColor = {"u_FlipbookTex"},
    flipbookColorAlpha = {"u_FlipbookTex"},
}

local function FindBonePath(paths, leafBoneName)
    for _, path in ipairs(paths) do
        if path == leafBoneName or string.startsWith(path, leafBoneName .. "|") then
            return path
        end
    end
    return nil
end

local function BakeNamedBones(actor, actorData, bakeData)
    if actorData.namedBones then
        bakeData.bonePaths = {}
        for name, leafBoneName in pairs(actorData.namedBones) do
            bakeData.bonePaths[name] = FindBonePath(ActorBonePaths[actor.name], leafBoneName)
        end
    end
end

local function BakeAttachments(actorData, bakeData)
    local attach = actorData.attachTo
    if attach then
        bakeData.attachTo = deepcopy(attach)
        if attach.parentName and attach.targetBone then
            bakeData.attachTo.bonePath = FindBonePath(ActorBonePaths[attach.parentName], attach.targetBone)
        end
    end
end

local function BakeFrameEvents(actor, actorData, bakeData)
    if actor.transforms and actor.nodeProps then
        local frameEvents = {}
        for i, transform in ipairs(actor.transforms) do
            local propString = actor.nodeProps[i]
            if propString and propString ~= "nil" then
                local props = loadstring("return " .. propString)()
                for event, value in pairs(props) do
                    local eventName = string.match(event, "frameEvent_(%a+)")
                    if eventName then
                        frameEvents[value]= { name = eventName, data = transform }
                    end
                end
            end
        end
        bakeData.frameEvents = frameEvents
    end
end

local function BakeMaterial(materialName, actor, actorData, bakeData)
    -- grab the material already in the actor definition if it exists
    local matData = {}
    local index
    if actorData.materials then
        index = table.find_index(actorData.materials, function(self) return self.name == materialName; end );
        matData = index and actorData.materials[index] or {}
    end

    local matInstance = actor:findMaterial(materialName)
    local technique =  matInstance:getTechniqueName()

    -- Update the material data's techique
    if not matData.dontOverrideTechnique then
        if technique ~= "default" then
            matData.technique = matData.technique or technique
        end
    end

      -- process material attributes
    local fbMaterial = false
    local props = matInstance:getPropertiesTable()
    for key, value in pairs(props) do
        if actorData.dumpMaterials then Log.message("\t\tproperty:"..tostring(key).."="..tostring(value)) end
    end
	
    matData.renderGroup = matData.renderGroup or props["renderGroup"]
    
     -- fbStart for flipbook frame offsets
    if props["fbStart"] and actorData.fbdrivers then
        fbMaterial = true
        for _, fbDriver in pairs(actorData.fbdrivers) do
        	
        	if(actorData.generateFlipbooks == false) then
        		fbDriver[materialName] = fbDriver[materialName]
        	else
            	fbDriver[materialName] 				= fbDriver[materialName] or {}
            end
			
			if(fbDriver[materialName]) then
                -- we're setting frameOffset to be the inverse of fbStart
                fbDriver[materialName].frameOffset 	= fbDriver[materialName].frameOffset or -(props["fbStart"] or 1) + 1
                fbDriver[materialName].fbStart = props.fbStart
                
                -- setup looping
                if(props["loopFlipbook"]) then  props["loopFlipbook"]=FBWrapMode.loop else  props["loopFlipbook"]=FBWrapMode.clamp end
                
                fbDriver[materialName].wrapMode = fbDriver[materialName].wrapMode or props["loopFlipbook"]
                fbDriver[materialName].frameStart = fbDriver[materialName].frameStart or props["loopStart"]
                fbDriver[materialName].frameEnd = fbDriver[materialName].frameEnd or props["loopEnd"]
                
                local exposureCount = fbDriver[materialName].exposureCount or 1
                if(fbDriver[materialName].frameStart) then
                	fbDriver[materialName].frameStart = math.max(0, fbDriver[materialName].frameStart-1)*exposureCount
                end
                if(fbDriver[materialName].frameEnd) then
                	fbDriver[materialName].frameEnd = math.max(0, fbDriver[materialName].frameEnd-1)*exposureCount
                end
            end
        end
    end

    -- process material textures
    local texCount = matInstance:getTextureCount()
    for i = 0, texCount - 1, 1 do
        local texName 	= matInstance:getTextureName(i)
        local texType	= matInstance:getTextureType(i)
        local texUri 	= "Invalid URI"

        if texType == MTT_TEXTURE then
            texUri = string.lower(matInstance:getTextureUri(i))
        elseif texType == MTT_FLIPBOOK then
            texUri = string.lower(matInstance:getFlipbookUri(i))
        else
            Log.message(">>>> BAD Texture URI = " .. texUri)
        end
        local set = index and matData.textures and matData.textures[texName] -- if the texture is overridden already
        if not set and
                actorData.defaultMaterial and
                actorData.defaultMaterial.textures and
                actorData.defaultMaterial.textures[texName] then

            matData.textures[texName] = actorData.defaultMaterial.textures[texName]
            set	= true
        end

        local index = i > 0 and i or ""
        local samplerName = (texName == "u_diffuseMap") and {"u_MainTex"} or {"u_MainTex"..index}

        if not set then
            if texType == MTT_TEXTURE then
                matData.textures = matData.textures or {}
                matData.textures[samplerName[1]] = string.lower(texUri)
            elseif texType == MTT_FLIPBOOK then
                matData.flipbooks = matData.flipbooks or {}
                samplerName = techniqueToFlipbookSampler[technique] or {"u_FlipbookTex"}
                for _, sampler in ipairs(samplerName) do
                    matData.flipbooks[sampler] = string.lower(texUri)
                end
            end
        end

        if fbMaterial and actorData.fbdrivers then
            local fbUri = texUri
            for fbAnim, fbMatTable in pairs(actorData.fbdrivers) do
                if fbMatTable[materialName] then
                    fbMatTable[materialName].uri 		= fbMatTable[materialName].uri or fbUri
                    fbMatTable[materialName].bindings 	= fbMatTable[materialName].bindings or   techniqueToFlipbookSampler[technique]  or {"u_FlipbookTex",}
                end
            end
        end
    end

    -- if the material is not in the actor data already, add it
    if not index then
        actorData.materials = actorData.materials or {}
        matData.name = materialName
        table.insert(actorData.materials, matData)
    end

    -- copy any properties from the default actor material that were not overridden
    if actorData.defaultMaterial then
        for name, value in pairs(actorData.defaultMaterial) do
            matData[name] = matData[name] or value
        end
    end

    -- set some global shader defaults in case they weren't set yet
    if not matData.parameters or (matData.parameters and not matData.parameters.u_MainTex_ST) then
        matData.parameters = matData.parameters or {}
        matData.parameters.u_MainTex_ST = {1.00,1.00,0.00,0.00}
    end

    -- Flipbook shaders have a u_Alpha uniform. This is set to 1 by default
    if not matData.parameters or (matData.parameters and not matData.parameters[techniqueToFlipbookSampler[technique]]) then
        matData.parameters = matData.parameters or {}
        matData.parameters["u_Alpha"] = 1
    end

    -- make sure all asset paths are lower-case
    if matData.textures then
        for tex, path in pairs(matData.textures) do
            matData.textures[tex] = string.lower(path)
        end
    end

end

local function BakeMaterials(actor, actorData, bakeData)
    -- import all the materials to the actor data so we can override their properties
    local materialNames = actor.listMaterials and actor:listMaterials()
    if materialNames then
        for _, materialName in ipairs(materialNames) do
            BakeMaterial(materialName, actor, actorData, bakeData)
        end
    end
    bakeData.materials = deepcopy(actorData.materials)
    bakeData.fbdrivers = deepcopy(actorData.fbdrivers)
end

local function BakeActors(actors, bakeData, scene)
    for actorName, actorData in pairs(actors) do
        Log.message("Baking " .. actorName)
        local actor
        if actorData.model then
            actor = scene:createActor(actorName, actorData, true)
            actor.nodeProps = actor:listProperties()
            actor.transforms = actor:listTransforms()
            ActorBonePaths[actorName] = actor.transforms
        else
            actor = scene:createTransform(actorName)
            actor.materials = {}
        end
        local bakedActor = bakeData[actorName] or {}
        bakeData[actorName] = bakedActor

        BakeNamedBones(actor, actorData, bakedActor)
        BakeMaterials(actor, actorData, bakedActor)
--        BakeFrameEvents(actor, actorData, bakedActor)

        actor:destroy()
    end

    for actorName, actorData in pairs(actors) do
        local bakedActor = bakeData[actorName] or {}
        BakeAttachments(actorData, bakedActor)
    end
end

local function BakeActorData()
    local bakeData = {}
    local scene = Scene.create()
    BakeActors(SceneData.actors, bakeData, scene)
    BakeActors(LaunchData.actors, bakeData, scene)

    scene:destroy()

    local file, err = io.open(Project.rawPersistSceneDir .. "baked_actor.lua", "w" )
    if not err and file then
        file:write("return " .. Serpent.dumppretty(bakeData))
        file:close()
    end
end

-- UI for run time baking of actor data
BakeTool = {}
BakeTool.__index = BakeTool

function BakeTool:new(params)
    local window = GUI.window(GUIAligned(ALIGN_CENTER), GUISize(400, 200), "Bake")
    window:anchor(ANCHOR_BOTTOM)

    local buttonCallbacks = {
        {
            {name = "Bake Actors", callback = BakeActorData },
        },
    }

    for _, rowButtons in ipairs(buttonCallbacks) do
        local row = GUI.frame(window, GUIAligned(ALIGN_TOP_LEFT), GUISize(SZ_FILL, SZ_WRAP), "widget.default")
        local buttonSize = GUIRelative(1 / #rowButtons, 0.95)
        for _, callbackData in ipairs(rowButtons) do
            local button = GUI.button(row, GUIAligned(ALIGN_BOTTOM_LEFT), buttonSize, callbackData.name)
            button:dock(DOCK_LEFT)
            button.onButtonDown = callbackData.callback
        end
        row:dock(DOCK_TOP)
    end

    return window
end
