
--[[
    blendTimeArray: This is an array of time points which define a function that changes the value of the the lockangle and innerlockangle over the time of the segment
                    there should be a corresponding number of values in both lockArray and innerLockArray.  Actormover will interpolate between the values according to the timesteps
                    If the first value is greater than 0, the angles will remain fixed at the first value until the first timestep.  if lastvalue is less than the total time of the segment,
                    the angles will remain fixed at the last value until the end of the segment.  
                    between the last and the second to the last value, the actor will also be moved back to the horizon, if the last angles are 180 (unlocking at the end)

    lockArray:      The values that lockAngle is interpolated through (see below)
    innerLockArray: The values that lockAngle is interpolated through (see below)
    
    Example:

    blendTimeArray        = {0.0, 6.0, 13.0, 19.0},
    lockArray             = {15.0, 0.0, 0.0, 180.0},
    innerLockArray        = {10.0, 0.0, 0.0, 180.0},

    this actormover starts with a Lock angle at 15 and innerLockAngle at 10, then immediately begins becoming more tightly locked until it is fully locked from 6 to 13 seconds.
    from 13 seconds until 19 seconds, the angles move towards fully unlocked and the actor also moves towards the horizon.  If the segment is longer than 19 segments, the actor
    will remain completely unlocked and at the horizon until the end.

    startAtCamera:      If this is set to true, the actormover willl be initialized in the first frame to be rotated to the same 
                            location as the camera.
    lockToTrackingBone: If this is set to true and there is a TargetBone node present in the actor, that rotation will be used 
                            instead of the root node.
    lerpSpeed:          The speed that the actor will lerp towards the camera when between the lockAngle and the innerLockAngle.
    stabilizefactor:    This is used in the stabilizeActor function, to make jitter of the camera have less noticable on the actor
                            when it is not completely unlocked or completely locked to the camera. By default this is set to 0.4

    usePolarSmoothing:  Boolean value to determine if we preserve "up" when passing through the down pole. "false" continually evaluates up, "true" does not.
    
    useRepositionTable: Boolean value to determine if an actor needs to be repositioned at startup.
    repositionKey:      Key in the global reposition table use to get the position for repositioning.
]]

local Logger = TaggedLog("ActorMover", true)

local function printQuat(q)
    return tostring( math.floor(q.x*100)/100) .. " " .. tostring( math.floor(q.y*100)/100) .. " " ..  tostring(math.floor(q.z*100)/100) .. " " .. tostring(math.floor(q.w*100)/100)
end

local function printVect(q)
    return tostring( math.floor(q.x*100)/100) .. " " .. tostring( math.floor(q.y*100)/100) .. " " ..  tostring(math.floor(q.z*100)/100)
end

local textStyle = { color = Color.new(0, 1.0, 0.0), font = "common:fonts/arial_default_14.fnt", scale = 1.0 }

local function renderAxes (inputRotation, inputLabel)
    local mySRT = {
        rotation=inputRotation,
        position= Vector.new ( 0, 0, -2000),
        scale= {1,1,1},
    }

    mySRT.position = Vector.rotateBy(mySRT.position, inputRotation);
   -- Log.message("rendering " .. inputLabel .. " position " .. printVect(mySRT.position) .. " rotation " .. printQuat(mySRT.rotation) .. " euler " .. printVect(Quaternion.getEuler(mySRT.rotation)));
    local labelposition = Util.getWindowPosition(mySRT.position, Global.curScreenWidth, Global.curScreenHeight)
    if (labelposition == nil) then 
        labelposition = {}
        labelposition.x = 0
        labelposition.y = 0
    end
    Renderer.drawText(labelposition.x, labelposition.y, textStyle, inputLabel)
    
    RenderDebug.axes(mySRT, 100.0)
end


ActorMover = 
{
    blendTimeArray        = nil,
    lockArray             = nil,
    innerLockArray        = nil,
    samples				  = nil,
    lerpSpeeds			  = nil,
    lockArrayIndex        = 1,
    frameStep        	  = 5,
    
    startAtCamera         = false,
    lockToTrackingBone    = false,
    lockAngle             = 180.0,
    innerLockAngle        = 180.0,  
    lerpSpeed             = 0.01,
    stabilizeFactor       = 0.4,
    debugOutput           = false,
    
    modifiedActorRot      = Quaternion(),
    previousRotation      = Quaternion(),
    currentRotation       = Quaternion(),
    previousRotationDelta = Quaternion(),
    currentRotationDelta  = Quaternion(),
    
    initDone              = false,
    blendTimer            = 0.0,
    lockAmount            = 0,
    lockedAngleToHorizon  = nil,
    paused		  	  	  = false,
    
    actorWorldUp          = nil,
    
    useRepositionTable    = false,
    usePolarSmoothing     = false,
	useNative             = actormover,
	finalRotation			= Quaternion()
}

ActorMover.new = function(self, data)    
    local o = data or {}
    setmetatable(o, self)
    self.__index = self    

	return o
end

ActorMover.onStart = function(self)
    self.blendTimer        = 0
    self.worldGoalRot      = Global.segmentOffset
--  self.initialYawValue   = Global.scene.mainCamera:getWorld().rotation:getYawAngle()
    self.initDone          = true
    self.ignoreFirstUpdate = false
    self.actorWorldUp       = nil

    Global.RepositionTable = Global.RepositionTable or {}

    -- If using actor repositioning 
    if (self.useRepositionTable) then
        -- Set the actors position so that it's looking a position previously stored in the reposition table
        if (self.repositionKey ~= nil and Global.RepositionTable[self.repositionKey] ~= nil ) then
	        local actorOrigin             = self.object:getWorld()
	        local targetBonePos           = Global.RepositionTable[self.repositionKey].position
	        local VecOriginToTarget       = (targetBonePos - actorOrigin.position):normalized()
	
	        local VecCamFwd               = Global.scene.mainCamera:getWorld():forward()
	        local axis                    = Vector.cross(VecCamFwd, VecOriginToTarget)
	        local angle                   = Vector.dot(VecCamFwd, VecOriginToTarget)
	        local rot                     = Quaternion.axisAngle(axis,math.deg(angle))            
	        
	        local newDir                  = Vector.rotateBy(VecCamFwd, rot)            
	        local modifiedCameraRot       = Quaternion.lookAtWithUp(newDir, actorOrigin:up())
	        actorOrigin.rotation          = modifiedCameraRot
	
	        self.object:setLocal(actorOrigin)
	        self.ignoreFirstUpdate = true
        end
    end

	if (self.useNative) then
		actormover.onStart(self)
	end
end

ActorMover.onDestroy = function(self)
    self.blendTimer        = 0
    self.lockArrayIndex    = 1

	if (self.useNative) then
		actormover.onDestroy(self)
	end
end

ActorMover.stabilizeActor = function(self, updatedTransform) 
    self.previousRotation       = self.currentRotation
    self.previousRotationDelta  = self.currentRotationDelta
    self.currentRotation        = updatedTransform.rotation
    self.currentRotationDelta   = self.currentRotation * Quaternion.inverse(self.previousRotation)
    local alignment             = self.stabilizeFactor
    local output                = Quaternion.slerp(math.clamp01(alignment), self.currentRotationDelta * self.modifiedActorRot, self.modifiedActorRot) 
    return output
end

ActorMover.adjustLockArrays = function (self)
    local index = self.lockArrayIndex 
    
    if (index == 1) then
        self.lockAmount = 0
        self.lockAngle = self.lockArray[index]
        self.innerLockAngle = self.innerLockArray[index]
        
    elseif (not self.blendTimeArray[index]) then
        self.lockAmount = 1
        self.lockAngle = self.lockArray[index-1]
        self.innerLockAngle = self.innerLockArray[index-1]
        
    else
        self.lockAmount = (self.blendTimer - self.blendTimeArray[index-1])/(self.blendTimeArray[index] - self.blendTimeArray[index-1])
        self.lockAngle = math.rad(math.lerp (self.lockArray[index-1], self.lockArray[index], self.lockAmount))
        self.innerLockAngle = math.rad(math.lerp (self.innerLockArray[index-1], self.innerLockArray[index], self.lockAmount))
    end
    
    --Global.AppDebugText2 = "lockArrayIndex " .. index .. " lockAmount " .. self.lockAmount
    --Log.message(Global.AppDebugText2);
end

ActorMover.adjustLockAmount = function (self)
    self.blendTimer = self.blendTimer + Time.dt()

    if (self.blendTimeArray) then
        -- Advance the table index to the next entry if necessary
        while self.blendTimeArray[self.lockArrayIndex] and self.blendTimeArray[self.lockArrayIndex] < self.blendTimer do
            self.lockArrayIndex  = self.lockArrayIndex + 1
        end

        ActorMover.adjustLockArrays(self)
    end
end

ActorMover.adjustToHorizon = function (self, baseRot)
    if (self.blendTimeArray and (self.lockArrayIndex ~= #self.blendTimeArray or self.lockArray[self.lockArrayIndex] < 180.0)) then
        return baseRot
    end

    local horizonActor = Quaternion.axisAngle(Vector.up, Quaternion.getYawAngle(baseRot))
    local horizonAngle = math.abs(Quaternion.angle(horizonActor,baseRot))

    if (not self.lockedAngleToHorizon) then
        self.lockedAngleToHorizon = horizonAngle
    end
    if ((horizonAngle/self.lockedAngleToHorizon) >  math.easeInOutQuart(1 -self.lockAmount) ) then
        return Quaternion.slerp( math.easeInOutQuart(1 -self.lockAmount) * self.lockedAngleToHorizon/horizonAngle, horizonActor, baseRot)
    else
        return baseRot

    end
end

ActorMover.updateActor = function(self, updatedTransform)
    
    if (not self.initDone) then
        return nil
    end

    local rotationChange    = Vector.angle(Vector.rotateBy(Vector.forward,updatedTransform.rotation), Vector.rotateBy(Vector.forward, self.previousRotation))
    self.previousRotation   = updatedTransform.rotation

    local actorWorldRot     = self.object:getWorld().rotation
    local actorWorldForward = Vector.rotateBy(Vector.forward, actorWorldRot)
   
    local actorWorldUp      = Vector.rotateBy(Vector.up, actorWorldRot)
    local dotUp             = Vector.dot(actorWorldUp, Vector.up)

    self.actorWorldUp       = actorWorldUp
    
    --Global.AppDebugText3 = string.format("dotUp %.4f : %.3f, %.3f, %.3f ", dotUp, actorWorldUp.x, actorWorldUp.y, actorWorldUp.z)
    --Log.message(Global.AppDebugText3)
    
    self.modifiedActorRot   = Quaternion.lookAtWithUp(actorWorldForward, actorWorldUp)

    self:adjustLockAmount()
    
    local finalSRT              = self.object:getLocal()
  
    local modifiedCameraRot     = updatedTransform.rotation
    local boneRotation          = nil 
    local boneRotationforward
    local boneInverse
    local bonePosition
    local modifiedBoneRotation

    self.modifiedActorRot       = self:stabilizeActor(updatedTransform) 
    
    local currentUp
    if (self.useFixedUp) then        
        currentUp = Vector.rotateBy(Vector.up,updatedTransform.rotation)
    else 
        -- Use the world "Up" when horizontal, use the actor's up when pointed stright down, and some blend when in-between
        -- This prevents us from the having really ugly flipping at the poles as we switch between "up"s
        local dotFwd      = Vector.dot(Vector.up, actorWorldForward)         
        local blendAmount = 1.0 - (math.asin(math.abs(dotFwd)) / (math.pi * 0.5))
        --blendAmount = math.easeInOutQuart(blendAmount)
        
        currentUp = Vector.normalized(Vector.lerp(blendAmount, actorWorldUp, Vector.up))
    end

    modifiedCameraRot = Quaternion.lookAtWithUp(Vector.rotateBy(Vector.forward, modifiedCameraRot), currentUp)
    if (self.object.namedBones and self.object.namedBones.targetBone ~= nil) then
        local targetBone = ActorManager:getTargetBone(self.object.name)

        if (targetBone) then
            bonePosition = targetBone:getLocal().position - self.object:getLocal().position
            bonePosition = Vector.normalized(bonePosition)
            boneRotation = Quaternion.lookAtWithUp(bonePosition,Vector.up)
            boneInverse  = boneRotation:inverse()
        end

        if (boneRotation and self.lockToTrackingBone) then
            modifiedCameraRot = modifiedCameraRot*(boneInverse) 
        end
    end

    local cameraRotDir          = Vector.rotateBy(Vector.forward, modifiedCameraRot) 
    local modActorForward       = Vector.rotateBy(Vector.forward, self.modifiedActorRot) 
    local angleDiff             = Vector.angle(cameraRotDir, modActorForward) 

    local outerThreshold =  self.lockAngle
    local innerThreshold =  self.innerLockAngle

    local lerpFactor     = 1 - self.lerpSpeed

    local finalLerpAmount 
    if (not angleDiff or angleDiff == 0) then 
        finalLerpAmount = 1
    elseif (angleDiff >  outerThreshold) then
        finalLerpAmount = ((outerThreshold-innerThreshold)*lerpFactor+innerThreshold)/angleDiff
    elseif (angleDiff > innerThreshold) then
        finalLerpAmount = ((angleDiff-innerThreshold)*lerpFactor+innerThreshold)/angleDiff
    else
        finalLerpAmount = 1
    end

    if (not self.created and self.startAtCamera) then
        finalLerpAmount  = 0
        self.created = true
    end

    actorWorldRot = self:adjustToHorizon(actorWorldRot)
    actorWorldRot = Quaternion.slerp(math.clamp01(finalLerpAmount), modifiedCameraRot, actorWorldRot)    

    --Global.AppDebugText3 = string.format("t %.3f : i %2d : finalLerp %.4f : diff %.4f : (%.3f, %.3f) : %2d", self.blendTimer, self.lockArrayIndex, finalLerpAmount, angleDiff, outerThreshold, innerThreshold)
    --Log.message(Global.AppDebugText3);
    
    finalSRT.rotation = actorWorldRot
    if(self.ignoreFirstUpdate ~= true) then
        self.object:setLocal(finalSRT) 
    else
        self.ignoreFirstUpdate = false
    end
    
    -- glow needs to use the actors updated position
    return finalSRT.rotation
end


