local LogDebug = false
local Debug = function(...) end
if (LogDebug) then Debug = Log.message end

-- Handle playing all the animations (including idles) for a given actor
AnimationDriver = {}
AnimationDriver.__index = AnimationDriver

function AnimationDriver.new()
    local this = {
        segmentName = "",
        idleNotification = false,
        mainAnim = nil,
        idleAnim = nil,
        playing = nil,
        flipbook = nil,
        flipbookMain = nil,
        flipbookIdle = nil,
        lastExposureFrame = -1,
    }
    return setmetatable(this, AnimationDriver)
end

local function idleSegmentName(segmentName)
    return segmentName .. "_Idle"
end

function AnimationDriver:init(segmentName, animations, notifyOnIdle, flipbook, frameEvents)
    self:stop();
    self.segmentName = segmentName
    self.idleNotification = notifyOnIdle

    if (animations == nil or animations[segmentName] == nil) then
        Log.error("AnimationDriver there is no animation \"" .. segmentName .. "\" on animated actor: " .. tostring(self.object.name))
        self.mainAnim = nil
        self.idleAnim = nil
    else
        self.mainAnim 		= animations[segmentName]
        self.idleAnim 		= animations[idleSegmentName(segmentName)]
        self.flipbookMain 	= flipbook[segmentName]
        self.flipbookIdle 	= flipbook[idleSegmentName(segmentName)]
    end
    
    self:initFlipbook(self.flipbookMain);
    self:initFlipbook(self.flipbookIdle);

    self:initFrameEvents(frameEvents)

    return self.mainAnim
end

function AnimationDriver:onUpdate()
    if (self.playing) then
        self:playIdleIfNeeded()
    end
    self:processFrameEvents()
end

function AnimationDriver:play(startTime)
    self.playing = self.mainAnim
    self.flipbook = self.flipbookMain;
    if (self.playing == nil) then return end
    self.playing:play()
    if(self.flipbook) then
    	self:playFlipbook(self.flipbook);
    end
    if (startTime) then
        -- Have to start slightly before the end otherwise
        -- it will just clamp it back to 0
        self.playing:setTime(startTime - 0.001)
    end
    
end

function AnimationDriver:stop()
    if (self.playing) then
        self.playing:stop()
    end
     if(self.flipbook) then
    	self:stopFlipbook(self.flipbook);
    end
    self.flipbook = nil
    self.playing = nil
end

function AnimationDriver:isFinished(checkLooping)
    return Util.isAnimFinished(self.playing, checkLooping)
end

function AnimationDriver:getLoopCount()
    local result = 0
    if (self.idleAnim) then
        result = self.idleAnim:getTime() / self.idleAnim:getDuration()
    end
    return result;
end

function AnimationDriver:playIdleIfNeeded()
    if (self.idleAnim and not self.idleAnim:isPlaying() and Util.isAnimFinished(self.mainAnim, false)) then
    	-- also play flipbook idle if it exists
    	if(self.flipbookIdle) then
    		self:stopFlipbook(self.flipbook)
    		self.flipbook = self.flipbookIdle;
    		self:playFlipbook(self.flipbook);
    	end
    	
        self.playing:stop()
        self.playing = self.idleAnim
        self.playing:play()
        if (self.idleNotification) then
            Events.send(self, "idlePlaying", idleSegmentName(self.segmentName))
        end
    end
end

function AnimationDriver:currentTime()
    if (self.mainAnim) then
        return self.mainAnim:getTime()
    end
    return 0
end

function AnimationDriver:currentFrame()
    return math.floor(self:currentTime() * 60)
end

function AnimationDriver:initFlipbook(flipbook)
	if(not flipbook) then return end
	
	for name, fb in pairs(flipbook) do
		fb:init() 
	end
end

function AnimationDriver:playFlipbook(flipbook)
	if(not flipbook) then return end

	for name, fb in pairs(flipbook) do
		fb:play()
	end
end

function AnimationDriver:stopFlipbook(flipbook)
	if(not flipbook) then return end
	
	for name, fb in pairs(flipbook) do
		fb:stop()
	end
end

function AnimationDriver:setLayer(layerNum)
	if(not self.flipbook) then return end
	
	for name, fb in pairs(self.flipbook) do
		fb.layerOffset = layerNum
	end
end

function AnimationDriver:initFrameEvents(events)
    self.frameEvents = events

    -- create a sorted list of frames for which events need to be fired
    self.pendingEvents = {}
    if self.frameEvents then
        for frame, _ in pairs(events) do
            table.insert(self.pendingEvents, frame)
        end
        table.sort(self.pendingEvents)
    end
end

function AnimationDriver:processFrameEvents()
    local frame = self:currentFrame()
    while self.pendingEvents and #self.pendingEvents > 0 and frame >= self.pendingEvents[1] do
        self:fireFrameEvent(self.pendingEvents[1])
        table.remove(self.pendingEvents, 1)
    end
end

function AnimationDriver:fireFrameEvent(frame)
    local event = self.frameEvents[frame]
    if not event then
        --Log.error("AnimationDriver: unable to find event for frame: " .. tostring(frame) .. " for actor " .. self.object.name)
        return
    end
    Events.send(self.object, event.name, self.object, event.data, frame)
    --Log.message("AnimationDriver: event sent: " .. self.object.name .. " - \"" .. event.name .. "\", data = " .. event.data)
end
