

local Logger = TaggedLog("CameraController", true)

CameraController = 
{
	cameraRig = nil,
	
	smoothCamera		= true;
	cameraSpeed			= 16;
	softClamp			= false;
	cameraOrbitSquashY	= 0.5;
	boomLength			= 1;
	cameraCrane			= Vector();
	lookAtOffset		= Quaternion();
	
	--layer handles
	proxyLayer 			= nil,
	inputLayer 			= nil,
	tiltLayer			= nil,
	headingClampLayer 	= nil,
	actorClampLayer 	= nil,
	trackingLayer		= nil,
	boundsLayer			= nil,
	lookAtLayer			= nil,
	boomLayer			= nil,
	squashLayer			= nil,
	actorsLayer			= nil,
	--actorMoverLayer		= nil,
    
    layersStarted       = false,
    horizonDebug		= false,
    -------------------------
    -- Tracking management --
    -------------------------
    cruiseControlDriver = {owner = nil},	-- the actor that owns the cruise control
    boundsDriverName = nil,
    targetActorName = nil,
    
	groups = {
		"boy",
		"girl",
		"dog"
	},
	
	maxAngularDist = 60,
	
	nearestActor = nil,
	nearestActorName = nil,
	
	enable90DegOffsets = true,
}

function CameraController:new(data)
	local o = shallowcopy(data) or {}
    setmetatable(o, self)
    self.__index = self;
    
    o:init()
    return o
end

function CameraController:init()
	self.cameraRig = CameraRig:new()
	
	-- Camera layers setup
	self.proxyLayer 		= self.cameraRig:addLayer(ProxyLayer:new());
	self.trackingLayer 		= self.cameraRig:addLayer(TrackingLayer:new());
	self.inputLayer 		= self.cameraRig:addLayer(InputLayer:new());
--	self.squashLayer		= self.cameraRig:addLayer(SquashLayer:new())
--	self.headingClampLayer 	= self.cameraRig:addLayer(HeadingClampLayer:new())
	self.actorClampLayer 	= self.cameraRig:addLayer(ActorClampLayer:new())
	self.lookAtLayer 		= self.cameraRig:addLayer(LookAtLayer:new())
--	self.boundsLayer 		= self.cameraRig:addLayer(BoundsLayer:new())
	self.boomLayer 			= self.cameraRig:addLayer(BoomLayer:new())
	self.starBlenderLayer   = self.cameraRig:addLayer(StarBlenderLayer:new())
	self.actorMoverLayer 	= self.cameraRig:addLayer(ActorMoverLayer:new())
	self.actorsLayer 		= self.cameraRig:addLayer(ActorsLayer:new())
	
	
	Events.listen(self, "segmentPlaying")
end

function CameraController:segmentPlaying(name)
	self.maxAngularDist = ActorTracking[name].characterOffset
end

function CameraController:lookAt(pos)
	
	-- when going to previous segement and then looking at the main actor - it is still positioned at the origin which causes strange behavior (camera tries to look at itself)
	if(pos.x == 0 and pos.y == 0 and pos.z == 0) then
		return;
	end
	
	local lookAt = pos - Vector(0,0,0)
	
	if(self.lookAtLayer) then
		self.lookAtLayer:setLookAt(lookAt)
	end
end

function CameraController:startLayers()
	MainCamera.camForward 	= Vector.forward:copy()
    MainCamera.camRight		= Vector.right:copy()
    MainCamera.camUp		= Vector.up:copy()
    
	self.cameraRig:startLayers()
    self.layersStarted = true
end

function CameraController:onLateUpdate()
	if(self.layersStarted) then
        self:trackingUpdate()
        self.cameraRig:updateLayers()
        
        -- !!!!!!!!!!!!!!!!!!
        -- HACK to avoid crash when show is interrupted. Ideally, the layers in cameraRig should not get populated with bogus values upon interruption
       	local lenSq = self.cameraRig.worldT.rotation:lengthSq()
       	if (lenSq > 0 and lenSq < 2) then
	       	self.object:setLocal(self.cameraRig.worldT)
	    end

        MainCamera.camForward 	= self.cameraRig.camForward:copy()
        MainCamera.camRight		= self.cameraRig.camRight:copy()
        MainCamera.camUp		= self.cameraRig.camUp:copy()
	    
	    if(CameraController.horizonDebug) then
	    	
	    	local dist = 20050
	    	RenderDebug:line(Vector(dist,0,dist), Vector(-dist,0,dist), Color.new(0,1,1))
	    	RenderDebug:line(Vector(-dist,0,dist), Vector(-dist,0,-dist), Color.new(0,1,1))
	    	RenderDebug:line(Vector(-dist,0,-dist), Vector(dist,0,-dist), Color.new(0,1,1))
	    	RenderDebug:line(Vector(dist,0,-dist), Vector(dist,0,dist), Color.new(0,1,1))
	    	
	    end
	
        
	end
end

function CameraController:onDestroy()
	self.cameraRig.destroyLayers()
end

function CameraController:getLayer(name)
	return self.cameraRig:getLayer(name)
end


function CameraController:trackingUpdate()
		------- check 90 degree offset - drag character if separation is greater -------
		if(not self.enable90DegOffsets) then return end
		
		self.nearestActor = ActorManager:getActor(self.nearestActorName);

		local isNativeAvailable = trackingmanager and trackingmanager.setCamData
		
		local camPos  = self.object:getWorld().position
		local camLook = MainCamera.camForward:copy()
		if ( isNativeAvailable ) then
			trackingmanager.setCamData( camPos, camLook )
		end

		for _, group in ipairs(self.groups) do
			local actor = ActorManager:getActor(ActorManager.proxyActors[group])
			if(actor and actor.targetBone) then
				-- angle to actor
				local toActor = {}
				local deltaAngle = 0
				if (isNativeAvailable) then
					actor.angularDist = trackingmanager.computeAngularDist(actor, actor.targetBone.bonePath)
				else
					toActor = actor.targetBone:getWorld().position - camPos
					deltaAngle = math.deg(Vector.angle(camLook, toActor))
					actor.angularDist = deltaAngle
				end

				-- assign default nearest actor if needed
				if(actor.actorGroup ~= "dog") then self.nearestActor = self.nearestActor or actor end
				
				if(actor.actorGroup ~= "dog" and self.nearestActor.angularDist and actor.angularDist and ((actor.angularDist < self.nearestActor.angularDist))) then
					self.nearestActorName = actor.name;
					self.nearestActor = actor
				end
			end
		end
		
		-- correct actor rotations
		if(self.nearestActor) then
			local nearestRot = {}
			local toNearest = {}
			local nearestRight = {}
			if ( isNativeAvailable ) then
				trackingmanager.setNearestActorData(self.nearestActor, self.nearestActor.targetBone.bonePath, self.maxAngularDist)
			else
				nearestRot = self.nearestActor:getWorld().rotation
				toNearest = self.nearestActor.targetBone:getWorld().position;
				nearestRight = Vector.rotateBy(Vector.right, self.nearestActor:getWorld().rotation)
			end
			
			for _, group in ipairs(self.groups) do
				
				local actor = ActorManager:getActor(ActorManager.proxyActors[group])
				if(actor and actor.targetBone) then
					
					if ( isNativeAvailable ) then
						trackingmanager.correctRotation(actor, actor.targetBone.bonePath, group)
					else
						local actorPosition = actor.targetBone:getWorld().position;
						local direction = (group == "boy" or group == "dog") and -1 or 1
						local deltaAngle = math.deg(Vector.angle(toNearest, actorPosition));
				
						if(deltaAngle > self.maxAngularDist ) then
							local localT = actor:getWorld();
							local target =  direction*(deltaAngle - self.maxAngularDist);
							localT.rotation = localT.rotation*Quaternion.axisAngle(Vector.up, target);
							actor:setLocal(localT);
						end
					end

				end
			end
		end
end

function CameraController:setBounds(driverName, leftBound, rightBound)
	if(self.boundsLayer and driverName == self.boundsDriverName and leftBound and rightBound) then
		self.boundsLayer.boundingActors.left = leftBound
		self.boundsLayer.boundingActors.right = rightBound
	end
end

function CameraController:clearBounds()
	if(self.boundsLayer) then
		self.boundsLayer.boundingActors.left = nil
		self.boundsLayer.boundingActors.right = nil
	end
end

-- try to take control of cruise - return true if successful
function CameraController:takeCruiseControl(driverName)
	
	if(driverName ~= nil and (self.cruiseControlDriver.owner == driverName or self.cruiseControlDriver.owner == nil)) then
		self.cruiseControlDriver.owner = driverName;
		self.boundsDriverName =  driverName;
		self.trackingLayer.driverName = driverName
		
		if(self.debug) then
			DrawText:color(0,1,1):addText("TakeCruise", ">>%s Took cruise control <<", driverName)
			Log.message(string.format(">>%s Took cruise control <<", driverName))
		end
		
		return true;
	end
	
	--DrawText:color(0,1,1):addText("TakeCruise", ">>%s FAILED to take cruise control <<", driverName)
	return false;
end

-- update cruise offset
function CameraController:setCruiseOffset(driverName, offset)
	if(self.trackingLayer and self.cruiseControlDriver.owner == driverName) then
		self.trackingLayer:setAdditiveOffset(offset, driverName);
	end
end

-- release cruise ownership
function CameraController:releaseCruiseControl(driverName)
	if(self.trackingLayer and self.cruiseControlDriver.owner == driverName) then
		self.cruiseControlDriver.owner = nil;
		self.boundsDriverName = nil
		self.trackingLayer:saveAdditiveOffset();
		
		if(self.debug) then
			DrawText:color(0,1,1):addText("Release Cruise", ">>%s Released cruise control <<", driverName)
			Log.message(string.format(">>%s Released cruise control <<", driverName))
		end
	end
end

