
local Logger = TaggedLog("LookAtDetectorCore", true)

local triggerLabelCounter = 1

mathSqrt = math.sqrt

LookAtDetectorCore = {
    worldPosT = nil,
    isTriggered = false,

    -- The closer to 1 this value is the more directly you have to look at it to trigger it
    lookAtThreshold = nil,
    lookedAt = false,
    toTarget = nil,
    colliderRadius = 0,
    dotDifferential = 0,
    debugSphere = nil,
    lookAtCallback = nil,
    debugMode = false,
    flip = 1,
    offset = Vector(),

    globalDebugMode = false,

	useNative = lookatdetector and lookatdetector.onStartNative,
}

LookAtDetectorCore.basenew = function(self)
	local o = {}
	setmetatable(o, self);
	self.__index = self;
	return o;
end

LookAtDetectorCore.new = function(self, params)
    local o = shallowcopy(params) or {}

    Logger:message("table = " .. table.tostring(o))
    setmetatable(o, self)
    self.__index = self
    return o
end

LookAtDetectorCore.onStart = function(self)
	if(self.label == nil) then
		if (self.name ~= nil) then
			self.label = self.name
		elseif (self.object ~= nil) then
			self.label = self.object.name
		else
			self.label = "Trigger_"..triggerLabelCounter
		end
		triggerLabelCounter = triggerLabelCounter + 1
	end
	self.name = self.label

	local scale = 1.0;
	
	if(self.object ~= nil and self.object.attachedTo ~= nil and self.object.attachmentBone ~= nil) then
    	self.flip = math.sign(self.object.attachedTo:getLocal().scale.x);
		scale = math.abs(self.object.attachedTo:getBoneLocal(self.object.attachmentBone).scale.x);
    end
    	
	-- set colliderRadius from the scale of the bone it's attached to, if not set manually
	if(self.colliderRadius == nil or self.colliderRadius == 0) then
		self.colliderRadius = scale;
	end

	if (self.useNative) then
		lookatdetector.onStartNative(self)
	end
end

--LookAtDetectorCore.getTransform = function(self)
--    return self.worldPosT
--end

--LookAtDetectorCore.getWorld = function(self)
--    return self.object:getWorld();
--end

LookAtDetectorCore.onUpdate = function(self)
	if (self.useNative) then
		lookatdetector.onUpdateNative(self)
	else
		local posT = self:getWorld()
    
		-- Don't use full camera pos because it leads to parallax error (since the cam is on a boom)
		-- Do account for the camera height however, so get the rig
		if (Global.scene.cameraRig == nil) then
			return
		end

		local toTarget = posT.position + self.offset - Global.scene.cameraRig:getWorld().position

		-- If the camera to target vector has changed then something moved, recalculate or goal value
		if (self.toTarget == nil or self.toTarget ~= toTarget) then
			self.toTarget = toTarget:copy()
			self.lookAtThreshold = self:calculateGoal()
		end

		--Inline normalize and re-use calculations.
		local lenSq = toTarget.x*toTarget.x + toTarget.y*toTarget.y + toTarget.z*toTarget.z
		if (lenSq ~= 0) then
			local scale = 1.0 / mathSqrt(lenSq)
			toTarget.x = toTarget.x * scale
			toTarget.y = toTarget.y * scale
			toTarget.z = toTarget.z * scale
		end
	
		local look = Global.scene.mainCamera:getWorld():forward()

		--inline dot()
		self.dotDifferential = look.x*toTarget.x + look.y*toTarget.y + look.z*toTarget.z
		local dotDifferential = self.dotDifferential

		if (dotDifferential >= self.lookAtThreshold and not self.isTriggered) then
			--self.object:setDrawDebug(true)
			--Logger:message("Looking at: " .. self.label)

			Events.send(self.object, "lookingAt", self.label)
			if(not self.isTriggered and self.lookAtCallback and self.lookAtCallback.lookedAt) then self.lookAtCallback:lookedAt(self, true) end
        
			self.isTriggered = true
			self.lookedAt = true
        
			-- overridden in child classes
			self:onLookAt(true);
        
		elseif (dotDifferential < self.lookAtThreshold and self.isTriggered) then
			--self.object:setDrawDebug(false)
			--Logger:message("Not looking at: " .. self.label)

			Events.send(self.object, "notLookingAt", self.label)
			if(self.isTriggered and self.lookAtCallback and self.lookAtCallback.lookedAt) then self.lookAtCallback:lookedAt(self, false) end
        
			self.isTriggered = false
			self.lookedAt = false
        
			-- overridden in child classes
			self:onLookAt(false);
		end
	
		local debug = self.debugMode or self.globalDebugMode
		if debug and not self.debugSphere then
			self:createDebugSphere()
		end
		if self.debugSphere then
			self.debugSphere:setVisibility(debug)

			local transform = self.debugSphere:getLocal()
			transform.position = posT.position + self.offset;
			transform.scale = Vector(self.colliderRadius, self.colliderRadius, self.colliderRadius)
			self.debugSphere:setLocal( transform )

	--        Logger:message("LookAt Debug " .. self.label .." : ("..transform.position.x..", "..transform.position.y..", "..transform.position.z..") : radius "..self.colliderRadius)
	--        Logger:message("LookAt WorldCam at: " .. self.object.name)
		end
	end
end

-- virtual override-able method
LookAtDetectorCore.onLookAt = function(self, lookedAt)
	Log.message(self.label..": looking at = ".. tostring(lookedAt));
end

LookAtDetectorCore.onDestroy = function(self)
    destroySafely(self.debugSphere)
    self.debugSphere = nil

	if (self.useNative) then
		lookatdetector.onDestroyNative(self)
	end
end

LookAtDetectorCore.calculateGoal = function(self)
    -- Calculate the minimum result of our dot product for the location and radius
    -- originally generated in the prototype.

    local targetLen	= self.toTarget:length()
    local maxAngle = math.atan2( self.colliderRadius, targetLen )

    -- Minimum dot-product we should need in our transition check
    local dotGoal = math.cos(maxAngle)

    return dotGoal
end

LookAtDetectorCore.getThreshold = function(self)
    return (self and self.lookAtThreshold) or ""
end

LookAtDetectorCore.wasLookedAt = function(self)
    return self.lookedAt
end

LookAtDetectorCore.debugDot = function(self)
    return self.dotDifferential
end

LookAtDetectorCore.createDebugSphere = function(self)
    SceneData.actors["DebugSphere"..self.label] = shallowcopy(SceneData.actors.DebugSphere)
    self.debugSphere = ActorManager:getActor("DebugSphere"..self.label, true)
    self.debugSphere:setVisibility(true)
end

