
--[[

    Description:
    Parameters to pass: 
    * applyDepthY           = boolean value. If set to true, if an actor has a
                               depthY bone, it's value will affect the opacity
                               of the particles
    * applyGlowIntensity    = boolean value. If set to true, if an actor has a
                               glowIntensity bone, it's value will be passed to
                               the focus particles shader to apply focal 
                               changes based on its value.
    --                            
    * applyFocus           = boolean value. If the technique supports a depth
                              based texture spritesheet, set to true

    * focusDriver          = boolean value. If the particle system needs to be
                              driven by the z component of "ParticleFocus" bone
                              of an actor in the current segment, then set this
                              to true

    * focusDriverActor     = string value. Name of the actor that has the
                             "ParticleFocus" bone defined
    --

]]

local Logger = TaggedLog("ParticlesController", true)

ParticlesController = {
	localTransform = nil,
	rotSpeed = 0.0,
	rotAxis = Vector(0.0,0.0,-1.0),
	debugMode = false,
}

function ParticlesController:new(params)
    local o = shallowcopy(params) or {}

    Logger:message("table = " .. table.tostring(o))
    setmetatable(o, self)
    self.__index = self

    return o
end

FocusParticlesTool = {}
local sliderLength = 300

function FocusParticlesTool:new()
    local window = GUI.window(GUIAligned(ALIGN_CENTER), GUISize(320, 180),
                               "Particles")
    window:anchor(ANCHOR_BOTTOM)

    local initialDOF = 5.0

    -- Depth of field slider
    local dofSlider = GUI.slider(window, GUIAligned(ALIGN_CENTER),
                                  sliderLength, "DOF", initialDOF, 0, 10000)
    dofSlider:dock(DOCK_TOP)
    local passiveHandler = {
        onPassiveUpdate = function(self)
            self:setValue(Global.dof)
        end,
    }
    dofSlider:addEventHandler(passiveHandler)
    dofSlider.onValueChanged = function(self, dof)
        Global.dof = dof
    end

    local initialFD = 0.0

    -- Focal Distance slider
    local fdSlider = GUI.slider(window, GUIAligned(ALIGN_CENTER), sliderLength,
                                              "FD", initialFD, -5000, 10000)
    fdSlider:dock(DOCK_TOP)
    local passiveHandler = {
        onPassiveUpdate = function(self)
            self:setValue(Global.focalDistance)
        end,
    }
    fdSlider:addEventHandler(passiveHandler)
    fdSlider.onValueChanged = function(self, fd)
        Global.focalDistance = fd
    end

    return window
end

function ParticlesController:onStart()
    
	
    Events.listen(self, "segmentPlaying")


    --Make emitter externally acessible by other components
    --self.object.emitter = self.emitter
	
	local initialPos = self.position or Vector(0,0,0)
	local initialScale = self.scale or Vector(1,1,1)

    self.object:setVisibility(true)
	self.object:setViewFlags(bit.bor( VF_MAIN_VIEW ))
	self.localTransform = SRT.new( initialPos, Quaternion.identity, 
                                   initialScale) 	
	self.object:setLocal( self.localTransform )    

    --Check if an actor with a focus driver exists in the current segment
    self.focusDriverActor = nil
    if(self.focusDriver == true) then
        for name, actor in pairs(Global.scene.actors) do
            if (actor.namedBones and actor.namedBones.ParticleFocus) then
                self.focusDriverActor = actor
            end
        end
    end

end

function ParticlesController:onUpdate()
    -- Native implementation	
end

function ParticlesController:segmentPlaying(segmentName)
    --Destroy particle system in the defined segment    
    if(self.timedDestroy and self.timedDestroy == true) then
        if(segmentName == self.destroyOnSegment) then
            self:asynchronousDestroy()

            --self.object:destroy()            
        end
    end

    -- Check to see if an actor with a focus driver exists in the current 
    --      segment
    self.focusDriverActor = nil
    if(self.focusDriver == true) then
        for name, actor in pairs(Global.scene.actors) do
            if (actor.namedBones and actor.namedBones.ParticleFocus) then
                self.focusDriverActor = actor
				if (particlescontroller) then
					particlescontroller.updateFocusDriver(self)
				end
            end
        end
    end

end

function ParticlesController:asynchronousDestroy()
    self.object:destroy()
    self.object = nil
end


function ParticlesController:onDestroy() 
    Events.stopListening(self, "segmentPlaying")      
end