
-- a plane that orients itself to the object this component is attached to
-- it use the ray source to create a ray from the source
-- callback can be set using the helper method or by passing callback = { target = obj, method="methodname" }
-- the callback method is called once per event - once for collision started and once for collision ended, the callback method is passed true or false to determine the state of collision
RayToPlane = {
	from = Vector(),
	dir = Vector(0, 0, -1, 0),
	raySource = nil,
	listener = nil,
	
	colliding = false;
	collided = false;
	
	offset = Vector(0,0,0,0),
	
	width = 0.1,
	height = 0.1,
	sidePlanes = {},
	worldPostion = Vector(0,0,0,0),
	
	callback = nil,
	
	-- debug
	debugQuad = nil,
	uniqueCounter = 0,
	debug = false,
	
	type = "RayToPane",
}

RayToPlane.new = function(self, params)
    local o = shallowcopy(params) or {}
    setmetatable(o, self)
    self.__index = self
    
	assert(o.raySource, "no ray source provided")
	
    if(o.debug) then
	    RayToPlane.uniqueCounter = RayToPlane.uniqueCounter + 1;
    end
    
    return o
end

function RayToPlane:setCollisionSize(width, height)
	
	self.width = width or self.width
	self.height = height or self.height
		
	if(width or height) then
		self:calculatePlane()
	end
end

function RayToPlane:calculatePlane()
	-- setup corners
	local pos = Vector(0,0,0,1); -- the center of the plane
	local tl = pos + Vector(-self.width/2, self.height/2, 0,0);
	local tr = pos + Vector(self.width/2, self.height/2, 0,0);
	local bl = pos + Vector(-self.width/2, -self.height/2, 0,0);
	local br = pos + Vector(self.width/2, -self.height/2, 0,0);
	
	
	-- setup planes with point and vector 'on the the plane' normal to be calculated later
	self.sidePlanes = { TLTR = { p = tl, v = (tr - tl):normalized() },
	 					TRBR = { p = tr, v = (br - tr):normalized()},
	 					BRBL = { p = br, v = (bl - br):normalized() },
	 					BLTL = { p = bl, v = (tl - bl):normalized() }};
	
end

-- update the planes orientation/position with the provided transform
function RayToPlane:updatePlane(worldT)

	local sWidth 	= Global.curScreenWidth;
	local sHeight 	= Global.curScreenHeight;
	
	local persp = (Global.nearFar.f - Global.nearFar.n);
	
	local pos = worldT.position;
	local rot = worldT.rotation;
	--Log.message(string.format("Plane rotation: <%.3f, %.3f, %.3f, %.3f>", rot.x, rot.y, rot.y, rot.w))
	
	local dist = pos - Global.scene.mainCamera:getWorld().position
	local length = dist:length()
	
	-- make scale relative to screen size - roughly
	local x = (persp);
	local y = ((sWidth)*50.0);
	local m = y/x;
	local scale = (length*m);
	
	local transformedPlanes = {}
	
	for name, plane in pairs(self.sidePlanes) do
		
		local s = self.height;
		if(name == "TLTR" or name == "BRBL") then
			s = self.width;
		end
		
		-- transform plane
		local point = plane.p:copy()
		point = point*scale
		point = point:rotateBy(rot)
		point = point + pos + self.offset*scale;
		transformedPlanes[name] = {p=point, v = plane.v:rotateBy(rot)*scale*s}
	end
	
	-- save the world position
	self.worldPostion = pos;
	
	return transformedPlanes;
end

function RayToPlane:onStart()
	if(type(self.raySource) == "string") then
		self.raySource = ActorManager:getActor(self.raySource)
	end
	self:calculatePlane();
end

function RayToPlane:onUpdate()
	if (self.object == nil) then
		return
	end
	
	if(not self.raySource) then
		return;
	end
	
	self.dir = Vector.rotateBy(Vector.forward, self.raySource:getWorld().rotation)
	self.from = self.raySource:getWorld().position

	local transformedPlanes = self:updatePlane(self.object:getWorld())
	
	-- draw edges
	if(self.debug) then
		-- collision ray
		RenderDebug:line(self.from, self.from + self.dir*math.floatMax, Color.new(1,1,0))
	end
	local testFailed = false
	
	--test ray against the planes
	for name, plane in pairs(transformedPlanes) do
	
		if(not testFailed or self.debug --[[falls through on debug for visualization purposes]]) then
			--create plane normal
			local n = Vector.cross(self.dir, plane.v):normalized();
			local d = Vector.dot(n, plane.p)
			
			local test = Vector.dot(n, self.from + self.dir)
			
			if(not testFailed and test - d < 0) then
				testFailed = true;
				if(self.debug) then
					RenderDebug:line(plane.p, plane.p + plane.v, Color.new(1,0,0))
				end
			elseif(self.debug) then
				RenderDebug:line(plane.p, plane.p + plane.v, Color.new(0,1,1))
			end
		end
	end
	
	self.collided = not testFailed;
	if(self.colliding ~= self.collided) then
		self.colliding = self.collided;
		if(self.callback ~= nil) then
--			assert(self.callback.target, "There is no 'target' provided to callback")
--			assert(self.callback.method, "There is no 'method' provided to callback")
			self.callback.target[self.callback.method](self.callback.target, self.colliding)
		end
	end
	
	if(self.debug) then
		-- render labels in screen space
		local wpos = Util.getWindowPosition(self.worldPostion, Global.curScreenWidth, Global.curScreenHeight)
		
		if(wpos) then
			local textStyle = { color = Color.new(1, 0.0, 1), font = "common:fonts/arial_default_14.fnt", scale = 1.0 }
			Renderer.drawText(wpos.x, wpos.y - 30, textStyle, "Col:"..tostring(not testFailed))
		end
	end
end


function RayToPlane:setCallback(targetObject, methodName)
	if(self.type == "RayToPane") then
		self.callback = {
			target = targetObject,
			method = methodName,
		}
	else
		assert(false, "setCallback method called without ':' or correct self value")
	end
end