
--[[

    Description:
    Parameters to pass: 
    * startAlpha = alpha value for material when actor is created  (0 .. 1)

    * blending   = whether to start alpha lerp when actor is created 
                   (true/false)

    * inout      = "in" or "out" - defines whether to lerp in or out. Must be 
                    initialized if blending is used.

    * fadetime   = Rate of fading in seconds. Default is 3.

    * delay      = Amount of delay before the assigned blend in or blend out 
                   takes place

    --
    * segmentData Table (Used to define behavior on segment transitions) 
      consists of :

        * segment     = Name of the segment to apply to. The actor to which
                        this component is attached should also be set in the
                        "active" table of that segment

        * startAlpha  = The alpha of the actor at the beginning of the 
                        specified segment

        * blending    = true enables blending, false by default (but should be
                        set)

        * inout       = "in" blend in from low opacity value
                        "out" blend out from higher opacity value to 0

        * fadetime    = Rate of fading. Default is 3 if no value is set

        * delay       = Amount of delay before the assigned blend in or blend
                        out takes place (seconds)
    --

]]

local Logger = TaggedLog("ShowSphereBlender", true)
--cache non-local functions
local getDt = Time.dt
local saturate = math.clamp01

ShowSphereBlender = {
    currentAlpha = 1,
    delay        = 0,
}

function ShowSphereBlender:new(params)
    local o = shallowcopy(params) or {}

    Logger:message("table = " .. table.tostring(o))
    setmetatable(o, self)
    self.__index = self

    return o
end

-- Checks if the segment name provided as a param has behaviors defined for it
--   and applies them and calls blend if needed.
function ShowSphereBlender:applySegmentBlend(currentSegment)
    for key, value in pairs(self.segmentData) do            
        if(currentSegment == value.segment) then
            self.currentAlpha = value.startAlpha
            if(value.blending ~= nil) then 
                self.lerpAlpha = value.blending
            end
            self.in_out = value.inout
            if (value.fadetime) then
                self.fadeTime = value.fadetime
            end
            if (value.delay) then
                self.delay = value.delay
                if (self.delay > 0 ) then
                    self:blend(0)
                end
            end
        end
    end
end

-- Function used locally to apply the blend 
-- (direction (1)-in, (-1)-out, (0)-none)
function ShowSphereBlender:blend(direction)
        
    self.currentAlpha = self.currentAlpha + (getDt()/self.fadeTime) * direction
    local blendValue = saturate(self.currentAlpha)

    
    self.material:setParameters({ u_blend = blendValue})
end

-- Applies blend behavior on creation of the parent actor.
function ShowSphereBlender:onStart()

    self.material = self.object.materials.starSphereBand_MAT
    Events.listen(self, "segmentPlaying")
    self.in_out = "none"
    if(not self.fadetime) then
        self.fadeTime = 3
    else
        self.fadeTime = self.fadetime
    end
    self.currentAlpha = 1
    self.lerpAlpha = false
    if(self.startAlpha) then 
        self.currentAlpha = self.startAlpha 
    end
    if(self.blending and self.inout) then
        self.lerpAlpha = self.blending
        self.in_out = self.inout
    end    

    if (self.delay and self.delay > 0) then
        self:blend(0)
    end

    if(self.segmentData ~= nil) then
        local currentSegment = SegmentManager:currentSegmentName()
        self:applySegmentBlend(currentSegment)
    end

    self.boy = ActorManager:getActor(ActorManager.proxyActors["boy"])
    self.girl = ActorManager:getActor(ActorManager.proxyActors["girl"])
    self.dog = ActorManager:getActor(ActorManager.proxyActors["dog"])

    
end

function ShowSphereBlender:onUpdate()
    
    if(self.material) then

        if(self.boy) then
            local targetPosition = self.boy:getBoneWorld(self.boy.targetBonePath).position
            if (targetPosition == nil) then
                targetPosition = self.boy:getBoneWorld(self.boy.glowBonePath).position            
            end
            local targetDirection = (targetPosition - Vector(0,0,0,1)):normalized()
            self.material:setParameters({ u_worldGlowOffset0 = targetPosition})
        end

        if(self.girl) then
            local targetPosition = self.girl:getBoneWorld(self.girl.targetBonePath).position
            if (targetPosition == nil) then
                targetPosition = self.girl:getBoneWorld(self.girl.glowBonePath).position            
            end
            local targetDirection = (targetPosition - Vector(0,0,0,1)):normalized()
            self.material:setParameters({ u_worldGlowOffset1 = targetPosition})
        end

        if(self.dog) then
            local targetPosition = self.dog:getBoneWorld(self.dog.targetBonePath).position
            if (targetPosition == nil) then
                targetPosition = self.dog:getBoneWorld(self.dog.glowBonePath).position            
            end
            local targetDirection = (targetPosition - Vector(0,0,0,1)):normalized()
            self.material:setParameters({ u_worldGlowOffset2 = targetPosition})
        end
        
    end

    if(self.lerpAlpha == false) then        
        self:blend(0)
    -- Process delay time if needed.    
    elseif (self.delay > 0) then        
        self.delay = self.delay - Time.dt()
    -- Blend out 
    elseif (self.currentAlpha > 0 and self.in_out == "out") then
        self:blend(-1)
    -- Blend in
    elseif (self.currentAlpha < 1 and self.in_out == "in") then
        self:blend(1)
    -- If blend mode is not defined, do not perform blending.
    end
end

-- Process segment transitions
function ShowSphereBlender:segmentPlaying(segmentName)

    self.boy = ActorManager:getActor(ActorManager.proxyActors["boy"])
    self.girl = ActorManager:getActor(ActorManager.proxyActors["girl"])
    self.dog = ActorManager:getActor(ActorManager.proxyActors["dog"])

    if(self.segmentData ~= nil) then
        self:applySegmentBlend(segmentName)
    end
end


function ShowSphereBlender:onDestroy()
    Events.stopListening(self, "segmentPlaying")
end