

---
-- Matches the position of its parent actor to the position of a bone in another actor
-- adds a soft-tracker component to its parent actor which will try to acquire a lock
-- from the tracking manager in order to drive the camera
TrackingTarget = {
	actorName = nil,
	bone = nil,
	proxyActorName = nil,
	name = nil,
	softTracker = nil,
	boundsManager = nil,
	billboard = nil,
	lookAtDetector = nil,
	rayToPlane = nil,
	clampAngle = 15,						-- the angle at which tracking becomes 100% enabled
	useBounds = true, 						-- turn on left and right bounds
	useTracking = true, 					-- turn tracking on/off
	boundsEnabled = false,
	colliding = false,
	localT = SRT(),
	segmentTransition = false,
	init = true,
	debug = false,
	segment = nil,
	cameraTransform = {},
	
	boundingEnabled = false,
	camerabounding = nil, -- holds per segment bounding data

	useNative = trackingtarget and trackingtarget.onStartNative,
}


TrackingTarget.new = function(self, data)
    local o = data or {};
    setmetatable(o, self);
    self.__index = self;
    
    return o;
end

function TrackingTarget:onStart()

	self.camerabounding = require "episode3/scripts/scenedata/persistent/camerabounding"
	self.camerabounding = self.camerabounding or {}
	
	--self.softTracker = self.object.components["SoftTrack"];
	local _actorName	= self.proxyActorName or self.actorName
	assert(_actorName ~= nil, "Track Man requires a name to track")
	
    local _clampAngle 	= self.clampAngle or 0
    local _width 		= self.width or .75
    local _height 		= self.height or 1
    local _offset 		= self.offset or Vector(0,0,0)
    
    -- setup a cocktail of components to detect ray collision and update tracking rotation
	-- the soft tracker manages a rotation offset
	self.softTracker = SoftTrack:new({debug = self.debug});
	self.lookAtDetector = LookAtDetector:new({debugMode = self.debug,  colliderRadius = 1000, lookAtCallback = self })

    Events.listen(self, "segmentPlaying");
    Events.listen(self, "segmentRunning");

	self.name = _actorName
    
    -- Method to simulate actor object behavior
    local objFunc = function(self, parent, localT, worldT, name)
    	
		self.object = {
			
			parent = parent,
			name = name,
			
			getLocal = function(self)
				return {position = parent.localT.position:copy(), rotation = parent.localT.rotation:copy()};
			end,
			
			setLocal = function(self, localT)
				-- write back to parent transform
				parent.localT.position.x = localT.position.x
				parent.localT.position.y = localT.position.y
				parent.localT.position.z = localT.position.z
				parent.localT.position.w = localT.position.w
				
				parent.localT.rotation.x = localT.rotation.x
				parent.localT.rotation.y = localT.rotation.y
				parent.localT.rotation.z = localT.rotation.z
				parent.localT.rotation.w = localT.rotation.w
			end,
			
			getWorld = function(self)
				return {position = parent.localT.position:copy(), rotation = parent.localT.rotation:copy()};
			end
		}
	end
    
    -- assign method to each component that would normally be attached to the actor
	self.lookAtDetector.setObject = shallowcopy(objFunc)
	self.softTracker.setObject = shallowcopy(objFunc)
	
	if (self.useNative) then
		trackingtarget.onStartNative(self)
	end
end

function TrackingTarget:segmentPlaying(name)
	self.segment = name;
	self.boundingEnabled = self.camerabounding[name] and self.camerabounding[name].enabled or false

	if (self.useNative) then
		trackingtarget.segmentPlayingNative(self)
	end
end

function TrackingTarget:onDestroy()
	if(not self.init) then
		self.lookAtDetector:onDestroy()
		self.softTracker:onDestroy()
	end

	if (self.useNative) then
		trackingtarget.onDestroyNative(self)
	end
end

function TrackingTarget:onUpdate(cameraTransform)
	if (self.useNative) then
		trackingtarget.onUpdateNative(self)
	else
		local actorName = ActorManager.proxyActors[self.proxyActorName]
	
		local actor = ActorManager:getActor(actorName)
	
		if(actor and not self.segmentTransition) then
		
			local bonePath = actor[self.bone.."Path"]
			if(bonePath and actor.getBoneWorld) then
			
				local targetT = actor:getBoneWorld(bonePath)
			
				if(targetT and targetT.copy) then
					self.localT = targetT:copy();
					self.lookAtDetector:setObject(self, self.localT, self.localT, self.proxyActorName)
					self.softTracker:setObject(self, self.localT, self.localT, self.proxyActorName);
				end
			end
		
			local didSwitch = false;
		
			if(actorName ~= self.lastActor) then
				self.lastActor = actorName
				didSwitch = true
			end
		
			if(didSwitch) then
				if(self.init) then
					self.init = false
					self.softTracker:onStart()
					self.lookAtDetector:onStart()
				end
			end
		
			self.softTracker:onUpdate(cameraTransform)
			self.lookAtDetector:onUpdate(cameraTransform)
		
			if(self.colliding) then
			
				if( MainCamera.controller:takeCruiseControl(self.proxyActorName)) then
				
					MainCamera.controller.targetActorName = actorName;
				
					if(not self.softTracker.doTrack) then
						self.softTracker:enableTracking(true, self.localT.position, actor.frameStep ~= nil, actor.frameStep);
					elseif(didSwitch and self.softTracker.doTrack) then
						self.softTracker.doSteppedUpdate = actor.frameStep ~= nil
						self.softTracker.frameStep = actor.frameStep
					end
				
					if(self.useTracking) then
						MainCamera.controller:setCruiseOffset(self.proxyActorName, self.softTracker.offset)
					end
				end
			end
		
		
		elseif( not actor and not self.segmentTransition) then
			MainCamera.controller:releaseCruiseControl(self.proxyActorName)
		end
		self.segmentTransition = false
	
		-- cruise control debug
		if(self.debug and MainCamera.controller.trackingLayer) then
			local yawAngle = MainCamera.controller.trackingLayer.offset:getYawAngle();
			local cruiseOwner = MainCamera.controller.cruiseControlDriver.owner or "none"
			if ActorTracking and ActorTracking[self.segment] then
				DrawText:color(1,1,0):addText("CruiseDebug", "**CruiseOwner %s angle <%.3f> strength <%.3f>**",
								MainCamera.controller.cruiseControlDriver.owner, yawAngle, ActorTracking[self.segment].strength)
    		end
		end
	end
end

function TrackingTarget:lookedAt(lookAtDetector, colliding)
	self.colliding = colliding;	
	if(not colliding) then
--		Log.message("NOT Colliding with: "..self.proxyActorName)
		if(self.doPause) then
			SegmentManager:pause()
		end
	
		self.trackingEnabled = false
		self.boundsEnabled = false;
		
		MainCamera.controller:releaseCruiseControl(self.proxyActorName);
		
		self.softTracker:enableTracking(false);
		
		if(self.debug) then
			DrawText:color(1,1,1):addText("ReleaseCollision", "**%s Lost Control **", self.proxyActorName)
		end
		
	else
--		Log.message("Colliding with: "..self.proxyActorName)
		if(self.pauseOnStopTracking) then
			self.doPause = true
		end
	end
end

