
local Logger = TaggedLog("WaveAnimator", false)
WaveAnimator = {
    numShaderTextures = 3, -- must match number of textures supported in the shader (waves.rsfx)
    shaderTexBaseName = "u_diffuseMap",
    shaderDecayBaseName = "u_decay",
	
    speed = 0.5,    -- seconds between cycling the textures
	
    current = {},   -- table of texture entries currently being used by the shader
    available = {}, -- table of texture entires that are available to be swapped in
}

function WaveAnimator:new(params)
    local newObj = shallowcopy(params) or {}
    setmetatable(newObj, self)
    self.__index = self

    return newObj
end

function WaveAnimator:onStart()
    self.material = self.object.materials[self.matName]
    Logger:message("Material = " .. tostring(self.material))

    self:updateSpeed()
    self:setupTextures()
    self:updateShaderProperties()

    self.timer = 0
end

function WaveAnimator:onUpdate()

    self:updateSpeed()

    -- update timers
    local dt = Time.dt()
    self.timer = self.timer + dt
    for _, entry in ipairs(self.current) do
        entry.timer = entry.timer + dt
    end

    -- swap out the textures if needed
    if self.timer > self.speed then
        local newIndex = math.random(#self.available)
        local newTexture = self.available[newIndex]
        local retireTexture = self.current[self.numShaderTextures]
        for i = self.numShaderTextures, 2, -1 do
            self.current[i] = self.current[i - 1]
        end
        self.current[1] = newTexture
        self.current[1].timer = 0
        self.available[newIndex] = retireTexture
        self.timer = 0
        Logger:message("Swapped textures current = " .. table.tostring(self.current) .. ", available = " .. table.tostring(self.available))
    end

    self:updateShaderProperties()
end

function WaveAnimator:decay(timer)
    -- Clamp it to 1 for the first speed inverval or 0 after the last interval
    if timer < self.speed then
        return 1
    elseif timer > (self.speed * self.numShaderTextures) then
        return 0
    end
    -- Do a linear ramp from 1 - 0 over number of texture speed intervals
    return (self.speed - timer) / ((self.numShaderTextures - 1) * self.speed) + 1
end

function WaveAnimator:updateSpeed()
    -- update the speed based on the wave speed bone's y position (if available)
    if self.object.waveSpeedBone then
        self.speed = self.object.waveSpeedBone:getWorld().position.y
    end
end

function WaveAnimator:updateShaderProperties()
    local props = {}
    for i, entry in ipairs(self.current) do
        self.material:setTexture(self.shaderTexBaseName..i, entry.texture)
        props[self.shaderDecayBaseName..i] = self:decay(entry.timer)
--        Global["AppDebugText" .. i] = "Tex" .. i .. ": " .. entry.texture .. ", decay = " .. string.format("%.2f", self:decay(entry.timer)) .. ", timer = " .. string.format("%.2f", entry.timer) .. ", speed = " .. tostring(self.speed)
    end
    self.material:setParameters(props)
end

function WaveAnimator:setupTextures()
    -- get the number of textures requested
    local propTable = self.material:getPropertiesTable()
    local waveCount = propTable.waves or 0

    -- pull out the base texture number
    local baseDigits = string.match(self.baseTextureURI, "%.(%d+)%.")
    if baseDigits == nil then
        Log.error("Unable to get base texture number for the wave animation on " .. self.object.name .. "'s material" .. tostring(self.matName))
        return
    end
    local baseTexNum = tonumber(baseDigits) or 0
    local numDigits = #baseDigits

    -- construct the list of texture URIs
    local textureURIs = {}
    for i = 0, waveCount - 1 do
        local texNum = string.format("%0"..numDigits.."d", tostring(baseTexNum + i))
        local texURI = string.gsub(self.baseTextureURI, "%.(%d+)%.", "." .. texNum .. ".")
        table.insert(textureURIs, { texture = texURI, timer = i * self.speed })
    end

    -- construct the list of current and available texture entries
    self.current = {}
    self.available = {}
    for i = 1, waveCount do
        if i <= self.numShaderTextures then
            table.insert(self.current, textureURIs[i])
        else
            table.insert(self.available, textureURIs[i])
        end
    end

    Logger:message("Textures = " .. table.tostring(textureURIs))
    Logger:message("Current = " .. table.tostring(self.current))
    Logger:message("Available = " .. table.tostring(self.available))
end
