
-- Google Analytics APIs

require "episode3/scripts/launcher/config"

-- This UID is property (aka episode) specific
-- as new episodes are released, this should correspondingly be updated
local analyticsUID			= "UA-40905299-1"

--  Session level custom dimensions
ANALYTICS_BUILD_DIMENSION 			= 6
ANALYTICS_PACKAGE_DIMENSION			= 29
ANALYTICS_MOXIE_DIMENSION			= 30
ANALYTICS_SUSTAINED_RUN_COUNT			= 16
ANALYTICS_SESSION_TYPE_DIMENSION		= 8
ANALYTICS_APP_TYPE_AT_USER_LAUNCH		= 9
ANALYTICS_CLIENT_LOCAL_HOUR			= 10
ANALYTICS_CLIENT_LOCAL_DAY			= 11
ANALYTICS_DAYS_SINCE_FIRST_PROMPT_DIMENSION	= 14
ANALYTICS_NUMBER_OF_HAT_PROMPTS_DIMENSION	= 15
ANALYTICS_REASON_FOR_ATTRITION			= 18
ANALYTICS_TOTAL_RUN_COUNT			= 26
ANALYTICS_TOTAL_SESSIONS_COMPLETE		= 31
ANALYTICS_SESSION_WAS_FROM_RESUME		= 19
ANALYTICS_SESSION_LAUNCH_REASON			= 24
ANALYTICS_DOGFOOD_BUCKET			= 33

-- Hit level custom dimensions
ANALYTICS_HEADSET_DIMENSION			= 1
ANALYTICS_SOUND_LEVEL_DIMENSION			= 22
ANALYTICS_ACTIVE_SCENE_DIMENSION		= 17

-- Used for time bins
local ANALYTICS_MILLIS_PER_TIME_BIN		= 5000
local startTime					= -1
local lastObservedScene				= ""
local showIsPlaying				= false
local timeAtStartOfFpsBin			= 0

-- for fps tracking
local framesThisBin 				= 0
local timeThisSegmentBegan			= 0

local curNumberOfTouchesInTimeBin		= 0
local curNumberOfSwipesInTimeBin		= 0

local curNumberOfTouchesInSceneBin		= 0
local curNumberOfMovesInSceneBin		= 0

-- Used for tracking GWI entry
local timeOfFirstGWI				= -1
local ANALYTICS_SECONDS_UNTIL_SUSTAINED_ENGAGEMENT = 15

local immAccessReported				= false

local promptReason				= nil

-- used to determine whether our interruption was user or system generated
local unmatchedOnUserLeaveHint 			= false

local touchIsDown				= false
local curMoveCount				= 0

local curLeadActor				= nil

-- Duet specific variables

-- running sum of focal values of each character
local focus_amts				= { 
	boy = 0,
	girl = 0,
	dog = 0,
	unfocused = 0,
}

-- running count of how many samples have been taken
local focus_counts				= { 
	boy = 0,
	girl = 0,
	dog = 0,
	unfocused = 0,
}

-- Make sure the persist tables are setup
Persist = Persist or {}
Persist.duet = Persist.duet or {}
Persist.duet.totalRunCount = Persist.duet.totalRunCount or 0 
Persist.duet.sustainedrunCount = Persist.duet.sustainedrunCount or 0
Persist.duet.episodesCompleted = Persist.duet.episodesCompleted or 0
Persist.duet.pendingUltimateAccess = Persist.duet.pendingUltimateAccess or false
Persist.duet.promptsSinceLastEngagement = Persist.duet.promptsSinceLastEngagement or 0

-- actor focus code provided by Tim
ActorFocus = {

	groups = {
		"boy",
		"girl",
		"dog"
	},

	getActor = function(self, groupName)
		return ActorManager:getActor(ActorManager.proxyActors[groupName])
	end,

	getFocus = function(self)

		local focusValues = {
			boy = 0,
			girl = 0,
			dog = 0,
		}
		for _, group in ipairs(self.groups) do

			local actor = self:getActor(group)

			if(actor and actor.targetBone) then
				local ndc = Util.toClipSpace(actor.targetBone:getWorld().position)
				if( ndc.y < 1.0 and ndc.y > -1.0 and ndc.x < 1.0 and ndc.x > -1.0 and ndc.z > 0 and ndc.z < 1) then
					focusValues[group] = (2.0 - ndc:lengthSq())/2.0
				end
			end
		end

		return focusValues;
	end
}


AnalyticsManager = {}

-- how many digits we will pad values out to (e.g. "00012" for 12 ) 
local ANALYTICS_ZERO_PAD_DEFAULT		= 5

-- helper function to calculate time deltas accounting for rollover
calculateTimeDelta = function ( startT, endT )
	-- if rollover happens endtime will actually be less then start time
	if endT < startT then
		-- HACK:  there was rollover.  lua does not offer a max value utility, and since 
		-- I don't know what the underlying architecture will be I'm just going to take a 
		-- subset of the delta until we have a proper native API in place.
		return 0
	else 
		return endT - startT
	end
end

mangleString = function( intToMangle, numDigits )
	local resultString = ""

	if numDigits == nil then
		numDigits = ANALYTICS_ZERO_PAD_DEFAULT
	end
	if intToMangle ~= nil then
		resultString = string.format("%0"..tostring(numDigits or 0).."i", intToMangle or 0)
	end

	return resultString
end

AnalyticsManager.getTimeBin = function()
	if startTime >= 0 then
		return math.floor(calculateTimeDelta(startTime, Time.milliseconds() ) / ANALYTICS_MILLIS_PER_TIME_BIN)
	else
		-- time bins begin at the hat prompt, we don't care about GWI load
		return 0
	end
end

AnalyticsManager.getTimeBinSceneBased = function()
	if startTime >= 0 then
		return math.floor(calculateTimeDelta(timeThisSegmentBegan, Time.milliseconds() ) / ANALYTICS_MILLIS_PER_TIME_BIN)
	else
		-- time bins begin at the hat prompt, we don't care about GWI load
		return 0
	end
end

AnalyticsManager.setAllHitDimensions = function()
	Analytics.setCustomDimension(ANALYTICS_HEADSET_DIMENSION, mangleString( JavaAudio.getHeadsetState()) )
	Analytics.setCustomDimension(ANALYTICS_SOUND_LEVEL_DIMENSION, mangleString( JavaAudio.getStreamVolume(3), 3))
	Analytics.setCustomDimension(ANALYTICS_ACTIVE_SCENE_DIMENSION, lastObservedScene or "")
end

AnalyticsManager.recordDayOfFirstPrompt = function()

	-- note:  the date is not returned human readable, which means the month is not 0 indexed
	-- our functions work off 0 offset...  I wish this was all just consistent
	local date = Calendar.getDate()
	date.month = date.month-1

	-- mark down the day of our first prompt so we can see how many days away from our first prompt we interacted
	Persist.duet.yearOfFirstRun = Persist.duet.yearOfFirstRun or date.year
	Persist.duet.monthOfFirstRun = Persist.duet.monthOfFirstRun or date.month
	Persist.duet.dayOfFirstRun = Persist.duet.dayOfFirstRun or date.day
	Persist.duet.hourOfFirstRun = Persist.duet.hourOfFirstRun or Calendar.getLocalTime().hour
end

AnalyticsManager.initialize = function()
	-- initialize analytics
	Log.message("Initialising Google Analytics.")
	Analytics.initialize(analyticsUID)
	Analytics.setVersionString(System.versionString())

	-- analytics would like to know what segments are playing for scene transitions
	Events.listen(AnalyticsManager, "segmentPlaying")
end

AnalyticsManager.sendInitialStats = function()
	-- tell the service this is a doodle session as opposed to a download session
	Analytics.setCustomDimension(ANALYTICS_BUILD_DIMENSION, System.versionString() or "unknown")
	Analytics.setCustomDimension(ANALYTICS_SESSION_TYPE_DIMENSION, "SESSION_TYPE_DOODLE")
	Analytics.setCustomDimension(ANALYTICS_PACKAGE_DIMENSION, Platform.package)
	Analytics.setCustomDimension(ANALYTICS_MOXIE_DIMENSION, Platform.version)

	Analytics.sendEvent("GeneralInfo " .. Platform.package, "CurrentShow", Platform.package, 0)

	-- set the dogfood bucket
	local config = Config.active()
	Analytics.setCustomDimension(ANALYTICS_DOGFOOD_BUCKET, config.name)

	-- build info
	Analytics.sendEvent("GeneralInfo " .. Platform.package, "BuildID", System.versionString(), 0)

	-- record how many times the user has previously completed the show
	Analytics.setCustomDimension( ANALYTICS_TOTAL_SESSIONS_COMPLETE, mangleString(Persist.duet.episodesCompleted))

	-- record the kind of app that was being used at launch
	-- (PixelPipe has been removed as it was Motorola only.  This is unlikely to return as there is no 'Android' way to do this)
	--[[
	local activeTargetInfo = PixelPipe.getActiveTargetInfo()
	local packageName = "";
	if (activeTargetInfo ~= nil) then
		packageName = activeTargetInfo.packageName;
	end
	Analytics.setCustomDimension(ANALYTICS_APP_TYPE_AT_USER_LAUNCH, AnalyticsManager.getAppType(packageName))
	--]]

	-- record the client time so we can filter based on the clients local time (analytics converts to GMT)
	local clientHour = Calendar.getLocalTime().hour
	local localDay = Calendar.getDate().day
	Analytics.setCustomDimension( ANALYTICS_CLIENT_LOCAL_HOUR, mangleString(clientHour, 2))
	Analytics.setCustomDimension( ANALYTICS_CLIENT_LOCAL_DAY, mangleString(localDay, 2))
	
	-- Also, because our .apk is not changing, manually keep track of episode viewings through the persist table
	-- (if we do not do this, it becomes very complicated to establish new vs. returning visitors)
	Analytics.sendEvent("GeneralInfo " .. Platform.package, "InitialVolume", "", JavaAudio.getStreamVolume(3))
	Analytics.sendEvent("GeneralInfo " .. Platform.package, "InitialHeadsetState", "", JavaAudio.getHeadsetState())
	Analytics.sendEvent("GeneralInfo " .. Platform.package, "TimeSubsegmentDuration", "", ANALYTICS_MILLIS_PER_TIME_BIN)
	Analytics.sendEvent("GeneralInfo " .. Platform.package, "RunCount", mangleString(Persist.duet.totalRunCount), 0)
end


-- listing of scenes we consider to be "main scenes"
-- each of these scenes will be represented as their own view as opposed to side shows and interactions
-- which will be captured through events
local mainScenes = {
	"Cell_Girl_IO_Intro",
	"Cell_Girl",

	"Birth_Girl",
	"Birth_Boy",

	"First_Meeting",
	"First_Meeting_Boy_Solo",
	"Second_Meeting1",
	"Second_Meeting3",

	"Third_Meeting1",
	"Third_Meeting3",
	"Third_Meeting3_Alt", 
	"Third_Meeting3_End_Girl",
	"Third_Meeting_End_Boy_Alt",

	"boytree_intro_a",
	"girltree_intro_a",
	"boytree",
	"girltree",

	"boytree_dance",

	"girltree_girloutro",
	"boytree_boyoutro",
	"Boy_4th_Solo_B",

	"Fifth_Meeting1",
	"Fifth_Meeting2",
	"Fifth_Meeting3",

	"Fifth_Meeting3_Boy_To_End",

	"End_Girl"
}

-- a listing of counts associated with each scene so we do not inadvertently end up blasting the 
-- server with repeat scenes 
local curCountTable = {} 

AnalyticsManager.whichActor = function(self, segmentName)
	local segment = SegmentManager.segments[segmentName]
	local leadActorName = segment.leadActor or nil
	if leadActorName ~= nil then
		--local actor = ActorManager:getActor(leadActorName, false)
		local actor = SceneData.actors[leadActorName]
		local group = "none"
		if(actor and actor.actorGroup) then
			return actor.actorGroup
		end
	end

	return nil
end

AnalyticsManager.segmentPlaying = function(self, segmentName)
	segmentName = segmentName or ""

	local found = false
	for i, name in ipairs(mainScenes) do
		if name == segmentName then
			found = true
		end
	end


	if found == false then
		-- this apparently isn't a segment we're interested in, skip it!
		return
	end

	curCountTable[segmentName] = curCountTable[segmentName] or 0
	if curCountTable[segmentName] == 0 then 
		-- let's also send some touch information for this scene
		self.setAllHitDimensions()
		Analytics.sendEvent("Ux " .. Platform.package, "touchesInSceneBucket", "", curNumberOfTouchesInSceneBin)
		curNumberOfTouchesInSceneBin = 0

		self.setAllHitDimensions()
		Analytics.sendEvent("Ux " .. Platform.package, "movesInSceneBucket", "", curNumberOfMovesInSceneBin)
		curNumberOfMovesInSceneBin = 0

		-- send updated info about actor tracking and then reset for the last scene
		self.sendBoyGirlDogViewCounts()
		self.resetBoyGirlDogViewCounts()

		-- and then mark down in our hash that we've seen the scene
		curCountTable[segmentName] = 1

		-- why are we appending Platform.package to the tail end of this?  due to issues with segmentation 
		-- we are simply naming events by appending their package name to isolate them.
		-- since episodes may share scene names, this helps isolate them 
		Analytics.sendView(segmentName .. " " .. Platform.package)

		-- if this is a sequence we're interested in, report fps for that scene before we switch
		local timeInSegment = Time.milliseconds() - timeThisSegmentBegan
		timeThisSegmentBegan = Time.milliseconds()

	else
		-- we've already seen this scene, no need to blast analytics with data
		curCountTable[segmentName] = curCountTable[segmentName]+1
	end

	lastObservedScene = segmentName
end

AnalyticsManager.update = function(self)
	-- don't start doing time binning until we've actually seen the prompt
	if startTime == -1 then
		return
	end

	framesThisBin = framesThisBin + 1

	-- the following code only applies to when the show begins
	-- TODO:  what's the proper mechanism for querying if the game is in show
	if startTime == -1 then 
		return
	end

	-- update how much each character has been observed
	local values = ActorFocus:getFocus()
	local some_actor_focused = false
	for _, group in ipairs({"boy", "girl", "dog"}) do
		-- rollovers are extremely unlikely, however, if it does just stop sampling data
		if focus_counts[group] + 1 < 0 then
			break
		end

		local focusValue = values[group]

		-- if there is any visibility at all increment this value
		-- (we are trying to calculate a 'time spent on screen' metric, the focus value itself is somewhat irrelevant)
		if focusValue > 0 then
			some_actor_focused = true
			focus_amts[group] = focus_amts[group] + 1
		end

		focus_counts[group] = focus_counts[group] + 1
	end

	-- lastly, there is no 'unfocused' group so do this manually
	if some_actor_focused ~= true then
		focus_amts['unfocused'] = focus_amts['unfocused'] + 1
	end
	focus_counts['unfocused'] = focus_counts['unfocused'] + 1

	-- see if the user has stayed within the app for at least the required engagement period
	-- TODO:  if the app persists past closure this needs a test to see if it's still actually rendering
	local elapsedTime = calculateTimeDelta(startTime, Time.milliseconds() )
	if immAccessReported ~= true and elapsedTime >= (ANALYTICS_SECONDS_UNTIL_SUSTAINED_ENGAGEMENT*1000) then

		--  this is the number of actually sustained runs we have made 
		Persist.duet.sustainedrunCount = Persist.duet.sustainedrunCount + 1
		Analytics.setCustomDimension(ANALYTICS_SUSTAINED_RUN_COUNT, mangleString(Persist.duet.sustainedrunCount))

		-- if the user clicked on the prompt within allotted time and stayed for allotted time 
		if immAccessReported ~= true then
			immAccessReported = true
			self.setAllHitDimensions()
			Analytics.sendEvent("GWI " .. Platform.package, "Success", "SustainedEngagement", Persist.duet.promptsSinceLastEngagement)
		end

		-- ultimate access and immediate access are NOT mutually exclusive by design
		if Persist.duet.pendingUltimateAccess == true then
			Persist.duet.pendingUltimateAccess = false

			self.setAllHitDimensions()
			Analytics.sendEvent("GWI " .. Platform.package, "Success", "UltimateAccess", 1)
		end

		-- reset our counter for how many prompts have been initiated since the last sustained engagement
		Persist.duet.promptsSinceLastEngagement = 0
	end
end

-- sends an event with the percentage of time spent viewing each type of character
AnalyticsManager.sendBoyGirlDogViewCounts = function(self)
	local timeSpentViewingGirl = (focus_amts.girl / focus_counts.girl) * 100
	local timeSpentViewingBoy = (focus_amts.boy / focus_counts.boy) * 100
	local timeSpentViewingDog = (focus_amts.dog / focus_counts.dog) * 100
	local timeSpentUnfocused = (focus_amts.unfocused / focus_counts.unfocused) * 100

	Analytics.sendEvent("InteractionInfo " .. Platform.package, "WatchPercentage", "Girl", tonumber(timeSpentViewingGirl))
	Analytics.sendEvent("InteractionInfo " .. Platform.package, "WatchPercentage", "Boy",  tonumber(timeSpentViewingBoy))
	Analytics.sendEvent("InteractionInfo " .. Platform.package, "WatchPercentage", "Dog",  tonumber(timeSpentViewingDog))
	Analytics.sendEvent("InteractionInfo " .. Platform.package, "WatchPercentage", "Unfocused",  tonumber(timeSpentUnfocused))
end

-- sends an event with the percentage of time spent viewing each type of character
AnalyticsManager.resetBoyGirlDogViewCounts = function(self)
	for key,value in pairs(focus_amts) do 
		focus_amts[key] = 0
	end

	for key,value in pairs(focus_counts) do 
		focus_counts[key] = 0
	end
end

AnalyticsManager.onShowBegan = function(self)

	-- used for time binning
	startTime = Time.milliseconds()
	timeAtStartOfFpsBin = Time.milliseconds()
	showIsPlaying = true

	-- total run count information
	if (Persist.duet.totalRunCount == nil) then
		Log.error("Persist.duet.totalRunCount is NIL! Copy /data/data/com.motorola.avatar/shared/persist.lua off the device and send to jritts")
	end
	if (type(Persist.duet.totalRunCount) ~= "number") then
		Log.error("Persist.duet.totalRunCount is not a number! Copy /data/data/com.motorola.avatar/shared/persist.lua off the device and send to jritts")
	end

	Persist.duet.totalRunCount = Persist.duet.totalRunCount + 1
	Analytics.setCustomDimension( ANALYTICS_TOTAL_RUN_COUNT, mangleString(Persist.duet.totalRunCount))

	--  add tracking information for the absolute run count (as opposed to the sustained engagement)
	if Persist.duet.totalRunCount == 0 then 
		Analytics.sendEvent("GeneralInfo " .. Platform.package, "FirstTimeRun", "", 1)
	end
	
	-- how was the show launched (hat prompt, splash, etc.)	local showLaunchReason = nil
	showLaunchReason = Context.getPromptReasonString()
	Analytics.setCustomDimension( ANALYTICS_SESSION_LAUNCH_REASON, showLaunchReason)

	-- mark down if this was a resume session or not
	if Context.getCheckpoint() == nil then
		Analytics.setCustomDimension( ANALYTICS_SESSION_WAS_FROM_RESUME, "FromBeginning" )
	else
		Analytics.setCustomDimension( ANALYTICS_SESSION_WAS_FROM_RESUME, "FromResume" )
		lastObservedScene = Context.getCheckpoint() or ""
	end

	self.setAllHitDimensions()
	Analytics.sendEvent("GWI " .. Platform.package, "ShowBegan", showLaunchReason, 0)
end

AnalyticsManager.onShowEnded = function(self)

	showIsPlaying = false

	local elapsedTime = calculateTimeDelta(startTime, Time.milliseconds() ) / 1000

	AnalyticsManager:sendBoyGirlDogViewCounts()
	Analytics.sendEvent("MangleStore " .. Platform.package, "ShowEnded_elapsedTime", mangleString(elapsedTime), elapsedTime)
end

AnalyticsManager.onDoNotify = function()
	AnalyticsManager.setAllHitDimensions()
	Analytics.sendEvent("GWI " .. Platform.package, "Impression", "Flourish", 0)
end

AnalyticsManager.setAppAtTimeOfPrompt = function(self)
    local appType = AnalyticsManager.getAppType(Query.getActiveActivity())
	self.setAllHitDimensions()
	Analytics.sendEvent("GWI " .. Platform.package, "AppAtTimeOfPromptAccepted", appType, 0)
end

AnalyticsManager.recordTouchDown = function(self)
	-- keep track of the number of touches for each time bucket
	curNumberOfTouchesInTimeBin = curNumberOfTouchesInTimeBin + 1 
	curNumberOfTouchesInSceneBin = curNumberOfTouchesInSceneBin + 1

	touchIsDown = true
end

AnalyticsManager.recordTouchUp = function(self)
	-- this wasn't just a touch but a proper swipe
	if curMoveCount >= 5 then 
		-- keep track of the number of touches for each time bucket
		curNumberOfSwipesInTimeBin = curNumberOfSwipesInTimeBin + 1 
		curNumberOfMovesInSceneBin = curNumberOfMovesInSceneBin + 1
	end 

	curMoveCount = 0
	touchIsDown = false
end

AnalyticsManager.recordMoveEvent = function(self)
	-- moves only occur on swiping motions in between a touch down/up
	curMoveCount = curMoveCount + 1
end

AnalyticsManager.hatPromptIsolated = function(animName)
	-- prior to this there was no way to differentiate hat prompts, so this will be an 
	-- additional piece of info in case we decide this proves useful in the future
	AnalyticsManager.setAllHitDimensions()
	Analytics.sendEvent("GWI " .. Platform.package, "HatPromptBreakdown", animName, 0)
end

AnalyticsManager.GWIPromptBegun = function(self)
	Analytics.sendView("GWI_prompt " .. Platform.package)
	self.setAllHitDimensions()
	Analytics.sendEvent("GWI " .. Platform.package, "Impression", "Prompt", 0)

	-- hack to circumvent the incredible crazy state machine that is GWI
	promptReason = "hat_prompt"

	-- mark down a time stamp so we can establish the time until click
	timeOfFirstGWI = Time.seconds()

	-- mark down the date if this was our first presentation
	AnalyticsManager:recordDayOfFirstPrompt()
	if Calendar.getDateDifferenceInDays ~= nil then 
		local year = Persist.duet.yearOfFirstRun;
		local month = Persist.duet.monthOfFirstRun;
		local day = Persist.duet.dayOfFirstRun;

		local daysSinceFirstPrompt = math.abs(Calendar.getDateDifferenceInDays(day, month, year))
		Analytics.setCustomDimension( ANALYTICS_DAYS_SINCE_FIRST_PROMPT_DIMENSION, mangleString(daysSinceFirstPrompt))
	end

	-- we are also interested in how many times the hat prompt has been presented, regardless of it being accepted
	Persist.duet.promptsSinceLastEngagement = Persist.duet.promptsSinceLastEngagement + 1
	Analytics.setCustomDimension( ANALYTICS_NUMBER_OF_HAT_PROMPTS_DIMENSION, mangleString(Persist.duet.promptsSinceLastEngagement) )
end

AnalyticsManager.kickupHappened = function(self)
	local timeUntilAccept = Time.seconds()-timeOfFirstGWI

	-- send an event describing how long it took to reach kickup
	-- note, this event will not be sent for show starts that had no hat, so it should not be used for user counts
	self.setAllHitDimensions()
	Analytics.sendEvent("GWI " .. Platform.package, "Success", "TimeBetweenGWIAndAccess", timeUntilAccept)
	Analytics.sendEvent("MangleStore " .. Platform.package, "GWI_Accept", mangleString(timeUntilAccept), timeUntilAccept)
end

AnalyticsManager.GWIAbandoned = function(self)
	-- previously we aborted the session here but now we are going to do it from the global shutdown
	Analytics.sendEvent("GWI " .. Platform.package , "Failure", "PromptFadedForever", 0)
	Persist.duet.pendingUltimateAccess = true
end

AnalyticsManager.showComplete = function(self)

	showIsPlaying = false
	
	-- increment the number of times the show has been completed, but don't update the custom dimension
	-- updating it now would retroactively update all of the hits in the session so far
	Persist.duet.episodesCompleted = Persist.duet.episodesCompleted + 1

	Analytics.setCustomDimension(ANALYTICS_REASON_FOR_ATTRITION, "ShowComplete")
	self.setAllHitDimensions()

	local timeInShow = math.floor(calculateTimeDelta(startTime, Time.milliseconds() ) / 1000)
	Analytics.sendEvent("GWO " .. Platform.package, "ShowComplete", "", timeInShow)

	self.onShowEnded()
end

AnalyticsManager.onActivityEvent = function(self, eventString)
	if (eventString == "onUserLeaveHint") then
		AnalyticsManager:onUserLeaveHint()
	end
end

AnalyticsManager.onUserLeaveHint = function(self)
	unmatchedOnUserLeaveHint = true
end

AnalyticsManager.userLeft = function(self)

	if showIsPlaying == true then 
		-- if user has hinted a leave, it is implied that the next onstop was triggered by the user
		if unmatchedOnUserLeaveHint == true then
			unmatchedOnUserLeaveHint = false
			Analytics.setCustomDimension(ANALYTICS_REASON_FOR_ATTRITION, "UserLeft")
		else
			Analytics.setCustomDimension(ANALYTICS_REASON_FOR_ATTRITION, "UserInterrupted")
		end

		self.setAllHitDimensions()

		local timeInShow = math.floor(calculateTimeDelta(startTime, Time.milliseconds() ) / 1000)
		Analytics.sendEvent("Attrition " .. Platform.package, lastObservedScene or "", mangleString(AnalyticsManager:getTimeBinSceneBased()), timeInShow)

		if timeInShow <= ANALYTICS_SECONDS_UNTIL_SUSTAINED_ENGAGEMENT then
			-- it's possible the user is going to return at a later point, and this allows us to qualify those users
			Persist.duet.pendingUltimateAccess = true

			-- report a quick abort
			self.setAllHitDimensions()
			Analytics.sendEvent("GWI " .. Platform.package, "Failure", "QuickAbort", timeInShow)
		end
	end

	self.onShowEnded()
end

AnalyticsManager.getAppType = function(packageName)
    if (packageName) then
        for i, appType in pairs(analyticsAppTypes) do
            if string.find(packageName, appType.appName) then
                return appType.appType or ""
            end
        end
    end
    return "APP_TYPE_OTHER"
end

AnalyticsManager.sendLookingUpEvent = function()
	-- warning this is not currently mutually exclusive with show starts so keeping it separate from ShowBegan action
	Analytics.sendEvent("GWI " .. Platform.package, "ShowStartByLookingUp", Platform.package, 0)
end
