--[[
  Motorola Confidential Proprietary
  Motorola Advanced Technology and Projects
  Copyright (c) 2013 Motorola, LLC. All Rights Reserved.
]]
local Logger = TaggedLog("ResourceLoader", false)

ResourceLoader = {
    segmentData = {},
    resourceGroup = {},
    currentSegmentName = nil,
    globalResourceGroup = nil,

    FSM = nil,
}

function ResourceLoader:init()
    Logger:message("initializing")

    self:_loadGlobalResources()

    if StoryManager then
        self.FSM = StoryManager:getFSM("episode3")
    end

    -- Initialize segments from scene data
    if SceneData.segments then
        for _, segment in ipairs(SceneData.segments) do
            self.segmentData[segment.name] = segment
        end
    end

    Events.listen(ResourceLoader, "segmentPlaying")
end

function ResourceLoader:shutdown()
    if self.globalResourceGroup then
        self.globalResourceGroup:destroy()
        self.globalResourceGroup = nil
    end
    if self.resourceGroup then
        for name, resources in pairs(self.resourceGroup) do
            resources:destroy()
        end
        self.resourceGroup = {}
    end
    if self.FSM then
        self.FSM:destroy()
    end
end

function ResourceLoader:_createResourceGroup(name)
    local group = System.createResourceGroup(name)
    group.name = name
    return group
end

function ResourceLoader:_loadGlobalResources()
    local globalResourceGroup = self:_createResourceGroup("GlobalResources")
    if SceneData.globalResources then
        for _, entry in ipairs(SceneData.globalResources) do
            local type = entry[1]
            local uri = entry[2]
            globalResourceGroup:addResource(type, uri)
        end
    end
    globalResourceGroup:prefetch()
    self.globalResourceGroup = globalResourceGroup
end

function ResourceLoader:prefetchInitialSegments()
    if SceneData.initialPrefetchSegments then
        for _, segmentName in ipairs(SceneData.initialPrefetchSegments) do
            Logger:message("Prefetching initial segment: " .. segmentName)
            self:prefetchSegment(segmentName)
        end
    end
end

function ResourceLoader:prefetchComplete()
    return System.getPendingResourceCount() == 0
end

local function transitionName(transition)
    local name = transition
    if type(transition) == 'table' then
        name = transition.nextSegment
    end
    return name
end

function ResourceLoader:segmentPlaying(segmentName)
    if System.getPendingResourceCount() ~= 0 then
        Log.error("ResourceLoader: starting segment \"" .. segmentName .. "\" with " .. System.getPendingResourceCount() .. " pending resources to load.")
    end

    if segmentName == self.currentSegmentName then return end

    Logger:message("segment playing: " .. segmentName)

    -- find the next segments that we could transition to and prefetch each of them
    local segment = self.segmentData[segmentName]
    local prefetchingSegment = {}
    if segment.transitions then
        for _, transition in pairs(segment.transitions) do
            self:prefetchSegment(transitionName(transition))
        end
    end

    if self.currentSegmentName then
        segment = self.segmentData[self.currentSegmentName]
        if segment.transitions then
            -- free all segments we didn't transition to
            for _, transition in pairs(segment.transitions) do
                local nextSegment = transitionName(transition)
                if nextSegment ~= segmentName and not prefetchingSegment[nextSegment] then
                    Logger:message("Releasing " .. nextSegment .. " from " .. segmentName)
                    self:releaseGroup(nextSegment)
                end
            end
        end
        Logger:message("Releasing " .. self.currentSegmentName .. " from " .. segmentName)
        if not prefetchingSegment[self.currentSegmentName] then
            self:releaseGroup(self.currentSegmentName)
        end
    end
    self.currentSegmentName = segmentName
end

function ResourceLoader:releaseGroup(name)
    local group = self.resourceGroup[name]
    if group then
        group:destroy()
        self.resourceGroup[name] = nil
    end
end

function ResourceLoader:_prefetchActors(actorList, group)
    if actorList then
        for _, actor in ipairs(actorList) do
            local actordef = SceneData.actors[actor]
            if actordef then
            	if(actordef.model) then
            		actordef.model = string.lower(actordef.model)
            	end
            	
            	-- make sure all animation paths are lower-case
			    if actordef.animation then
			        for _, anim in pairs(actordef.animation) do
			            anim[1] = string.lower(anim[1])
			        end
			    end
            	
                Logger:message("pre-fetching actor: " .. tostring(actor))
                group:addActor(actordef)

                -- Handle any idle flipbook materials
                if actordef.fbdrivers then
                    for _, fbdata in pairs(actordef.fbdrivers) do
                        for _, fbmatdata in pairs(fbdata) do
                            if fbmatdata.uri then
                                group:addFlipbook(fbmatdata.uri)
                            end
                        end
                    end
                end

                if actordef.prefetch then
                    for _, entry in ipairs(actordef.prefetch) do
                        local type = entry[1]
                        local uri = entry[2]
                        Logger:message("Prefetching: "  .. tostring(type) .. " - " .. tostring(uri))
                        group:addResource(type, uri)
                    end
                end
            end
        end
    end
end

function ResourceLoader:_prefetchAudio(segmentName, group)
    if Global.audioPlayer then
        Global.audioPlayer:prefetchSegment(segmentName, group)
    end
end

function ResourceLoader:getResourceGroupForSegment(segmentName)
    local group
    if not segmentName then
        segmentName = SceneData and SceneData.startSegment
    end
    local segment = self.segmentData[segmentName]
    if segment then
        group = self:_createResourceGroup(segmentName)

        self:_prefetchActors(table.flatten(segment.active), group)
        self:_prefetchActors(table.flatten(segment.active_postRot), group)
        self:_prefetchActors(table.flatten(segment.prefetch), group)

        if Global.audioEnabled then
            self:_prefetchAudio(segmentName, group)
        end

        if segment.particleSystems then
            for _, uri in ipairs(segment.particleSystems) do
                group:addParticleSystem(uri)
            end
        end
    end
    return group
end

-- prefetch a segment - prefetches all the resources
-- for all the actors, sounds, materials, etc in the segment
function ResourceLoader:prefetchSegment(segmentName)
    Logger:message("prefetchSegment: " .. tostring(segmentName))
    if not segmentName then
        segmentName = SceneData and SceneData.startSegment
    end
    if self.resourceGroup[segmentName] then return end
    local segment = self.segmentData[segmentName]
    if segment then
        local group = self:getResourceGroupForSegment(segmentName)
        self.resourceGroup[segmentName] = group

        Logger:message("prefetching segment " .. tostring(segment.name) .. " during " .. tostring(self.currentSegmentName))
        group:prefetch()

        -- If the segment is too short to prefetch its next segments, then we need
        -- to prefetch them when we prefetch this one
        if segment.prefetchTransitions and segment.transitions then
            for _, transition in pairs(segment.transitions) do
                local nextSegment = transition
                if type(transition) == 'table' then
                    nextSegment = transition.nextSegment
                end
                self:prefetchSegment(nextSegment)
            end
        end

        -- Prefetch special groups with this segment, but hold on
        -- to them until manually told to release them (via ResourceLoader:releaseGroup())
        if segment.prefetchGroups then
            for groupName, resources in pairs(segment.prefetchGroups) do
                if self.resourceGroup[groupName] then break end
                self.resourceGroup[groupName] = self:_createResourceGroup(groupName)
                for _, resource in ipairs(resources) do
                    local type = resource[1]
                    local uri = resource[2]
                    self.resourceGroup[groupName]:addResource(type, uri)
                end
                self.resourceGroup[groupName]:prefetch()
            end
        end

        if self.FSM then
            self.FSM:prefetchState(segmentName)
        end

        if self.FSM and segment.transitions then
            -- have the story manager prefetch the chunks for the next segments
            for _, transition in pairs(segment.transitions) do
                self.FSM:prefetchState(transitionName(transition))
            end
        end

    else
        Logger:message("unable to find segment name = " .. tostring(segmentName))
    end
end
