local Logger = TaggedLog.new("Segment", true)

local activatedObjs = {}

Segment = {
    hasBegun = false,
    playing = false,

    delayTimer  		= 0,
    motionMin			= 0,
    transitionBlocked	= false,

    offsetRots = {},

	fromAngle = nil, -- temporary data saved during transitions for alignment
}

function Segment:new(data)
    local o = data or {}

    o.animated = table.flatten(o.animated)
    o.active = table.flatten(o.active)

    Logger:message(o.name .. " active = " .. table.tostring(o.active))
    Logger:message(o.name .. " animated = " .. table.tostring(o.animated))

    if (o.active_postRot) then
        o.active_postRot = table.flatten(o.active_postRot)
        Logger:message(o.name .. " active_postRot = " .. table.tostring(o.active_postRot))
    end
    if (o.counterRot) then
        o.active_postRot = table.flatten(o.active_postRot)
        Logger:message(o.name .. " active_postRot = " .. table.tostring(o.active_postRot))
    end
    
    -- Add any components needed
    if (o.components) then
        for actorName, actorCompList in pairs(o.components) do
            
            -- add conditional segment data (makes this component active for given segments)
            
            for _, component in ipairs(actorCompList) do
            	
            	local class = _G[component.name]
	            if (class == nil) then
	                Log.error("Unknown component type: " .. tostring(component.name))
	            end
	            
            end
                
        end
    end

    setmetatable(o, self)
    self.__index = self

    return o
end

function Segment:update()
    -- offset debug
	--[[
    local count = 0;
    for segment, transition in pairs(self.transitions) do
        local leadActor = ActorManager:getActor(transition.leadActor)
        if(leadActor) then
        	
        	count = count + 1;
	        local rot = leadActor:getWorld().rotation;
			local textStyle = { color = Color.new(1.0, 0.0, 0.0), font = "common:fonts/arial_default_24.fnt", scale = 1.0 }
	    	Renderer.drawText(10, 190 + 30*(count - 1), textStyle, string.format("%s YPR <%.3f, %.3f, %.3f>", transition.leadActor, rot:getYawAngle(), rot:getPitch(), rot:getRollAngle()))
    	end
    end
    --]]

    if (self.playing) then
    	local actor = nil
	    local lookedAt = self:lookingAtTrigger()	
    	
    	if(lookedAt) then    		
    		local trigger = ActorManager:getActor(lookedAt)
    		if(trigger.keyActor) then
	        	actor = ActorManager:getActor(trigger.keyActor)
    		end
    	end
    	
	    local currentFinished = false	
    	
    	if(actor == nil) then 
    		if (self.animated[1]) then 
    			actor = ActorManager:getActor(self.animated[1])
    		else
            	currentFinished = true
    		end
        end
        
        if(actor ~= nil) then
    		currentFinished = actor.animDriver:isFinished(true) or self.ignoreAnimEnd
        end
            
        -- If the following conditions are met then we're ready for the next
        if (currentFinished) then
			self:goToNextSegmentIfReady(lookedAt)
        end
    end
end

function Segment:goToNextSegmentIfReady(lookedAt)
    local dt = Time.dt()
    self.delayTimer = self.delayTimer - dt
            
	if ((not self.needsFocus or lookedAt) and self.transitionBlocked == false and self.delayTimer <= 0 and (math.abs(SegmentManager.deltaYaw) / dt >= self.motionMin) )then
        Logger:message(self.name .. ": currentFinished")
        self:prepareNextSegment(lookedAt)
    end
end


function Segment:startPlaying(startingCheckpoint, startFrameOffset)
    Logger:message("Starting to play: " .. self.name .. " at frame " .. tostring(startFrameOffset))

    self:activateObjects(self.active, true)
 
    self:setupOmniLights()
    self:setupVignettes()

    self.playing = true

    local startTime = 0
    if startFrameOffset then
        startTime = startFrameOffset * SceneData.defaultFrameRate
    end

    self:initializeAnimDrivers()
    self:playAnimations(startTime)

    self.hasBegun = true

    self.delayTimer = self.delay or 0

    -- Look at the main actors if we are starting from a checkpoint
    if self.checkpointTriggered and self.checkpoint and self.checkpointLocation then
        Global.scene.mainCamera.controller:lookAt(self.checkpointLocation)
    end
end

function Segment:stopPlaying()
    self.playing = false
end


function Segment:initializeAnimDrivers()
    for i, actorName in ipairs(self.animated) do
        if (actorName) then
            local actor = ActorManager:getActor(actorName, false)
            if (actor and actor.animDriver) then
                local anim = actor.animDriver:init(self.name, actor.animation, i == 1, actor.fbdrivers, actor.frameEvents)
            end
        end
    end
end

function Segment:playAnimations(startTime)
    for _, actorName in ipairs(self.animated) do
        if (actorName) then
            local actor = ActorManager:getActor(actorName, false)
            if (actor and actor.animDriver) then
                Logger:message("Playing " .. self.name .. " on " .. actorName)
                actor.animDriver:play(startTime)
            end
        end
    end
end


function Segment:lookAtMainActor()
	
    if (self.animated[1] and Global.scene.mainCamera and Global.scene.mainCamera.controller) then
        local actor = ActorManager:getActor(self.animated[1])
        if (actor) then
            -- TODO: this might need to do a lookup on bone, rather than the actor root
            
            Global.scene.mainCamera.controller:lookAt(actor.targetBone:getWorld().position)
        end
    end
end


function Segment:lookingAtTrigger()
    if (self.transitions == nil) then
        return nil
    end

    for target,_ in pairs(self.transitions) do
        if (target) then
            local detector = ActorManager:getActor(target, false)
            if (detector and detector.components) then
                detector = detector.components["LookAtDetector"]
                if (detector and detector:wasLookedAt()) then
                    return target
                end
            end
        end
    end

    return nil
end


function Segment:prepareNextSegment(target)
    if (target == nil) then
        target = "default"
    end

    if (self.transitions[target]) then
    	if(type(self.transitions[target]) == "string") then
        	SegmentManager:startSegment(self.transitions[target])
        else
        	SegmentManager:startSegment(self.transitions[target].nextSegment, self.transitions[target].leadActor)
        end
    else
        self:stopPlaying()
        Events.send(self, "FinalSegment")
    end
end

function Segment:currentTime()
    if (self.animated and #self.animated > 0) then
        local actor = self.animated[1]
        if (actor and actor.animDriver) then
            return actor.animDriver:currentTime()
        end
    end
    return 0
end

function Segment:activateObjects(names, removeExceptions)
    if (names == nil) then
        return
    end

    local remove = shallowcopy(activatedObjs)
    local activate = {}

    -- Find the objects that need to be activated (and not removed)
    for _, name in ipairs(names) do
        if (name) then
            if (not activatedObjs[name]) then
                activate[name] = true
            end
            remove[name] = nil
        end
    end

    -- Activate any new objects
    local newActivations = {}
    if (activate) then
        for name, value in pairs(activate) do
            local actor = ActorManager:getActor(name, true)
            newActivations[name] = true
            if (actor) then
                
                if (Global.useDotCharacters) then
                    -- Special mode for a debug build that disables the art and has only colored spheres
                    -- This will likely become unncessary and removably very quickly
                    actor:setVisibility(false) 
                    if (actor["targetBone"] ~= nil) then
                        local debugName = "debugDot_"..name.."_girl"
                        actor:addComponent(debugName, DebugDot:new({ boneName = "targetGirl", label = debugName, dotColor = Color(0.8, 0.1, 0.6, 0.4), }))
                        debugName = "debugDot_"..name.."_boy"
                        actor:addComponent(debugName, DebugDot:new({ boneName = "targetBoy", label = debugName, dotColor =  Color(0.2, 0.2, 0.8, 0.4), }))
                        debugName = "debugDot_"..name.."_dog"                        
                        actor:addComponent(debugName, DebugDot:new({ boneName = "targetDog", label = debugName, dotColor = Color(0.1, 0.6, 0.6, 0.4), radius = 70 }))
                    end             
                else
                    actor:setVisibility(true)
                end
                
            end
            activatedObjs[name] = actor
            Log.message("Activating: " .. name)
        end
    end
    
    -- Apply offset to all actors from the segment's leadActor
    local lastOffset = Global.segmentOffset:copy()
    
	if self.leadActor then
        local leadActor = ActorManager:getActor(self.leadActor)
        if leadActor and leadActor.offsetBonePath and not self.ignoreOffset then
            Global.segmentOffset =  leadActor:getBoneWorld(leadActor.offsetBonePath).rotation
        end
    end
    
    -- force hiding/showing of actor
    if (self.setVisibility) then
    	for name, vis in pairs(self.setVisibility) do
    		local actor = ActorManager:getActor(name, false)
    		if(actor) then
    			actor:setVisibility(vis);
    		end
    	end
    end
	
	if(Global.segmentOffset) then	
        -- Discard non-yaw rotation and store only that
		local y, yq = Global.segmentOffset:getYaw()

		Global.segmentOffset = yq
	
		for _, name in ipairs(names) do
			if (name) then
				local actor = ActorManager:getActor(name, false)
				if (actor and actor.attachedTo == nil) then
					local newOffset = SRT.new()
					
					self.useLastOffset = self.useLastOffset or {}
					if(self.useLastOffset[actor.name]) then
						newOffset.rotation = lastOffset:copy()
					else
						newOffset.rotation = Global.segmentOffset:copy()
					end
					
					local offsetBy90Table = self.offsetBy90 or {}
					if(offsetBy90Table[actor.name]) then
						local direction = offsetBy90Table[actor.name] == "left" and 1 or -1
						local leadActor = ActorManager:getActor(self.leadActor)
						if(leadActor and actor.actorGroup ~= leadActor.actorGroup) then
							local y, qy = leadActor:getWorld().rotation:getYaw()
							newOffset.rotation = qy*Quaternion.axisAngle(Vector.up, direction*90);
						end
					end

                    -- offset debug
                    --[[
                    Log.message(string.format("-------------------- "..self.name.." applying offset <%3.f, %3.f, %3.f>", newOffset.rotation:getPitch(), newOffset.rotation:getYawAngle(), newOffset.rotation:getRollAngle()))
                    --]]

					-- if ignoring offsets do nothing unless this is a new activation
					if( (newActivations[name] or self.useLastOffset[actor.name] or not self.ignoreOffset) and not actor.ignoreSegmentOffsets) then
						actor:setLocal(newOffset);
					end
				end
			end
		end
	end
	
    -- attach components specified in the segment
    self:configureComponents(names, self.components);
    
    -- run any methods attached to the segment
    self:runMethods(self.methods)
    
    -- Remove any unneeded objects
    if (removeExceptions and remove) then
        for name, _ in pairs(remove) do
            ActorManager:destroyActor(name)
            activatedObjs[name] = nil
            Logger:message("Removing: " .. name)
        end
    end
    
end

function Segment:setupOmniLights()
	-- destroy any lights in destroy list
	if(self.destroyOmniLights) then
		for _, light in ipairs(self.destroyOmniLights) do
			OmniLightBlendManager:destroyOmniLight(light)
		end
	end
	
	-- create any new lights
	if(self.createOmniLights) then
		for _, light in ipairs(self.createOmniLights) do
			OmniLightBlendManager:createOmniLight(light)
		end
	end
end

function Segment:setupVignettes()
    -- Update the vignette component for each active actor to give them unique index values
    local index = 0
    if self.active then
        for _, actorName in ipairs(self.active) do
            local actor = ActorManager:getActor(actorName, false)
            if actor and actor.components then
                local vignette = actor.components.Vignette
                if vignette then
                    vignette.index = index
                    index = index + vignette.numTargets
                end
            end
        end
    end
end

function Segment:runMethods(methods)
	methods = methods or {}
	for actorName, methodList in pairs(methods) do
	
		local actor = ActorManager:getActor(actorName, false);
		
		if(methodList) then
    		-- run all methods found in the data
    		for name, method in pairs(methodList) do
    			if(type(methodList[name]) == "function") then
    				methodList[name](self, actor);
    			end
    		end
    	end
	end
end

function Segment:configureComponents(names, segmentComponents)
    	
	local actorTable = {}
	local compRemoveList = {}
	segmentComponents = segmentComponents or {}
	
	if(names) then
		for _, name in ipairs(names) do
	        if (name) then
	            local actor = ActorManager:getActor(name, false);
	            
	            -- get all existing components that have been added through segments on an actor
	            for name, comp in pairs(actor.components) do
					-- get string up to first '_' (underscore)
					if(string.match(name, "^.-_") == "segcomp_") then
						compRemoveList = compRemoveList or {}
						compRemoveList[name] = actor.name
					end
				end
	        end
	    end
    end
	
	for actorName, actorCompList in pairs(segmentComponents) do
	
		local actor = ActorManager:getActor(actorName, false);
		
		if(actor) then
		
			-- flag all segment components for removal unless they are re-added
--			local compRemoveList = {}
			
			if(actorCompList) then
		    	for _, data in ipairs(actorCompList) do
		    		
		    		local compName = "segcomp_"..data.name;
		    		
		    		-- do not remove this component
		    		compRemoveList[compName] = nil
		    		
		    		-- add the component, if it has not been added
		    		if(actor.components[compName] == nil) then
		    			local class = _G[data.name]
		    			actor:addComponent(compName, class:new(data.params))
		        		Logger:message("Segment: "..self.name.." ADDED component: " .. data.name .. " to " .. actorName)
		    		else
		        		Logger:message("Segment: "..self.name.." KEEPING component: " .. data.name .. " on " .. actorName)
		        		local comp = actor.components[compName];
		        		if(comp.onReinitialize) then
		        			comp:onReinitialize(data.params)
		        		end
		    		end
		    		
		    	end
	    	end
		end
	end
	
	-- remove anything still flagged for removal
	for name, owner in pairs(compRemoveList) do
		if(owner ~= nil) then
			local actor = ActorManager:getActor(owner, false);
		
			actor:removeComponent(name)
			Logger:message("Segment: "..self.name.." REMOVED component: " .. name .. " from " .. owner)
		end
	end
end
