local Logger = TaggedLog.new("SegmentManager", false)

-- Used to decide whether to reset to beginning of current segment
-- or to jump to the one that came before when using "Previous"
local goToPrevThreshold = 1.5

SegmentManager = {
    name = "SegmentManager",

    currentYaw  = 0,
    prevYaw     = 0,
    deltaYaw    = 0,
    leadActor 	= nil,
 	anchors		= {},
 	segmentComponents = {}, -- { Segmentname = {actorname = {componentData1, componentData2, componentData3}} }
    showStarted = false,
	useNative   = segmentmanager and segmentmanager.updateDeltaYaw,
    waitForStreamingState  = nil,
}
SegmentManager.__index = SegmentManager

function SegmentManager:initSegments()
    self.current = nil
    self.segments = {}
    self.history = {}
    self.starterSegment = nil

    if StoryManager then
        self.FSM = StoryManager:getFSM("episode3")
    end

    local firstSegmentName
    -- Initialize segments from scene data
    if (SceneData.segments) then 
        if (SceneData.startSegment) then
			firstSegmentName = SceneData.startSegment
            Persist.episode3.lastStartSegment = SceneData.startSegment
		end
        for _, segmentData in ipairs(SceneData.segments) do
            local segment = Segment:new(segmentData)
            self.segments[segment.name] = segment
            firstSegmentName = firstSegmentName or segment.name
        end
    end

    self.starterSegment = firstSegmentName
    if(Context.getCheckpoint()) then
    	self.starterSegment = Context.getCheckpoint()
    end

    if Platform.segment and Platform.segment ~= "" then
        self.starterSegment = Platform.segment
    end
    
    self.segments[self.starterSegment].checkpointTriggered = true
end

function SegmentManager:onStart()
    if not self.segments then
        self:initSegments()	
    end

    if Global.pauseOnStart then
        self:pause()
    end

    self.initialized = true
end

function SegmentManager:onUpdate()
    if not self.initialized then return end

    -- Wait to start the initial segment until we have loaded everything needed to start the show
    if (Platform.scene and Platform.scene ~= "") and not self.showStarted and not ResourceLoader:prefetchComplete() then return end    

    if not self.showStarted then
        self.showStarted = true
        if (self.starterSegment) then
            Logger:message("Starting segment: " .. self.starterSegment)
            self:startSegment(self.starterSegment)
        end
    end

    if self.waitForStreamingState then
        self:startSegment(self.waitForStreamingState[1], self.waitForStreamingState[2])
    end

    -- Used for determining user "intention" by camera direction
	if ( self.useNative ) then
		self.deltaYaw = segmentmanager.updateDeltaYaw(Global.scene.mainCamera:getWorld().rotation)
	else
		self.prevYaw = self.currentYaw
		self.currentYaw = Global.scene.mainCamera:getWorld().rotation:getYawAngle()

		local prevRot       = Quaternion.axisAngle(Vector(0.0, 1.0, 0.0, 0.0), self.prevYaw)
		local currentRot    = Quaternion.axisAngle(Vector(0.0, 1.0, 0.0, 0.0), self.currentYaw)
		local deltaRot      = prevRot:inverse() * currentRot

		self.deltaYaw = deltaRot:getYawAngle()
		if (self.deltaYaw > 180) then
			self.deltaYaw = -360 + self.deltaYaw
		end
	end

    if (Input.isKeyDown(KB_N, true)) then
        Logger:message("going next")
        self:next()
    elseif (Input.isKeyDown(KB_P, true)) then
        Logger:message("going previous")
        self:previous()
    elseif (Input.isKeyDown(KB_SPACE, true)) then
        Logger:message("toggling pause")
        self:togglePause()
    elseif (Input.isKeyDown(KB_M, true)) then
        Logger:message("stepping forward")
        self:step()
    end

    if (self.current) then
        self.current:update()
    end
end

function SegmentManager:next()
    if (self.current == nil or self.current.transitions == nil) then return end

    local firstTransitionKey, firstTransitionName = next(self.current.transitions)
    if (firstTransitionName) then
    	local transName = firstTransitionName;
    	if(type(firstTransitionName) == "table") then
    		transName = firstTransitionName.nextSegment;
    		Logger:message("next transition to: " .. transName)
    		
        	self:startSegment(firstTransitionName.nextSegment, firstTransitionName.leadActor)
    	else
	        Logger:message("next transition to: " .. transName)
	        self:startSegment(firstTransitionName)
        end
    end
end


function SegmentManager:previous()
    if (self.current == nil) then return end

    local goToPrevious = (self.current:currentTime() < goToPrevThreshold)
    local numHistory = #self.history

    -- If we are below the threshold and we have previous segment history then go to it
    if (numHistory > 1 and goToPrevious) then
        local prevSegment = self.history[numHistory - 1]

        table.remove(self.history)
        numHistory = numHistory - 1
        self.current = prevSegment
--        if (Stat) then
--            Stat.segment:set(self:currentDisplayName())
--        end
    end

    -- Play normally
    self.current:startPlaying()

    self:notifySegmentPlaying()
end

function SegmentManager:togglePause()
    if self.paused then
        self:resume()
    else
        self:pause()
    end
end

function SegmentManager:pause()
    Time.pause()
    self.paused = true
end

function SegmentManager:resume()
    Time.resume()
    self.paused = false
end

function SegmentManager:step()
    if (not self.paused) then
        self:pause()
    end
    Time.step()
end

function SegmentManager:blockTransition(val)
	if (self.current == nil) then return end

	self.current.transitionBlocked = val
end

function SegmentManager:startSegment(name, leadActor)

	if(Global.loadDelay ~= nil and System.resetWaitCounter ~= nil) then
		System.resetWaitCounter()
		Log.message("Wait Counter Reset")
	end

    local segment = self.segments[name]
    if (segment == nil) then
        Log.error("startSegment fail, segment not found: " .. name)
        return
    end

    if self.FSM then
        local resourcesLoaded = self.FSM:setState(name)
        if not resourcesLoaded then
            Logger:message("waiting for segment resources: " .. name)
            self.waitForStreamingState = { name, leadActor }
            return
        else
            self.waitForStreamingState = nil
        end
    end

    Logger:message("startSegment: " .. name)

    -- figure out which actor to pivot on
    local newLeadActors = segment.leadActors or {}
    local index = table.find_index( newLeadActors, function(name) return self.leadActor == name; end );
    
    -- the new way to find lead actor
	self.leadActor = nil;
    if(leadActor) then 
    	self.leadActor = leadActor;
    elseif(self.current and self.current.leadActor) then
		-- if there's a previous segment, grab the leadActor from there
		self.leadActor = self.current.leadActor;		
	end
    
	-- the only way to get a lead actor now is through the argument (usually from transition parameters)
    segment.leadActor = leadActor;
    
    self.current = segment
    table.insert(self.history, self.current)

    local isStartingCheckpoint = (Context and Context.getCheckpoint() == self.current.name)
    local frameOffset = self.current.name == Platform.segment and Platform.startFrameOffset
    self.current:startPlaying(isStartingCheckpoint, frameOffset)
    
    self:notifySegmentPlaying()

	if(Global.loadDelay ~= nil and System.fetchWaitCounter ~= nil) then
		Global.loadDelay = System.fetchWaitCounter()
		Log.message("Loading delay: "..Global.loadDelay)
	end
end

function SegmentManager:lookAtCurrent()
    if (self.current) then
        self.current:lookAtMainActor()
    end
end

function SegmentManager:currentSegmentName()
    return (self.current and self.current.name) or "EMPTY_SEGMENT"
end

function SegmentManager:notifySegmentPlaying()
    local segmentName = self:currentSegmentName()
    if System.getPendingResourceCount() ~= 0 then
        Logger:message("Starting segment " .. segmentName .. " with " .. System.getPendingResourceCount() .. " pending resources to load")
    end

    -- let the engine know the segment for diagnostics
    if Scene.setActiveSegment then
        Global.scene:setActiveSegment(segmentName)
    end

    -- update the stats with the latest segment
    if System.isStatManagerEnabled() then
        Stat.segment:set(segmentName)
        Stat.segmentTime:set(0)
    end

    if (self.current and self.current.checkpoint) then
        Context.onCheckpoint(self.current.name)
    end
	
	if(Replay.setMarker) then
		Replay.setMarker(segmentName);
	end
	
    Logger:message("Segment playing: " .. segmentName)
    Events.send(self, "segmentPlaying", segmentName, self.current)
end
