
require "episode3/scripts/util/stableids"

SegmentGraph = {}

local DefaultStartSegments = { "Cell_Girl", "Cell_Boy" }

function SegmentGraph:new(startSegments)
    local instance = { startSegments = startSegments or DefaultStartSegments }

    setmetatable(instance, self)
    self.__index = self

    instance:_init()
    return instance
end

function SegmentGraph:_init()
    self:_initSegments()
    self:patchActorWithBaked()
    self.scene = Scene.create()
    self.soundEmitter = AudioEmitter.create()
end

function SegmentGraph:_initSegments()
    self.segments = {}

    -- Initialize segments from scene data
    if (SceneData and SceneData.segments) then
        for _, segment in ipairs(SceneData.segments) do
            self.segments[segment.name] = segment
        end
    end
end

function SegmentGraph:_cleanup()
    if self.scene then
        self.scene:destroy()
        self.scene = nil
    end

    if self.soundEmitter then
        self.soundEmitter:destroy()
        self.soundEmitter = nil
    end
end

local function transitionSegment(transition)
    if type(transition) == "table" then
        return transition.nextSegment
    end
    return transition
end

local function leadActor(transition)
    if type(transition) == "table" then
        return transition.leadActor
    end
    return nil
end

-- Breadth first traversal of the segments.  Will call back the
-- passed in visit function with the name and segment for each segment
-- the show.  I.e. visit(name, segment).
function SegmentGraph:breadthFirstTraverseSegments(visit)
    local visited = {}
    local pending = deepcopy(self.startSegments)

    while #pending > 0 do
        local segmentName = table.remove(pending, 1)
        if not visited[segmentName] then
            local segment = self.segments[segmentName]
            if not segment then
                Log.error("SEGMENT " .. segmentName .. " NOT FOUND")
            else
                visit(segmentName, segment)
                visited[segmentName] = true

                if segment.transitions then
                    for _, transition in pairs(segment.transitions) do
                        table.insert(pending, transitionSegment(transition))
                    end
                end
            end
        end
    end
end

local function loadPersistentFile(baseFileName)
    local result
    pcall(function() result = require(Project.persistSceneDir .. baseFileName) end)
    return result
end

local function writePersistentFile(value, baseFileName)
    local luaValue = "return " .. Serpent.dumppretty(value)
    local file, err = io.open(Project.rawPersistSceneDir .. baseFileName .. ".lua", "w" )
    if not err and file then
        file:write(luaValue)
        file:close()
    end
end

-- GraphVis segment graph
function SegmentGraph:dumpSegmentGraph(filename, params)
    filename = filename or "segmentgraph.gv"
    params = shallowcopy(params) or {}
    if params.showTransitionNames == nil then params.showTransitionNames = true end
    if params.showAudio == nil then params.showAudio = false end

    local segmentIds = StableIDs:new({ name = "Segment", fileName = "segment_ids", prefix = "s", })
    local animIds = StableIDs:new({ name = "Animation", fileName = "anim_ids", prefix = "a", })

    local ignoreSegments = {}

    local file, error = io.open(filename, "w")

    file:write("digraph Segments {\n")
    file:write("  rankdir=TB\n")
    file:write("  node [shape=\"record\" margin=0, fontname=\"helvitica\"]\n")
    file:write("  edge [fontname=\"helvetica\" fontsize=\"14\" fontcolor=\"#888888\"]\n")

    local function dumpSegment(name, segment)
        if not segment then Log.message("SEGMENT "..name.." Not FOUND") return end
        if ignoreSegments[name] then return end

        local segmentId = segmentIds:getId(name)
        local segmentColor = "#00aa00"
        if segment.needsFocus then
            segmentColor = "#aa0000"
        end
        local segmentCheckpoint = ""
        if segment.checkpoint then
            segmentCheckpoint = " C"
        end
        file:write("  " .. name .. " [label=<<TABLE BORDER=\"0\" CELLBORDER=\"1\" CELLSPACING=\"0\"><TR><TD COLSPAN=\"3\" BGCOLOR=\"" .. segmentColor .. "\"><FONT COLOR=\"white\">" .. name .. " [" .. segmentId .. "]" .. segmentCheckpoint .. "</FONT></TD></TR>\n")
        if segment.animated then
            for _, actorName in ipairs(segment.animated) do
                local actorData = SceneData.actors[actorName]
                if actorData then
                    local anim = actorData.animation and actorData.animation[name] or nil
                    if anim then
                        local animName = string.match(anim[1], ".*/(.+)%.mxa")

                        local animId = animIds:getId(string.lower(animName))
                        local startFrame = anim[2]
                        local endFrame = anim[3]
                        if endFrame == -1 then
                            -- load the actor so we can get the animation's real end frame
                            local actor = self.scene:createActor(actorName, actorData)
                            if actor and actor.animation and actor.animation[name] then
                                endFrame = actor.animation[name]:getStopFrame()
                                actor:destroy()
                                Util.fullGC()
                            end
                        end
                        file:write("    <TR><TD ALIGN=\"LEFT\">[" .. animId .. "] " .. tostring(animName) .. "</TD><TD>" .. tostring(startFrame) .. "</TD><TD>" .. tostring(endFrame) .. "</TD></TR>\n")
                    end
                end
            end
        end

        if params.showAudio and SceneData.audio and SceneData.audio.segments[name] and SceneData.audio.segments[name].events then
            file:write("   <TR><TD COLSPAN=\"3\" BGCOLOR=\"#ffffdd\" FONTCOLOR=\"white\"><TABLE BORDER=\"0\" CELLBORDER=\"1\" CELLSPACING=\"0\">\n")
            file:write("   <TR><TD COLSPAN=\"2\">Audio</TD><TD>start</TD><TD>dur</TD><TD>vol</TD></TR>\n")
            for _, event in ipairs(SceneData.audio.segments[name].events) do
                local name = event.play or event.fade
                local eventType = (event.play and "play") or (event.fade and "fade") or "unknown"
                local color = "#eeeeee"
                if name then
                    local frame = (event.frame or 0) + (event.frameOffset or 0)
                    local duration = (event.duration and tostring(event.duration)) or ""
                    local volumeValue = event.volume or event.to
                    local volume = tostring(volumeValue or 0)
                    
                    if type(volumeValue) == "table" then
                        volume = ""
                        for targetName, value in pairs(volumeValue) do
                            volume = volume .. " " .. tostring(targetName) .. "=" .. tostring(value)
                        end
                    end
                    if event.fade then
                        if event.autorelease then
                            color = "#ffeeee"
                        else
                            color = "#eeffee"
                        end
                    end
                    file:write("    <TR><TD ALIGN=\"LEFT\" BGCOLOR=\"" .. color .. "\">" .. eventType .. "</TD><TD ALIGN=\"LEFT\" BGCOLOR=\"" .. color .. "\">" .. name .. "</TD><TD BGCOLOR=\"" .. color .. "\">" .. tostring(frame) .. "</TD><TD BGCOLOR=\"" .. color .. "\">" .. tostring(duration) .. "</TD><TD BGCOLOR=\"" .. color .. "\">".. volume .. "</TD></TR>\n")
                end
            end
            file:write("   </TABLE></TD></TR>\n")
        end
        file:write("  </TABLE>>]\n")

        if segment.transitions then
            local transitions = {}
            for _, transition in pairs(segment.transitions) do
                local transName = transitionSegment(transition)
                local transActor = leadActor(transition)
                transitions[transName] = transitions[transName] or {}
                if transActor then
                    transitions[transName][transActor] = true
                end
            end
            for transition, actors in pairs(transitions) do
                if not ignoreSegments[transition] then
                    local anyActor = false
                    if params.showTransitionNames then
                        for actor, _ in pairs(actors) do
                            file:write("  " .. name .. " -> " .. transition .. " [label = \" " .. actor .. "\"];\n")
                            anyActor = true
                        end
                    end
                    if not anyActor then
                        file:write("  " .. name .. " -> " .. transition .. "\n")
                    end
                end
            end
        end
        file:write("\n")
    end
    self:breadthFirstTraverseSegments(dumpSegment)

    file:write("}\n")
    file:close()

    segmentIds:save()
    animIds:save()

    self:_cleanup()
end

function SegmentGraph:dumpOrphanData()
    -- Walk the segment graph to find all referenced segments and actors
    local referencedSegments = {}
    local referencedActors = table.create_set(SceneData.keepActors) or {}
    local ignoreFBDriversFor = table.create_set(SceneData.keepFBDriversFor) or {}

    local function visitSegment(name, segment)
        referencedSegments[name] = true
        if segment.active then
            for _, actorName in ipairs(table.flatten(segment.active)) do
                referencedActors[actorName] = true
            end
        end
        if segment.animated then
            for _, actorName in ipairs(table.flatten(segment.animated)) do
                referencedActors[actorName] = true
            end
        end
    end
    self:breadthFirstTraverseSegments(visitSegment)

    -- Find any segments in the segment data not referenced by existing segments
    local orphanSegments = {}
    for _, segment in pairs(self.segments) do
        if not referencedSegments[segment.name] then
            table.insert(orphanSegments, segment.name)
        end
    end
    table.sort(orphanSegments)
    local isOrphanSegment = table.create_set(orphanSegments)

    -- Find any actors in the segment data not referenced by existing segments
    local orphanActors = {}
    local orphanAnimations = {}
    local orphanFbDrivers = {}
    for actorName, actorData in pairs(SceneData.actors) do
        if not referencedActors[actorName] then
            table.insert(orphanActors, actorName)
        else
            if actorData.animation then
                for animationName, _ in pairs(actorData.animation) do
                    local referencedAnim = false
                    local segName = animationName
                    if string.endsWith(segName, "_Idle") then
                        -- treat idles as the same as the base animation
                        segName = string.match(segName, "(.*)_Idle")
                    end
                    if self.segments[segName] then
                        if self.segments[segName].animated then
                            for _, animatedActor in ipairs(self.segments[segName].animated) do
                                if animatedActor == actorName then
                                    referencedAnim = true
                                    break
                                end
                            end
                        end
                    end
                    if not referencedAnim then
                        table.insert(orphanAnimations, actorName .. "." .. animationName)
                    end
                end
            end
            if actorData.fbdrivers and (not ignoreFBDriversFor[actorName])then
                for fbName, _ in pairs(actorData.fbdrivers) do
                    local segName = fbName
                    if string.endsWith(segName, "_Idle") then
                        -- treat idles as the same as the base animation
                        segName = string.match(segName, "(.*)_Idle")
                    end
                    if not self.segments[segName] then
                        table.insert(orphanFbDrivers, actorName .. "." .. fbName)
                    end
                end
            end
        end
    end
    table.sort(orphanActors)
    table.sort(orphanAnimations)
    table.sort(orphanFbDrivers)

    -- Find any transitions defined, but don't match any actor in the scene
    local orphanTransitions = {}
    for segmentName, _ in pairs(referencedSegments) do
        if self.segments[segmentName].transitions then
            for name, _ in pairs(self.segments[segmentName].transitions) do
                if not referencedActors[name] and name ~= "default" then
                    table.insert(orphanTransitions, segmentName .. "." .. name)
                end
            end
        end
    end
    table.sort(orphanTransitions)

    local file, error = io.open("orphans.txt", "w")
    if not error and file then
        file:write("Orphaned Segments: \n")
        for _, segment in ipairs(orphanSegments) do
            file:write("  " .. segment .. "\n")
        end
        file:write("\n")

        file:write("Orphaned Actors: \n")
        for _, actor in ipairs(orphanActors) do
            file:write("  " .. actor .. "\n")
        end
        file:write("\n")

        file:write("Orphaned actor animations: \n")
        for _, anim in ipairs(orphanAnimations) do
            file:write("  " .. anim .. "\n")
        end
        file:write("\n")

        file:write("Orphaned actor fbdrivers: \n")
        for _, fb in ipairs(orphanFbDrivers) do
            file:write("  " .. fb .. "\n")
        end
        file:write("\n")

        file:write("Orphaned segment transitions: \n")
        for _, transition in ipairs(orphanTransitions) do
            file:write("  " .. transition .. "\n")
        end
        file:write("\n")

        file:close()
    end
end

function SegmentGraph:dumpAnimationIDs()
    local animIds = StableIDs:new({ name = "Animation", fileName = "anim_ids", prefix = "a", })
    animIds:writeCSV("animIds")
end

function SegmentGraph:_resourceDuration(uri)
    local duration
    if string.endsWith(uri, ".mfb") then
        local fbSampler = FlipbookSampler.create()
        fbSampler:setFlipbook(uri)
        duration = fbSampler:getNumFrames() / 60
        fbSampler:destroy()
    elseif string.endsWith(uri, ".ogg") then
        local sound = self.soundEmitter:addSound(uri, uri, false, 0, -1)
        duration = sound:getDurationSeconds()
        self.soundEmitter:removeSound(uri)
    end
    return duration
end

function SegmentGraph:_animationDuration(actorName, animName)
    local actorData = SceneData.actors[actorName]
    if actorData and actorData.animation and actorData.animation[animName] then
        local anim = actorData.animation[animName]
        local startFrame = anim[2]
        local endFrame = anim[3]
        if endFrame == -1 then
            -- load the actor so we can get the animation's real end frame
            local actor = self.scene:createActor(actorName, actorData)
            if actor and actor.animation and actor.animation[animName] then
                endFrame = actor.animation[animName]:getStopFrame()
                actor:destroy()
                Util.fullGC()
            end
        end
        return (endFrame - startFrame) / 60
    end
end

function SegmentGraph:_minAnimationDuration(actorNames, segmentName)
    local minDuration
    for _, actorName in ipairs(actorNames) do
        local animDuration = self:_animationDuration(actorName, segmentName)
        if minDuration then
            minDuration = math.min(minDuration, animDuration)
        else
            minDuration = animDuration
        end
    end
    return minDuration
end

function SegmentGraph:dumpChunkData()
    local states = {}
    local resources = {}
    local order = 1

    local function convertSegmentResources(resourceGroup)
        local segmentResources = {}
        if resourceGroup.uris then
            for uri, _ in pairs(resourceGroup.uris) do
                local path = uri:gsub(":", "/")
                if not resources[path] then
                    resources[path] = { path = path, duration = self:_resourceDuration(uri) }
                end
                table.insert(segmentResources, path)
            end
        end
        return segmentResources
    end

    -- Determine show order for each segment by doing a breadth first traversal of the graph
    local function visitSegment(segmentName, segmentData)
        if not states[segmentName] then
            local state = {}
            states[segmentName] = state

            state.order = order
            order = order + 1

            state.resources = convertSegmentResources(ResourceLoader:getResourceGroupForSegment(segmentName))

            state.transitions = {}
            local minSegmentDuration
            if segmentData.transitions then
                for targetName, transition in pairs(segmentData.transitions) do
                    local keyActors
                    if targetName == "default" then
                        -- need to find the look-at target actor to figure out which animation is being tracked for
                        -- the min duration for this transition
                        for _, activeActorName in ipairs(segmentData.active) do
                            local actor = SceneData.actors[activeActorName]
                            if actor and actor.components then
                                for _, component in ipairs(actor.components) do
                                    if component.name == "LookAtDetector" then
                                        keyActors = { actor.keyActor }
                                        break
                                    end
                                end
                            end
                        end
                        keyActors = keyActors or segmentData.animated
                    else
                        keyActors = { SceneData.actors[targetName].keyActor }
                    end

                    local minTransitionDuration = self:_minAnimationDuration(keyActors, segmentName)
                    local childSegmentName = transitionSegment(transition)
                    state.transitions[targetName] = {
                        toSegment = childSegmentName,
                        minDuration = minTransitionDuration,
                    }
                    if minSegmentDuration then
                        minSegmentDuration = math.min(minSegmentDuration, minTransitionDuration)
                    else
                        minSegmentDuration = minTransitionDuration
                    end

                    if states[childSegmentName] then
                        -- child has already been populated.  Make sure it has a higher
                        -- order than this parent
                        states[childSegmentName].order = order
                        order = order + 1
                    end
                end
            end
            state.minDuration = minSegmentDuration
        end
    end
    self:breadthFirstTraverseSegments(visitSegment)

    -- Compute the base resources needed to start the show
    local baseResourceGroup = ResourceLoader.globalResourceGroup

    -- Special case the launcher and show start assets
    baseResourceGroup:addActor(LaunchData.actors.LauncherToNotification)
    baseResourceGroup:addActor(LaunchData.actors.GoodWayIn)
    baseResourceGroup:addActor(LaunchData.actors.GWI_Title)
    baseResourceGroup:addFlipbook("episode3:flipbooks/title.mfb")
    baseResourceGroup:addAudio("episode3:audio/mxloop_intro.ogg")
    for _, actorName in ipairs(SceneData.ShowStartActors) do
        baseResourceGroup:addActor(SceneData.actors[actorName])
    end

    local rootState = {}
    rootState.order = 0
    rootState.resources = convertSegmentResources(baseResourceGroup)
    states["BaseResources"] = rootState


    local filename = "episode3.xml"
    local file, error = io.open(filename, "w")

    file:write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
    file:write("<StoryManifest xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:noNamespaceSchemaLocation=\"ResourceManifestSchema.xsd\">\n")

    file:write("\t<Chunks>\n")
    for _, state in pairs(states) do
        local url = string.format("http://dl.google.com/spotlight/episode3/1/episode3_%03d.zip", state.order)
        file:write("\t\t<Chunk ID=\"" .. tostring(state.order) .. "\" URL=\"" .. url .. "\">\n")
        if state.resources then
            for _, name in ipairs(state.resources) do
                local resource = resources[name]
                if resource then
                    local duration = ""
                    if resource.duration then
                        duration = " duration=\"" .. string.format("%.2f", resource.duration) .. "\""
                    end
                    file:write("\t\t\t<Resource Name=\"" .. resource.path ..  "\"" .. duration .. "/>\n")
                end
            end
        end
        file:write("\t\t</Chunk>\n")
    end
    file:write("\t</Chunks>\n")

    file:write("\t<StateMachines>\n")
    file:write("\t\t<StateMachine Name=\"episode3\" Root=\"BaseResources\">\n")

    file:write("\t\t\t<States>\n")
    for name, state in pairs(states) do
        local minDuration = ""
        if state.minDuration then
            minDuration = string.format(" MinDuration=\"%.2f\"", state.minDuration)
        end
        file:write("\t\t\t\t<State Name=\"" .. name .. "\" Order=\"" .. tostring(state.order) .. "\"" .. minDuration .. ">\n")
        file:write("\t\t\t\t\t<Chunks>\n")
        file:write("\t\t\t\t\t\t<ChunkID>0</ChunkID>\n")
        file:write("\t\t\t\t\t\t<ChunkID>" .. tostring(state.order) .. "</ChunkID>\n")
        file:write("\t\t\t\t\t</Chunks>\n")
        if state.transitions then
            file:write("\t\t\t\t\t<Transitions>\n")
            for targetName, transition in pairs(state.transitions) do
                file:write("\t\t\t\t\t\t<Transition To=\"" .. transition.toSegment .. "\" MinDuration=\"" .. string.format("%.2f", transition.minDuration or 0) .. "\" Target=\"" .. targetName .. "\"/>\n")
            end
            file:write("\t\t\t\t\t</Transitions>\n")
        end
        file:write("\t\t\t\t</State>\n")
    end
    file:write("\t\t\t</States>\n")

    file:write("\t\t</StateMachine>\n")
    file:write("\t</StateMachines>\n")

    file:write("</StoryManifest>\n")
    file:close()
    self:_cleanup()

    Log.message("DMA: States = \n" .. Serpent.dumppretty(states))
    Log.message("DMA: Resources = \n" .. Serpent.dumppretty(resources))
end

function SegmentGraph:patchActorWithBaked()
    -- patch actor defs with the baked data
    for actorName, actorDef in pairs(SceneData.actors) do
        local bakedData = SceneData.bakedActorData and SceneData.bakedActorData[actorName] or {}
        if bakedData then
            if bakedData.materials then
                actorDef.materials = bakedData.materials
            end

            if bakedData.fbdrivers then
                actorDef.fbdrivers = bakedData.fbdrivers
            end
        end
    end
end


function SegmentGraph:computeSegmentTimes()
    for _, segmentData in pairs(self.segments) do
        local minSegmentTime
        local segmentName = segmentData.name
        local transitionTimes = {}
        if segmentData.transitions then
            segmentData.transitionInfo = {}
            for targetName, transition in pairs(segmentData.transitions) do
                local transitionInfo = {}
                local keyActors = {}
                if targetName == "default" then
                    -- need to find the look-at target actor to figure out which animation is being tracked for
                    -- the min duration for this transition
                    for _, activeActorName in ipairs(segmentData.active) do
                        local actor = SceneData.actors[activeActorName]
                        if actor and actor.components then
                            for _, component in ipairs(actor.components) do
                                if component.name == "LookAtDetector" then
                                    keyActors = { actor.keyActor }
                                    break
                                end
                            end
                        end
                    end
                    keyActors = keyActors or segmentData.animated
                else
                    keyActors = { SceneData.actors[targetName].keyActor }
                end

                local minTransitionFrames
                if keyActors then
                    for _, keyActor in ipairs(keyActors) do
                        local animFrames = animationLength(keyActor, segmentName, self.scene)
                        if minTransitionFrames then
                            minTransitionFrames = math.min(minTransitionFrames, animFrames)
                        else
                            minTransitionFrames = animFrames
                        end
                    end
                end
                local minTransitionTime = (minTransitionFrames or 0) / 60
                segmentData.transitionInfo[targetName] = {
                    segment = transitionSegment(transition),
                    minTime = minTransitionTime
                }
                if minSegmentTime then
                    minSegmentTime = math.min(minSegmentTime, minTransitionTime)
                else
                    minSegmentTime = minTransitionTime
                end
            end
        end
        segmentData.time = { minSegmentTime = minSegmentTime or 0, minStartTime = nil }
    end

    -- Propagate the min times throughout the show
    local function propagateTimes(name, segment)
        local time = segment.time
        local childMinStartTime = time.minSegmentTime + (time.minStartTime or 0)
        if segment.transitions then
            for _, transition in pairs(segment.transitions) do
                local childTime = self.segments[transitionSegment(transition)].time
                if childTime.minStartTime then
                    childTime.minStartTime = math.min(childTime.minStartTime, childMinStartTime)
                else
                    childTime.minStartTime = childMinStartTime
                end
            end
        end
    end
    self:breadthFirstTraverseSegments(propagateTimes)
end

function SegmentGraph:generateMinSegmentTimes()
    self:patchActorWithBaked()
    self:computeSegmentTimes()

    local filename = "min-times.txt"
    local file, error = io.open(filename, "w")

    local times = {}
    for _, segmentData in ipairs(SceneData.segments) do
        times[segmentData.name] = segmentData.time
    end

    file:write(Serpent.dumppretty(times))
    file:close()
end

function SegmentGraph:generateSegmentPlaythroughs()
    local playthroughs = {}
    local function traverse(startTransition, path)
        local segmentName = transitionSegment(startTransition)
        local segment = self.segments[segmentName]
        if not segment then
            Log.error("SEGMENT " .. tostring(segmentName) .. " NOT FOUND")
            return
        end

        table.insert(path, startTransition)
        local endSegment = true
        if segment.transitions then
            local transitions = {}
            for _, transition in pairs(segment.transitions) do
                local transName = transitionSegment(transition)
                local transActor = leadActor(transition)
                transitions[transName] = transitions[transName] or {}
                if transActor then
                    transitions[transName][transActor] = true
                end
            end
            for transition, actors in pairs(transitions) do
                local anyActor = false
                for actor, _ in pairs(actors) do
                    traverse({ nextSegment = transition, leadActor = actor }, path)
                    anyActor = true
                end
                if not anyActor then
                    traverse(transition, path)
                end
                endSegment = false
            end
        end
        if endSegment then
            -- we have hit an end, output the path
            table.insert(playthroughs, deepcopy(path))
        end
        table.remove(path)
    end

    for _, startSegmentName in ipairs(self.startSegments) do
        traverse(startSegmentName, {})
    end

    return playthroughs
end

-- UI to generate the graphs from running show
SegmentGraphTool = {}
SegmentGraphTool.__index = SegmentGraphTool

function SegmentGraphTool:new(params)
    local window = GUI.window(GUIAligned(ALIGN_CENTER), GUISize(400,350), "Generate segment files")
    window:anchor(ANCHOR_BOTTOM)

    local buttonCallbacks = {
        {
            {name = "Segment graph", callback = function() local graph = SegmentGraph:new(); graph:dumpSegmentGraph() end },
            {name = "Audio graph", callback = function() local graph = SegmentGraph:new(); graph:dumpSegmentGraph("segmentgraph-audio.gv", { showAudio = true}) end },
        },
        {
            {name = "Animation IDs", callback = function() local graph = SegmentGraph:new(SceneData.keepSegments); graph:dumpAnimationIDs() end },
            {name = "Orphan data", callback = function() local graph = SegmentGraph:new(SceneData.keepSegments); graph:dumpOrphanData() end },
        },
        {
            {name = "Chunk data", callback = function() local graph = SegmentGraph:new(SceneData.keepSegments); graph:dumpChunkData() end },
            {name = "Segment time data", callback = function() local graph = SegmentGraph:new(SceneData.keepSegments); graph:generateMinSegmentTimes() end },
        },
    }

    for _, rowButtons in ipairs(buttonCallbacks) do
        local row = GUI.frame(window, GUIAligned(ALIGN_TOP_LEFT), GUISize(SZ_FILL, SZ_WRAP), "widget.default")
        local buttonSize = GUIRelative(1 / #rowButtons, 0.95)
        for _, callbackData in ipairs(rowButtons) do
            local button = GUI.button(row, GUIAligned(ALIGN_BOTTOM_LEFT), buttonSize, callbackData.name)
            button:dock(DOCK_LEFT)
            button.onButtonDown = callbackData.callback
        end
        row:dock(DOCK_TOP)
    end

    return window
end
