

-- Google Analytics APIs

require "help/scripts/segmentmanager"

-- This UID is property (aka episode) specific
-- as new episodes are released, this should correspondingly be updated
local analyticsUID			= "UA-40905299-1"

--  Session level custom dimensions
ANALYTICS_BUILD_DIMENSION 			= 6
ANALYTICS_PACKAGE_DIMENSION			= 29
ANALYTICS_MOXIE_DIMENSION			= 30
ANALYTICS_SUSTAINED_RUN_COUNT			= 16
ANALYTICS_SESSION_TYPE_DIMENSION		= 8
ANALYTICS_APP_TYPE_AT_USER_LAUNCH		= 9
ANALYTICS_CLIENT_LOCAL_HOUR			= 10
ANALYTICS_CLIENT_LOCAL_DAY			= 11
ANALYTICS_DAYS_SINCE_FIRST_PROMPT_DIMENSION	= 14
ANALYTICS_NUMBER_OF_HAT_PROMPTS_DIMENSION	= 15
ANALYTICS_REASON_FOR_ATTRITION			= 18
ANALYTICS_TOTAL_RUN_COUNT			= 26
ANALYTICS_TOTAL_SESSIONS_COMPLETE		= 31
ANALYTICS_SESSION_WAS_FROM_RESUME		= 19
ANALYTICS_SESSION_LAUNCH_REASON			= 24
ANALYTICS_DOGFOOD_BUCKET			= 33

-- Hit level custom dimensions
ANALYTICS_HEADSET_DIMENSION			= 1
ANALYTICS_SOUND_LEVEL_DIMENSION			= 22
ANALYTICS_ACTIVE_SCENE_DIMENSION		= 17

-- Used for time bins
local ANALYTICS_MILLIS_PER_TIME_BIN		= 5000
local startTime					= -1
local lastObservedScene				= ""
local showIsPlaying				= false
local timeAtStartOfFpsBin			= 0

-- for fps tracking
local framesThisBin 				= 0
local timeThisSegmentBegan			= 0

local curNumberOfTouchesInTimeBin		= 0
local curNumberOfSwipesInTimeBin		= 0

local curNumberOfTouchesInSceneBin		= 0
local curNumberOfMovesInSceneBin		= 0

-- Used for tracking GWI entry
local timeOfFirstGWI				= -1
local ANALYTICS_SECONDS_UNTIL_SUSTAINED_ENGAGEMENT = 15

local immAccessReported				= false

local promptReason				= nil

-- used to determine whether our interruption was user or system generated
local unmatchedOnUserLeaveHint 			= false

local touchIsDown				= false
local curMoveCount				= 0

local curLeadActor				= nil


-- running sum of focal values of each character
local focus_amts				= { 
	girl = 0,
	monster = 0,
	cop = 0,
	unfocused = 0,
}

-- running count of how many samples have been taken
local focus_counts				= { 
	girl = 0,
	monster = 0,
	cop = 0,
	unfocused = 0,
}


-- Make sure the persist tables are setup
Persist = Persist or {}
Persist.help = Persist.help or {}
Persist.help.totalRunCount = Persist.help.totalRunCount or 0 
Persist.help.sustainedrunCount = Persist.help.sustainedrunCount or 0
Persist.help.episodesCompleted = Persist.help.episodesCompleted or 0
Persist.help.promptsSinceLastEngagement = Persist.help.promptsSinceLastEngagement or 0

AnalyticsManager = {}

-- how many digits we will pad values out to (e.g. "00012" for 12 ) 
local ANALYTICS_ZERO_PAD_DEFAULT		= 5

-- helper function to calculate time deltas accounting for rollover
calculateTimeDelta = function ( startT, endT )
	-- if rollover happens endtime will actually be less then start time
	if endT < startT then
		-- HACK:  there was rollover.  lua does not offer a max value utility, and since 
		-- I don't know what the underlying architecture will be I'm just going to take a 
		-- subset of the delta until we have a proper native API in place.
		return 0
	else 
		return endT - startT
	end
end

mangleString = function( intToMangle, numDigits )
	local resultString = ""

	if numDigits == nil then
		numDigits = ANALYTICS_ZERO_PAD_DEFAULT
	end
	if intToMangle ~= nil then
		resultString = string.format("%0"..tostring(numDigits or 0).."i", intToMangle or 0)
	end

	return resultString
end

AnalyticsManager.getTimeBin = function()
	if startTime >= 0 then
		return math.floor(calculateTimeDelta(startTime, Time.milliseconds() ) / ANALYTICS_MILLIS_PER_TIME_BIN)
	else
		-- time bins begin at the hat prompt, we don't care about GWI load
		return 0
	end
end

AnalyticsManager.getTimeBinSceneBased = function()
	if startTime >= 0 then
		return math.floor(calculateTimeDelta(timeThisSegmentBegan, Time.milliseconds() ) / ANALYTICS_MILLIS_PER_TIME_BIN)
	else
		-- time bins begin at the hat prompt, we don't care about GWI load
		return 0
	end
end

AnalyticsManager.setAllHitDimensions = function()
	Analytics.setCustomDimension(ANALYTICS_HEADSET_DIMENSION, mangleString( JavaAudio.getHeadsetState()) )
	Analytics.setCustomDimension(ANALYTICS_SOUND_LEVEL_DIMENSION, mangleString( JavaAudio.getStreamVolume(3), 3))
	Analytics.setCustomDimension(ANALYTICS_ACTIVE_SCENE_DIMENSION, lastObservedScene or "")
end

AnalyticsManager.initialize = function()
	-- initialize analytics
	Log.message("Initialising Google Analytics.")
	Analytics.initialize(analyticsUID)
	Analytics.setVersionString(System.versionString())

	AnalyticsManager.sendInitialStats()
end

AnalyticsManager.sendInitialStats = function()

	-- tell the service this is a doodle session as opposed to a download session
	Analytics.setCustomDimension(ANALYTICS_BUILD_DIMENSION, System.versionString() or "unknown")
	Analytics.setCustomDimension(ANALYTICS_SESSION_TYPE_DIMENSION, "SESSION_TYPE_DOODLE")
	Analytics.setCustomDimension(ANALYTICS_PACKAGE_DIMENSION, Platform.package)
	Analytics.setCustomDimension(ANALYTICS_MOXIE_DIMENSION, Platform.version)

	Analytics.sendEvent("GeneralInfo " .. Platform.package, "CurrentShow", Platform.package, 0)

	-- set the dogfood bucket
	--local config = Config.active()
	--Analytics.setCustomDimension(ANALYTICS_DOGFOOD_BUCKET, config.name)

	-- build info
	Analytics.sendEvent("GeneralInfo " .. Platform.package, "BuildID", System.versionString(), 0)

	-- record how many times the user has previously completed the show
	Analytics.setCustomDimension( ANALYTICS_TOTAL_SESSIONS_COMPLETE, mangleString(Persist.help.episodesCompleted))

	-- record the client time so we can filter based on the clients local time (analytics converts to GMT)
	local clientHour = Calendar.getLocalTime().hour
	local localDay = Calendar.getDate().day
	Analytics.setCustomDimension( ANALYTICS_CLIENT_LOCAL_HOUR, mangleString(clientHour, 2))
	Analytics.setCustomDimension( ANALYTICS_CLIENT_LOCAL_DAY, mangleString(localDay, 2))
	
	-- Also, because our .apk is not changing, manually keep track of episode viewings through the persist table
	-- (if we do not do this, it becomes very complicated to establish new vs. returning visitors)
	Analytics.sendEvent("GeneralInfo " .. Platform.package, "InitialVolume", "", JavaAudio.getStreamVolume(3))
	Analytics.sendEvent("GeneralInfo " .. Platform.package, "InitialHeadsetState", "", JavaAudio.getHeadsetState())
	Analytics.sendEvent("GeneralInfo " .. Platform.package, "TimeSubsegmentDuration", "", ANALYTICS_MILLIS_PER_TIME_BIN)
	Analytics.sendEvent("GeneralInfo " .. Platform.package, "RunCount", mangleString(Persist.help.totalRunCount), 0)
end


-- optimisation to avoid scanning the whole table
local lastObservedSceneIdx				= 0

-- listing of scenes we consider to be "main scenes"
-- each of these scenes will be represented as their own view as opposed to side shows and interactions
-- which will be captured through events
local mainScenes = {
}

mainScenes[0] = {} 
mainScenes[0][0] = "intro"
mainScenes[0][1] = 10

mainScenes[1] = {} 
mainScenes[1][0] = "asteroids"
mainScenes[1][1] = 500

mainScenes[2] = {} 
mainScenes[2][0] = "alien_intro"
mainScenes[2][1] = 1500

mainScenes[3] = {} 
mainScenes[3][0] = "confusion"
mainScenes[3][1] = 2000

mainScenes[4] = {} 
mainScenes[4][0] = "chase"
mainScenes[4][1] = 2450

mainScenes[5] = {} 
mainScenes[5][0] = "subway_intro"
mainScenes[5][1] = 3000

mainScenes[6] = {} 
mainScenes[6][0] = "subway_run"
mainScenes[6][1] = 4000

mainScenes[7] = {} 
mainScenes[7][0] = "hallway_chase"
mainScenes[7][1] = 4700

mainScenes[8] = {} 
mainScenes[8][0] = "outside"
mainScenes[8][1] = 5400

mainScenes[9] = {} 
mainScenes[9][0] = "emotional_reconciliation"
mainScenes[9][1] = 7150

mainScenes[10] = {} 
mainScenes[10][0] = "handoff"
mainScenes[10][1] = 8665

mainScenes[11] = {} 
mainScenes[11][0] = "credits"
mainScenes[11][1] = 3000000

-- totes robust
local CREDITS_SCENE_IDX = 11;

-- a listing of counts associated with each scene so we do not inadvertently end up blasting the 
-- server with repeat scenes 
local curCountTable = {} 

AnalyticsManager.whichActor = function(self, segmentName)
	local segment = SegmentManager.segments[segmentName]
	local leadActorName = segment.leadActor or nil
	if leadActorName ~= nil then
		--local actor = ActorManager:getActor(leadActorName, false)
		local actor = SceneData.actors[leadActorName]
		local group = "none"
		if(actor and actor.actorGroup) then
			return actor.actorGroup
		end
	end

	return nil
end

AnalyticsManager.segmentPlaying = function(self, segmentIdx, segmentName)
	segmentName = segmentName or ""

	curCountTable[segmentName] = curCountTable[segmentName] or 0
	if curCountTable[segmentName] == 0 then 
		-- let's also send some touch information for this scene
		self.setAllHitDimensions()
		Analytics.sendEvent("Ux " .. Platform.package, "touchesInSceneBucket", "", curNumberOfTouchesInSceneBin)
		curNumberOfTouchesInSceneBin = 0

		self.setAllHitDimensions()
		Analytics.sendEvent("Ux " .. Platform.package, "movesInSceneBucket", "", curNumberOfMovesInSceneBin)
		curNumberOfMovesInSceneBin = 0

		-- and then mark down in our hash that we've seen the scene
		curCountTable[segmentName] = 1

		-- why are we appending Platform.package to the tail end of this?  due to issues with segmentation 
		-- we are simply naming events by appending their package name to isolate them.
		-- since episodes may share scene names, this helps isolate them 
		Analytics.sendView(segmentName .. " " .. Platform.package)

		-- if this is a sequence we're interested in, report fps for that scene before we switch
		local timeInSegment = Time.milliseconds() - timeThisSegmentBegan
		timeThisSegmentBegan = Time.milliseconds()

		local timeSpentViewingGirl = (focus_amts.girl / focus_counts.girl) * 100
		local timeSpentViewingMonster = (focus_amts.monster / focus_counts.monster) * 100
		local timeSpentViewingCop = (focus_amts.cop / focus_counts.cop) * 100
		local timeSpentUnfocused = (focus_amts.unfocused / focus_counts.unfocused) * 100

		focus_amts.girl = 0
		focus_counts.girl = 0
		focus_amts.monster = 0
		focus_counts.monster = 0
		focus_amts.cop = 0
		focus_counts.cop = 0
		focus_amts.unfocused = 0
		focus_counts.unfocused = 0

		Analytics.sendEvent("InteractionInfo " .. Platform.package, "WatchPercentage", "Girl", tonumber(timeSpentViewingGirl))
		Analytics.sendEvent("InteractionInfo " .. Platform.package, "WatchPercentage", "Monster",  tonumber(timeSpentViewingMonster))
		Analytics.sendEvent("InteractionInfo " .. Platform.package, "WatchPercentage", "Cop",  tonumber(timeSpentViewingCop))
		Analytics.sendEvent("InteractionInfo " .. Platform.package, "WatchPercentage", "Unfocused",  tonumber(timeSpentUnfocused))

	else
		-- we've already seen this scene, no need to blast analytics with data
		curCountTable[segmentName] = curCountTable[segmentName]+1
	end

	lastObservedScene = segmentName
end

AnalyticsManager.setActorFocus = function( girlFocus, copFocus, monsterFocus, unfocused)
	focus_amts.girl = focus_amts.girl + girlFocus
	focus_counts.girl = focus_counts.girl + 1
	focus_amts.cop = focus_amts.cop + copFocus
	focus_counts.cop = focus_counts.cop + 1
	focus_amts.monster = focus_amts.monster + monsterFocus
	focus_counts.monster = focus_counts.monster + 1
	focus_amts.unfocused = focus_amts.unfocused + unfocused
	focus_counts.unfocused = focus_counts.unfocused + 1

	--[[
	Log.message("girl focus "..tostring(girlFocus))
	Log.message("cop focus "..tostring(copFocus))
	Log.message("monster focus "..tostring(monsterFocus))
	Log.message("unfocused focus "..tostring(unfocused))
	--]]
end

AnalyticsManager.update = function(self, video)
	local frame = video:getFrame()
	
        local trackingTables = Tracking_Controller:getTrackingTables(video)


	-- this works as the show is linear and so scenese will always be incremental
	if mainScenes[lastObservedSceneIdx][1] < frame then
		lastObservedSceneIdx = lastObservedSceneIdx + 1
		local sceneName = mainScenes[lastObservedSceneIdx][0]
		AnalyticsManager:segmentPlaying(lastObservedSceneIdx, sceneName)

		if CREDITS_SCENE_IDX == lastObservedSceneIdx then
			AnalyticsManager:showComplete()
		end
	end


	-- see if the user has stayed within the app for at least the required engagement period
	-- TODO:  if the app persists past closure this needs a test to see if it's still actually rendering
	local elapsedTime = calculateTimeDelta(startTime, Time.milliseconds() )

	if immAccessReported ~= true and elapsedTime >= (ANALYTICS_SECONDS_UNTIL_SUSTAINED_ENGAGEMENT*1000) then

		--  this is the number of actually sustained runs we have made 
		Persist.help.sustainedrunCount = Persist.help.sustainedrunCount + 1
		Analytics.setCustomDimension(ANALYTICS_SUSTAINED_RUN_COUNT, mangleString(Persist.help.sustainedrunCount))

		-- if the user clicked on the prompt within allotted time and stayed for allotted time 
		if immAccessReported ~= true then
			immAccessReported = true
			self.setAllHitDimensions()
			Analytics.sendEvent("GWI " .. Platform.package, "Success", "SustainedEngagement", Persist.help.promptsSinceLastEngagement)
		end

		-- reset our counter for how many prompts have been initiated since the last sustained engagement
		Persist.help.promptsSinceLastEngagement = 0
	end
end

AnalyticsManager.onShowBegan = function(self)

	-- used for time binning
	startTime = Time.milliseconds()
	timeAtStartOfFpsBin = Time.milliseconds()
	showIsPlaying = true

	-- total run count information
	if (Persist.help.totalRunCount == nil) then
		Log.error("Persist.help.totalRunCount is NIL! Copy /data/data/com.motorola.avatar/shared/persist.lua off the device and send to jritts")
	end
	if (type(Persist.help.totalRunCount) ~= "number") then
		Log.error("Persist.help.totalRunCount is not a number! Copy /data/data/com.motorola.avatar/shared/persist.lua off the device and send to jritts")
	end

	Persist.help.totalRunCount = Persist.help.totalRunCount + 1
	Analytics.setCustomDimension( ANALYTICS_TOTAL_RUN_COUNT, mangleString(Persist.help.totalRunCount))

	--  add tracking information for the absolute run count (as opposed to the sustained engagement)
	if Persist.help.totalRunCount == 0 then 
		Analytics.sendEvent("GeneralInfo " .. Platform.package, "FirstTimeRun", "", 1)
	end
	
	-- how was the show launched (hat prompt, splash, etc.)	local showLaunchReason = nil
	showLaunchReason = Context.getPromptReasonString()
	Analytics.setCustomDimension( ANALYTICS_SESSION_LAUNCH_REASON, showLaunchReason)

	-- mark down if this was a resume session or not
	if Context.getCheckpoint() == nil then
		Analytics.setCustomDimension( ANALYTICS_SESSION_WAS_FROM_RESUME, "FromBeginning" )
	else
		Analytics.setCustomDimension( ANALYTICS_SESSION_WAS_FROM_RESUME, "FromResume" )
		lastObservedScene = Context.getCheckpoint() or ""
	end

	self.setAllHitDimensions()
	Analytics.sendEvent("GWI " .. Platform.package, "ShowBegan", showLaunchReason, 0)
end

AnalyticsManager.onShowEnded = function(self)

	showIsPlaying = false

	local elapsedTime = calculateTimeDelta(startTime, Time.milliseconds() ) / 1000

	Analytics.sendEvent("MangleStore " .. Platform.package, "ShowEnded_elapsedTime", mangleString(elapsedTime), elapsedTime)
end

AnalyticsManager.recordTouchDown = function(self)
	-- keep track of the number of touches for each time bucket
	curNumberOfTouchesInTimeBin = curNumberOfTouchesInTimeBin + 1 
	curNumberOfTouchesInSceneBin = curNumberOfTouchesInSceneBin + 1

	touchIsDown = true
end

AnalyticsManager.recordTouchUp = function(self)
	-- this wasn't just a touch but a proper swipe
	if curMoveCount >= 5 then 
		-- keep track of the number of touches for each time bucket
		curNumberOfSwipesInTimeBin = curNumberOfSwipesInTimeBin + 1 
		curNumberOfMovesInSceneBin = curNumberOfMovesInSceneBin + 1
	end 

	curMoveCount = 0
	touchIsDown = false
end

AnalyticsManager.recordMoveEvent = function(self)
	-- moves only occur on swiping motions in between a touch down/up
	curMoveCount = curMoveCount + 1
end

AnalyticsManager.showComplete = function(self)

	showIsPlaying = false
	
	-- increment the number of times the show has been completed, but don't update the custom dimension
	-- updating it now would retroactively update all of the hits in the session so far
	Persist.help.episodesCompleted = Persist.help.episodesCompleted + 1

	Analytics.setCustomDimension(ANALYTICS_REASON_FOR_ATTRITION, "ShowComplete")
	self.setAllHitDimensions()

	local timeInShow = math.floor(calculateTimeDelta(startTime, Time.milliseconds() ) / 1000)
	Analytics.sendEvent("GWO " .. Platform.package, "ShowComplete", "", timeInShow)
end

AnalyticsManager.onActivityEvent = function(self, eventString)
	if (eventString == "onUserLeaveHint") then
		AnalyticsManager:onUserLeaveHint()
	end
end

AnalyticsManager.onUserLeaveHint = function(self)
	unmatchedOnUserLeaveHint = true
end

AnalyticsManager.userLeft = function(self)

	if showIsPlaying == true then 
		-- if user has hinted a leave, it is implied that the next onstop was triggered by the user
		if unmatchedOnUserLeaveHint == true then
			unmatchedOnUserLeaveHint = false
			Analytics.setCustomDimension(ANALYTICS_REASON_FOR_ATTRITION, "UserLeft")
		else
			Analytics.setCustomDimension(ANALYTICS_REASON_FOR_ATTRITION, "UserInterrupted")
		end

		self.setAllHitDimensions()

		local timeInShow = math.floor(calculateTimeDelta(startTime, Time.milliseconds() ) / 1000)
		Analytics.sendEvent("Attrition " .. Platform.package, lastObservedScene or "", mangleString(AnalyticsManager:getTimeBinSceneBased()), timeInShow)

		if timeInShow <= ANALYTICS_SECONDS_UNTIL_SUSTAINED_ENGAGEMENT then
			-- report a quick abort
			self.setAllHitDimensions()
			Analytics.sendEvent("GWI " .. Platform.package, "Failure", "QuickAbort", timeInShow)
		end
	end

	self.onShowEnded()
end
