


AudioPlayer = {
    enabled = true,

    emitter = nil,
    sounds = {},
    segmentEventData = nil,
}
AudioPlayer.__index = AudioPlayer

function AudioPlayer.new(app, screenplay)
    if AudioPlayer.enabled then
        local instance = { app = app, screenplay = screenplay or (def and def.audio) }
        setmetatable( instance, AudioPlayer )
        instance:_init()

        return instance
    else
        return {
            destroy = function() end,
            update = function() end,
            renderInfo = function() end,
        }
    end
end

function AudioPlayer:_init()
    self.maxFrame = 0
    self.emitter = AudioEmitter.create()
    self.emitter:setHRTF("help:audio/biquad_hrtf_subject_015_order_12.mxhrtf")

    if not self.screenplay then self.screenplay = {} end
    if not self.screenplay.sounds then self.screenplay.sounds = {} end
    if not self.screenplay.segments then self.screenplay.segments = {} end

    Events.listen(self, "segmentStarted")
end

function AudioPlayer:destroy()
    if self.emitter then
        self.emitter:destroy()
        self.emitter = nil
    end
    Events.stopListening(self, "segmentStarted")
end

function AudioPlayer:update()
    HeadingVolume:clearCache()
    self:_processEvents()
    self:_updateSounds()
end

function AudioPlayer:renderInfo(nextLine, font_style)
    if self.sounds then
        for name, sound in pairs(self.sounds) do
            Renderer.drawText( 10, nextLine(), font_style, string.format("%s : volume = %.02f", name, sound:getVolume()))
        end
    end
end

function AudioPlayer:segmentStarted(segmentName)
--    Log.message("DMA: audioplayer segment started: " .. segmentName .. " at " .. tostring(self.app:getFrame()))

    local segmentEventData = self.screenplay.segments[segmentName]
    if not segmentEventData then return end

    local pendingEvents = {}
    for _, event in ipairs(segmentEventData) do
        table.insert(pendingEvents, event)
    end
--    Log.message("DMA: pending events for segment " .. segmentName .. ": " .. table.tostring(pendingEvents))
    self.pendingEvents = pendingEvents
end

function AudioPlayer:_processEvents()
    if not self.pendingEvents or #self.pendingEvents <= 0 then return end

    local frame = self.app:getFrame()
    if self.screenplay.frameLatency then
        frame = frame + self.screenplay.frameLatency
    end

    local stillPending = {}
    for _, event in ipairs(self.pendingEvents) do
        local eventFrame = event.at or 0
        if eventFrame <= frame then
            if event.play then
                self:_playSound(event)
            elseif event.fade then
                self:_fadeSound(event)
            elseif event.stop then
                self:_stopSound(event)
            end
        else
            table.insert(stillPending, event)
        end
    end
    self.pendingEvents = stillPending
end

function AudioPlayer:getSound(name)
    local sound = self.sounds[name]
    if not sound then
        local soundDef = self.screenplay.sounds[name]
        if not soundDef then
            Log.error("Undefined sound: " .. tostring(name))
            return
        end

        -- soundName, uriPath, preCache, startTimeMillis, stopTimeMillis
        sound = self.emitter:addSound(name, soundDef.uri, false, 0, -1)
        if sound then
            sound.name = name
            sound.control = SoundControl.new(sound)
            if soundDef.volume   then sound.control:setVolume(soundDef.volume) end
            if soundDef.track    then sound.control:track(soundDef.track) end
            if soundDef.binaural then sound.control:setBinaural(soundDef.binaural) end
            self.sounds[name] = sound
        else
            Log.error("Unable to create sound for " .. table.tostring(soundDef))
        end
    end
    return sound
end


function AudioPlayer:_playSound(event)
--    Log.message("DMA: playing sound event: " .. table.tostring(event) .. " at frame " .. tostring(self.app:getFrame()))
    local sound = self:getSound(event.play)
    if sound then
        if event.startTime then
            sound.control:setSeekPosition(event.startTime)
        end
        if event.loop then 
            sound.control:setLooping(event.loop)
        end
        
        sound:play()

        if event.volume then
            sound.control:setVolume(0)    
            local time = event.fadeTime or 0
            sound.control:setVolume(event.volume,time)          
        end
    end
end


function AudioPlayer:_stopSound(event)
    local sound = self:getSound(event.stop)
    if sound then
        local time = event.fadeTime or 0
        sound.control:setVolume(0,time)
        sound.control.expire = true
    end
end


function AudioPlayer:_updateSounds()
    for _, sound in pairs(self.sounds) do
        sound.control:update()

        -- Cleanup this sound if it was set to expire and is close enough
        if (sound.control.expire and sound:getVolume() == 0.0) then
            sound:stop()
            self.emitter:removeSound(sound.name)
            self.sounds[sound.name] = nil
        end
    end
end
