
tracking_states = {
	on = 0,
	off = 1,
	disabled = 2,
}

AnalyticTargetController = {
	tracking_objects = nil,
	debug = false,
	activeGroups = {},
	rowCount = 0,
	distFromCameraRecord = {
		frame = {}
	},
	enableAnalytics = false,
	
	-- global object for controlling tracking strength scale
	globalScaleTracker = nil,
	enableGlobalScaleTracker = false,
}

-- stack method for creating a tracking object
function AnalyticTargetController.createTrackingActor(objDef, optGroupName, timeOffset)
	local actor = Global.scene:createActor(objDef.model, objDef, false);
	Global.scene:setupActor(actor, objDef)
	actor:setVisibility(false)
	
	actor.name = actor:getName()
	actor.targetLocator = objDef.targetLocator
	actor.groupName = optGroupName
	actor.material = actor:findMaterial(objDef.materials[1].name)
	
	actor.animation["tracking_clip"]:play()
	actor.animation["tracking_clip"]:setTime(timeOffset or 0)
	actor.animation["tracking_clip"]:setPaused(true);
	actor.trackingState = tracking_states.disabled
	
	return actor
end

function AnalyticTargetController:init(timeOffset)

	-- create all tracking objects
	local i = 1
	local objectList = {}
	for groupName, group in pairs(self.tracking_objects) do
		
		for _, objDef in ipairs(group) do
			if(objDef) then
				objectList[i] = AnalyticTargetController.createTrackingActor(objDef, groupName, timeOffset)
			end
			i = i + 1
		end
	end
	
	self.tracking_objects = objectList
	
	Events.listen(self, "videoIsPaused")
end


-- tracking onUpdate monitors the scale of the actors and adds/removes the actors from tracking
-- based on animated scale values
function AnalyticTargetController:onUpdate(video)

	for i, actor in ipairs(self.tracking_objects) do
		
		-- get the transform
		local localT = actor:getBoneLocal(actor.targetLocator)
		
		-- check the scale and decided to turn tracking on/off for this actor
		if(actor.trackingState ~= tracking_states.disabled) then
			if(localT.scale.x < 0.011)then
				
				-- remove tracking
				actor.trackingState = tracking_states.off
				
			elseif(actor.trackingState == tracking_states.off) then
			
				-- add tracking
				actor.trackingState = tracking_states.on
			end
		end
	end
	
end

function AnalyticTargetController:enableTrackingGroup(groupName, isEnabled)
	
	self.activeGroup = groupName
	self.activeGroups[groupName] = isEnabled
	
	for i, actor in ipairs(self.tracking_objects) do
		if(isEnabled and actor.groupName == groupName) then
			
			-- enable tracking by adding the actor to tracking list
			actor.trackingState = tracking_states.on
			if(self.debug) then
				actor:setVisibility(true)
			end
		elseif(actor.groupName == groupName) then
			-- disable tracking by removing the actors from tracking list
			actor.trackingState = tracking_states.disabled
			actor:setVisibility(false)
		end
	end
end

function AnalyticTargetController:videoIsPaused(isPaused)
	if(self.globalScaleTracker and self.enableGlobalScaleTracker) then
		self.globalScaleTracker.animation["tracking_clip"]:setPaused(isPaused);
	end
	for i, actor in ipairs(self.tracking_objects) do
		if(actor and actor.animation["tracking_clip"]:isPaused() ~= isPaused) then
			actor.animation["tracking_clip"]:setPaused(isPaused);
		end
	end
end

function AnalyticTargetController:updateAnalytics(video)
	
	--if(self.enableAnalytics) then
	if(true) then
		
		local frame = video:getFrame()
		
		if(frame % 2 == 0 and frame ~= self.lastFrame) then
		
			self.lastFrame = frame
			
			local cameraFwrd = Vector.rotateBy(Vector.forward, MainCamera:getWorld().rotation)
			
			table.insert(self.distFromCameraRecord.frame, frame)
			local girlFocus = 0	
			local copFocus = 0	
			local monsterFocus = 0	
			local unfocused = 0	

			for i, actor in ipairs(self.tracking_objects) do
			
				if(actor.groupName == self.activeGroup) then
					self.distFromCameraRecord[actor.name] = self.distFromCameraRecord[actor.name] or {}
					local bone = actor:getBoneWorld(actor.targetLocator)
					
					local value = 0
					if(actor.trackingState == tracking_states.on) then
						value = Vector.dot(cameraFwrd, bone.position:normalized())
					end


					if value > 0.7 then
						value = 1
					else
						value = 0
					end

					if(actor.name == "help:models/track_girl_30fps_500cut.mxm") then 
						girlFocus = value
					elseif(actor.name == "help:models/track_monster_30fps_500cut.mxm") then 
						monsterFocus = value
					elseif(actor.name == "help:models/track_cop_30fps_500cut.mxm") then
						copFocus = value
					end
					
					table.insert(self.distFromCameraRecord[actor.name], value)
				end
			end
		
			if(girlFocus == 0 and copFocus == 0 and monsterFocus == 0 ) then
				unfocused = 1
			end

			AnalyticsManager.setActorFocus(girlFocus, copFocus, monsterFocus, unfocused)

			self.rowCount = self.rowCount + 1
		end
		
	end
end

function AnalyticTargetController:writeAnalytics()

	local stringData = "frame,"
	for i, actor in ipairs(self.tracking_objects) do
		if(actor.groupName == self.activeGroup) then
			stringData = stringData..actor.name..","
		end
	end
	
	stringData = stringData .. "\n"
	
	for i=1, self.rowCount, 1 do
		stringData = stringData..self.distFromCameraRecord.frame[i]..","
		for j, actor in ipairs(self.tracking_objects) do
			if(actor.groupName == self.activeGroup) then
				stringData = stringData .. string.format("%.3f", self.distFromCameraRecord[actor.name][i]) .. ","
			end
		end
		
		stringData = stringData .. "\n"
	end
	
	local file,err = io.open( "actortracking.csv", "w" )
	if( file ~= nil ) then
		file:write(stringData)
		file:close()
	end
end
