

local Logger = TaggedLog("CameraRig", true)

CameraRig = 
{
    -- Each layer is is a transform that represents a camera rotation offset
    layers        = {},
    
    -- Resulting transform of all layers
    worldT        = SRT.new(),
    camForward    = Vector.new(),
    camUp         = Vector.new(),
    camRight      = Vector.new(),
    initialized   = false,
}

function CameraRig:new(data)
    local obj = shallowcopy(data) or {}
    setmetatable(obj, self)
    self.__index = self;

    return obj
end

function CameraRig:updateLayers(transform)
	if(not self.initialized) then return end

	-- Calculate time independently
	local currTimeSecs = os.clock()
	if(self.lastUpdateSecs == nil) then
		self.lastUpdateSecs = currTimeSecs
	end

	local dt = (currTimeSecs - self.lastUpdateSecs)
	self.lastUpdateSecs = currTimeSecs

	-- Use inplace copies in order to reduce memory allocations.
	self.worldT:reset()

	self.camForward:copy_inplace(Vector.forward)
	self.camRight:copy_inplace(Vector.right)
	self.camUp:copy_inplace(Vector.up)

	for i, layer in ipairs(self.layers) do
	
		-- Use inplace copies in order to reduce memory allocations.
		layer.worldT:copy_inplace(self.worldT)
		layer.camForward:copy_inplace(self.camForward)
		layer.camRight:copy_inplace(self.camRight)
		layer.camUp:copy_inplace( self.camUp )
	
		if(layer.onUpdate) then
			layer:onUpdate(dt)
		end

		if(layer.leftMult) then
			self.worldT:mul_inline(layer.localT, self.worldT)
		else
		    --local prevRot = self.worldT.rotation
		    --local testT = self.worldT.rotation * layer.localT.rotation
            --local testString = string.format("newRot %.3f %.3f %.3f %.3f : prevRot %.3f %.3f %.3f %.3f [%s:%i] ", testT.x, testT.y, testT.z, testT.w, prevRot.x, prevRot.y, prevRot.z, prevRot.w, layer.name, i) 
	        --Logger:message(testString)
		
			self.worldT:mul_inline(self.worldT, layer.localT)
		end
		
		-- Update basis vectors
		self.camForward:rotateByInline(Vector.forward, self.worldT.rotation)
		self.camRight:rotateByInline(Vector.right, self.worldT.rotation)
		self.camUp:rotateByInline(Vector.up, self.worldT.rotation)

		if(layer.onPostUpdate) then
			layer:onPostUpdate(self.worldT) 
		end
	end
end

function CameraRig:destroyLayers()
	if(self ~= nil and self.layers ~= nil) then
		for i, layer in ipairs(self.layers) do
			
			if(layer.onDestroy) then
				layer:onDestroy()		
			end

		end
	end
end

function CameraRig:addLayer(layer, object)

	assert(layer.verifyInterface, "Layer with name '"..layer.name.."' does not support CameraLayer interface");
	
	layer:verifyInterface()
	
	layer.object = object
	layer.worldT = SRT.new();
	layer.camForward = Vector.new();
	layer.camUp = Vector.new();
	layer.camRight = Vector.new();
	
	table.insert(self.layers, layer)

	if(layer.onCreate) then
		layer:onCreate()
	end	

	return layer;
end

function CameraRig:startLayers()
    -- Initialize basis vectors
    self.initialized = true
    self.camForward = Vector.rotateBy(Vector.forward, self.worldT.rotation);
	self.camRight = Vector.rotateBy(Vector.right, self.worldT.rotation);
	self.camUp = Vector.rotateBy(Vector.up, self.worldT.rotation);
	
	for i, layer in ipairs(self.layers) do
		if(layer.onStart) then
			layer:onStart()
		end
	end
end