

local Logger = TaggedLog("InputLayer", true)

InputLayer = createClass({
	touchAnchor           = nil,
	touchStartTime        = 0.0,
    touchTime             = 0.0,

    rotY                  = 15.0,
    rotX                  = 0.0,
    rotationScaleX        = 0.15,
    rotationScaleY        = 0.15,

    prevRotX              = 0.0,
    prevSensorXZ          = nil,

    constrainedTouch      = true,
    isUp                  = true,
    isDown                = false,

    quatY      = Quaternion.new(),
    quatX      = Quaternion.new(),
    scrollQuat = Quaternion.new(),

}, deepcopy(CameraLayer));

InputLayer.name	= "InputLayer"

function InputLayer:onStart()
	if (Platform.win32) then
        self.constrainedTouch = false
    end
    
    self:enableTouchControls(true)
end

function InputLayer:onUpdate()
	
	if (self.rotX ~= self.prevRotX) then
        self.quatY    = Quaternion.axisAngle(Vector.up, self.rotX)
        self.prevRotX = self.rotX
    end
    
	if (self.constrainedTouch) then
		-- Disable pitching the camera with touch
        self.rotY  = 0
        self.quatX = Quaternion()
		
        self.scrollQuat = self.quatY:copy()
    else
    	self.scrollQuat = self.quatY*self.quatX
    end
    
    self.localT.rotation = self.scrollQuat*InputManager:getSensorRotation()
    
    -- Project current sensor "direction" to the XZ plane
    local currentSensorXZ  = Vector.rotateBy(Vector.forward, InputManager:getSensorRotation())
    currentSensorXZ.y = 0.0
    currentSensorXZ:normalize()
    
    local angle = 0  
    if (self.prevSensorXZ ~= nil) then 
        -- Determine angle delta of our sensor projection
        angle     = -1.0 * math.deg(math.atan2(self.prevSensorXZ.x,self.prevSensorXZ.z) - math.atan2(currentSensorXZ.x,currentSensorXZ.z))

        -- Account for the fact that atan2 is discontinous
        if (math.abs(angle) > 180) then
            angle = -1 * math.sign(angle) * 360 + angle
        end
    end
	
    -- Add a possible "boost" to the xz-plane rotation by addeding to sensor delta
    local yawMult   = Global.scene.yawMult or 0.0
    self.rotX = self.rotX + angle * yawMult
    self.prevSensorXZ = currentSensorXZ
	
    --local prevRot  = MainCamera:getWorld().rotation
    --local deltaRot = prevRot:inverse() * self.localT.rotation
    --local deltaYaw = deltaRot:getYawAngle()
    --Global.AppDebugText1 = string.format("rotX %.5f : angle %.5f", self.rotX, angle)
    --Global.AppDebugText2 = string.format("currentSensorXY %.3f, %.3f, %.3f ", currentSensorXZ.x, currentSensorXZ.y, currentSensorXZ.z)
    --Logger:message(Global.AppDebugText1)
end

--- enableTouchControls
-- If true - start listening for touch events
-- else stop listening to touch events
function InputLayer:enableTouchControls(enabled)
    self.touchControls = enabled;

    if (self.touchControls) then
        Events.listen(self, "touchDown")
        Events.listen(self, "touchUp")
        Events.listen(self, "drag")
    else
        Events.stopListening(self, "touchDown")
        Events.stopListening(self, "touchUp")
        Events.stopListening(self, "drag")
    end
end


--- touchDown
-- Event listener - receives the touch down position
function InputLayer:touchDown(touchPosition)
    if(self.isDown) then
        Log.message("Touch: UP NOT RECVD")
        self.isUp = true
        self:resetInput(touchPosition)
    end

    self.isDown         = true
    self.touchAnchor    = touchPosition
    self.touchStartTime = os.time()
    
    self.isUp = false
end

--- resetInput
-- At the end of an input session (on 'touchUp') input controls a reset
-- and readied for the next input event
function InputLayer:resetInput(touchPosition)
    if (self.touchAnchor == nil) then return end

    self.touchTime = os.time() - self.touchStartTime

    local toAnchor = (self.touchAnchor - touchPosition)

    self.dragPosition = nil
    self.isDown = false
end

--- touchUP
-- An event listener - receives the touchUp position
function InputLayer:touchUp(touchPosition)
    self.isUp = true;
    self:resetInput(touchPosition)
end

--- drag
-- An event listener - receives the the change in touch position
function InputLayer:drag(touchDelta)
    if(not self.dragPosition) then
        self.dragPosition = Vector(touchDelta.x, touchDelta.y)
    end

    local pos = Vector(touchDelta.x, touchDelta.y)

    self.touchDelta = Vector(0.5*(pos.x - self.dragPosition.x), 0.5*(pos.y - self.dragPosition.y), 0.0, 0.0)

    self.dragPosition = pos
    
    --Global.AppDebugText1 = string.format("dragPosition %.3f, %.3f ", pos.x, pos.y)

    local rotation = InputManager:getSensorRotation():copy()
    local rollAngle, rollQuat 	= rotation:getRoll()
    self.touchDelta = self.touchDelta:rotateBy(rollQuat:inverse())
    
    -- Update rotation
    self.rotX = self.rotX + (self.touchDelta.x * self.rotationScaleY)
        
    -- Vertical drag is disabled on device to prevent confusing horizion behavior
    if(Platform.win32) then
    	local yawAngle, yawQ 	= rotation:getYaw()
    	local right 			= Vector.rotateBy(Vector.right, yawQ)
    	self.rotY 				= self.rotY + (self.touchDelta.y  * self.rotationScaleX)
    	self.quatX              = Quaternion.axisAngle(right, self.rotY)
    end
end
