
tracking_states = {
	on = 0,
	off = 1,
	disabled = 2,
}

Tracking_Controller = {
	tracking_objects = nil,
	debug = false,
	activeGroups = {},
	rowCount = 0,
	distFromCameraRecord = {
		frame = {}
	},
	
	-- global object for controlling tracking strength scale
	globalScaleTracker = nil,
	enableGlobalScaleTracker = false,
}

-- stack method for creating a tracking object
function Tracking_Controller.createTrackingActor(objDef, optGroupName, timeOffset)
	local actor = Global.scene:createActor(objDef.model, objDef, false);
	Global.scene:setupActor(actor, objDef)
	actor:setVisibility(false)
	
	actor.name = actor:getName()
	actor.targetLocator = objDef.targetLocator
	actor.groupName = optGroupName
	actor.material = actor:findMaterial(objDef.materials[1].name)
	
	actor.animation["tracking_clip"]:play()
	actor.animation["tracking_clip"]:setTime(timeOffset or 0)
	actor.animation["tracking_clip"]:setPaused(true);
	actor.trackingState = tracking_states.disabled
	
	return actor
end

function Tracking_Controller:init(timeOffset)

	-- create all tracking objects
	local i = 1
	local objectList = {}
	for groupName, group in pairs(self.tracking_objects) do
		
		for _, objDef in ipairs(group) do
			if(objDef) then
				objectList[i] = Tracking_Controller.createTrackingActor(objDef, groupName, timeOffset)
			end
			i = i + 1
		end
	end
	
	self.tracking_objects = objectList
	
	Events.listen(self, "videoIsPaused")
end

function Tracking_Controller:enableDebug(isEnabled)

	self.debug = isEnabled
	
	for i, actor in ipairs(self.tracking_objects) do
		if(actor and self.activeGroups[actor.groupName]) then
			actor:setVisibility(isEnabled)
		end
	end
	
	if(isEnabled) then
	
		-- enable update
		Tracking_Controller.saveOnUpdate = Tracking_Controller.onUpdate
		Tracking_Controller.onUpdate = function(self, video)
		
			-- called the saved method
			Tracking_Controller:saveOnUpdate(video)
			
			local trackedActor = MainCamera.controller:getTrackedObject()
			
			for i, actor in ipairs(self.tracking_objects) do
				-- reset colors or set to 'tracked' colors
				if(trackedActor and trackedActor:getName() == actor:getName()) then
					actor.material:setParameters({u_Color = Vector(0,1,0,1)})
				else
					actor.material:setParameters({u_Color = Vector(1,1,1,1)})
				end
				
			end
			
		end
	else
		-- clear colors
		for i, actor in ipairs(self.tracking_objects) do
			-- reset colors
			self.tracking_objects[i].material:setParameters({u_Color = Vector(1,1,1,0.25)})
		end
		
		-- disable debug update loop
		if(Tracking_Controller.saveOnUpdate) then
			Tracking_Controller.onUpdate = Tracking_Controller.saveOnUpdate
		end
	end
end

-- tracking onUpdate monitors the scale of the actors and adds/removes the actors from tracking
-- based on animated scale values
function Tracking_Controller:onUpdate(video)

	if(self.globalScaleTracker and self.enableGlobalScaleTracker) then
		-- get the transform
		local localT = self.globalScaleTracker:getBoneLocal(self.globalScaleTracker.targetLocator)
		local strength = localT.scale.x
		
		if(not Platform.candidate and Settings) then
			Settings.trackingStrength:setValue(strength > 0.01 and strength or 0)
			
--			Global.AppDebugText3 = string.format("Global Strength: %.4f", localT.scale.x)
		else
			MainCamera.controller:setGlobalTrackingStrength(strength > 0.01 and strength or 0)
		end
	end

	for i, actor in ipairs(self.tracking_objects) do
		
		-- get the transform
		local localT = actor:getBoneLocal(actor.targetLocator)
		
		-- check the scale and decided to turn tracking on/off for this actor
		if(actor.trackingState ~= tracking_states.disabled) then
			if(localT.scale.x < 0.011)then
				
				-- remove tracking
				actor.trackingState = tracking_states.off
				MainCamera.controller:removeTrackingObject(actor)
				
			elseif(actor.trackingState == tracking_states.off) then
			
				-- add tracking
				actor.trackingState = tracking_states.on
				MainCamera.controller:addTrackingObject(actor, actor.targetLocator, 0.25)
			end
		end
	end
	
	
	-- check timing
--	Global.AppDebugText2 = "Video time: "..video:getTime().." fbx time: "..self.globalScaleTracker.animation["tracking_clip"]:getTime()

	
end

function Tracking_Controller:enableTrackingGroup(groupName, isEnabled)
	
	self.activeGroup = groupName
	self.activeGroups[groupName] = isEnabled
	
	for i, actor in ipairs(self.tracking_objects) do
		if(isEnabled and actor.groupName == groupName) then
			
			-- enable tracking by adding the actor to tracking list
			MainCamera.controller:addTrackingObject(actor, actor.targetLocator, 0.25)
			actor.trackingState = tracking_states.on
			if(self.debug) then
				actor:setVisibility(true)
			end
		elseif(actor.groupName == groupName) then
			-- disable tracking by removing the actors from tracking list
			MainCamera.controller:removeTrackingObject(actor)
			actor.trackingState = tracking_states.disabled
			actor:setVisibility(false)
		end
	end
end

function Tracking_Controller:videoIsPaused(isPaused)
	if(self.globalScaleTracker and self.enableGlobalScaleTracker) then
		self.globalScaleTracker.animation["tracking_clip"]:setPaused(isPaused);
	end
	for i, actor in ipairs(self.tracking_objects) do
		if(actor and actor.animation["tracking_clip"]:isPaused() ~= isPaused) then
			actor.animation["tracking_clip"]:setPaused(isPaused);
		end
	end
end

function Tracking_Controller:enableGlobalScaleTrackAnimation(enabled)
	Log.message(" USE TRACK ANIMNATION ")
	self.enableGlobalScaleTracker = enabled
end

function Tracking_Controller:getTrackingTables(video)
	return self.distFromCameraRecord
end


