

require "common/scripts/moxie/moxie"
require "common/scripts/moxie/replay"

require "help/scripts/fixup/fixup"
require "help/scripts/tools/logger"
require "help/scripts/util"

require "help/scripts/data/actors"
require "help/scripts/data/lights"
require "help/scripts/data/segments"
require "help/scripts/data/audio"

require "help/scripts/components/camera/camerarig"
require "help/scripts/components/camera/layers/cameralayer"
require "help/scripts/components/camera/layers/inputlayer"
require "help/scripts/components/camera/layers/trackinglayer"
require "help/scripts/components/camera/layers/targetwarplayer"
require "help/scripts/components/camera/cameracontroller"

require "help/scripts/components/sensor_controller"
require "help/scripts/components/mouse_controller"
require "help/scripts/components/actortracker_controller"
require "help/scripts/components/lookatdetector"
require "help/scripts/components/fadetitle"
require "help/scripts/components/tracking_controller"
require "help/scripts/components/analytictargetcontroller"
require "help/scripts/components/fbdriver"
require "help/scripts/components/titleparallax"
require "help/scripts/components/cameralock"

require "help/scripts/gui/settings"
require "help/scripts/inputmanager"
require "help/scripts/resource_loader"
require "help/scripts/segmentmanager"
require "help/scripts/segment"
require "help/scripts/analyticsmanager"

require "help/scripts/audio/audio_player"
require "help/scripts/audio/sound_control"
require "help/scripts/audio/heading_volume"

enum( "VF_MAIN_VIEW", 1 )

local font_style =
{
    color   = Color.new( 1.0, 1.0, 1.0, 1.0 ),
    font    = "common:fonts/arial_default_24.fnt",
    scale   = ( Platform.android and 1 or 0 ) * 0.5 + 1.0
}

Persist = Persist or {}
Persist.help = Persist.help or { showTime = 0 }

HelpApp =
{
    scene               = nil,
    view                = nil,
    video               = nil,
    settings            = nil,
    material            = nil,
    loading             = true,
    width               = 0,
    height              = 0,
    audioPlayer         = nil,
    boom                = 0.0,
    heading             = 79.5,
    startupHeadingLock  = true,
    pauseKeyUp          = true,
    paused              = false,
    centeringStrength   = 0.5,
    tclEnabled          = true, -- target capture latency
    tclDebugEnabled     = false, -- show tracking strength
    showFPS             = false,
    trackingAnalytics   = false,
    precisionMin		= .3,
    precisionMax		= 7,
    
    enablePitchRemap	= true,
    projectiveOffset	= {y=-.98, z=1.95},
    projectiveRemap		= {thresholdIn = .85, thresholdOut = .2},
  
    globalTrackingStrength          = .4, -- overall scale of tracking
    enableTrackingControlAnimation  = true, -- use animation or control manually
    
    -- pitch clamp controls
    pitchUnclampDuration    = 13.7,
    elapsedPitchTime        = 0,
	pitchFrame 				= 0,
	maxPitchReached			= -1,
    
    videoFadeTimer      = 0.0,
    videoFadeDuration   = 0.0,
    videoFromColor      = Color.new( 1.0, 1.0, 1.0, 1.0 ),    
    videoToColor        = Color.new( 1.0, 1.0, 1.0, 1.0 ),  
    
    showTimeSaveFreq	= 1,
    elapsedSaveTime		= 0,
}

HelpApp.__index = HelpApp

function HelpApp.create()
    local instance = {}
    setmetatable( instance, HelpApp )
    return instance
end

function HelpApp:onStartup()

end

function HelpApp:onInitialize()

	Persist.help.showTime = Persist.help.showTime or 0
	Persist.help.showTime = Platform.reason == "browser_play" and 0 or Persist.help.showTime
	
	--resuming the show
	if(Persist.help.showTime > 0) then
    	self.enablePitchRemap = false
    
		-- start from main show when resuming
		def.firstSegment      = "SegmentMainShow"
		
		-- dump active list (to remove title card out)
		def.SegmentMainShow.active = {}
		
		-- update sound start times
        if (def.audio.segments.SegmentMainShow) then
            for _, soundDef in ipairs(def.audio.segments.SegmentMainShow) do
    			if(soundDef.startTime) then
    				soundDef.startTime = Persist.help.showTime
    			end
    		end
        end
	end
	
    -- RenderDebug.setEnabled( Application.debug )
    self:createScene()
    self:createView()

    InputManager:init(self.scene)
    ResourceLoader:init()
    AnalyticsManager.initialize()

    self.audioPlayer = AudioPlayer.new(self)
    self.segmentManager = SegmentManager.new(self.scene);

    MainCamera = self.scene.cameras.main
    MainCamera.controller = TrackingCamera.create(MainCamera)

    self.view:setCamera(MainCamera)

    -- create tracking reference objects
    Tracking_Controller.tracking_objects = {

        hero = {
            def.TrackGirl_Hero,
            def.TrackMonster_Hero,
            def.TrackCop_Hero,
        },

        composition = {
            def.TrackGirl,
            def.TrackMonster,
            def.TrackCop,
        }
    }

    Tracking_Controller.globalScaleTracker = Tracking_Controller.createTrackingActor(def.TrackGlobalScale4)
    Tracking_Controller:init(Persist.help.showTime/1000)
    
    AnalyticTargetController.tracking_objects = {
        hero = {
            def.TrackGirl_Hero,
            def.TrackMonster_Hero,
            def.TrackCop_Hero,
        },
    }
    
    AnalyticTargetController:init(Persist.help.showTime/1000)
    AnalyticTargetController:enableTrackingGroup("hero", true)
	
    MainCamera.controller:setBoom(self.boom)
    MainCamera.controller:lookAtHeadingDegrees(self.heading)
    MainCamera.controller:setTargetCentering(self.centeringStrength)
    MainCamera.controller:setGlobalTrackingStrength(self.globalTrackingStrength)
    MainCamera.controller:setIsTCLEnabled(self.tclEnabled)
    if(self.enablePitchRemap) then
    	MainCamera.controller:setProjectiveOffset(Vector(0, self.projectiveOffset.y, self.projectiveOffset.z), self.projectiveRemap.thresholdIn, self.projectiveRemap.thresholdOut)
    end
    self:enableTrackingGroup("composition", true)
    self:enableTrackingAnalytics(self.trackingAnalytics)
    self:enableGlobalTrackScaleAnimation(self.enableTrackingControlAnimation)
    self:setCameraPrecisionFilterMinMax(self.precisionMin, self.precisionMax)

    --  self:selectVideo( def_videos[1].name )
    Renderer.setRenderDebug(true)

    -- configuring settings from one place to reduce the number of files with special flags that need setting
    if(Settings) then
        Settings.onInitialized = function()
            if(self.tclEnabled) then
                Settings.tclEnabled:check()
            else
                Settings.tclEnabled:uncheck()
            end

            if(self.showFPS) then
                Settings.fps:check()
            else
                Settings.fps:uncheck()
            end

            Settings.tracking_comp:check()
            --            Settings.tracking_hero:check()
            --            Settings.tracking_debug:check()

            if(self.trackingAnalytics) then
                Settings.trackingAnalytics:check()
            end

            Settings.target_centering:setValue(self.centeringStrength)
            Settings.trackingStrength:setValue(self.globalTrackingStrength)
            if(self.enableTrackingControlAnimation) then
                Settings.trackingStrengthAnimation:check()
            else
                Settings.trackingStrengthAnimation:uncheck()
            end
            Settings.precisionMin:setValue(self.precisionMin)
            Settings.precisionMax:setValue(self.precisionMax)
        end
    end


    if(Platform.win32) then
        MainCamera.controller:startSensorClient();
    end
    
    
    Sensor.setSensorFusionLowpassFilterEnabled(true, 1.5)
    
    self:createGUI()

    --Prefetch audio resources
    for _, sound in pairs(def.audio.sounds) do
        self.audioPlayer:getSound(sound)
    end
    
    -- resuming the show
    if(Persist.help.showTime > 0) then
    
	    self:playVideo(movieName or "help:videos/help_1024_v20", false)
	    
	    -- setup sound seek time
	    self.audioPlayer:getSound(def.audio.primary):setSeekPosition(Persist.help.showTime)
        if (def.audio.segments.SegmentMainShow) then
            for _, soundDef in ipairs(def.audio.segments.SegmentMainShow) do
    	    	local name = soundDef.play
    	    	if(name) then
    				self.audioPlayer:getSound(name):setSeekPosition(Persist.help.showTime)
    			end
    		end
        end
		
	    self:pauseVideo()
    end
    
    Global.scene.app:startFadeVideo(1.5, 0.4, Color.new( 0.0, 0.0, 0.0, 1.0 ), Color.new( 1.0, 1.0, 1.0, 1.0 ))

    return true
end

function HelpApp:onShutdown()

	-- write show time to persistent
    System.flushPersist()

	-- write analytics
	AnalyticsManager:onShowEnded()
	-- Pause the audio system before destroying it
	System.pauseAllAudio();
    if self.audioPlayer then
        self.audioPlayer:destroy()
    end
    if self.video then
        self.video:close()
    end

    if self.scene then
        self.scene:destroy()
    end

    if self.view then
        self.view:destroy()
    end

    ResourceLoader:shutdown()

    if self.storyManager then
        self.storyManager:destroy()
        self.storyManager = nil
    end
    
end


function HelpApp:onUpdate()
	
	local dt = Time.dt();
    
    if(not Platform.candidate) then
  		Global.AppDebugText1 = string.format("Version 7 (Extended Intro)")-- %s", MainCamera.controller:getCameraVersion())
  	end
    
    self.loading = ( System.getPendingResourceCount() > 0 )

    if(self.showTCLDebug) then
        Global.AppDebugText3 = string.format(" TCL Strength: %.4f", MainCamera.controller:getTCLDebugStrength())
    end

--    if(self.startupHeadingLock) then
--        -- ALT_TITLE: comment out the line below for the alt Title Card behavior
--        MainCamera.controller:lookAtHeadingDegrees(self.heading)
--        if(self:getFrame() > 10) then
--            self.startupHeadingLock = false
--        end
--    end
	
	if(self.enablePitchRemap and self.elapsedPitchTime < self.pitchUnclampDuration) then
	
		local rawSensor = MainCamera.controller:getSensorRotation()
		
		-- change pitch limit if device is face up
		local pitchAngle = MainCamera:getLocal().rotation:getPitch()
		local rawPitch = rawSensor:getPitch()
		
		local pitchPercent =  -1
		
		if(self.pitchFrame > 2) then
		   pitchPercent = math.clamp(rawPitch/90, -1, 0)
		end
		
		if(pitchPercent <= 0 and self.maxPitchReached < pitchPercent) then
		   	self.maxPitchReached = pitchPercent
		end
		
		if(self:getFrame() > 10) then
			
			self.elapsedPitchTime = self.elapsedPitchTime + dt
			self.elapsedPitchTime = math.clamp(self.elapsedPitchTime, 0, self.pitchUnclampDuration)
		end
		
		self.elapsedPitchTime = math.clamp(self.elapsedPitchTime, 0, self.pitchUnclampDuration)
		
--		Global.AppDebugText1 = string.format("pitch %.4f, pitchPrcnt %.4f, cam angle %.4f", rawPitch, pitchPercent, pitchAngle)
--		Global.AppDebugText2 = string.format("maxPitchPrcnt %.4f, time %.4f", self.maxPitchReached, self.elapsedPitchTime)
		
		local t = 1.0 - math.easeInOutQuad((self.elapsedPitchTime/self.pitchUnclampDuration))
		local ease = 1.0 - math.clamp01(math.easeInOutQuad((1.0 - math.abs(self.maxPitchReached)), 0, 1, 1.6))
		
		MainCamera.controller:setProjectiveRemappingStrength(math.min(ease, ease*t))
		
--		Global.AppDebugText3 = string.format("ease %.4f, t %.4f", ease, t)
		
		self.pitchFrame = self.pitchFrame + 1
	end
	
    if (self.settings and self.settings.adaptive_toggle.checked) then
        if (self.settings.qual_slider) then
            self.settings.qual_slider:setValue(self.video:getQualityLevel())
        end
    end


    -- handle pause key
    if (Input.isKeyDown(KB_P) and HelpApp.pauseKeyUp) then

        HelpApp.pauseKeyUp = false

        HelpApp.paused = not HelpApp.paused
        self:pause(HelpApp.paused)
        Global.scene.app:pauseVideo(HelpApp.paused)
    end

    if (Input.isKeyUp(KB_P)) then
        HelpApp.pauseKeyUp = true
    end

    if self.settings then
        Toolbox:update()
    end

    if self.segmentManager then
        self.segmentManager:onUpdate();
    end

    self.scene:update()

    if self.audioPlayer then
        self.audioPlayer:update()
    end

    if Tracking_Controller and Tracking_Controller.onUpdate then
        Tracking_Controller:onUpdate(self.video)
        AnalyticTargetController:updateAnalytics(self.video)
    end
    
    if (self.videoFadeTimer > 0.0) then
        self:fadeVideoUpdate()
    end

    AnalyticsManager:update(self.video)
    
    -- save show time
    self.elapsedSaveTime = self.elapsedSaveTime + dt
    if(self.elapsedSaveTime > self.showTimeSaveFreq) then
    	
    	-- reset time
    	self.elapsedSaveTime = 0
    	
    	-- get the audio time
    	local showTime = self.audioPlayer:getSound(def.audio.primary):getPlaybackTime()
    	
    	if(self.segmentManager.currentSegment == "SegmentMainShow") then
    		
    		-- if in the main segment save the time
    		Persist.help.showTime = self.audioPlayer:getSound(def.audio.primary):getPlaybackTime()
--    		Persist.help.showTime = self.video:getTime()*1000

			System.flushPersist()
    		
    	elseif(self.segmentManager.currentSegment  == "SegmentFadeToCredits" 
	    	or self.segmentManager.currentSegment  == "SegmentCredits"
	    	or self.segmentManager.currentSegment  == "SegmentCreditsFade"
	    	or self.segmentManager.currentSegment  == "SegmentExit") then
    	
    		-- if at the show end reset to start
    		Persist.help.showTime = 0
    	end
    end
	
    --if self.video:isPlaying() then
    --  self.view:saveScreenshot()
    --end

    --System.gc()
end

function HelpApp:onReshape( width, height )
    assert( width > 0 and height > 0 )

    self.width  = width
    self.height = height

    self.view:resize( width, height )

    if GUI and not self.settings then
        self:createGUI()
    end

    self:updateCamera()
end

function HelpApp:onRender()
    if self.loading then
        if not Platform.candidate then
            Renderer.clear()
            Renderer.drawText( 30, 1280/2-10, font_style, "Loading..." )
            Renderer.display( self.view )
        end

        return
    end

    if Toolbox then
        local textScale = 1
        if (Platform.android) then
            textScale = 1.5
        end
        local y         = 42
        local yStep     = 28 * textScale
        local nextLine  = function()
            y = y + yStep
            return y - yStep
        end

        if(self.showFPS) then
            Renderer.drawText( 10, nextLine(), font_style, string.format( "fps: %s", FPS.string() ) )
            Renderer.drawText( 10, nextLine(), font_style, string.format( "frame: %d", self:getFrame() ) )
            Renderer.drawText( 10, nextLine(), font_style, string.format( "quality level: %d", self.video:getQualityLevel() ) )
            local debugMode = self.video:getDebugMode();
            if (debugMode > 0) then
                if (debugMode == 1) then
                    Renderer.drawText( 10, nextLine(), font_style, "debug mode: FRAME" )
                elseif (debugMode == 2) then
                    Renderer.drawText( 10, nextLine(), font_style, "debug mode: LOOKAHEAD" )
                elseif (debugMode == 3) then
                    Renderer.drawText( 10, nextLine(), font_style, "debug mode: MIP" )
                elseif (debugMode == 4) then
                    Renderer.drawText( 10, nextLine(), font_style, "debug mode: RECT" )
                end
            end
        end
        Renderer.drawText( 10, nextLine(), font_style, Global.AppDebugText1 or "")
        Renderer.drawText( 10, nextLine(), font_style, Global.AppDebugText2 or "")
        Renderer.drawText( 10, nextLine(), font_style, Global.AppDebugText3 or "")

        --[[
        if self.renderAudioInfo and self.audioPlayer then
        nextLine()
        self.audioPlayer:renderInfo(nextLine, font_style)
        end
        --]]
    end

    self.scene:render( self.view )
    Renderer.display( self.view )
end

function HelpApp:createScene()
    if StoryManager then
        HelpApp.storyManager = StoryManager:getFSM("help")
    end

    self.scene = Scene.create()
    self.scene.app = self

    local camera = self.scene:createCamera( "main" )

    camera.near = 0.01
    camera.far  = 100.0
    camera.fov  = 75.65 -- official FOV for Help final via Alex
    -- SensorController and MouseController eplaced by the InputLayer in the CameraController, should be removed after testing that sufficiently
    --camera:addComponent( "controller", Platform.android and SensorController or MouseController )
    --camera:addComponent( "controller2", ActorTracker_Controller)
    --NOTE: not sure this aspect it all correct
    camera.aspect = self.width / self.height

    -- Cylinder Test
    ---[[
    self.scene:createActor( "Cube", def.Cube )
    local material = self.scene.actors.Cube.materials["Mat_Cube"]
    self.scene.actors.Cube:setScale(Vector(0,0,0))

    self.video = self.scene.actors.Cube:addVideoCubePlayer()
    self.video:setCamera( camera )
    --]]

    -- TO DO: Should we have a list of global flags instead?
    self.scene.useDebugTargets = false
    self.scene.useFollowCam    = true
    self.scene.yawMult         = 0.95 --1.0
    self.scene.pitchMult       = 1.0

    self.renderAudioInfo = false

    if (not Global) then
        Global = {}
    end
    Global.scene = self.scene

end

function HelpApp:createView()
    local renderTargetDesc =
    {
        surfaces =
        {
            {
                attachment          = RTA_COLOR0,
                type                = RTF_TEXTURE2D,
                format              = PF_RGBA8,
            },
            {
                attachment          = RTA_DEPTH_STENCIL,
                type                = RTF_TEXTURE2D,
                format              = PF_DEPTH24_STENCIL8,
            }
        }
    }

    self.view = Renderer.createView( "main" )
    self.view:setFlags( VF_MAIN_VIEW )
    self.view:setRenderTarget( renderTargetDesc )
    --    self.view:setAlphaRange( 1, 1 )
    self.view:setClearColor(Color.new( 0.0, 0.0, 0.0, 1.0 ));

end

function HelpApp:createGUI()
    if not Toolbox then
        return
    end

    Toolbox:init()
    self.settings = Settings.create( self )

    local window = self.settings.window
    Toolbox:add( "settings", window )

    --      window:show()
    --  RenderDebug.setEnabled(true) -- Application.debug and self.checked )

    function Toolbox.uitoggle:onclick()
        if self.checked then
            window:show()
        else
            window:hide()
        end

        RenderDebug.setEnabled( Application.debug and self.checked )
    end

    --  window:show()
end

function HelpApp:pause( value )
    if (value) then
        Time.pause()
    else
        Time.resume()
    end
    self:pauseVideo(value)
end

function HelpApp:enableFps(value)
    self.showFPS = value
end

function HelpApp:setAdaptiveQuality( value )
    self.video:setAdaptiveQuality( value )
end

function HelpApp:setVideoDebugMode( value )
    self.video:setDebugMode( value )
end

function HelpApp:setVideoColor( value )
    self.video:setColor( value )
end

function HelpApp:setVideoQualityLevel( level )
    self.video:setQualityLevel( level )
end

function HelpApp:setDebugTimeline( value )
    self.video:setDebugTimeline( value );
end

function HelpApp:setVideoCubeShader(enableLUT)
    local material = self.scene.actors.Cube.materials["Mat_Cube"]
    local technique = "cubemap";

    if (Platform.videoSplitYUV) then technique = technique.."YUV" end
    if (enableLUT) then technique = technique.."LUT" end

    material:setShader( "help:shaders/video.rsfx", technique )
    if (enableLUT) then
        if (Platform.android) then
            material:setTexture( "u_lut", "help:textures/lut_galaxys5_srgb_v1.dds" )
        else
            material:setTexture( "u_lut", "help:textures/lut_identity.dds" )
        end
    end

    self.video:setMaterial( material )
end

function HelpApp:playVideo(videoUri, looping)
    if (not self.video) then
        return
    end

    -- Cylinder Test
    ---[[
    --self.video:stop()
    self.video:setVideo( videoUri )
    self.video:setLooping( looping )
    self:setVideoCubeShader( false )

    -- Sync video playback to a sound file
    if (self.audioPlayer and self.video) then
        local sound = self.audioPlayer:getSound(def.audio.primary)
        self.video:setSound(sound)
    end

    self.video:play()
    --self:setVideoColor(Vector.new(0,1,0,1))
	Events.send(self, "videoIsPaused", false)
    Replay:init(MainCamera)
    --]]
end

function HelpApp:pauseVideo( value )
    if (not self.video or not self.video:isPlaying()) then
        return
    end

    -- Cylinder Test
    ---[[
    if (value==true and not self.video:isPaused()) then
        self.video:pause()
        Events.send(self, "videoIsPaused", true)
    elseif (value==false and self.video:isPaused()) then
        self.video:play()
        Events.send(self, "videoIsPaused", false)
    end
    --]]
end

function HelpApp:getFrame()
    -- Cylinder Test
    ---[[
    if self.video then
        return self.video:getFrame()
    end
    --]]
    return 0
end

function HelpApp:isVideoPlaying()
    -- Cylinder Test
    ---[[
    if self.video then
        return self.video:isPlaying()
    end
    --]]
    return false
end

function HelpApp:startFadeVideo(duration, delay, fromColor, toColor)
    self.videoFadeTimer    = duration + delay
    self.videoFadeDuration = duration
    
    self.videoToColor   = toColor
    self.videoFromColor = fromColor
end

function HelpApp:fadeVideoUpdate()
    if (not self.video) then
        return
    end
    
    self.videoFadeTimer = self.videoFadeTimer - Time.dt()
    
    -- blendAmount goes 1 to 0, so flip the lerp
    local blendAmount   = math.easeInOutSine(math.clamp01(self.videoFadeTimer / self.videoFadeDuration))
    local matColor      = Color.lerp(blendAmount, self.videoToColor, self.videoFromColor)

    -- setColor() needs a vec4 right now
    local matColorVector = Vector.new(matColor.r, matColor.g, matColor.b, matColor.a)
    self.video:setColor(matColorVector)
    
    self.view:setAlphaRange(matColor.a, matColor.a)
    
    --Global.AppDebugText1 = string.format("%.5f : %.5f : %4.4f %4.4f %4.4f %4.4f", self.videoFadeTimer, blendAmount, matColor.r, matColor.g, matColor.b, matColor.a)    
end


function getFocalLengthFromFOV( sensorWidth, sensorHeight, cropValue, fov )
    local x = math.sqrt( sensorWidth*sensorWidth + sensorHeight*sensorHeight )
    local f = x / ( 2.0 * math.tan( math.pi * fov / 360.0 ) )
    return f / cropValue
end

function HelpApp:setFOV( fov )
    local focal_length = getFocalLengthFromFOV( 24.89, 18.66, 1.39, fov) -- super 35mm
    Log.message("FOV "..fov..", Focal Length: "..focal_length)
    MainCamera.fov = focal_length
    self:updateCamera()
end

function HelpApp:setBoom( boom )
    MainCamera.controller:setBoom(boom)
end

function HelpApp:enableDebugTargets(enable)
    self.scene.useDebugTargets = enable
end

function HelpApp:setTargetCenteringSpeed(strength)
    MainCamera.controller:setTargetCentering(strength)
end

function HelpApp:updateGlobalTrackingStrength(strength)
    MainCamera.controller:setGlobalTrackingStrength(strength)
end

function HelpApp:enableTrackingDebug(enable)
    Tracking_Controller:enableDebug(enable)
end

function HelpApp:enableTracking(enable)
    Tracking_Controller:enableTracking(enable)
end

function HelpApp:isTCLEnabled(enable)
    MainCamera.controller:setIsTCLEnabled(enable)
end

function HelpApp:setTCLDebug(enable)
    self.showTCLDebug = enable
end

function HelpApp:enableGlobalTrackScaleAnimation(enable)
    Tracking_Controller.enableGlobalScaleTracker = enable
end

function HelpApp:enableTrackingGroup(groupName, enabled)
    Log.message("Setting group: "..groupName.."to "..tostring(enabled))
    Tracking_Controller:enableTrackingGroup(groupName, enabled)
end

function HelpApp:enableFollowCam(enable)
    self.scene.useFollowCam = enable
end

function HelpApp:enableTrackingAnalytics(enable)
    Tracking_Controller.enableAnalytics = enable
end

function HelpApp:setCameraPrecisionFilterMinMax(min, max)
    MainCamera.controller:setPrecisionFilterMinMax(min, max)
end

function HelpApp:updateCamera()
    local camera = MainCamera

    local aspect = self.width / self.height
    MainCamera.aspect = aspect

    -- Use largest dimension to calculate FOV (engine by default uses narrowest)
    local projMatrix
    if aspect > 1.0 then
        projMatrix = Matrix.perspectiveFovX( math.rad( camera.fov ), aspect, camera.near, camera.far )
    else
        projMatrix = Matrix.perspectiveFovY( math.rad( camera.fov ), aspect, camera.near, camera.far )
    end

    camera:setProjectionMatrix( projMatrix )
end

function HelpApp:doDisablePitchClamp()
    if(not self.pitchClampDisabled) then
        local localT = MainCamera:getLocal()
        local pitch = localT.rotation:getPitch()
        if(pitch > 0) then
            self.pitchClampDisabled = true
--            MainCamera.controller:setPitchDownClamp(0.0)
        end
    end
end

Shell.attach("help", HelpApp);
