
--local Logger = TaggedLog("ResourceLoader", false)

ResourceLoader = {
    segmentData = {},
    resourceGroup = {},
    currentSegmentName = nil,
    globalResourceGroup = nil,
}

function ResourceLoader:init()
--    Logger:message("initializing")

    self:_loadGlobalResources()

    -- Initialize segments from scene data
    if def then
        for _, obj in pairs(def) do
            if obj.objectType == "Segment" then
                self.segmentData[obj.name] = obj
            end
        end
    end

    self:prefetchSegments()
    self:prefetchFlipbooks()    

    Events.listen(ResourceLoader, "segmentPlaying")
end

function ResourceLoader:shutdown()
    if self.globalResourceGroup then
        self.globalResourceGroup:destroy()
        self.globalResourceGroup = nil
    end
    if self.resourceGroup then
        for _, resources in pairs(self.resourceGroup) do
            resources:destroy()
        end
        self.resourceGroup = {}
    end
end

function ResourceLoader:_loadGlobalResources()
    local globalResourceGroup = System.createResourceGroup("GlobalResources")
    if def.globalResources then
        for _, entry in ipairs(def.globalResources) do
            local type = entry[1]
            local uri = entry[2]
            globalResourceGroup:addResource(type, uri)
        end
    end
    globalResourceGroup:prefetch()
    self.globalResourceGroup = globalResourceGroup
end

function ResourceLoader:prefetchSegments()
    if def.prefetchSegments then
        for _, segmentName in ipairs(def.prefetchSegments) do
--            Logger:message("Prefetching initial segment: " .. segmentName)
            self:prefetchSegment(segmentName)
        end
    end
end

function ResourceLoader:prefetchFlipbooks()
    if def.prefetchFlipbooks then
        groupName = "FlipbooksResourceGroup"
        local group = System.createResourceGroup(groupName)
        self.resourceGroup[groupName] = group
        for _, flipbookUri in ipairs(def.prefetchFlipbooks) do
            group:addFlipbook(flipbookUri)
        end
        self.resourceGroup[groupName]:prefetch()        
    end
end

function ResourceLoader:prefetchComplete()
    return System.getPendingResourceCount() == 0
end

function ResourceLoader:segmentPlaying(segmentName)
    if System.getPendingResourceCount() ~= 0 then
        Log.error("ResourceLoader: starting segment \"" .. segmentName .. "\" with " .. System.getPendingResourceCount() .. " pending resources to load.")
    end

    if segmentName == self.currentSegmentName then return end

--    Logger:message("segment playing: " .. segmentName)

    -- find the next segments that we could transition to and prefetch each of them
    local segment = self.segmentData[segmentName]
    local prefetchingSegment = {}
    if segment.transitions then
        for _, transition in ipairs(segment.transitions) do
            local nextSegment = transition.goTo
            prefetchingSegment[nextSegment] = true
            self:prefetchSegment(nextSegment)
        end
    end

    if self.currentSegmentName then
        segment = self.segmentData[self.currentSegmentName]
        if segment.transitions then
            -- free all segments we didn't transition to
            for _, transition in ipairs(segment.transitions) do
                local nextSegment = transition.goTo
                if nextSegment ~= segmentName and not prefetchingSegment[nextSegment] then
--                    Logger:message("Releasing " .. nextSegment .. " from " .. segmentName)
                    self:releaseGroup(nextSegment)
                end
            end
        end
--        Logger:message("Releasing " .. self.currentSegmentName .. " from " .. segmentName)
        if not prefetchingSegment[self.currentSegmentName] then
            self:releaseGroup(self.currentSegmentName)
        end
    end
    self.currentSegmentName = segmentName
end

function ResourceLoader:releaseGroup(name)
    local group = self.resourceGroup[name]
    if group then
        group:destroy()
        self.resourceGroup[name] = nil
    end
end

function ResourceLoader:_prefetchActors(actorList, group)
    if actorList then
        for _, actor in ipairs(actorList) do
            local actordef = def[actor]
            if actordef then
                if(actordef.model) then
                    actordef.model = string.lower(actordef.model)
                end

                -- make sure all animation paths are lower-case
                if actordef.animation then
                    for _, anim in pairs(actordef.animation) do
                        anim[1] = string.lower(anim[1])
                    end
                end

                --Log.message("prefetching actor: " .. tostring(actor))
                group:addActor(actordef)

                -- Handle any idle flipbook materials
                if actordef.fbdrivers then
                    for _, fbdata in pairs(actordef.fbdrivers) do
                        for _, fbmatdata in pairs(fbdata) do
                            if fbmatdata.uri then
                                group:addFlipbook(fbmatdata.uri)
                            end
                        end
                    end
                end

                if actordef.prefetch then
                    for _, entry in ipairs(actordef.prefetch) do
                        local type = entry[1]
                        local uri = entry[2]
--                        Logger:message("prefetching: "  .. tostring(type) .. " - " .. tostring(uri))
                        group:addResource(type, uri)
                    end
                end
            end
        end
    end
end

function ResourceLoader:_prefetchAudio(segmentName, group)
    local segmentEvents = def.audio and def.audio.segments[segmentName]
    if segmentEvents then
        for _, event in ipairs(segmentEvents) do
            local soundName = event.play or event.fade
            if soundName then
                local soundDef = def.audio.sounds[soundName]
                if soundDef then
                    group:addAudio(soundDef.uri)
--                    Logger:message("prefeching sound: " .. sound.name)
                end
            end
        end
    end
end

-- prefetch a segment - prefetches all the resources
-- for all the actors, sounds, materials, etc in the segment
function ResourceLoader:prefetchSegment(segmentName)
    -- Log.message("prefetchSegment: " .. tostring(segmentName))
    if not segmentName then
        segmentName = def and def.segments and def.segments.firstSegment
    end
    local segment = self.segmentData[segmentName]
    if segment then
        if self.resourceGroup[segmentName] then return end
        local group = System.createResourceGroup(segmentName)
        self.resourceGroup[segmentName] = group

        self:_prefetchActors(segment.active, group)
        self:_prefetchActors(segment.active_postRot, group)
        self:_prefetchActors(segment.prefetch, group)

        if AudioPlayer.enabled then
            self:_prefetchAudio(segmentName, group)
        end

        if segment.particleSystems then
            for _, uri in ipairs(segment.particleSystems) do
                group:addParticleSystem(uri)
            end
        end

        if segment.video then
            group:addVideoCube(segment.video)
        end        

        --Log.message("prefetching segment " .. segmentName .. " during " .. tostring(self.currentSegmentName))
        group:prefetch()

        -- If the segment is too short to prefetch its next segments, then we need
        -- to prefetch them when we prefetch this one
        if segment.prefetchTransitions and segment.transitions then
            for _, transition in pairs(segment.transitions) do
                local nextSegment = transition
                if type(transition) == 'table' then
                    nextSegment = transition.nextSegment
                end
                self:prefetchSegment(nextSegment)
            end
        end

        -- Prefetch special groups with this segment, but hold on
        -- to them until manually told to release them (via ResourceLoader:releaseGroup())
        if segment.prefetchGroups then
            for groupName, resources in pairs(segment.prefetchGroups) do
                if self.resourceGroup[groupName] then break end
                self.resourceGroup[groupName] = System.createResourceGroup(groupName)
                for _, resource in ipairs(resources) do
                    local type = resource[1]
                    local uri = resource[2]
                    self.resourceGroup[groupName]:addResource(type, uri)
                end
                self.resourceGroup[groupName]:prefetch()
            end
        end
    else
--        Logger:message("unable to find segment name = " .. tostring(segmentName))
    end
end
