
Segment = {}
Segment.__index = Segment

Segment.new = function(def, theScene)
	local ob = {};
	ob.def = def or {};
	ob.name = def.name or "";
	ob.playing = def.playing or false;
	ob.hasBegun = def.hasBegun or false;
	ob.duration = def.duration or 0;
	ob.video = def.video;
	ob.looping = def.looping or false;
	ob.active = def.active or {};
    ob.methods = def.methods or {};
    ob.scene = theScene;
    ob.target = nil;
    ob.videoDuration = def.videoDuration or ob.duration;
    ob.videoFrameCutoff = def.videoFrameCutoff or nil;
	return setmetatable(ob, Segment);
end

function Segment:onUpdate()
	if (not self.playing) then
		return
	end
	
	self.timeElapsed = self.timeElapsed + Time.dt()
            

    -- Stop the video if its time has elapsed (often this is the same as segment duration)    
    if (self.timeElapsed > self.videoDuration) then
        if self.scene.app then
            self.scene.app:pauseVideo(true)
        end
    end
    
    local tempCutoff = -1    
    local frameCutoffReached = false
    if (self.videoFrameCutoff) then
        tempCutoff =  self.videoFrameCutoff
        local currentFrame = self.scene.app:getFrame()
        -- This is potentially unreliable at high framerates
        -- We need a proper "video is done" function in the API
        frameCutoffReached = (currentFrame > self.videoFrameCutoff) or (self.videoFrameCutoff and self.scene.app:isVideoPlaying() == false)
    end

 	if ((self.timeElapsed > self.duration) or frameCutoffReached) then
 		if (self.looping) then
 			self:startPlaying();
 		else    		
            --self.playing = false;
            Events.send(self, "segmentFinished", self);
 		end	
 	end

    --Global.AppDebugText2 = self.name .. string.format(" %.3f / %.3f : %4d / %4d : ", self.timeElapsed, self.duration, self.scene.app:getFrame(), tempCutoff )..tostring(frameCutoffReached) 	
 	
    self.target = self:lookingAtTrigger();
    
    --[[
    if (self.target) then
        Global.AppDebugText2 = self.target.name
    else
        Global.AppDebugText2 = ""
    end
    --]]
    --Log.error(" Animation time"  .. self.scene.actors["Gracie"].animation[self.name]:getTime())
end

function Segment:isPlaying() 
	return self.playing;
end

function Segment:getName()
	return self.name;
end

function Segment:startPlaying()
    Events.send(self, "segmentStarted", self.name);

    self:activateObjects();
    self.playing = true;
    self.startTime = self:currentTime();
    self.endTime = self.startTime + self.duration * 1000
    self.timeElapsed = 0.0;
    self:playAnimations();

    if self.video and self.scene.app then
        self.scene.app:playVideo(self.video, false)
    end
    
    if self.scene.app then
        self.scene.app:pauseVideo(false)
    end

    self:runMethods();
    self.hasBegun = true;
    Log.message("Starting to play: " .. self.name .. " at " .. tostring(self.scene.app:getFrame()));
end

function Segment:runMethods()
    for objectName, table in pairs(self.methods) do
        if (self.scene and self.scene.actors and self.scene.actors[objectName]) then        
            for _, fn in pairs(table) do
                fn(self,self.scene.actors[objectName] )
            end
        elseif (objectName == "NoActor") then        
            for _, fn in pairs(table) do
                fn(self,nil )
            end
        else
            Log.message("actor for method not found: " .. objectName)
        end
    end
end

function Segment:stopPlaying()
    self.playing = false;
end

function Segment:deactivateObjects(oldactorlist, newactorlist)

end

creator = {
    Actor = function(self, objectName, def)
        local actor = self.scene:createActor(objectName, def[objectName])
        actor.name = objectName
        actor.targetLocator = def[objectName].targetLocator
        actor:setVisibility(true)
        return actor
    end,
    Trigger = function(self, objectName, def)
    return self.scene:createActor(objectName, def[objectName]);
    end,
    Camera = function(self, objectName,def)
    return self.scene:createActor(objectName, def[objectName]);
    end,
    Set = function(self, objectName,def)
    return self.scene:createActor(objectName, def[objectName]);
    end,
    Light = function(self, objectName,def)
    return self.scene:createLight(objectName, def[objectName]);
    end,
}

function Segment:createObject(objectName)
    Log.message("Creating " .. objectName);
    Log.message(table.tostring(def[objectName]));
    Log.message(  " type " .. def[objectName].objectType);
    local a = creator[def[objectName].objectType](self, objectName, def);
    self.scene.activatedObjs[objectName] = a;
     -- Add any components needed
    if (def[objectName].components) then
        Log.message(table.tostring(def[objectName].components));
        for _, compData in ipairs(def[objectName].components) do
            Log.message("entry " ..table.tostring( compData) );
            local class = _G[compData.name]
            --Log.message("class " .. class );
            if (class == nil) then
                Log.message("Unknown component type: " .. tostring(compData.name))
            else
                a:addComponent(compData.name, class:new(compData.params))
                Log.message("Added component: " .. compData.name .. " to " .. a.name)
            end
        end
    end
    return a;
end

function Segment:activateObjects() 
    self.scene.activatedObjs = self.scene.activatedObjs or {};
    --TODO check against multiple self.scene lists, not just actors    
    
    local remove = shallowcopy(self.scene.activatedObjs)
    
	for _, objectName in ipairs(self.active) do
        if (self.scene and self.scene.actors and self.scene.actors[objectName]) then		
            self.scene.activatedObjs[objectName] = self.scene.actors[objectName]
        elseif (self.scene and self.scene.lights and self.scene.lights[objectName]) then        
            self.scene.activatedObjs[objectName] = self.scene.lights[objectName] 
        else
             self.scene.activatedObjs[objectName] = self:createObject(objectName);
        end
        
        remove[objectName] = nil
	end
	
    -- Remove any unneeded objects
    -- NB: This method of removal assume only one segment active at a time
    if (remove) then
        for name, _ in pairs(remove) do
            local object = self.scene:find(name)
            if (object) then
                object:destroy()
            end
            self.scene.activatedObjs[name] = nil
        end
    end
end

function Segment:lookingAtTrigger()
    for _,target in pairs(self.scene.activatedObjs) do
		if (target and target.components) then
			local detector = target.components["LookAtDetector"]
			if (detector and detector:lookedAt()) then
				return target
			end
		end
    end

    return nil
end

function Segment:playAnimations()
    for _, actorName in ipairs(self.active) do
    	if (self.scene.activatedObjs[actorName] and self.scene.activatedObjs[actorName].animation and self.scene.activatedObjs[actorName].animation[self.name]) then
      	 	self.scene.activatedObjs[actorName].animation[self.name]:play();
    	end
    end
end


function Segment:currentTime()
    return Time.milliseconds()
end
