

-- Simple wrapper around the system's Log service that allows us
-- to easily tag the output from a single file in the logs as
-- well as turn off logging for a given file without needed to
-- comment out or remove the log function calls

-- Usage: at the top of the file add a:
-- local Logger = TaggedLog("ModuleName", true)
--
-- then use the Logger:message(), Logger:warning(), or Logger:error()
-- methods in the same way you would with Log.
--
-- The output for each log message will be prepended with the tag
-- for easy filtering.  If you want to turn off output for a given file
-- just pass false as the second parameter to Logger.new.  That will
-- turn off all of the logging functions
--
-- Note, you should really only use this for debugging purposes, as
-- they will most likely be turned off for release builds.

TaggedLog = {}
TaggedLog.__index = TaggedLog

function TaggedLog.new(tag, enabled)
    return setmetatable({ tag = tag, enabled = enabled }, TaggedLog)
end

local log = function(logger, message, fn)
    if (logger.enabled) then
        if (logger.tag) then
            message = logger.tag .. ": " .. message
        end
        Log[fn](message)
    end
end

function TaggedLog:message(message) log(self, message, "message") end
function TaggedLog:warning(message) log(self, message, "warning") end
function TaggedLog:error(message) log(self, message, "error") end

setmetatable(TaggedLog, { __call = function(_, ...) return TaggedLog.new(...) end })
