----------------------------------------------------------------------------------------
-- inpect functions and utilities
----------------------------------------------------------------------------------------


-- specialized __inspect functions for common types
local prettify_b = function(b)
	if(b) then
		return "true";
	else
		return "false";
	end
end

local prettify_v2 = function(self)
	return 	string.format("%.2f", self.x)..", "..
			string.format("%.2f", self.y);
end

local prettify_v3 = function(self)
	return 	string.format("%.2f", self.x)..", "..
			string.format("%.2f", self.y)..", "..
			string.format("%.2f", self.z);
end

local prettify_v4 = function(self)
	return 	string.format("%.2f", self.x)..", "..
			string.format("%.2f", self.y)..", "..
			string.format("%.2f", self.z)..", "..
			string.format("%.2f", self.w);
end

Vector.__inspect2 = function(self, pgrid, group)
	pgrid:addProperty(group, "x", self.x);
	pgrid:addProperty(group, "y", self.y);
end

Vector.__inspect3 = function(self, pgrid, group)
	pgrid:addProperty(group, "x", self.x);
	pgrid:addProperty(group, "y", self.y);
	pgrid:addProperty(group, "z", self.z);
end

Vector.__inspect = function(self, pgrid, group)
	pgrid:addProperty(group, "x", self.x);
	pgrid:addProperty(group, "y", self.y);
	pgrid:addProperty(group, "z", self.z);
	pgrid:addProperty(group, "w", self.w);
end

function Quaternion.__inspect(self, pgrid, group)
	pgrid:addProperty(group, "x", self.x);
	pgrid:addProperty(group, "y", self.y);
	pgrid:addProperty(group, "z", self.z);
	pgrid:addProperty(group, "w", self.w);
end

function Color.__inspect(self, pgrid, group)
	pgrid:addProperty(group, "r", self.r);
	pgrid:addProperty(group, "g", self.g);
	pgrid:addProperty(group, "b", self.b);
	pgrid:addProperty(group, "a", self.a);
end


SRT.__inspect = function(self, pgrid, group)
	pgrid:addProperty(group, "position",	prettify_v3(self.position));	
	pgrid:addProperty(group, "rotation",	prettify_v4(self.rotation));
	pgrid:addProperty(group, "scale",		prettify_v3(self.scale));
end

Rect = {}
Rect.__inspect = function(self, pgrid, group)
	pgrid:addProperty(group, "left", self.left);
	pgrid:addProperty(group, "top", self.top);
	pgrid:addProperty(group, "right", self.right);
	pgrid:addProperty(group, "bottom", self.bottom);
end

GUIFrame.__inspect = function(self, pgrid, group)
	local r = self:getLocalRect();
	local subgroup = pgrid:addGroup(group, "Local Rectangle");	
	Rect.__inspect(r, pgrid, subgroup);
	
	r = self:getWorldRect();
	subgroup = pgrid:addGroup(group, "World Rectangle");	
	Rect.__inspect(r, pgrid, subgroup);
	
	subgroup = pgrid:addGroup(group, "Dimensions");	
	local pos = self:getWorldPosition();
	pgrid:addProperty(subgroup, "x", pos.x);
	pgrid:addProperty(subgroup, "y", pos.y);
	pgrid:addProperty(subgroup, "width", self:getWidth());
	pgrid:addProperty(subgroup, "height", self:getHeight());
end

Animation.__inspect = function(self, pgrid, group)
	if(group and not(group:isVisible())) then return end;
	
	pgrid:addProperty(group, "anim", self.uri);
	pgrid:addProperty(group, "frame", self:getFrame());
	pgrid:addProperty(group, "clipStart", self:getStartFrame());
	pgrid:addProperty(group, "clipEnd", self:getStopFrame());
	pgrid:addProperty(group, "clipLength", self:getNumFrames());
	pgrid:addProperty(group, "framesPerSec", self:getFramesPerSecond());
	pgrid:addProperty(group, "time", self:getTime());
	pgrid:addProperty(group, "duration", self:getDuration());
	pgrid:addProperty(group, "progress", self:getProgress());
	pgrid:addProperty(group, "rate", self:getRate());
	pgrid:addProperty(group, "blend", self:getWeight());	
	
	local state = "inactive";
	if(self:isPlaying()) then
		state = "playing";
		if(self:isPaused()) then
			state = "paused";
		end
	end
	pgrid:addProperty(group, "state", state);
	pgrid:addProperty(group, "looping", prettify_b(self:isLooping()));
	pgrid:addProperty(group, "interpolated", prettify_b(self:isInterpolated()));
end

Light.__inspect = function(self, pgrid, group)
	local lightInfo = self:inspect();
	
	local lightTypeNames = {
		"directional",
		"point",
		"spot",
	};
	
	local subgroup = pgrid:addGroup(group, "Light", true);
	pgrid:addProperty(subgroup, "type", lightTypeNames[lightInfo.type+1]);
	if(lightInfo.type == LT_SPOT) then
		pgrid:addProperty(subgroup, "spotAngleMin", math.deg(lightInfo.spotAngle.x));
		pgrid:addProperty(subgroup, "spotAngleMax", math.deg(lightInfo.spotAngle.y));
	end
	
	pgrid:addProperty(subgroup, "ambient", prettify_v4(lightInfo.ambient));
	pgrid:addProperty(subgroup, "diffuse", prettify_v4(lightInfo.diffuse));
	pgrid:addProperty(subgroup, "rangeMin", lightInfo.range.x);
	pgrid:addProperty(subgroup, "rangeMax", lightInfo.range.y);
	pgrid:addProperty(subgroup, "wrap", lightInfo.wrap);
	
	Transform.__inspect(self, pgrid, group);
end

function table.not_empty (self)
    for _, _ in pairs(self) do
        return true;
    end
    return false;
end

Actor.__inspect = function(self, pgrid, group)
	local actorGroup = pgrid:addGroup(group, "Actor", true);
	
	pgrid:addProperty(actorGroup, "model", self.uri);
	
	if(self.animation) then
		local animTableGroup = pgrid:addGroup(actorGroup, "Animation", false);
		if(animTableGroup:isVisible()) then
			if(table.not_empty(self.animation)) then
				for k, anim in pairs(self.animation) do
					local animgroup = pgrid:addGroup(animTableGroup, k);
					anim:__inspect(pgrid, animgroup);		
				end
			end
		end
	end
		
	local skeleton = pgrid:addGroup(actorGroup, "Skeleton", false);
	if(skeleton:isVisible()) then	
		local transformList = self:inspectSkeleton();
		for k, v in ipairs(transformList) do 
			local boneGroup = pgrid:addGroup(skeleton, "bone["..k.."]");
			if(boneGroup and boneGroup:isVisible()) then
				local name = string.gsub(string.gsub(v.name, "%w+:", ""), "|%w+", ""); -- strip namespace
				pgrid:addProperty(boneGroup, "name", name);
				pgrid:addProperty(boneGroup, "visible", prettify_b(v.visibility));		
				v.worldTransform:__inspect(pgrid, boneGroup);
			end
			-- break after the first bone if the group is collapsed.
		end
	end
	
-- TODO: material support
--	if(self.materials) then
--		if(table.not_empty(self.materials)) then
--			local subgroup = pgrid:addGroup(group, "Materials");
--			table.__inspect(self.materials, pgrid, subgroup);		
--		end
--	end

	Transform.__inspect(self, pgrid, group);

end

Object.__inspect = function(self, pgrid, group)
	local objectGroup = pgrid:addGroup(group, "Object", true);
	
	local mt = getmetatable(self);
	if(mt) then
		pgrid:addProperty(objectGroup, "classname", mt.__name);		
		if(mt.__index.__base) then
			pgrid:addProperty(objectGroup, "basename", mt.__index.__base.__name);
		end
	end

	if(self.__object) then
		pgrid:addProperty(objectGroup, "handle", System.handleToString(self.__object));
	end
end


Transform.__inspect = function(self, pgrid, group)
	self:getWorld():__inspect(pgrid, pgrid:addGroup(group, "World Transform", true));
	self:getLocal():__inspect(pgrid, pgrid:addGroup(group, "Local Transform", true));
	
	if(self.components) then
		local subgroup = pgrid:addGroup(group, "Components", true);
		for k, v in pairs(self.components) do
			local compgroup = pgrid:addGroup(subgroup, k);
			table.__inspect(v, pgrid, compgroup);
		end
	end
	
	local transformGroup = pgrid:addGroup(group, "Node", true);
	
	local parent = self:getParent();
	local parentName = "nil";
	if(parent) then
		parentName = parent:getName();
	end	
	
	pgrid:addProperty(transformGroup, "parent", parentName);
	pgrid:addProperty(transformGroup, "active", prettify_b(self:isActive()));
	pgrid:addProperty(transformGroup, "visible", prettify_b(self:isVisible()));		
	
	Object.__inspect(self, pgrid, group);
end

table.__inspect = function(self, pgrid, group, visited)
	visited = visited or {}
	if(visited[self]) then
		return;
	end
	
	for k, v in pairs(self) do
		if(type(k) == "number") then
			k = "index["..k.."]";	
		elseif(type(v)=="table") then
			if(v.__base) then
				local name = v.name;
				if(name == nil and v.getName) then
					name = v:getName();
				else
					name = "unknown";
				end
				pgrid:addProperty(group, k, name);								
			else
				local subgroup = pgrid:addGroup(group, k);	
				if(v.__inspect~=nil) then
					v:__inspect(pgrid, subgroup);
				else
					table.__inspect(v, pgrid, subgroup);
				end			
			end
		elseif(type(v)=="boolean") then
			if(v) then
				pgrid:addProperty(group, k, "true");	
			else
				pgrid:addProperty(group, k, "false");	
			end
		elseif(type(v)=="string" or type(v)=="number" ) then		
			pgrid:addProperty(group, k, v);	
		elseif(type(v)=="function") then
			pgrid:addProperty(group, k, "function");			
		else
			-- unsupported
		end
	end
	visited[self] = self;	
end

----------------------------------------------------------------------------------------
-- Object inspector
-- A simple read-only implementation of an object property viewer. Currently property 
-- values can only be displayed in string form. Generic lua tables are inspected using 
-- table.__inspect(), in which case properties will not be nicely ordered (not yet). 
-- For a more structured layout, objects may have a specialized __inspect function.
-- NOTE: The inspect function acts like immediate mode layout creation. 
-- __inspect is called every frame. A property is added to the grid if it is not 
-- already present, otherwise the property value is simply updated to the current 
-- value. The current implementation does not automatically clean up stale GUI frames
-- (groups or properties that are no longer being referenced during population). 
-- As a result, the property grid will never shrink. 
----------------------------------------------------------------------------------------
ObjectInspector = {}
ObjectInspector.__index = ObjectInspector;
ObjectInspector.defaultPos = GUIPos(300,0);
ObjectInspector.defaultSize = GUISize(300, 512);

ObjectInspector.init = function(self, name)
	self.window	= GUI.window(ObjectInspector.defaultPos, ObjectInspector.defaultSize, "Object Inspector");	
	self.scroll	= GUI.scrollframe(self.window, GUIPos(0, 0), GUISize(SZ_FILL,SZ_FILL), "vertical", "");
	self.pgrid	= GUI.propertygrid(self.scroll, GUIPos(0,0), GUISize(SZ_FILL, SZ_WRAP));
end

ObjectInspector.new = function()	
	local inspector = setmetatable({}, ObjectInspector);
	inspector:init("Object Inspector");	
	return inspector;
end

ObjectInspector.destroy = function(self)
	self.window:destroy();
end

ObjectInspector.setObject = function(self, object)
	self.window:setTitle(object:getName());
	if(self.object ~= object) then
		self.object = object;
		
		-- update property grid
		self.pgrid:clear();
		object:__inspect(self.pgrid, nil);
		--
	end
	-- show the window.
	self.window:show();
end

ObjectInspector.update = function(self)
	if(status(self.object)==nil) then
		self.window:close();
		return;
	end
	self.object:__inspect(self.pgrid, nil);
end

ObjectInspector.hide = function(self)
	self.window:hide();
end

ObjectInspector.show = function(self)
	self.window:show();
end

----------------------------------------------------------------------------------------
-- Scene Inspector
-- Manages a hierarchy view of the scene and a list of inspector windows. 
-- How to use:
-- To select an object, single click the name of an object in the hierarchy view.
-- To inspect an object, double click the name of an object in the hierarchy view. 
-- Hold the control key and double click to bring up additional inspector windows to 
-- inspect multiple objects simultaneously. 
-- To view the scene from a different camera, select from any available cameras in 
-- the dropdown menu. 
----------------------------------------------------------------------------------------
SceneInspector = {}
SceneInspector.__index = SceneInspector;
SceneInspector.defaultPos = GUIPos(0,0);
SceneInspector.defaultSize = GUIPos(300,512);

SceneInspector.initCameraSelectionTool = function(self, window, scene)
	local toolbar = GUI.frame(window, GUIPos(0, 0), GUISize(SZ_FILL, 48), "");
	toolbar:dock(DOCK_TOP);
--[[	
	local inspectorCam = scene:createCamera("InspectorCamera");
	inspectorCam:setPerspective(35 * (1184 / 1280), 1.0, 1000) -- match long axis to unity's aspect
	inspectorCam.controller = inspectorCam:addComponent("InspectorCamera", FlyCameraController.new())
	inspectorCam:setActive(false);
]]	
	local cameraMenuFrame = GUI.frame(toolbar, GUIAligned(ALIGN_CENTER), GUISize(SZ_FILL, 32), "");	
	cameraMenuFrame:anchor(true, true, true, false);
	
	local cameraMenuLabel = GUI.label(cameraMenuFrame, GUIPos(0,0), GUISize(72, SZ_FILL), "Camera:", "h1");
	cameraMenuLabel:setTextAlign(ALIGN_CENTER_LEFT);
	cameraMenuLabel:dock(DOCK_LEFT);
	
	-- build menu selection
	local cameras = {};
	local selectedCam = 1;
	for k, v in pairs(scene.cameras) do
		table.insert(cameras, k);
		if(v:isActive()) then
			selectedCam = #cameras;
		end
	end	
	
	-- create camera menu
	local cameraMenu = GUI.dropdownmenu(cameraMenuFrame, GUIPos(0,0), GUISize(32, 32), cameras, selectedCam);
	cameraMenu:dock(DOCK_FILL);	
	
	cameraMenu.onItemDeselected = function(self, name)
		Log.message("deactivate: "..name);
		local camera = scene:find(name);
		if(camera) then
			camera:setActive(false);
		end
	end
	
	cameraMenu.onItemSelected = function(self, name)
		Log.message("activate: "..name);
		local camera = scene:find(name);
		if(camera) then
			camera:setActive(true);
		end
		Toolbox.mainView:setCamera(camera);
	end
end

SceneInspector.addInspector = function(self, scene, name)
	self.inspectors = self.inspectors or {};
	local object = scene:find(name);
	if(object == nil) then
		Log.error("SceneInspector: could not find object: "..name); 
		return;
	end

	local inspectorIndex = self:findInspectorIndex(name);
	if(inspectorIndex > 0) then
		self.inspectors[inspectorIndex][2].window:bringToFront();
	elseif(#self.inspectors > 0 and not(Input.isModifierDown(KB_RCTRL) or Input.isModifierDown(KB_LCTRL))) then			
		self.inspectors[#self.inspectors][2]:setObject(object);
	else
		local inspector = ObjectInspector.new();
		local sceneInspector = self;
		inspector.window.onClose = function(self)
			sceneInspector:removeInspector(name);
			inspector:destroy();
		end
		
		if(#self.inspectors > 0) then
			local top = self.inspectors[#self.inspectors][2];
			local pos = top.window:getWorldPosition();
			inspector.window:moveTo(pos.x, pos.y); 
			inspector.window:move(0, 32);
		end
	
		-- insert by name and index
		table.insert(self.inspectors, { name, inspector });
		
		inspector:setObject(object);
		--			
	end		
end

SceneInspector.removeInspector = function(self, name)
	local inspectorIndex = self:findInspectorIndex(name);
	table.remove(self.inspectors, inspectorIndex);
end

SceneInspector.findInspectorIndex = function(self, name)
	for index, inspector in ipairs(self.inspectors) do
		if(inspector[1]==name) then
			return index;
		end
	end
	return 0;
end

SceneInspector.initSceneHierarchyView = function(self, window, scene)
	local sceneInspector = self;
	self.scrollframe = GUI.scrollframe(window, GUIPos(0, 0), GUISize(SZ_FILL,32), "vertical", "treelist.frame");
	self.treelist = GUI.treelist(self.scrollframe, parent, GUIPos(0,0), GUISize(SZ_FILL, SZ_FILL));
	self.treelist:populate(scene:inspect());
	self.treelist:expand(); -- expand root only
	
	-- when selecting an item, create an object inspector
	self.treelist.onItemSelected = function(self, item)
		local object = scene:find(item.name);
		if(object) then
			if(object and object.setRenderDebug) then
				object:setRenderDebug(true);
			end
		end
	end
	
	self.treelist.onItemDeselected = function(self, item)
		local object = scene:find(item.name);
		if(object) then
			if(object and object.setRenderDebug) then
				object:setRenderDebug(false);
			end
		end
	end
	
	self.treelist.onItemDblClick = function(self, item)
		sceneInspector:addInspector(scene, item.name);
	end
end

SceneInspector.initWindow = function(self, scene)
	self:initCameraSelectionTool(self.window, scene);
	self:initSceneHierarchyView(self.window, scene);	
	--self.scene = Global.scene;
end

SceneInspector.onUpdate = function(self)
	if(self.scene == nil) then
		-- wait for Global.scene to be set.
		--if(Global.scene) then
		--	self:initWindow(Global.scene);
		--else
		--	return;
		--end
	end
	
	-- update scene hierarchy view
	if(self.treelist) then
		self.treelist:populate(self.scene:inspect());
	end

	-- update inspectors
	if(self.inspectors) then
		for index, inspector in ipairs(self.inspectors) do
			inspector[2]:update();
		end
	end
end

SceneInspector.hide = function(self)
	self.window:hide();
	if(self.inspectors) then
		for name, inspector in pairs(self.inspectors) do
			inspector:hide();
		end
	end	
end

SceneInspector.show = function(self)
	self.window:show();
	if(self.inspectors) then
		for name, inspector in pairs(self.inspectors) do
			inspector:show();
		end
	end	
end

SceneInspector.new = function()
	-- create the window
	local inspector = setmetatable({}, SceneInspector);		
	inspector.window = GUI.window(SceneInspector.defaultPos, SceneInspector.defaultSize, "Scene Hierarchy");		
    inspector.window.onClose = function(self)
        if inspector.onClose then
            inspector:onClose()
        end
    end
	return inspector;
end
