
local Logger = TaggedLog.new("AnimationManager", true)

local previousThreshold = 1.5

AnimationManager = {}

function AnimationManager:init(camera, forceStartSegment)
	self.camera = camera
    self.current = nil
    self.segments = {}
    self.history = {}
    self.startSegment = nil
    self.waitForStreamingState  = nil

    local firstSegmentName
    -- initialize segments from scene data
	if (segments and type(segments) == "table") then
		for _, seg in ipairs(segments) do
            local segment = Segment.new(seg, self.camera)
            segment:onInitialize()
			self.segments[segment.name] = segment
            firstSegmentName = firstSegmentName or segment.name
		end
	end

	-- Listen for control events
	Events.listen(self, "nextAnimation")
	Events.listen(self, "previousAnimation")

    self.startSegment = firstSegmentName
    if (Context.getCheckpoint()) then
        self.startSegment = Context.getCheckpoint()
    end

    if (forceStartSegment) then
        self.startSegment = forceStartSegment
    end

    if (self.startSegment) then
        self:goToSegment(self.startSegment)
        
        if (forceStartSegment) then
	        if (Global.scene.mainCamera.controller.enableSensor) then
	            Global.scene.mainCamera.controller:enableSensor()
	        end
            if Global.scene.actors.WhiteScreen then
                Global.scene.actors.WhiteScreen:setVisibility(false)
            end
            Shell.start("windyday")
        end
        if (Context.getCheckpoint()) then
            SoundManager:startGlobalFadeIn()
        end
    end
end

function AnimationManager:update()
    if (Input.isKeyDown(KB_N, true)) then
        self:next()
    elseif (Input.isKeyDown(KB_P, true)) then
        self:previous()
    elseif (Input.isKeyDown(KB_SPACE, true)) then
        self:pause()
    end

    if self.waitForStreamingState then
        self:goToSegment(self.waitForStreamingState)
    end

	if (self.current) then
		self.current:update()
	end
	
	if(System.isStatManagerEnabled()) then
		local segTime = Stat.segmentTime:get() + Time.dt();
		Stat.segmentTime:set(segTime);
	end
end

function AnimationManager:next()
    if (self.current == nil or self.current.nextSegments == nil or table.getn(self.current.nextSegments) <= 0) then
        return
    end
    self.current:resetToEnd()
    self:goToSegment(self.current.nextSegments[1])

	AsyncLoader.prefetch();
end

function AnimationManager:previous()
    if (self.current == nil) then
        return
    end

    local goToPrevious = (self.current.currentTime() < previousThreshold);
    local prevSegment
    local numHistory = table.getn(self.history)
    if (numHistory > 1) then
        prevSegment = self.history[numHistory - 1]
    end
    self.current:resetToBeginning(prevSegment);
    if (numHistory ~= 1 and goToPrevious) then
        table.remove(self.history)
        numHistory = table.getn(self.history)
        self.current = self.history[numHistory]
        Stat.segment:set(self:currentDisplayName());

        prevSegment = nil
        if (numHistory > 1) then
            prevSegment = self.history[numHistory - 1]
        end
        self.current:resetToBeginning(prevSegment)
    end

    if (self.current.debugTransparent) then
        -- Don't play transparent segments, skip to the one before
        self:previous()
    else
        -- Play normally
        self.current:startPlaying()
        self:lookAtMainTarget()
    end
    self:notifySegmentPlaying()

	-- jump the async loader back
	AsyncLoader.jump(-1);
end

function AnimationManager:pause()
	Time.pause();
end

function AnimationManager:goToSegment(name)
    local lastName = self.current and self.current.name or nil

    if Global.storyManager then
        local resourcesLoaded = Global.storyManager:setState(name)
        if not resourcesLoaded then
            Logger:message("waiting for segment resources: " .. name)
            self.waitForStreamingState = name
            return
        else
            self.waitForStreamingState = nil
        end
    end

    Logger:message("goToSegment: " .. name)
	local segment = self.segments[name]
	if (segment == nil) then
        Log.error("Attempting to go to an unknown segment: " .. name)
        return
	end

    Logger:message("Starting segment: " .. name)
	self.current = segment
    if(System.isStatManagerEnabled()) then
        Stat.segment:set(self:currentDisplayName());
    end
	table.insert(self.history, self.current)

    local startingCheckpoint = (Context.getCheckpoint() == self.current.name)
	self.current:startPlaying(startingCheckpoint)
	self:notifySegmentPlaying()

	if(System.isStatManagerEnabled()) then
		Stat.segment:set(name);
		Stat.segmentTime:set(0.0);
	end
	AsyncLoader.prefetch()
    if lastName then
        AsyncLoader.freeResourceGroup( lastName )
    end
end

function AnimationManager:lookAtMainTarget()
    if (self.current) then
        local mouseBone = ActorManager:getBone({actor="Mouse", bone="mouse_mox_cn_pelvis|mouse_MOXIE_skeleton|RootNode"})
        if (mouseBone) then
            self.camera.controller:lookAt(mouseBone:getWorld())
        end
    end
end

function AnimationManager:currentSegmentName()
    return (self.current and self.current.name) or ""
end

function AnimationManager:currentDisplayName()
return (self.current and self.current.displayName) or ""
end

function AnimationManager:currentWindMagnitude()
    return (self.current and self.current.windMagnitude) or ""
end

function AnimationManager:notifySegmentPlaying()
    if (self.current and self.current.checkpoint) then
        Context.onCheckpoint(self.current.name)
        -- TODO: call Context.onCheckpoint(self.current.name)
    end
    if SceneControlPanel then
        SceneControlPanel:setText(self:currentDisplayName())
    end
    Events.send(self, "segmentPlaying", self:currentSegmentName())
    Events.send(self, "windMagnitude", self:currentWindMagnitude())
end
