
local Logger = TaggedLog("MusicPlayer", false)

MusicPlayer = {
    disableMusicBeds = false,
    disableMusicStabs = false,
    disableAmbientBeds = false,
}

function MusicPlayer:onStart()
    Logger:message("onStart")

    self.currentMusicEvent = nil

    Events.listen(self, "segmentPlaying")

    local startSegment = AnimationManager:currentSegmentName()
    if (startSegment ~= "") then
        self:segmentPlaying(startSegment)
    end
end

function MusicPlayer:onDestroy()
    Events.stopListening(self, "segmentPlaying")
end

function MusicPlayer:playIntroMusic()
    self.currentMusicEvent = musicIntro
    self:playEvent(self.currentMusicEvent)
end

function MusicPlayer:segmentPlaying(segmentName)
    Logger:message("segment playing: " .. segmentName)
    self:playSegmentMusic(segmentName)
end

function MusicPlayer:playSegmentMusic(segmentName)
    self.currentMusicEvent = nil
    for _, event in ipairs(music) do
        if (event.name == segmentName) then
            self.currentMusicEvent = event
            break
        end
    end
    if (self.currentMusicEvent == nil) then
        Logger:error("Didn't find music for " .. segmentName)
        return
    end

    self:playEvent(self.currentMusicEvent)
end

function MusicPlayer:playEvent(event)
    -- queue up music bed events
    Logger:message("Playing event: " .. event.name)
    for _, bed in ipairs(event.musicBedEvents) do
        local startTime = event.segmentStartFrame
        local playTime = bed.startTime
        local startDelay = (playTime - startTime) / 24
        local quantizedDelay = self:getQuantizedDelay(bed.quantized);

        self:playAudioBed(
            bed.name,
            (startDelay + quantizedDelay),
            bed.fadeTime,
            bed.fadeStart,
            bed.fadeEnd,
            bed.ambient)
    end

    -- queue up music stab events
    if (not self.disableMusicStabs) then
        for _, stab in ipairs(event.musicStabEvents) do
            local startTime = event.segmentStartFrame
            local playTime = stab.startTime
            local stabDelay = (playTime - startTime) / 24
            Logger:message("Setting up stab: " .. tostring(stab.name) .. ", time = " .. tostring(totalDelay))
            DelayScheduler:runAfter(stabDelay, function() MusicPlayer:playQuantizedStab(stab) end, stab.name)
        end
    end
end

function MusicPlayer:playAudioBed(musicBedName, startDelay, fadeTime, fadeStart, fadeEnd, ambient)

    if (ambient and self.disableAmbientBeds) then return end
    if (not ambient and self.disableMusicBeds) then return end

    Logger:message("Playing bed: " .. musicBedName .. ", startDelay = " .. startDelay .. ", fadeTime = " .. fadeTime .. ", fadeStart = " .. fadeStart .. ", fadeEnd = " .. fadeEnd .. ", ambient = " .. tostring(ambient))

    local sound = SoundManager:getMusicBed(musicBedName, ambient)
    if (sound == nil) then return end

    sound.volumeControl:setTimedFade(startDelay, fadeTime, fadeStart, fadeEnd)
end

function MusicPlayer:playQuantizedStab(stab)
    local delay = MusicPlayer:getQuantizedDelay(stab.quantized)
    DelayScheduler:runAfter(delay, function() SoundManager:playStab(stab.name, stab.volume) end, stab.name)
end

function MusicPlayer:getQuantizedDelay(interval)
    local quantizedDelay = 0
    if (interval > 0) then
        local currentTime = self:getMusicBedTime() / interval
        local nextBlock = math.ceil(currentTime)
        quantizedDelay = (nextBlock - currentTime) * interval
        Logger:message("quantized music delay = " .. self:getMusicBedTime() .. ", interval = " .. interval .. ", nextBlock = " .. nextBlock .. ", delay = " .. quantizedDelay)
    end
    return quantizedDelay
end

function MusicPlayer:getMusicBedTime()
    local currentMusicBed = SoundManager.currentMusicBed
    if (currentMusicBed and currentMusicBed:isPlaying()) then
        return currentMusicBed.volumeControl:getSeekPosition()
    else
        return 0
    end
end

function MusicPlayer:getMusicBedLength()
    local currentMusicBed = SoundManager.currentMusicBed
    if (currentMusicBed and currentMusicBed:isPlaying()) then
        return currentMusicBed:getDurationSeconds()
    else
        return 0
    end
end
