
local Logger = TaggedLog.new("SoundManager", false)

local function audioFileURI(baseName)
    return "windyday:audio/" .. string.lower(baseName) .. ".ogg"
end

local function volumeControlName(sound)
    return sound.name .. "-volume"
end

local SoundManagerGameObj = nil

SoundManager = {
    globalVolumes = {
        Music = 1,
        SFX = 1,
        Stab = 1,
        Ambient = 1,
        AlwaysOn = 1,
    },

    emitter = nil,
    sounds = {},
    volumeTimer = 0,
    adjustAmbient = true,
    ambientFade = 0,
    globalFadeTimer = 0,
    globalFadeAmount = 2, -- 2 seconds

    currentMusicBed = nil,
}

function SoundManager.prefetch(audioName, group)
    Logger:message("Prefetching: " .. audioName .. " from segment: " .. tostring(AnimationManager:currentSegmentName()))
    group:addAudio( audioFileURI(audioName) )
    group:prefetch()
end

function SoundManager.init(scene)
    Logger:message("init started")

    if (SoundManagerGameObj) then SoundManager.shutdown() end
    SoundManagerGameObj = scene:createTransform("SoundManager")

    if (SoundManagerGameObj.components == nil or SoundManagerGameObj.components.SoundManager == nil) then
        SoundManagerGameObj:addComponent("SoundManager", SoundManager)
    end
    if (SoundManagerGameObj.components == nil or SoundManagerGameObj.components.MusicPlayer == nil) then
        SoundManagerGameObj:addComponent("MusicPlayer", MusicPlayer)
    end
    if (SoundManagerGameObj.components == nil or SoundManagerGameObj.components.SoundFXPlayer == nil) then
        SoundManagerGameObj:addComponent("SoundFXPlayer", SoundFXPlayer)
    end

    SoundManager.emitter = AudioEmitter.create()
    Events.listen(SoundManager, "segmentPlaying")
    Logger:message("init completed")
end

function SoundManager.shutdown()
    Logger:message("shutdown started")

    if (SoundManagerGameObj) then
        SoundManagerGameObj:destroy()
        SoundManagerGameObj = nil
    end

    Logger:message("shutdown completed")
end

function SoundManager:onDestroy()
    Events.stopListening(SoundManager, "segmentPlaying")

    if (SoundManager.emitter) then
        SoundManager.emitter:destroy()
        SoundManager.emitter = nil
    end

    SoundManagerGameObj = nil
end

function SoundManager:onUpdate()
    if (MainActorsDetector:anyOnScreen()) then
        if (self.volumeTimer < 1) then self.volumeTimer = self.volumeTimer + Time.dt() / 2 end
    else
        if (self.volumeTimer > 0) then self.volumeTimer = self.volumeTimer - Time.dt() / 3 end
    end

    if (self.adjustAmbient) then
        if (self.ambientFade < 1) then self.ambientFade = self.ambientFade + Time.dt() / 5 end
    else
        if (self.ambientFade > 0) then self.ambientFade = self.ambientFade - Time.dt() end
    end

    self.globalVolumes.SFX = math.lerp(0.3, 1, self.volumeTimer)
    self.globalVolumes.Stab = math.lerp(0.3, 1, self.volumeTimer)
    self.globalVolumes.Music = math.lerp(0.3, 1, self.volumeTimer)

    local globalAmbientVolume2 = math.lerp(0.3, 0.7, 1 - self.volumeTimer)
    self.globalVolumes.Ambient = math.lerp(0.7, globalAmbientVolume2, self.ambientFade)

    if (self.globalFadeTimer > 0) then
        local fadeAmount = (self.globalFadeAmount - self.globalFadeTimer) / self.globalFadeAmount
        self.globalVolumes.SFX = self.globalVolumes.SFX * fadeAmount
        self.globalVolumes.Stab = self.globalVolumes.Stab * fadeAmount
        self.globalVolumes.Music = self.globalVolumes.Music * fadeAmount
        self.globalVolumes.Ambient = self.globalVolumes.Ambient * fadeAmount
        Logger:message("Setting volumes = " .. table.tostring(self.globalVolumes))
        self.globalFadeTimer = self.globalFadeTimer - Time.dt()
        if (self.globalFadeTimer < 0) then self.globalFadeTimer = 0 end
    end
end

function SoundManager:startGlobalFadeIn()
    self.globalFadeTimer = self.globalFadeAmount
    self.globalVolumes.SFX = 0
    self.globalVolumes.Stab = 0
    self.globalVolumes.Ambient = 0
    self.globalVolumes.Music = 0
end

function SoundManager:getSound(name)
    local sound = self.sounds[name]
    if (sound) then return sound end

    -- soundName, uriPath, preCache, startTimeMillis, stopTimeMillis
    sound = self.emitter:addSound(name, audioFileURI(name), false, 0, -1)
    if (sound) then
        sound.name = name
        sound.volumeControl = SoundManagerGameObj:addComponent(volumeControlName(sound), VolumeControl:new(sound))
        self.sounds[name] = sound
        Logger:message("Created sound " .. tostring(name) .. " (" .. tostring(Util.numKeys(self.sounds)) .. " sounds allocated)")
    else
        Log.error("Unable to create sound for: " .. tostring(name))
    end
    return sound
end

function SoundManager:releaseSound(sound)
    --the SoundManager has already been destroyed... (bad lifetime management?)
	if (SoundManager.emitter == nil or SoundManagerGameObj == nil) then
		Logger:message("Released sound but the SoundManager has already been shutdown.")
		return
	end

	--validate that the sound itself is valid.
    if (sound and sound.name) then
		local name = sound.name

        if (sound == self.currentMusicBed) then
            self.currentMusicBed = nil
        end
        if (sound.volumeControl) then
            SoundManagerGameObj:removeComponent(volumeControlName(sound))
            sound.volumeControl = nil
        end
        self.emitter:removeSound(name)
        self.sounds[name] = nil
        Logger:message("Released sound " .. tostring(name) .. " (" .. tostring(Util.numKeys(self.sounds)) .. " sounds allocated)")
    end
end

function SoundManager:getMusicBed(name, ambient)
    local sound = self.sounds[name]
    if (sound) then return sound end

    sound = self:getSound(name)
    if (sound ~= nil) then
        sound:setLooping(true)
        if (ambient) then
            sound.volumeControl:setGlobalVolumeType("Ambient")
        else
            sound.volumeControl:setGlobalVolumeType("Music")
        end
    end

    return sound
end

function SoundManager:playStab(name, volume)
    self:playSound(name, volume, false, "Stab")
    Logger:message("Playing stab: " .. name)
end

function SoundManager:playSFX(name, volume, looping)
    self:playSound(name, volume, looping, "SFX")
    Logger:message("Playing sound FX = " .. name)
end

function SoundManager:playAmbient(name, volume)
    self:playSound(name, volume, false, "Ambient")
    Logger:message("Playing ambient: " .. name)
end

function SoundManager:playSound(name, volume, looping, type)
    local sound = self:getSound(name)
    if (sound == nil) then return end

    sound:setLooping(looping)
    sound.sfx = (type == "SFX")
    sound.volumeControl:setGlobalVolumeType(type)
    sound.volumeControl:setVolume(volume)
    sound:play()
end

function SoundManager:segmentPlaying(segmentName)
    if (segmentName == "PepeIntro") then
        self.adjustAmbient = true
    elseif (segmentName == "HatIntro") then
        self.adjustAmbient = false
    end

    for _, sound in pairs(self.sounds) do
        if (sound.sfx and sound:getLooping()) then
            sound:stop()
        end
    end
end
