
require "windyday/scripts/scenedata/tornado"
require "windyday/scripts/components/blackholealpha"
require "windyday/scripts/components/drivenparticles"
require "windyday/scripts/components/drivenrotator"

local LogDebug = false
local Debug = function(...) end
if (LogDebug) then Debug = Log.message end

-- Utility that converts a bone data table {actor = name, bone = path} into a bone from the actormanager
local function convertBoneDataToBone(table, datafield)
    if (table[datafield]) then
        table[datafield] = ActorManager:getBone(table[datafield])
    end
end

Tornado = {
}

function Tornado.configTornado(proxy)
    Debug("   *** Configurating tornado    ****")

    local tornadoExtras = ActorManager:getActor("Tornado_Extras")
    local tornadoDriver = tornadoExtras:addComponent("tornadodriver", TornadoDriver:new(scenedata.tornado.tornadodriver))

    tornadoDriver.tornadoTube = ActorManager:getActor("Tornado_Tube")

    tornadoDriver["tornadoBlendDriver"] = tornadoDriver.tornadoTube:addComponent("tornadoBlendDriver", BlendDriver:new(scenedata.tornado.blenddrivers.tornadoBlendDriver))
    tornadoDriver["outerColorsBlendDriver"] = tornadoExtras:addComponent("outerColorsBlendDriver", BlendDriver:new(scenedata.tornado.blenddrivers.outerColorsBlendDriver))

    -- Add the black hole alpha drivers
    for driver, data in pairs(scenedata.tornado.blackholeAlphas) do
        tornadoExtras:addComponent(driver, BlackholeAlpha:new(data))
    end

    for particleGroup, data in pairs(scenedata.tornado.particleGroups) do
        tornadoDriver.particleParents[particleGroup] = Global.scene:createTransform(particleGroup)
        if (data.drivenRotator) then
            tornadoDriver.particleParents[particleGroup]:addComponent(particleGroup.."Rotator", DrivenRotator:new(data.drivenRotator))
        end
        for i = 1, #data.drivenParticles do
            tornadoDriver.particleParents[particleGroup]:addComponent(particleGroup.."Particles"..i, DrivenParticles:new(data.drivenParticles[i]))
        end
    end
    
    Debug("   *** DONE Configurating tornado    ****")
end

TornadoDriver = {
    tornadoProxyTopStartPos = 0,
    tornadoProxyBottomStartPos = 0,

    tornadoTube = nil,
    tornadoExtras = nil,

    particleParents = {},

    enabled = false,
}

function TornadoDriver:new(data)
    local this = shallowcopy(data) or {}
    setmetatable(this, self)
    self.__index = self
    return this
end

function TornadoDriver:onStart()
    -- Convert the bone data into bone actor objects
    convertBoneDataToBone(self, "tornadoProxy")
    convertBoneDataToBone(self, "tornadoProxyTop")
    convertBoneDataToBone(self, "tornadoProxyBottom")
    convertBoneDataToBone(self, "tornadoSky")
    convertBoneDataToBone(self, "tornadoSkyDriver")
    convertBoneDataToBone(self, "tornadoSkyScaleDriver")
    convertBoneDataToBone(self, "tornadoScaleGroup")
    convertBoneDataToBone(self, "tornadoBlackCap")
    convertBoneDataToBone(self, "tornadoBlackCapScaler")
    convertBoneDataToBone(self, "tornadoEnableDriver")
    convertBoneDataToBone(self, "tornadoAnimModelSpeed")

    self.tornadoProxyTopStartPos = self.tornadoProxyTop:getLocal().position.y
    self.tornadoProxyBottomStartPos = self.tornadoProxyBottom:getLocal().position.y
    self.tornadoProxyStartHeight = self.tornadoProxyTopStartPos - self.tornadoProxyBottomStartPos
end

function TornadoDriver:onUpdate()
    if (self.tornadoEnableDriver:getWorld().position.y < 0.5) then
        self:enableTornado(false)
        Events.send(self, "tornadoEnd")
    elseif (self.tornadoProxy:getLocal().scale.y > 0.05) then
        if (not self.enabled) then
            self:enableTornado(true)
        end

        -- Cosine curve to multiply in for observation
        -- local debugScalar = ((math.cos(Time.seconds() * 2 ) * 0.5) + 0.5) * 15.0

        local tornadoTransform = self.tornadoScaleGroup:getLocal()
        tornadoTransform.position = self.tornadoProxy:getWorld().position
        tornadoTransform.scale = self.tornadoProxy:getLocal().scale
        self.tornadoScaleGroup:setLocal(tornadoTransform)

        local skyDriverPos = self.tornadoSkyDriver:getWorld().position
        local skyDriverScale = self.tornadoSkyScaleDriver:getLocal().scale
        local skyTransform = self.tornadoSky:getLocal()
        skyTransform.position = Vector(0, skyDriverPos.y, 0)
        skyTransform.scale = skyDriverScale

        self.tornadoSky:setLocal(skyTransform)

        local blackCapScalerPos = self.tornadoBlackCapScaler:getLocal().position
        -- local blackCapScale = Vector.new(blackCapScalerPos.y, blackCapScalerPos.y, blackCapScalerPos.y)
        -- Scaling Y here has a VERY undesirable effect on bounding boxes in relation to the camera
        local blackCapScale = Vector.new(blackCapScalerPos.y, 1.0, blackCapScalerPos.y)
        self.tornadoBlackCap:setLocalScale(blackCapScale)

        local alphaTileX = (self.tornadoProxyBottom:getLocal().position.y - self.tornadoProxyBottomStartPos) / self.tornadoProxyStartHeight
        local alphaTileY = (self.tornadoProxyTop:getLocal().position.y - self.tornadoProxyBottomStartPos) / self.tornadoProxyStartHeight
        if (self.tornadoBlendDriver) then
            self.tornadoBlendDriver.masterAlphaTile.x = alphaTileX
            self.tornadoBlendDriver.masterAlphaTile.y = alphaTileY
        end
        if (self.outerColorsBlendDriver) then
            self.outerColorsBlendDriver.masterAlphaTile.x = alphaTileX
            self.outerColorsBlendDriver.masterAlphaTile.y = alphaTileY
        end

        -- controls speed of tornado wobbly animation
        if (self.tornadoTube and self.tornadoTube.animation and self.tornadoTube.animation.WindUpB) then
            -- This probably should not be specifically named animation...
            self.tornadoTube.animation.WindUpB:setRate(self.tornadoAnimModelSpeed:getWorld().position.y)
        end

    elseif (self.enabled) then
        self:enableTornado(false)
    end
end

function TornadoDriver:onDestroy()
    for _, p in pairs(self.particleParents) do
        p:destroy()
    end
end

function TornadoDriver:enableTornado(enabled)
    self.enabled = enabled
    Debug("Tornado enabled = " .. tostring(enabled))

    self.object:setVisibility(enabled)

    if (self.tornadoTube) then
        self.tornadoTube:setVisibility(enabled)
    end
    if (self.tornadoBlendDriver) then
        self.tornadoBlendDriver:enable(enabled)
    end
    if (self.outerColorsBlendDriver) then
        self.outerColorsBlendDriver:enable(enabled)
    end
end

BlendDriver = {}

function BlendDriver:new(data)
    local this = shallowcopy(data) or {}
    setmetatable(this, self)
    self.__index = self

    Debug("BlendDriver: new returning " .. tostring(this))

    return this
end

function BlendDriver:enable(enabled)
    if (self.rootBone) then
        -- TODO: need support from the engine for this
        -- self.rootBone:setVisible(enabled)
    end
end

function BlendDriver:onStart()
    self.lastOffset = 0

    self.MainTex_offset = Vector.zero
    self.BlendTex1_offset = Vector.zero
    self.BlendTex2_offset = Vector.zero
    self.BlendTex3_offset = Vector.zero

    -- Convert the bone data into bone actor objects
    convertBoneDataToBone(self, "rootBone")
    convertBoneDataToBone(self, "alphaOverride")
    convertBoneDataToBone(self, "colorOverride")
    convertBoneDataToBone(self, "heightScrollMultiplier")
    convertBoneDataToBone(self, "tornadoRotationMultiplier")
    convertBoneDataToBone(self, "heightScrollMover")
    convertBoneDataToBone(self, "invert")

    self.material = self.object:findMaterial(self.materialName)
    if (self.material == nil) then Log.error("Unable to get material " .. self.materialName .. " for BlendDriver.") end

end

function BlendDriver:onUpdate()
    local rotationMult = 1
    if (self.tornadoRotationMultiplier) then rotationMult = self.tornadoRotationMultiplier:getWorld().position.y end
    local invert = 0
    if (self.invert) then invert = self.invert:getWorld().position.y end
    local alphaO = 1
    if (self.alphaOverride) then alphaO = self.alphaOverride:getWorld().position.y end
    local colorO = Vector.new(1,1,1)
    if (self.colorOverride) then colorO = self.colorOverride:getLocal().scale end

    local heightMult = 1
    if (self.heightScrollMultiplier) then heightMult = self.heightScrollMultiplier:getWorld().position.y end
    local heightMove = 0
    if (self.heightScrollMover) then heightMove = self.heightScrollMover:getWorld().position.y end
    self.heightScrollOffset = Time.dt()
    local offset = self.heightScrollOffset + heightMove

    local mat
    local fixForAnimationDrift

	local dt = Time.dt()
    -- main texture scrolling
    self.MainTex_offset = self.MainTex_offset + self.mainTexSpeed * rotationMult * dt
    fixForAnimationDrift = ((offset - self.lastOffset + self.mainTexSpeed.y) * heightMult) * dt
    if (fixForAnimationDrift > .015) then
        self.MainTex_offset.y = self.MainTex_offset.y + fixForAnimationDrift
    end
    self:setTextureParameters("_MainTex", self.MainTex_offset, self.mainTexTile)

    -- blend1 texture scrolling
    self.BlendTex1_offset = self.BlendTex1_offset + (self.blend1TexSpeed * rotationMult * dt)
    fixForAnimationDrift = ((offset - self.lastOffset + self.blend1TexSpeed.y) * heightMult) * dt
    if (fixForAnimationDrift > .015) then
        self.BlendTex1_offset.y = self.BlendTex1_offset.y + fixForAnimationDrift
    end
    self:setTextureParameters("_BlendTex1", self.BlendTex1_offset, self.blend1TexTile)

    -- blend2 texture scrolling
    self.BlendTex2_offset = self.BlendTex2_offset + (self.blend2TexSpeed * rotationMult * dt)
    fixForAnimationDrift = ((offset - self.lastOffset + self.blend2TexSpeed.y) * heightMult) * dt
    if (fixForAnimationDrift > .015) then
        self.BlendTex2_offset.y = self.BlendTex2_offset.y + fixForAnimationDrift
    end
    self:setTextureParameters("_BlendTex2", self.BlendTex2_offset, self.blend2TexTile)

    -- blend3 texture scrolling
    self.BlendTex3_offset = self.BlendTex3_offset + (self.blend3TexSpeed * rotationMult * dt)
    fixForAnimationDrift = ((offset - self.lastOffset + self.blend3TexSpeed.y) * heightMult) * dt
    if (fixForAnimationDrift > .015) then
        self.BlendTex3_offset.y = self.BlendTex3_offset.y + fixForAnimationDrift
    end
    self:setTextureParameters("_BlendTex3", self.BlendTex3_offset, self.blend3TexTile)

    local currentColor = Color.new(colorO.x, colorO.y, colorO.z, alphaO)
    if (self.material) then
        self.material:setParameters({ u_Fade2Black = invert, u_Color = currentColor })
    end

    -- gradient texture scrolling
    self:setTextureParameters("_Gradient", self.gradientValue, self.gradientTile)

    -- masterAlpha texture scrolling
    self:setTextureParameters("_MasterAlpha", self.masterAlphaValue, self.masterAlphaTile)
    
    --Log.message("MasterAlpha tile: "..string.format ("<%f, %f>", self.masterAlphaTile.x, self.masterAlphaTile.y))

    self.lastOffset = offset
end

function BlendDriver:setTextureParameters(texName, textureOffset, textureScale)
    if (self.material) then
        local props = {}
        props["u" .. texName .. "_ST"] = Vector.new(textureScale.x, textureScale.y, textureOffset.x, textureOffset.y)
--        Debug("Setting texture params " .. self.materialName .. ": " .. texName .. table.tostring(props))
        self.material:setParameters(props)
    end
end
