

OmniLightBlendManager = {
	
	activeLightsMap = {}; -- maps a target (ie 'mouse') to its light data
	exclusivityMap = {}; -- map of targets inside a light probe (deprecated)
	
	litHistory = {}; -- keeps track of any lit objects, object that have had a material change
	
	matTargetToExclusiveLight = {}; --new Dictionary<GameObject, OmniLightProbe>();
	
	numberOfEffectiveLights = 2;
	maxEffectiveLight = 4;
	lightColorUniforms = "_AmbientLightColor";
	
	omniLightWeights = Vector.new(0,0,0,0);
	omniLightColors = Matrix.identity;
	
	scene = nil;
	
	probeList = {};
	activeProbes = {};
	runningSet = nil;
}

local compare = function(a, b)
	if(a.hierarchyDepth < b.hierarchyDepth) then
		return true;
	end
	
	return false;
end

local comp = {}
comp[1] = "x"
comp[2] = "y"
comp[3] = "z"
comp[4] = "w"


function OmniLightBlendManager:init(scene)
	self.scene = scene;
	
	local sortedList = {}
	for key, probeData in pairs(OmniLightProbes) do
		
		if(self.probeList[probeData.setName] == nil) then
			self.probeList[probeData.setName] = {}
		end
		table.insert(self.probeList[probeData.setName], probeData)
	end
	
	Events.listen(self, "onActorCreated")
	Events.listen(self, "onActorDestroyed")
	
end

function OmniLightBlendManager:onActorCreated(actor)
	if(self.probeList[actor.name] ~= nil) then
	
		-- destroy old probes if on a new set
		if(self.runningSet ~= actor.name) then
			for _, probe in ipairs(self.activeProbes) do
				probe.object:destroy()
			end
		end
		
		self.runningSet = actor.name;
		
		Log.message("CREATED OMNI LIGHT FOR: "..actor.name)
		
		local probe = self.probeList[actor.name][1];
		local transform = self.scene:createTransform(probe.name);
		self.activeProbes[probe.setName] = transform:addComponent(probe.name, OmniLightProbe:new(probe))
	end
end

function OmniLightBlendManager:onActorDestroyed(name)
	if(self.probeList[name] ~= nil) then
		self.probeList[name] = nil
	end
	
	if(self.activeProbes[name] ~= nil)then
		self.activeProbes[name] = nil
	end
end

	
function OmniLightBlendManager:onStart ()
	self.numberOfEffectiveLights = math.min(self.numberOfEffectiveLights, self.maxEffectiveLight)
	self:clearWeights();
end
	
function OmniLightBlendManager:clearWeights()
	self.omniLightWeights.x = 0
	self.omniLightWeights.y = 0
	self.omniLightWeights.z = 0
	self.omniLightWeights.w = 0
end
	
function OmniLightBlendManager:clearColor()
	self.omniLightColors = Matrix.new(1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1)
end
	
function OmniLightBlendManager:update()

		-- reset any changes
		for key, material in pairs(self.litHistory) do
			material:setParameters({u_AmbientLightColor = Vector.one});
		end
		
		-- dump history now
		self.litHistory = {}
		
		self:clearWeights();
		self:clearColor();
		
		
		for key, value in pairs(self.activeLightsMap) do
			
			local targetData = value.data;
			local materialTarget = targetData.lightTargetMat;
			local activeLights = value.list;
			
			local numLights = math.min(table.getn(activeLights), self.numberOfEffectiveLights);
			
			-- setup blend weights
			for i = 1, numLights, 1 do
			
				local cutoff2 			 = activeLights[i].cutOffDist;
				cutoff2 		    	 = cutoff2*cutoff2; -- squared cutoff distance
				local dist2 			 = activeLights[i]:findTargetInThisProbe(targetData).currentDistanceSqrd;
				self.omniLightWeights[comp[i]] = math.clamp(1.0 - dist2/cutoff2, 0.0, 1.0);
				
				-- setup inverted colors
				self.omniLightColors:setColumn(i, activeLights[i]:getInvertedLightProbeColor());
			end
			
			if(materialTarget ~= nil) then
			
				-- multiply inverted colors by weights and then 'un'-invert
				local color = (Vector.one - self.omniLightColors:mulVec(self.omniLightWeights));
				materialTarget:setParameters({ u_AmbientLightColor = color});
				
				-- since we are changing this material keep track of it
				if(not self.litHistory[materialTarget.name]) then
					self.litHistory[materialTarget.name] = materialTarget;
				end
			end
		end
		
		self.activeLightsMap = {}
		self.exclusivityMap = {}
		self.matTargetToExclusiveLight = {}
end
	
-- deprecated
function OmniLightBlendManager:inVolume(target, probe)
		
		local targetName = target.hashname
		if( self.exclusivityMap[targetName] ~= nil and (self.exclusivityMap[targetName]:findTargetInThisProbe(targetName).currentDistanceSqrd > target.currentDistanceSqrd)) then
			self.exclusivityMap[targetName] = probe;
			self.matTargetToExclusiveLight[target.lightTargetMat.name] = probe;
		end
		
end
	
function OmniLightBlendManager:addProbe(target, probe)

	local activeLights = self.activeLightsMap[target.hashname]; -- list of light probes
	
	if(activeLights == nil) then
		activeLights = {list={}, data=nil};
		self.activeLightsMap[target.hashname] = activeLights;
		activeLights.data = target;
	end
	
	table.insert(activeLights.list, probe);
	
	-- insert lights so the nearest light is at the top of the list
	
	-- insertion sort
	local swap = nil;
	local compareProbe = nil;
	local activeLightCount = table.getn(activeLights.list);
	local probeTestIndex = activeLightCount;
	for i = activeLightCount - 1, 1, -1 do
		compareProbe = activeLights.list[i];
		
		if(target.currentDistanceSqrd < compareProbe:findTargetInThisProbe(target).currentDistanceSqrd) then
			
			-- save old
			swap = activeLights.list[i];
			
			-- replace with new
			activeLights.list[i] = activeLights.list[probeTestIndex];
			
			-- put old in new's previous slot
			activeLights.list[probeTestIndex] = swap;
			
			-- update test index
			probeTestIndex = i;
			
		else
			i = 0; -- break
		end
	end
end
	
function OmniLightBlendManager:reset(target, probe)
	
	if(target.lightTargetMat ~= nil) then
		local material = target.lightTargetMat;
		--material.setProperty(lightColorUniforms, Vector.one);
		material:setParameters({u_AmbientLightColor = Vector.one});
	end
	
end

function OmniLightBlendManager:shutdown()
	self.activeLightsMap = {}; 				
	self.exclusivityMap = {}; 				
	self.litHistory = {}; 					
	self.matTargetToExclusiveLight = {}; 	
	self.scene = nil;
	self.probeList = {};
	self.activeProbes = {};
	self.runningSet = nil;
end
