
local Logger = TaggedLog.new("Segment", true)

Segment = {
    hasBegun = false,
    playing = false,
    beginOnStart = false,
    camera = nil,
}
Segment.__index = Segment

function Segment.new(data, camera)
    local this = Util.copy(data) or {}
    this.camera = camera
    return setmetatable(this, Segment)
end

function Segment:onInitialize()
end

function Segment:update()
    if (self.beginOnStart and not self.hasBegun) then
        self:startPlaying(false)
    end
	
    if(self.actors[1] == nil) then
        return;
    end	

    local lookedAt = self:lookingAtTarget()
	

    if (self.playing) then
        local currentFinished = false
        if (self.actors and #self.actors > 0) then
            local actor = ActorManager:getActor(self.actors[1].name, self.name)
            currentFinished = actor.animDriver:isFinished(true)
        end

        -- If the following conditions are met then we're ready for the next
        if (currentFinished and (not self.needsFocus or lookedAt)) then
            Logger:message(self.name .. ": currentFinished")
            self:prepareNextSegment()
        end
    elseif (not self.hasBegun and self.beginOnSight and lookedAt) then
        self:startPaying(false)
    end
end

function Segment:startPlaying(startingCheckpoint)
    Logger:message("Starting to play: " .. self.name)

    self:hideObjects(self.objectsToHide)
    self:showObjects(self.objectsToShow)

    if (startingCheckpoint and self.hideOnCheckpoint) then
        Logger:message("Hiding checkpoint objects: " .. table.tostring(self.hideOnCheckpoint))
        self:hideObjects(self.hideOnCheckpoint)
    end

    self.playing = true
    self:initializeAnimDrivers()

    if (self.setName) then
        SetManager:showSet(self.setName)
    end

    self:playAnimations()
    self.hasBegun = true
end

function Segment:stopPlaying()
    self.playing = false
end

function Segment:playAnimations()
    for _, actorName in ipairs(self.actors) do
        if (actorName) then
            local actor = ActorManager:getActor(actorName.name, self.name)
            if (actor and actor.animDriver) then
                Logger:message("Playing " .. self.name .. " on " .. actorName.name)
                actor.animDriver:play()
            end
        end
    end
end

function Segment:initializeAnimDrivers()
    self.currentAnim = nil
    for i, actorName in ipairs(self.actors) do
        if (actorName) then
            local actor = ActorManager:getActor(actorName.name, self.name)
            if (actor and actor.animDriver) then
                local anim = actor.animDriver:init(self.name, actor.animation, i == 1)
                Logger:message("Initing " .. self.name .. " on " .. tostring(actor.name) .. " = " .. tostring(anim))
                if (self.currentAnim == nil) then
                    self.currentAnim = anim
                end
            end
        end
    end
end

function Segment:lookingAtTarget()
	if(self.trackingTargets == nil) then
		return false;
	end
    for _, target in ipairs(self.trackingTargets) do
        if (target) then
            local detector = ActorManager:getActor(target)
            if (detector.components) then
                detector = detector.components["lookatdetector"]
                if (detector and detector.lookedAt) then
                    return true
                end
            end
        end
    end
    return false
end

function Segment:prepareNextSegment()
    local targetBranch = 1

    if (self.nextSegments and table.getn(self.nextSegments) > 0) then
        AnimationManager:goToSegment(self.nextSegments[targetBranch], false)
    else
        self:stopPlaying()
        Events.send(self, "IrisClose")
    end
end

function Segment:resetToBeginning(previousSegment)
    self:stopPlaying()

    self.hasBegun = false
    self:hideObjects(self.objectsToShow)
    self:showObjects(self.objectsToHide)
end

function Segment:resetToEnd()
    self:hideObjects(self.objectsToHide)
    self:showObjects(self.objectsToShow)
end

function Segment:currentTime()
    -- TODO: implement this
    return 0
end

function Segment:hideObjects(names)
    self:activateObjects(names, false)
end

function Segment:showObjects(names)
    self:activateObjects(names, true)
end

function Segment:activateObjects(names, activate)
    if (names) then
        for _, name in ipairs(names) do
            local actor
            if (activate) then
                actor = ActorManager:getActor(name, self.name)
            else
                -- Don't want to create the actor if we are only going to hide it
                actor = Global.scene.actors[name]
            end
            if (actor) then
                if (activate) then
                    actor:setVisibility(true)
                else
                    ActorManager:destroyActor(name)
                end
            end
        end
    end
end

function Segment:lookAtFirstTarget()
    -- TODO: implement this
end

-- For Debugging lookAtTargets, remove once unnecessary
-- function Segment:debugDot()
--     local angle = nil
--     for _, target in ipairs(self.trackingTargets) do
--         if (target) then
--             local detector = ActorManager:getActor(target)
--             if (detector.components) then
--                 detector = detector.components["lookatdetector"]
--                 if (detector) then
--                     angle = detector:debugDot()
--                     -- Logger:message(" **** Has LookAtDetector "..angle)
--                 end
--             end
--         end
--     end
--     return angle
-- end
