
local LogDebug = true
local Debug = function(...) end
if (LogDebug) then Debug = Log.message end

Util = {}


Util.isAnimFinished = function(anim, checkLooping)
    if (anim) then
        if (checkLooping and anim:isLooping()) then
            -- Looping animations are "done" when near the loop point, and "not done" at all other times
            if (anim:getDuration() == 0) then
                return true
            end
            local remainder = anim:getTime() % anim:getDuration()
            local epsilon = Time.dt()
            return (remainder < epsilon or remainder > (anim:getDuration() - epsilon))
        else
            return anim:getProgress() >= 1.0
        end
    end
    return false
end


Util.fullGC = function()
    if (System and System.gc) then
        local allobjgroups = 0xffffffff
        local collected = System.gc(allobjgroups)
        while (collected > 0) do
            Debug("Garbage collected " .. collected .. " objects")
            collected = System.gc(allobjgroups)
        end
    end
end

-- Perform a shallow copy of the given table
Util.copy = function(t)
    if (t) then
        local newT = {}
        for key, value in pairs(t) do
            newT[key] = value
        end
        return newT
    end
    return nil
end

-- Randomly shuffles the given table in place
Util.shuffle = function(tab)
    local n = #tab
    while n >= 2 do
        local i = math.random(n)
        tab[n], tab[i] = tab[i], tab[n]
        n = n - 1
    end
end

Util.hash = function(str)
	local hash = 2166136261
	for i = 1, #str do
		local c = str:sub(i,i)
		hash = bit.bxor(hash, string.byte(c, 1)) * 16777619
	end
	return hash
end

Util.numKeys = function(tbl)
	local numItems = 0
	for k,v in pairs(tbl) do
	    numItems = numItems + 1
	end
	return numItems
end

Util.getKeyIdx = function(tbl, idx)
	for k,v in pairs(tbl) do
		if (idx == 1) then return v end
		if (idx < 1) then break end
		idx = idx - 1;
	end
	return nil
end

math.clamp = function(value, min, max)
	if (value < min) then
		value = min
	elseif (value > max) then
        value = max
	end
	return value
end

math.clamp01 = function(value)
    return math.clamp(value, 0, 1)
end

math.sign = function(value)
	if( value < 0.0) then
		return -1.0
	end
	
	return 1.0
end

math.easeOutExpo = function (t) 
	return -2^(-10 * t) + 1
end

math.easeInOutSine = function(t) 
	return -0.5 * (math.cos(math.pi*t) - 1)
end

math.easeInOutQuart = function (t) 
	t = t / 0.5
	if (t < 1) then
		return 1/2*t*t*t*t
	end
	t = t - 2
	return -0.5 * (t*t*t*t - 2)
end

math.easeInOutQuad = function (t) 
	t = t * 0.5;
	if (t < 1) then return 1/2*t*t end
	t = t - 1;
	return -1 * t *(t-2)
end

math.easeOutQuad = function (t) 

	return -1 * t *(t-2)
end

math.lerp = function(from, to, value)
	return from + (to - from) * math.clamp01(value)
end

math.lerpEaseIn = function(from, to, value)

	local t = math.clamp01(value)
	local _value = t*t*t*2.5
	return from + (to - from) * math.clamp01(_value)
end


math.inverseLerp = function(from, to, value)
    if (from == to) then return 0 end
    value = math.clamp(value, from, to)
    return (value - from) / (to - from)
end

math.minArr = function( arr )
	local min = nil
	if(type(arr) == "table") then
		for i, value in ipairs(arr) do
			if(value < min or min == nil) then
				min = value
			end
		end
	end
	return min;
end

math.maxArr = function( arr )
	local max = nil
	if(type(arr) == "table") then
		for i, value in ipairs(arr) do
			if(value > max or max == nil) then
				max = value
			end
		end
	end
	return max;
end

math.floatMax = 5.6e300;
math.floatMin = 5.6e-300;

math.spline = function(sampleTime, knots, knotCount)
	if(knotCount >= 4) then
		-- find the appropriate 4-point span of the spline
		knotCount				= knotCount - 3;
		sampleTime				= math.clamp(sampleTime, .0, 1.0) * knotCount;
	
		local s					= math.floor(sampleTime);
		if(s >= knotCount)then
			s = knotCount - 1;
		end
	
		sampleTime		   	= sampleTime - s;
		local index			= s;
	
		local k0			= knots[index+1];
		local k1			= knots[index+2];
		local k2			= knots[index+3];
		local k3			= knots[index+4];
	
		-- evaluate the span cubic at sampleTime using Horner's rule
		local c3			= -0.5 * k0 +  1.5 * k1 + -1.5 * k2 +  0.5 * k3;
		local c2			=  1.0 * k0 + -2.5 * k1 +  2.0 * k2 + -0.5 * k3;
		local c1			= -0.5 * k0 + 0.5 * k2;
		local c0			= k1;
	
		return ((c3 * sampleTime + c2) * sampleTime + c1) * sampleTime + c0;
	end
	
	-- if you did it wrong you will blow up
	return nil
end

-- not ready for fixemups yet
Quaternion.rotatePoint = function(self, point)
	local num = self.x * 2.0;
	local num2 = self.y * 2.0;
	local num3 = self.z * 2.0;
	local num4 = self.x * num;
	local num5 = self.y * num2;
	local num6 = self.z * num3;
	local num7 = self.x * num2;
	local num8 = self.x * num3;
	local num9 = self.y * num3;
	local num10 = self.w * num;
	local num11 = self.w * num2;
	local num12 = self.w * num3;
	local result = Vector(0,0,0,1);
	
	result.x = (1 - (num5 + num6)) * point.x + (num7 - num12) * point.y + (num8 + num11) * point.z;
	result.y = (num7 + num12) * point.x + (1 - (num4 + num6)) * point.y + (num9 - num10) * point.z;
	result.z = (num8 - num11) * point.x + (num9 + num10) * point.y + (1 - (num4 + num5)) * point.z;
	return result
end
table.insert_sort = function(list, item, compareFunc)
	-- insertion sort
	local swap = nil;
	local compareProbe = nil;
	
	table.insert(list, item)
	
	local count = table.getn(list);
	
	local test = count;
	
	for i = count - 1, 1, -1 do
		compareProbe = list[i];
		
		if(compareFunc(item, compareProbe)) then
			
			-- save old
			swap = list[i];
			
			-- replace with new
			list[i] = list[test];
			
			-- put old in new's previous slot
			list[test] = swap;
			
			-- update test index
			test = i;
			
		else
			i = 0;
		end
	end
	
end

table.find_index = function(list, search_delegate)
	if(list and type(list) == "table") then
		for i, value in ipairs(list) do
			if search_delegate(value) then
				return i;
			end
		end
	end
	
	return nil
end

string.startsWith = function(s, prefix)
    return string.sub(s, 1, string.len(prefix)) == prefix
end

string.endsWith = function(s, suffix)
    return suffix == '' or string.sub(s, -string.len(suffix)) == suffix
end
